/*
 * client.c
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <signal.h>
#include <syslog.h>
#ifdef TIME_WITH_SYS_TIME
 #include <sys/time.h>
 #include <time.h>
#else
 #ifdef HAVE_SYS_TIME_H
  #include <sys/time.h>
 #else
  #include <time.h>
 #endif
#endif
#include <errno.h>

#include "alist.h"
#include "sentry.h"
#include "img.h"
#include "pool.h"
#include "client.h"


int     debug=0;
int     verbose=0;

/* sentry protocol parameter */

char    server_host[128];
int     server_port = SENTRY_CTRL_PORT;

u_char  client_type = OPEN_UDP_UNICAST;
int     data_port   = SENTRY_DATA_PORT;

int     server_size = SIZE_160x120;
int     server_zoom = 1;                /* zoom is invalid parameter */
int     server_color = FULL_COLOR;
int     server_format = SEN_F_JPEG;
int     server_rate = DEFAULT_RATE;
int     request_interval = REQUEST_INTERVAL;

int     hivision_flag = 0;
int     voffset = -1;


#if 0
extern sentry_shmrec *sentry_shm_chunk;
extern sentry_shmrec *sentry_shm_info;
#endif

int     use_shm_flag = 1;
int     shm_size = SHM_SIZE;
int     save_type = SEN_F_JPEG;        /* save with JPEG format */

int     wait_time = WAIT_TIME;
int     rateview_interval = RATEVIEW_INTERVAL;
int     report_interval=10000;

int     output_format;
int     output_corder;

int     ctrlfd = -1;
int     datafd = -1;

#ifdef USE_LOCKFILE
char    lockfile[] = "/var/tmp/sentryc.pid";
#endif

static int new_udp_socket();
void    init_sentryc();
void    usage();

extern  void request_call();


void complete()
{
/*
    extern int sentry_shmid;
    extern key_t sentry_shmkey;
*/
 
    client_type = CLOSE;
    request_call();

#if 0
#ifdef DELETE_IPC
    delete_pool(sentry_shmid);
#endif

    syslog(LOG_NOTICE, "sentryc closed key=%d (%#x)\n",
        sentry_shmkey, sentry_shmkey);
    closelog();

#ifdef USE_LOCKFILE
    unlink(lockfile);
#endif
#endif
    exit (0);
}


static
int
new_udp_socket()
{
  int sockfd;
  int flag=1;

  if ((sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    perror("sentryc: cannot make new UDP socket");
    return -1;
  }
#if 0
  setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, (char *)&flag, sizeof(flag));
#endif

  return sockfd;
}

int build_ctrl_port(char *host, int port) 
{
    int sockfd;
    struct sockaddr_in server;
    struct hostent *hp;

    if(debug) {
        fprintf(stderr, "build_ctrl_port: host '%s', port# %d\n",
            host, port);
    }

    sockfd = new_udp_socket();

    memset(&server, 0, sizeof(server));
    /*
    server.sin_family = hp->h_addrtype;
    */
    server.sin_family = AF_INET;
    server.sin_port = htons(port);
    hp = gethostbyname(host);
    if (!hp) {
      fprintf(stderr, "%s: invalid address type.\n", server_host);
      close(sockfd);
      return -1;
    }
    server.sin_addr = *(struct in_addr*)*(hp->h_addr_list);

    if (connect(sockfd, (struct sockaddr *)&server, sizeof(server)) < 0) {
        printf("fail connect '%s:%d' (%d)\n", server_host, port, errno);
        perror("cannot connect");
        close(sockfd);
        return -1;
    }
    if(debug>=5) {
        fprintf(stderr, "prepared ctrl socket# %d\n", sockfd);
    }

    return sockfd;
}

int build_data_port(int port)
{
    int sockfd;
    struct sockaddr_in me;
    int namelen;
  
    if(debug) {
        fprintf(stderr, "build_data_port: port# %d\n",
            port);
    }

    sockfd = new_udp_socket();
  
    bzero((char *)&me, sizeof(me));
    me.sin_family = AF_INET;
    me.sin_addr.s_addr = htonl(INADDR_ANY);
    me.sin_port = htons(port);

    if (bind(sockfd, (struct sockaddr *)&me, sizeof(me)) < 0) {
        fprintf(stderr, "fail bind() to port# %d (%d)\n", port, errno);
        close(sockfd);
        return -1;
    }

#if 0
    /* get data port# */
    if (getsockname(sockfd, (struct sockaddr *)&me, &namelen) < 0) {
        fprintf(stderr, "fail getsockname() (%d)\n", errno);
        close(sockfd);
        return -1;
    }
#endif

    if(debug>=5) {
        fprintf(stderr, "prepared data port# %d, ctrlfd# %d.\n", 
            data_port, sockfd);
    }

    return sockfd;
}


int build_data_port_try(int sport, int *rport)
{
    int tc;
    int xport;
    int xdatafd;

    if(sport<=-1) {
        tc = 0;
        xport = data_port;
        do {
            fprintf(stderr, "ntry %d %d\n", tc, xport);
            fflush(stderr);
            xdatafd = build_data_port(xport);
            if(xdatafd>=0) {
                break;
            }
            else {
                fprintf(stderr, "fail setup data fd\n");
            }
            tc++;
            xport = rand()%64000+1000;
        } while(xdatafd<0 && tc<5);

        if(xdatafd>=0) {
            *rport = xport;
        }
        else {
        }
#if 1
        fprintf(stderr,
            "build_data_port_try: sport %d-> xport %d, xdatafd %d\n",
            sport, xport, xdatafd);
#endif
        return xdatafd;
    }
    else {
#if 1
        fprintf(stderr,
            "build_data_port_try: sport %d\n", sport);
#endif
        return build_data_port(sport);
    }
}






sentry_protocol3_request_hdr req_hdr;

void
init_request()
{
  struct itimerval itval;

    if(debug>=10) {
        fprintf(stderr,
            "init_reqeust: reqeust_interval %d\n", request_interval);
    }

  itval.it_value.tv_sec = request_interval;
  itval.it_value.tv_usec = 0;
  itval.it_interval.tv_sec = request_interval;
  itval.it_interval.tv_usec = 0;

  setitimer(ITIMER_REAL, &itval, NULL);
}

void
_request_call(int serv_recept_fd, int here_receive_port)
{
  if (serv_recept_fd < 0) {
    fprintf(stderr, "no dest., set serv_recept_fd as destination fd.\n");
    return;
  }
#if 0
  else {
    fprintf(stderr,
        "_request_call: send request to server fd# %d, with here port# %d\n",
        serv_recept_fd, here_receive_port);
  }
#endif

  /*
   * I don't know it is proper to to change to hivision size here. :)
   */
  if (hivision_flag) {
    switch(server_size) {
    case SIZE_640x480:
      server_size = SIZE_640x360;
      break;
    case SIZE_320x240:
      server_size = SIZE_320x180;
      break;
    case SIZE_212x160:
      server_size = SIZE_212x120;
      break;
    case SIZE_160x90:
      server_size = SIZE_160x90;
      break;
    }
  }

  if (debug>=10) 
    fprintf(stderr,
"request_call: type# %x, here_receive_port# %d, size# %d, zoom# %d, color# %d, format# %d, rate# %d\n", 
        client_type, here_receive_port, server_size, server_zoom, server_color, server_format, server_rate);

  if (debug>=10) 
    fprintf(stderr,
"_request_call: fd# %d, comm. type %s, here_receive_port# %d, rate# %.3f\n\
              image size %s, zoom# %d, color %s, format %s\n", 
        serv_recept_fd, assoc_v(req_type_alist, client_type), here_receive_port,
            (double)server_rate/100.0,
        assoc_v(size_alist, server_size),
        server_zoom,
        assoc_v(color_alist, server_color), 
        assoc_v(format_alist, server_format));

  req_hdr.hd_ver = SENTRY_PROTOCOL_VERSION;
  req_hdr.type = client_type;
  req_hdr.port = htons(here_receive_port);
  req_hdr.size = htonl(server_size);
  req_hdr.zoom = htonl(server_zoom);
  req_hdr.color = htonl(server_color);
  req_hdr.format = htonl(server_format);
  req_hdr.rate = htonl(server_rate);

  write(serv_recept_fd, (char *)&req_hdr, sizeof(req_hdr));
}





/*
 * for upper compatible
 */
void
request_call()
{
    signal(SIGALRM, SIG_IGN);
    _request_call(ctrlfd, data_port);
    signal(SIGALRM, request_call);
}
