/*==
 ==*/
/*
 * $Id$
 */

#include <stdio.h>
#include <stdlib.h>

#ifndef MAX 
#define MAX(a,b)    ((a)>(b) ? (a) : (b))
#endif
#ifndef MIN
#define MIN(a,b)    ((a)<(b) ? (a) : (b))
#endif

typedef struct {
    int level;
    int cval;
} levelmap_t;

levelmap_t map_far[] = {
    {1L<<0  ,   0x00},
    {1L<<1  ,   0x11},
    {1L<<2  ,   0x22},
    {1L<<3  ,   0x33},
    {1L<<4  ,   0x44},
    {1L<<5  ,   0x55},
    {1L<<6  ,   0x66},
    {1L<<7  ,   0x77},
    {1L<<8  ,   0x88},
    {1L<<9  ,   0x99},
    {1L<<10 ,   0xaa},
    {1L<<11 ,   0xbb},
    {1L<<12 ,   0xcc},
    {1L<<13 ,   0xdd},
    {1L<<14 ,   0xee},
    {-1,        0xff},
};

levelmap_t map_near[] = {
    {1L<<0  ,   0xff},
    {1L<<1  ,   0xee},
    {1L<<2  ,   0xdd},
    {1L<<3  ,   0xcc},
    {1L<<4  ,   0xbb},
    {1L<<5  ,   0xaa},
    {1L<<6  ,   0x99},
    {1L<<7  ,   0x88},
    {1L<<8  ,   0x77},
    {1L<<9  ,   0x66},
    {1L<<10 ,   0x55},
    {1L<<11 ,   0x44},
    {1L<<12 ,   0x33},
    {1L<<13 ,   0x22},
    {1L<<14 ,   0x11},
    {-1,        0x00},
};

int
resolve_cval(levelmap_t *m, int l)
{
    register int i;
    register int r;

    i = 0;
    while(m[i].level>0 && l>m[i].level) {
        i++;
    }

    r = m[i].cval;

    return r;
}



int
convert_RGB2XBGR(unsigned char *dest, unsigned char *src, int len)
{
    int i;
    unsigned char *p, *q;

    p = src;
    q = dest;
    i = 0;
    while(i<len) {
        *q++ = 0xff;
        *q++ = *(p+2);
        *q++ = *(p+1);
        *q++ = *(p+0);
        p += 3;
        i++;
    }

    return len;
}

int
convert_RGB2BGRX(unsigned char *dest, unsigned char *src, int len)
{
    volatile int i;
    volatile unsigned char *p, *q;

    p = src;
    q = dest;
    i = 0;
    while(i<len) {
        *q++ = *(p+2);
        *q++ = *(p+1);
        *q++ = *(p+0);
        *q++ = 0xff;
        p += 3;
        i++;
    }

    return len;
}

int 
convert_none(char *cur, int size, int xsize, int ysize,
    int *nxsize, int *nysize, char **new)
{
    *new = cur;
    *nxsize = xsize;
    *nysize = ysize;

    return size;
}

void
convert_BGRA2RGB_s(unsigned char *cur, int xsize, int ysize)
{
    unsigned char *ptr1, *ptr2, tmp;
    int n;

    /* covert RGB lines. BGRa -> RGB */
    for(n=0, ptr1=cur, ptr2=cur; n<xsize*ysize; n++, ptr1+=4, ptr2+=3){
        tmp=ptr1[0]; ptr2[0]=ptr1[2]; ptr2[1]=ptr1[1]; ptr2[2]=tmp;
    }
}

void
convert_BGR2RGB_s(unsigned char *cur, int xsize, int ysize)
{
    unsigned char *p, *pend;
    unsigned char tmp;
    
    p = cur;
    pend = cur + xsize*ysize*3;
    while(p<pend) {
        tmp = *p;
        *p = *(p+2);
        p++;
        p++;
        *p = tmp;
        p++;
    }
}


int
subghost_rgb(unsigned char *abuf, unsigned char *bbuf,
    unsigned char *cbuf, int wid, int ht)
{
    unsigned char *p, *q, *s;
    unsigned char *pr, *pg, *pb, *qr, *qg, *qb;
    unsigned char gp, gq;
    int iv;
    unsigned char *pend;
    unsigned char *vb, *ve;
    int dv;
    int len;
    
    len = wid*ht*3;

    p = abuf;
    q = bbuf;
    s = cbuf;
    pend = p+len;
    vb = p;
    ve = p + len;

    while(p<pend) {
        pr = p; pg = p+1; pb = p+2;
        qr = q; qg = q+1; qb = q+2;
        iv = (*pr+*pg+*pb)/3;
        gp = (iv>255? 255: iv);
        iv = (*qr+*qg+*qb)/3;
        gq = (iv>255? 255: iv);

        dv = (*pr-*qr)*(*pr-*qr)+(*pg-*qg)*(*pg-*qg)+(*pb-*qb)*(*pb-*qb);
/*DIFF*/
        if(dv<800) {
            *s++ = gp;
            *s++ = gp;
            *s++ = gp;
        }
        else {
            *s++ = 0xff;
            *s++ = (gp+gq)/2;
            *s++ = (gp+gq)/2;
        }
        p+=3;
        q+=3;
    }
}

int
subghost_bgrx(unsigned char *abuf, unsigned char *bbuf,
    unsigned char *cbuf, int wid, int ht)
{
    unsigned char *p, *q, *s;
    unsigned char *pr, *pg, *pb, *qr, *qg, *qb;
    unsigned char gp, gq;
    int iv;
    unsigned char *pend;
    unsigned char *vb, *ve;
    int dv;
    int len;
    
    len = wid*ht*4;

    p = abuf;
    q = bbuf;
    s = cbuf;
    pend = p+len;
    vb = p;
    ve = p + len;

    while(p<pend) {
        pr = p; pg = p+1; pb = p+2;
        qr = q; qg = q+1; qb = q+2;
        iv = (*pr+*pg+*pb)/3;
        gp = (iv>255? 255: iv);
        iv = (*qr+*qg+*qb)/3;
        gq = (iv>255? 255: iv);

        dv = (*pr-*qr)*(*pr-*qr)+(*pg-*qg)*(*pg-*qg)+(*pb-*qb)*(*pb-*qb);

#if 0
        if(p>=vb&&p<=ve) {
            fprintf(stderr, "p %08x %5d %03dx%03d: %02x %02x %02x : %6d\n",
                p, p-abuf, ((p-abuf)/3)%wid, ((p-abuf)/3)/wid,
                *pr, *pg, *pb, *qr, *qg, *qb, dv);
        }
#endif
    
/*DIFF*/
        if(dv<800) {
            *s++ = gp;
            *s++ = gp;
            *s++ = gp;
            s++;
        }
        else {
#if 0
            *s++ = 0xff;
            *s++ = gp;
            *s++ = gq;
#endif
#if 1
            *s++ = gp;
            *s++ = gq;
            *s++ = (gp+gq)/2;
#endif
#if 0
            *s++ = 0xff;
            *s++ = (gp+gq)/2;
            *s++ = (gp+gq)/2;
#endif
            s++;
        }
        p+=4;
        q+=4;
    }
}

/* return a percentage of active on the area */
int
count_covered_neighbors(unsigned char *base, unsigned char *cur,
    int wid, int ht, register int el, int vl, int vr) 
{
    int x, y;
    int gx, gy;
    int gxs, gxe;
    int gys, gye;
    register int a;
    register unsigned char *f;
    register unsigned char *fe;

    x = ((cur-base)/el)%wid;
    y = ((cur-base)/el)/wid;
    gxs = MAX(0, x-vr);
    gxe = MIN(x+vr, wid);
    gys = MAX(0,y-vr);
    gye = MIN(y+vr,ht);
    a = 0;

    for(gy=gys; gy<=gye; gy++) {
        f = base+(gy*wid+gxs)*el;
#if 1
        for(gx=gxs;gx<=gxe;gx++) {
            if(*f>vl) a++;
            f += el;
        }
#endif
#if 0
        fe = f+(gxe-gxs+1)*el;
        for(;f<fe;f+=el) {
            if(*f>vl) a++;
        }
#endif
    }

    return a*100/((gxe-gxs+1)*(gye-gys+1));
}

/* XXX buggy */
int
merge_neighbors(unsigned char *base, unsigned char *cur,
    int wid, int ht, register int el, int vl, int vr, int vu) 
{
    int x, y;
    int gx, gy;
    int gxs, gxe;
    int gys, gye;
    register unsigned char *f;

    x = ((cur-base)/el)%wid;
    y = ((cur-base)/el)/wid;
    gxs = MAX(0, x-vr);
    gxe = MIN(x+vr, wid);
    gys = MAX(0,y-vr);
    gye = MIN(y+vr,ht);

#if 0
fprintf(stderr, "x %d-%d y %d-%d\n", gxs, gxe, gys, gye);
#endif

    for(gy=gys; gy<=gye; gy++) {
        f = base+(gy*wid+gxs)*el;
        for(gx=gxs;gx<=gxe;gx++) {
            if(count_covered_neighbors(
                base, f, wid, ht, el, vl, vr)>vu)
                *(f+1) = 1;
            else
                *(f+1) = 0;

            f += el;
        }
    }

    return;
}




int
subwipebg_rgb(unsigned char *abuf, unsigned char *bbuf,
    unsigned char *cbuf, int wid, int ht)
{
    register unsigned char *p, *q, *s;
    register unsigned char *pend;
    int len;
    unsigned char *tmp;
    int cv;
    int vr, vl, vu;

#if 0
    for(cv=0;cv<256;cv++) {
        cc[cv] = 0;
    }
#endif
    
    len = wid*ht*3;
    tmp = alloca(len);

    /*
     * match fg and bg images
     */
    s = tmp;
    p = abuf;
    q = bbuf;
    pend = p+len;
    while(p<pend) {
        /* be dark as different image */
        *s = resolve_cval(map_far, 
                (*p-*q)*(*p-*q)+(*(p+1)-*(q+1))*(*(p+1)-*(q+1))
                    +(*(p+2)-*(q+2))*(*(p+2)-*(q+2))
                );
        *(s+1) = 0;
/*
        *s = cv;
*/
        s += 3;
        p += 3;
        q += 3;
    }



    vr = 5;
    vl = 0xa0;
    vu = 33;

    /*
     * check points arround focus point and store that result
     * to G-part of the pixel.
     */
    p = tmp;
    pend = p+len;
    while(p<pend) {
#if 1
        if(*p>vl || *(p+1)==0) {
        *(p+1) =
          ( count_covered_neighbors(tmp, p, wid, ht, 3, vl, vr)>vu ?
            2 : 1 );
        }
        else {
            *(p+1) = 1;
        }
#endif

#if 0
        if(*p>vl) {
            merge_neighbors(tmp, p, wid, ht, 3, vl, vr, vu);
        }
        else {
            *(p+1) = 1;
        }
#endif

        p += 3;
    }


    /*
     * according to a G-part of a pixel, masking background image
     */

    s = cbuf; 
    p = tmp;
    q = bbuf;
    pend = p+len;
    p++;
    while(p<pend) {
        if(*p==2) {
            *s++ = *q++;
            *s++ = *q++;
            *s++ = *q++;
        }
        else {
            *s++ = 0x99;
            *s++ = 0x99;
            *s++ = 0x99;
            q += 3;
        }
        p += 3;
    }

#if 0
    for(cv=0;cv<256;cv++) {
        if(cc[cv]) {
            fprintf(stderr, " %3d 0x%02x %6d\n", cv, cv, cc[cv]);
        }
    }
#endif

}


int
Ysubwipebg_rgb(unsigned char *abuf, unsigned char *bbuf,
    unsigned char *cbuf, int wid, int ht)
{
    unsigned char *tmp;
    unsigned char *p, *q, *s, *u;
    unsigned char *pr, *pg, *pb, *qr, *qg, *qb;
    unsigned char gp, gq;
    int iv;
    unsigned char *pend;
    unsigned char *vb, *ve;
    int dv;
    int len;
    
    len = wid*ht*3;

    tmp = alloca(len);

    s = tmp;

    p = abuf;
    q = bbuf;
    pend = p+len;
    vb = p;
    ve = p + len;

    while(p<pend) {
        pr = p; pg = p+1; pb = p+2;
        qr = q; qg = q+1; qb = q+2;
        iv = (*pr+*pg+*pb)/3;
        gp = (iv>255? 255: iv);
        iv = (*qr+*qg+*qb)/3;
        gq = (iv>255? 255: iv);

        dv = (*pr-*qr)*(*pr-*qr)+(*pg-*qg)*(*pg-*qg)+(*pb-*qb)*(*pb-*qb);

        /*DIFF*/
        if(dv<800) {
            *s++ = 0x00;
            *s++ = 0x00;
            *s++ = 0x00;
        }
        else {
            *s++ = 0xff;
            *s++ = 0xff;
            *s++ = 0xff;
        }
        p+=3;
        q+=3;
    }

    s = cbuf; 
    p = tmp;
    q = bbuf;
    pend = p+len;
    vb = p;
    ve = p + len;

    while(p<pend) {
        if(*p) {
            *s++ = *q++;
            *s++ = *q++;
            *s++ = *q++;
        }
        else {
            *s++ = 0xff;
            *s++ = 0xff;
            *s++ = 0xff;
            q += 3;
        }

        p+=3;
    }


}





int
Xsubwipebg_rgb(unsigned char *abuf, unsigned char *bbuf,
    unsigned char *cbuf, int wid, int ht)
{
    unsigned char *p, *q, *s;
    unsigned char *pr, *pg, *pb, *qr, *qg, *qb;
    unsigned char gp, gq;
    int iv;
    unsigned char *pend;
    unsigned char *vb, *ve;
    int dv;
    int len;
    
    len = wid*ht*3;

    p = abuf;
    q = bbuf;
    s = cbuf;
    pend = p+len;
    vb = p;
    ve = p + len;

    while(p<pend) {
        pr = p; pg = p+1; pb = p+2;
        qr = q; qg = q+1; qb = q+2;
        iv = (*pr+*pg+*pb)/3;
        gp = (iv>255? 255: iv);
        iv = (*qr+*qg+*qb)/3;
        gq = (iv>255? 255: iv);

        dv = (*pr-*qr)*(*pr-*qr)+(*pg-*qg)*(*pg-*qg)+(*pb-*qb)*(*pb-*qb);

/*DIFF*/
        if(dv<800) {
            *s++ = 0xff;
            *s++ = 0xff;
            *s++ = 0xff;
        }
        else {
            *s++ = *qr;
            *s++ = *qg;
            *s++ = *qb;
        }
        p+=3;
        q+=3;
    }
}


int
engray_rgb(unsigned char *dst, unsigned char *src,
    int wid, int ht, int el)
{
    int x, y;
    unsigned char *p, *q;
    int i;
    int g;

    q = dst;
    p = src;
    for(x=0;x<wid;x++) {
        for(y=0;y<ht;y++) {
            p = src+(y*wid+x)*el;
            q = dst+(y*wid+x)*el;
            g = 0;  
            for(i=0;i<el;i++)
                g += *p++;
            g = g/3;
            for(i=0;i<el;i++)
                *q++ = g;
        }
    }
}

int
flip_rgb(unsigned char *dst, unsigned char *src,
    int wid, int ht, int el)
{
    int x, y;
    unsigned char *p, *q;
    int i;

    q = dst;
    p = src;
    for(x=0;x<wid;x++) {
        for(y=0;y<ht;y++) {
            p = src+(y*wid+x)*el;
            q = dst+(y*wid+(wid-x))*el;
            for(i=0;i<el;i++)
                *q++ = *p++;
        }
    }
}

int
flop_rgb(unsigned char *dst, unsigned char *src,
    int wid, int ht, int el)
{
    int x, y;
    unsigned char *p, *q;
    int i;

    q = dst;
    p = src;
    for(y=0;y<ht;y++) {
        for(x=0;x<wid;x++) {
            p = src+(y*wid+x)*el;
            q = dst+((ht-y)*wid+x)*el;
            for(i=0;i<el;i++)
                *q++ = *p++;
        }
    }
}

/* with high cut */
int
sub3hc_rgb(unsigned char *dst, unsigned char *src1, unsigned char *src2,
    int wid, int ht, int el)
{
    int x, y;
    unsigned char *p1, *p2, *s;
    int i;
    int g1, g2;
    int d;
    int count;

    count = 0;
    for(x=0;x<wid;x++) {
        for(y=0;y<ht;y++) {
            p1 = src1+(y*wid+x)*el;
            p2 = src2+(y*wid+x)*el;
            s = dst+(y*wid+x)*el;
            g1 = 0; for(i=0;i<el;i++) g1 += *p1++; g1 = g1/3;
            g2 = 0; for(i=0;i<el;i++) g2 += *p2++; g2 = g2/3;
            if(g1>g2)   d = g1 - g2;
            else        d = g2 - g1;

            if(g1<0xd0 && d>0x30) {
                *s++ = 0xff;
                for(i=0;i<el-1;i++)
                    *s++ = 0x00;
                count++;
            }
            else {  
                for(i=0;i<el;i++)
                    *s++ = g1;
            }
        }
    }

    return count;
}

int
sub3_rgb(unsigned char *dst, unsigned char *src1, unsigned char *src2,
    int wid, int ht, int el)
{
    int x, y;
    unsigned char *p1, *p2, *s;
    int i;
    int g1, g2;
    int d;

    for(x=0;x<wid;x++) {
        for(y=0;y<ht;y++) {
            p1 = src1+(y*wid+x)*el;
            p2 = src2+(y*wid+x)*el;
            s = dst+(y*wid+x)*el;
            g1 = 0; for(i=0;i<el;i++) g1 += *p1++; g1 = g1/3;
            g2 = 0; for(i=0;i<el;i++) g2 += *p2++; g2 = g2/3;
            if(g1>g2)   d = g1 - g2;
            else        d = g2 - g1;

            if(d>0x20) {
                *s++ = 0xff;
                for(i=0;i<el-1;i++)
                    *s++ = 0x00;
            }
            else {  
                for(i=0;i<el;i++)
                    *s++ = g1;
            }
        }
    }
}

int
hline_rgb(unsigned char *dst, unsigned char *src,
    int wid, int ht, int el)
{
    int x, y;
    unsigned char *p, *q;
    int i;
    int g1, g2;
    int d;

    q = dst;
    p = src;
    for(y=0;y<ht;y++) {
        for(x=0;x<wid-1;x++) {
            p = src+(y*wid+x)*el;
            g1 = 0; for(i=0;i<el;i++) g1 += *p++; g1 = g1/3;
            p = src+(y*wid+(x+1))*el;
            g2 = 0; for(i=0;i<el;i++) g2 += *p++; g2 = g2/3;

            if(g1>=g2)  d = g1-g2;  
            else        d = g2-g1;

            q = dst+(y*wid+x)*el;
            if(d<=0x20) {
                for(i=0;i<el;i++)
                    *q++ = 0xff;
            }
            else {
                for(i=0;i<el;i++)
                    *q++ = 0x55;
            }
        }
    }
}

int
vline_rgb(unsigned char *dst, unsigned char *src,
    int wid, int ht, int el)
{
    int x, y;
    unsigned char *p, *q;
    int i;
    int g1, g2;
    int d;

    q = dst;
    p = src;
    for(y=0;y<ht-1;y++) {
        for(x=0;x<wid;x++) {
            p = src+(y*wid+x)*el;
            g1 = 0; for(i=0;i<el;i++) g1 += *p++; g1 = g1/3;
            p = src+((y+1)*wid+x)*el;
            g2 = 0; for(i=0;i<el;i++) g2 += *p++; g2 = g2/3;

            if(g1>=g2)  d = g1-g2;  
            else        d = g2-g1;

            q = dst+(y*wid+x)*el;
            if(d<=0x20) {
                for(i=0;i<el;i++)
                    *q++ = 0xff;
            }
            else {
                for(i=0;i<el;i++)
                    *q++ = 0x55;
            }
        }
    }
}

int
hvline_rgb(unsigned char *dst, unsigned char *src,
    int wid, int ht, int el)
{
    int x, y;
    unsigned char *p, *q;
    int i;
    int g, gh, gv;
    int dh, dv;

    q = dst;
    p = src;
    for(y=0;y<ht-1;y++) {
        for(x=0;x<wid-1;x++) {
            p = src+(y*wid+x)*el;
            g = 0; for(i=0;i<el;i++) g += *p++; g = g/3;
            p = src+(y*wid+(x+1))*el;
            gh = 0; for(i=0;i<el;i++) gh += *p++; gh = gh/3;
            p = src+((y+1)*wid+x)*el;
            gv = 0; for(i=0;i<el;i++) gv += *p++; gv = gv/3;

            if(g>=gh)   dh = g-gh;  
            else        dh = gh-g;
            if(g>=gv)   dv = g-gv;  
            else        dv = gv-g;

            q = dst+(y*wid+x)*el;
            if(dh<=0x20 && dv<=0x20) {
                for(i=0;i<el;i++)
                    *q++ = 0xff;
            }
            else {
                for(i=0;i<el;i++)
                    *q++ = 0x55;
            }
        }
    }
}

int
brightcircle_rgb(unsigned char *dst, unsigned char *src,
    int wid, int ht, int el, int rr)
{
    int x, y, r;
    int dr;
    unsigned char *p, *q;
    unsigned char *en;
    int i;
    int g;

    q = dst;
    p = src;
    en = src+(wid*ht*el);
    dr = rr*rr;
    while(p<en) {
        x = ((en-p)/el)%wid;
        y = ((en-p)/el)/wid;
        r = (wid/2-x)*(wid/2-x)+(ht/2-y)*(ht/2-y);

        if(r<dr) {
            for(i=0;i<el;i++)
                *q++ = *p++;
        }
        else {
            g = 0;  
            for(i=0;i<el;i++)
                g += *p++;
            g = g/3;
            for(i=0;i<el;i++)
                *q++ = g;
        }
    }
}

int
clipcircle_rgb(unsigned char *dst, unsigned char *src,
    int wid, int ht, int el, int rr)
{
    int x, y, r;
    int dr;
    unsigned char *p, *q;
    unsigned char *en;
    int i;
    int g;

    q = dst;
    p = src;
    en = src+(wid*ht*el);
    dr = rr*rr;
    while(p<en) {
        x = ((p-src)/el)%wid;
        y = ((p-src)/el)/wid;
        r = (wid/2-x)*(wid/2-x)+(ht/2-y)*(ht/2-y);

        if(r<dr) {
            for(i=0;i<el;i++)
                *q++ = *p++;
        }
        else {
            p += el;
            for(i=0;i<el;i++)
                *q++ = 0x99;
        }
    }
}

int
zoomcenter_rgb(unsigned char *dst, unsigned char *src,
    int wid, int ht, int el)
{
    int x, y, r;
    int dr;
    unsigned char *p, *q;
    unsigned char *en;
    int i;
    int g;

    p = src;
    q = dst;
    en = dst+(wid*ht*el);
    while(q<en) {
        x = ((q-dst)/el)%wid;
        y = ((q-dst)/el)/wid;

        p = src+((y/2+ht/4)*wid+(x/2+wid/4))*el;
        for(i=0;i<el;i++)
            *q++ = *p++;
    }
}


int
dope_rgb(unsigned char *dst, unsigned char *src,
    int wid, int ht, int el, int dp)
{
    int x, y;
    unsigned char *p, *q;
    int i;

    q = dst;
    p = src;
    for(y=0;y<ht;y++) {
        p = src+(y*wid)*el;
        q = dst+(y*wid)*el;
        for(x=0;x<wid;x++) {
            for(i=0;i<el;i++) {
                *q = (unsigned char)
                        ((((unsigned int)*q)*dp
                            + (unsigned int)*p)/(dp+1));
                p++;
                q++;
            }
        }
    }
}

int
Xadd3_rgb(unsigned char *dst, unsigned char *src1, unsigned char *src2,
    int wid, int ht, int el)
{
    int x, y;
    unsigned char *p1, *p2, *q;
    int i;

    q = dst;
    p1 = src1;
    p2 = src2;
    for(y=0;y<ht;y++) {
        for(x=0;x<wid;x++) {
            p1 = src1+(y*wid+x)*el;
            p2 = src2+(y*wid+x)*el;
            q  = dst+(y*wid+x)*el;
            for(i=0;i<el;i++) {
                *q = (unsigned char)(((int)*p1+(int)*p2)/2);
                p1++;
                p2++;
                q++;
            }
        }
    }
}

int
add3_rgb(unsigned char *dst, unsigned char *src1, unsigned char *src2,
    int wid, int ht, int el)
{
    register unsigned char *p1, *p2, *q;
    register unsigned char *e;

    e = src1+(wid*ht*el);
    p1 = src1;
    p2 = src2;
    q = dst;
    while(p1<e) {
        *q++ = (unsigned char)(((int)*p1+(int)*p2)/2);
        p1++;
        p2++;
    }
}

int
mul3_rgb(unsigned char *dst, unsigned char *src1, unsigned char *src2,
    int wid, int ht, int el)
{
    register unsigned char *p1, *p2, *q;
    register unsigned char *e;

    e = src1+(wid*ht*el);
    p1 = src1;
    p2 = src2;
    q = dst;
    while(p1<e) {
        *q++ = (unsigned char)(((int)*p1*(int)*p2)>>8);
        p1++;
        p2++;
    }
}


