/*==
 ==*/
/*
 * $Id$
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#ifdef HAVE_INTTYPES_H
#include <inttypes.h>
#endif
#include <sys/time.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <setjmp.h>
#include <errno.h>

#include <jpeglib.h>

#include "alist.h"
#include "sentry.h"


#define OUTBUF_SIZE (16*1024)

unsigned char outbuf[OUTBUF_SIZE];

void x_src_init(j_decompress_ptr cinfo)
{
/*
  PASS;
*/
}

boolean x_src_isempty(j_decompress_ptr cinfo)
{
/*
  PASS;
*/
  return FALSE;
}

void x_src_input_data(j_decompress_ptr cinfo, long any)
{
/*
  PASS;
*/
}

boolean x_src_sync(j_decompress_ptr cinfo, int any)
{
/*
  PASS;
*/
  return FALSE;
}

void x_src_term(j_decompress_ptr cinfo)
{
/*
  PASS;
*/
}


jmp_buf error_recover_jmp;

void
err_trap(j_common_ptr cinfo)
{
    fprintf(stderr, "err_trap: meet JPEG error\n");
    (*cinfo->err->output_message) (cinfo);
    fflush(stderr);

    longjmp(error_recover_jmp, 77);
}


int
decode_jpeg(unsigned char *raw_data, int raw_len,
    unsigned char *out_area, int out_len, int *rwid, int *rht)
{

    struct jpeg_decompress_struct cinfo;
    struct jpeg_error_mgr jerr;
    struct jpeg_source_mgr src_mgr;
    JSAMPROW bufptr[1];
    int wid, ht;
    int chk;

    JDIMENSION num_scanlines;

    if(debug>=100) {
        fprintf(stderr, 
"decode_jpeg: from %p, len %d\n\
             to   %p, &wid %p, &ht %p\n",
            raw_data, raw_len, out_area, rwid, rht);
    }

#if 0
dump_to_file("/var/tmp/x.jpg", raw_data, raw_len);
#endif


    src_mgr.next_input_byte     = raw_data;
    src_mgr.bytes_in_buffer     = raw_len;
    src_mgr.init_source         = x_src_init;
    src_mgr.fill_input_buffer   = x_src_isempty;
    src_mgr.skip_input_data     = x_src_input_data;
    src_mgr.resync_to_restart   = x_src_sync;
    src_mgr.term_source         = x_src_term;


    cinfo.err = jpeg_std_error(&jerr);
    jerr.error_exit = err_trap;
    if(setjmp(error_recover_jmp)) {
        fprintf(stderr, "recover JPEG decording\n");
        jpeg_destroy_decompress(&cinfo);
        return -1;
    }

    jpeg_create_decompress(&cinfo);

    cinfo.src = &src_mgr;
    cinfo.out_color_space = JCS_RGB;

    chk = jpeg_read_header(&cinfo, TRUE);

    chk = jpeg_start_decompress(&cinfo);

    *rwid = cinfo.output_width;
    *rht  = cinfo.output_height;

    if(debug>=100) {
        fprintf(stderr, "wid %d, ht %d\n",
            cinfo.output_width, cinfo.output_height);
    }
    if(out_len<cinfo.output_width*cinfo.output_height*3) {
        fprintf(stderr,
            "not enough out area size: given %d require %d [bytes]\n",
            out_len, cinfo.output_width*cinfo.output_height*3);
        return -1;
    }

    while (cinfo.output_scanline < cinfo.output_height) {
        bufptr[0] =
            (JSAMPROW) &out_area[cinfo.output_width*3*cinfo.output_scanline];
        num_scanlines = jpeg_read_scanlines(&cinfo, bufptr, 1);
        if(num_scanlines!=1) {
            fprintf(stderr, "meet JPEG decording error\n");
            jpeg_destroy_decompress(&cinfo);
            return -1;
        }
    }

    (void) jpeg_finish_decompress(&cinfo);
    jpeg_destroy_decompress(&cinfo);

#if 0
 {
    static int pno=0;
    char fname[128];
    char tmp[128];
    int fd;

    sprintf(tmp, "P6\n%d\n%d\n255\n",
        cinfo.output_width, cinfo.output_height);

    sprintf(fname, "/var/tmp/p%d", pno%8);
    fd = open(fname, O_WRONLY|O_CREAT|O_TRUNC, 0777);
    if(fd<0) {
    }
    else {
        chk = write(fd, tmp, strlen(tmp));
        chk = write(fd, out_area, cinfo.output_width*cinfo.output_height*3);
        printf("file %3d: chk %d\n", pno, chk);
        close(fd);
    }

    pno++;
 }
#endif

    return 0;
}

#ifdef TEST_DEJPEG
int debug;

#define IBUFLEN (128*1024)

int
main(int argc, char *argv[])
{
    unsigned char imgbuf[IBUFLEN];
    unsigned char outbuf[IBUFLEN];
    FILE *fp;
    char *fname;
    int nread;
    int wid, ht;
    int chk;

    if(argc<2) {
        fprintf(stderr, "usage: %s <file>\n", argv[0]);
        exit(5);
    }
    
    fname = strdup(argv[1]);
    fp = fopen(fname, "rb");
    if(fp==NULL) {
        fprintf(stderr, "fail fopen '%s' (%d)\n", fname, errno);
        exit(7);
    }
    nread = fread(imgbuf, 1, IBUFLEN, fp);
    if(nread<=0) {
        fprintf(stderr, "error or empty '%s' (%d)\n", fname, errno);
        exit(7);
    }

    chk = decode_jpeg(imgbuf, nread, outbuf, IBUFLEN, &wid, &ht);
    
    if(chk) {
        fprintf(stdout, "fail decode jpeg %s\n", fname);
    }
    else {
        fprintf(stdout, "'%s' %dbytes : %dx%d %dbytes : scale %.2f : %d\n",
            fname, nread, wid, ht, wid*ht*3, (double)wid*ht*3/nread, chk);
    }
    
}
#endif
