/*
 * for check routine
 */

#include <stdio.h>
#include <fcntl.h>

extern int debug;

void
dump_to_file(char *filename, char *image, int size)
{
    int fd;

    if ( (fd = open(filename, (O_WRONLY|O_CREAT|O_TRUNC), 0644)) < 0 ) {
        fprintf(stderr, "dump_to_file: cannot open %s:", filename);
        perror("open");
    }
    write(fd, image, size);
    close(fd);

    if (debug>=100)
        fprintf(stderr, "dump_to_file: imagePtr# %x\n", image);
}

void
dump_rgb_to_file(char *filename, char *image, int wid, int ht)
{
    int fd;
    char head[64];
    int nw;

    if ( (fd = open(filename, (O_WRONLY|O_CREAT|O_TRUNC), 0644)) < 0 ) {
        fprintf(stderr, "dump_rgb_to_file: cannot open %s:", filename);
        perror("open");
    }
    sprintf(head,"P6 %d %d %d\n", wid, ht, 255);
    nw = write(fd, head, strlen(head));
    nw = write(fd, image, wid*ht*3);
    close(fd);

    if (debug>=100)
        fprintf(stderr, "dump_rgb_to_file: imagePtr# %x\n", image);
}

#ifndef Trace
#define Trace   printf
#endif

#ifndef STRING_SIZE
#define STRING_SIZE (BUFSIZ)
#endif

void
dump_stringsegment(char *prefix, unsigned char *str, int len)
{
    int i;

    Trace("%s[", prefix);
    for(i=0;i<len;i++) {
        if(str[i]=='\0')  {
            Trace("<NULL>");
        }
        else
        if(str[i]=='\r')  {
            Trace("<CR>");
        }
        else
        if(str[i]=='\n')  {
            Trace("<LF>");
        }
        else {
            if(str[i]<' ' || str[i]>0x7f) {
                Trace("<%02x>",str[i]);
            }
            else {
                Trace("%c", str[i]);
            }
        }
    }
    Trace("]\n");
}


void dump_string(char *prefix, char *line)
{
    unsigned char *p;

    Trace("%s[", prefix);
    p = (unsigned char*) line;
    if(*p) {
        while(*p&&*p!='\r'&&*p!='\n') {
            if(*p<' ' || *p>0x7f) {
                Trace("<%02x>",*p);
            }
            else {
                Trace("%c",*p);
            }
            p++;
        }
        if(*p=='\0')
            Trace("<NULL>");

        if(*p=='\r') {
            Trace("<CR>");
            p++;
        }
        if(*p=='\n')
            Trace("<LF>");
        Trace("]\n");
    }
    else {
        Trace("<EMPTY>]\n");
    }
}

void dump_stringchunk(char *prefix, char *chunk)
{
    unsigned char *p;
    char *l;
    char label[STRING_SIZE];
    int  c;

    if(strlen(prefix)>STRING_SIZE) {
        strcpy(label, "CHUNK ");
    }
    else {
        memset(label, ' ', strlen(prefix));
        label[strlen(prefix)] = '\0';
    }

    c = 0;

    p = (unsigned char*) chunk;
    if(!*p) {
         Trace("%s %03d[<EMPTY>]\n", prefix, c);
        return;
    }

    l = prefix;
    while(*p) {
        Trace("%s %03d[", l, c);
        c++;

        while(*p&&*p!='\r'&&*p!='\n') {
#if 0
            Trace("%c",*p);
#endif
            if(*p<' ' || *p>0x7f) {
                Trace("<%02x>",*p);
            }
            else {
                Trace("%c",*p);
            }
            p++;
        }
        if(*p=='\0') {
            Trace("<NULL>");
        }

        if(*p=='\r') {
            Trace("<CR>");
            p++;
        }
        if(*p=='\n') {
            Trace("<LF>");
            p++;
        }
        Trace("]\n");

        l = label;
    }
}

void
dump_hexstringsegment(char *prefix, unsigned char *str, int len)
{
    int i;

    for(i=0;i<len;i++) {
        if(i%16==0) {
            Trace("%s ", prefix);
        }
        Trace("%02x ", str[i]);
        if(i%16==15) {
            Trace("\n");
        }
    }
    if(i%16)
        Trace("\n");
}



