/*
 * str_draw.c: string drawing
 *
 * --- Id: vfdraw.c,v 1.4 1997/12/29 14:51:00 k-chinen Exp k-chinen ---
 * $Id$
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xproto.h>
#include <X11/cursorfont.h>

#include <X11/Xmu/WinUtil.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/Scrollbar.h>



#ifdef USE_VFLIB
#include <VF.h>
#include <VFenc.h>
#endif

#include "hook.h"
#include "alist.h"

#if 0 
#include "xpobj.h"
#endif

#if 0
#include "xpfont.h"
#endif

typedef struct {
    double csize;
    char *afont;
    char *kfont;
} xp_font_t;


#if 0
#include "xpoptions.h"
#endif

#include "ds_misc.h"

#if 0 
#include "img.h"
#include "img_file.h"
#endif

#if 0 
#include "dvi.h"
#endif
#include "str_draw.h"



#define LINELEN 2048
#define STRLEN 1024

#define U_NOFONT    (-1)
#define U_XFONT     (0)
#define U_VFONT     (1)
#define U_TTF       (2)




/*
 *
 */
int guess_fonttype(char *name)
{
    int ret;

    switch(*name) {
    case '^':   ret = U_VFONT;  break;
    case '@':   ret = U_TTF;    break;
    case '-':   ret = U_XFONT;  break;
    default:    ret = U_NOFONT; break;
    }

    return ret;
}




int
normarlize_expstr(char *dst, char *src)
{
    char *p, *q;
    char *r;

    p = src;
    q = dst;

    r = NULL;
    while(*p) {
        if(*p==' ') {
            if(r && !(*r=='@' || (*r>='A' && *r<='Z') || (*r>='a' && *r<='z') ) ) {
                p++;    /* skip */
                continue;
            }
#if 0
            if(r && *r==' ') {
                p++;    /* skip */
                continue;
            }
#endif
        }
        r = p;
        *q++ = *p++;
    }
    *q = '\0';

#if 0
    Trace("src '%s'\n", src);
    Trace("dst '%s'\n", dst);
#endif

    return 0;
}

int
chop_doller(char *dst, char *src)
{
    char *p, *q;

    p = src;
    q = dst;
    
    if(*p=='$') {
        p++;
        while(*p && *p!='$') {
            *q++ = *p++;
        }
        *q = '\0';

        return 0;
    }
    else {
        return -1;
    }
}

int size_fit_range=12;


/*
 * See also PutBitmap() of tools/ktest.c in VFlib .
 *
 */

int
_DrawString(Display *dpy, Drawable win, GC gc,
        int sx, int sy, Pixel fg, Pixel bg, int depth,
        char *_afontname, char *_kfontname, int csize,
        unsigned char *line, int exp_expand)
{
    static int _count_ = 0;

    int acsize;

    long code;
    int i;
    int gx, gy;
    unsigned char *p, *body;
    unsigned char *q;
    int clen;
    int aclen;
    int vfd = -1;
    XFontStruct *afinfo = NULL;
    XFontStruct *kfinfo = NULL;
    int bsize, cbytes;
    unsigned char *buf, *cbuf;
    int ctype;
    XWindowAttributes win_attr;
    Status status;
    char afontname[XP_FONTNAME_LEN], kfontname[XP_FONTNAME_LEN];
    char expbuf[BUFSIZ];

    int afont_type=U_NOFONT;
    int kfont_type=U_NOFONT;


    int dist;
    int tlen;

    int x, y, b;
    int px, py;
    register char u;

    Pixmap pix;
    Pixmap mask;
    unsigned char tmpline[LINELEN];

    void* fchk;
    int cc;
    int iwidth, iheight;

    static unsigned char
        tb1[8] = {0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01},
        tb2[8] = {0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80};

    unsigned char d;

    int chk;
    int raise;

    extern int default_dviscale;
    extern int dviscale;

/*
    TracePush;
*/


    dist = 0;
    buf = NULL;
    cbuf = NULL;

    Trace("_DrawString: [%d] dpy %p, win %d, x %d, y %d\n",
        ++_count_, dpy, win, sx, sy);
    if(_trace_flag_) dump_jstring("  string", line);

    if(line==NULL || *line=='\0') {
        return -1;
    }



#if 1
    Trace("  sz %3d\n\
  mask\n\
    _a %p %s\n\
    _k %p %s\n",
        csize,
        _afontname, _afontname, _kfontname, _kfontname);
#endif

  /*
   * ASCII
   */

    if(_afontname[0]=='-') {
        afont_type = U_XFONT;

        sprintf(afontname, _afontname, csize);
        if((afinfo = XLoadQueryFont(dpy, afontname)) == NULL) {
            Error("_DrawString: fail XLoadQueryFont '%s' as ASCII font\n", afontname);

            if((afinfo = XLoadQueryFont(dpy, "a14")) == NULL) {
                Error("_DrawString: fail XLoadQueryFont '%s' as recovery ASCII font\n", "a14");
                goto out;
            }
            Error("_DrawString: use '%s'\n", "a14");
        }
    }
    else
    if(_afontname[0]=='@') {
        afont_type = U_TTF;
        Error("_DrawString: TTF (%s) is not support\n", _afontname);
        goto out;
    }
    else {
        Error("_DrawString: unknow font (%s) for ASCII\n", _afontname);
        goto out;
    }



  /*
   * KANJI
   */
    if(_kfontname[0]=='-') {
        kfont_type = U_XFONT;

        for(acsize=csize; acsize>=csize-size_fit_range; acsize--) {
            sprintf(kfontname, _kfontname, acsize);
            if((kfinfo = XLoadQueryFont(dpy, kfontname)) == NULL) {
                Error("_DrawString: fail XLoadQueryFont '%s' as KANJI font\n", kfontname);
            }
            else {
                break;
            }
        }
        if(kfinfo==NULL) {
            if((kfinfo = XLoadQueryFont(dpy, "k14")) == NULL) {
                Error("_DrawString: fail XLoadQueryFont '%s' as recovery KANJI font\n", "k14");
                goto out;
            }
            Error("_DrawString: use '%s'\n", "k14");
        }
    }
    else
    if(_kfontname[0]=='@') {
        kfont_type = U_TTF;

        strcpy(kfontname, &_kfontname[1]);
        Error("_DrawString: unknow TTfont (%s) for KANJI\n", kfontname);
        goto out;
    }
    else {
        Error("_DrawString: unknow font (%s) for KANJI\n", _kfontname);
        goto out;
    }

#if 1
    Trace("\n\
  actual\n\
    a %p (%d) %4d %s\n\
    k %p (%d) %4d %s\n",
        afontname, afont_type, afinfo ? afinfo->fid : -1, afontname,
        kfontname, kfont_type, kfinfo ? kfinfo->fid : -1, kfontname);
#endif

#if 0
    Trace("  depth %d\n", depth);
#endif


    q = line;
    gx = sx;
    gy = sy;
    while(*q) {
        ctype = get_chunk(q, tmpline, &clen);
        aclen = clen;

        if(!exp_expand && ctype==EXP_CHUNK) {
            ctype = ASCII_CHUNK;
        }

#if 1
        Trace("  :type %d, '%s' %d\n", ctype, tmpline, clen);
#endif
#if 0
        Trace("  len %d, value '%s'\n", aclen, tmpline);
#endif

        switch(ctype) {
        case KANJI_CHUNK:
          if(kfont_type == U_XFONT) {
            Trace("++xfont for KANJI\n");
            Trace("  <%s>\n", tmpline);

            XSetFont(dpy, gc, kfinfo->fid);
            XSetForeground(dpy, gc, fg);
            /*
            XSetBackground(dpy, gc, bg);
            */

#if 1
            /* convert EUC to JIS */
            {
                char *u;
                u =  tmpline;
                while(*u) {
                    *u &= 0x7f;
                    u++;
                }
            }
            chk = XDrawString16(dpy, win, gc, gx, gy+afinfo->ascent,
                (XChar2b*)tmpline, aclen/2);
            if(chk) printf("_DrawString: KANJI DrawString16 chk %d\n", chk);
#endif

            tlen = XTextWidth16(kfinfo, (XChar2b*)tmpline, aclen/2);
            dist += tlen;
            gx += tlen;

          }
          else {
            Trace("unsupport kanji font type %d\n", kfont_type);
          }
            break;

        case ASCII_CHUNK:
          if(afont_type == U_XFONT) {
            Trace("++xfont for ASCII\n");
#if 0
            Trace("  <%s>\n", tmpline);
#endif
            if(_trace_flag_) dump_jstring("  chunk", tmpline);

            XSetFont(dpy, gc, afinfo->fid);
            XSetForeground(dpy, gc, fg);
            /*
            XSetBackground(dpy, gc, bg);
            */

            tlen = XTextWidth(afinfo, tmpline, aclen);

            XDrawString(dpy, win, gc,
                gx, gy+afinfo->ascent, tmpline, aclen);

            dist += tlen;
            gx += tlen;

          }
          else {
            Trace("unsupport ascii font type %d\n", afont_type);
          }

            break;


#if 0
        case EXP_CHUNK:
            if(_trace_flag_) dump_jstring("  chunk", tmpline);

            chk = normarlize_expstr(expbuf, tmpline);
            Trace("  expbuf '%s'\n", expbuf);

            chop_doller(tmpline, expbuf);
            Trace("  tmpline '%s'\n", tmpline);

            cc = 0;
            do {
                iwidth = csize/2;
                iheight = csize;
Trace("  b iwidth %d, iheight %d\n", iwidth, iheight);

                fchk = XPImageFind(image_cache, expbuf, -1, -1);
                if(fchk==NULL) {
                    Trace("  exp-'%s' not found\n", expbuf);

                    chk = compile_platex(tmpline, "#EXPNONUM");

                    fchk = (void*)read_dvi_file(dpy, win,
                        gx, gy, gc, NULL, -1, -1,
                        "#EXPNONUM.dvi", &iwidth, &iheight, -1, 1, 0);
                    if(fchk) {
Trace("  a iwidth %d, iheight %d\n", iwidth, iheight);
                        chk = XPImageStore(image_cache, expbuf, -1, -1, fchk);
                        break;
                    }
                    else {
                        Error("fail read dvi file\n");
                    }
                }

                cc++;
            } while (fchk==NULL && cc<3);


            XSetForeground(dpy, gc, fg);

#if 0
            iwidth = -1;
            iheight = afinfo->ascent;
#endif
            dviscale = 600000/csize;
Trace(" csize %d -> dviscale %d\n", csize, dviscale);

            CalcDVIsize((xp_dvi_t*)fchk, -1, &iwidth, &iheight);
Trace("  u iwidth %d, iheight %d\n", iwidth, iheight);
            raise = iheight;

            iwidth = -1;
            iheight = -1;

            DrawDVI(dpy, win, gc, NULL, -1, -1,
                        gx, gy+afinfo->ascent-raise, &iwidth, &iheight,
                        (xp_dvi_t*)fchk , -1, -1, XP_OPT_FRAME & 0);
Trace("  e iwidth %d, iheight %d\n", iwidth, iheight);

            tlen = iwidth;

            dist += tlen;
            gx += tlen;

            break;

#endif

        default:
            break;
        }

#if 1
        Trace("    length charactor/dot %d/%d : %d\n", aclen, tlen, dist);
#endif

        q += clen;
    }


out:

#if 0
    if(vfd != -1) {
        VF_CloseFont(vfd);
    }
#endif

    if(afinfo != NULL) {
        XFreeFontInfo(NULL, afinfo, 1);
    }
    if(kfinfo != NULL) {
        XFreeFontInfo(NULL, kfinfo, 1);
    }

#if 0
    Trace(" dist %d \n", dist);
#endif

/*
    TracePop;
*/

    return dist;
}



int
_GetStringLength(Display *dpy,
    char *_afontname, char *_kfontname, int csize, unsigned char *line,
    int exp_expand)
{
    static int _count_ = 0;

    unsigned long code;
    int i;
    unsigned char *p, *body;
    unsigned char *q;
    int clen;
    int vfd = -1;
    XFontStruct *afinfo = NULL, *kfinfo = NULL;
    int bsize, cbytes;
    unsigned char *buf, *cbuf;
    int ctype;
    char afontname[XP_FONTNAME_LEN];
    char kfontname[XP_FONTNAME_LEN];

    int chk;

    int dist;
    int tlen;

    int x, y, b;
    int px, py;
    register char u;

    unsigned char tmpline[LINELEN];
    unsigned char d;

    int afont_type=U_NOFONT;
    int kfont_type=U_NOFONT;

    void* fchk;
    int iwidth, iheight;
    char expbuf[BUFSIZ];

    extern int default_dviscale;
    extern int dviscale;


/*
TracePush;
*/

    dist = 0;

#if 0
    Trace("_GetStringLength: [%d] dpy %p, win %d, x %d, y %d\n",
        ++_count_, dpy, win, sx, sy);
#endif
#if 0
    Trace("_GetStringLength: exp_expand %d\n", exp_expand);
    dump_jstring("_GetStringLength/enter  ", line);
#endif

    if(line==NULL || *line=='\0') {
        return -1;
    }


    afont_type = guess_fonttype(_afontname);
    kfont_type = guess_fonttype(_kfontname);


    sprintf(afontname, _afontname, csize);
    sprintf(kfontname, _kfontname, csize);
    Trace(" _GetStringLength: sz %3d\n\
      mask   _a %s\n\
      actual  a %s\n\
      actual  k %s\n", csize, _afontname, afontname, kfontname);

    if((afinfo = XLoadQueryFont(dpy, afontname)) == NULL) {
        Trace("_GetStringLength: fail XLoadQueryFont '%s'\n", afontname);
        goto out;
    }
    if(kfont_type==U_XFONT) {
        if((kfinfo = XLoadQueryFont(dpy, kfontname)) == NULL) {
            Trace("_GetStringLength: fail XLoadQueryFont '%s'\n", kfontname);
            goto out;
        }
    }

    q = line;
    while(*q) {

        ctype = get_chunk(q, tmpline, &clen);
        if(!exp_expand && ctype==EXP_CHUNK) {
            ctype = ASCII_CHUNK;
        }

#if 0
        Trace("  len %d, value '%s'\n", clen, tmpline);
        if(_trace_flag_) dump_jstring("  chunk", tmpline);
        Trace("  type %s(%d)\n", rassoc(chunktype_alist, ctype), ctype);
#endif

        Trace("  %s ", rassoc(chunktype_alist, ctype));
        if(_trace_flag_) dump_jstring(" ", tmpline);

        switch(ctype) {
        case KANJI_CHUNK:
          if(kfont_type==U_XFONT) {
            p = tmpline;
            tlen = 0;

            /* convert EUC to JIS */
            {
                char *u;
                u =  tmpline;
                while(*u) {
                    *u &= 0x7f;
                    u++;
                }
            }
            tlen = XTextWidth16(kfinfo, (XChar2b*)tmpline, clen/2);

            dist += tlen;

          }
          else
          if(kfont_type==U_VFONT) {
            p = tmpline;
            tlen = 0;

            while(*p) {
                if(*p>0x7f) {
                    code = (*p & 0x7f) * 256 + (*(p+1) & 0x7f);

                    p++;
                }
                else {
                    code = *p;
                }

                p++;

                tlen += csize;
            }

            dist += tlen;

          }
          else {
          }
            break;

#if 0
        case EXP_CHUNK:

            chk = normarlize_expstr(expbuf, tmpline);
            chop_doller(tmpline, expbuf);
#if 0
            Trace("  expbuf '%s'\n", expbuf);
            Trace("  tmpline '%s'\n", tmpline);
#endif

            fchk = XPImageFind(image_cache, expbuf, -1, -1);
            if(fchk==NULL) {
#if 1
                Trace("unknow size exp-'%s'\n", expbuf);
                break;
#else
                chk = compile_platex(tmpline, "#EXPNONUM");

                fchk = (void*)read_dvi_file(dpy, -1,
                    -1, -1, -1, NULL, -1, -1,
                    "#EXPNONUM.dvi", &iwidth, &iheight, -1, 1, 0);
                if(fchk) {
Trace("  a iwidth %d, iheight %d\n", iwidth, iheight);
                    chk = XPImageStore(image_cache, expbuf, -1, -1, fchk);
                }
                else {
                    Error("fail read dvi file\n");
                    break;
                }
#endif
            }

            dviscale = 600000/csize;
            CalcDVIsize((xp_dvi_t*)fchk, -1, &iwidth, &iheight);

            
            tlen = iwidth;
            dist += tlen;


            break;

#else
        case EXP_CHUNK:
#endif
        case ASCII_CHUNK:

            tlen = XTextWidth(afinfo, tmpline, clen);
            dist += tlen;

            break;


        default:
            break;
        }

#if 0
    Trace("    bit length %d\n", tlen);
#endif

        Trace("   tlen %d, dist %d\n", tlen, dist);
        q += clen;
    }


out:
    if(afinfo != NULL) {
        XFreeFontInfo(NULL, afinfo, 1);
    }
    if(kfinfo != NULL) {
        XFreeFontInfo(NULL, kfinfo, 1);
    }

    Trace(" dist %d \n", dist);

/*
TracePop;
*/

    return dist;
}


int
_GetStringNchar(Display *dpy,
    char *_afontname, char *kfontname, int csize, unsigned char *line,
    int exp_expand, int qx)
{
    static int _count_ = 0;

    long code;
    int i;
    int gx, gy;
    unsigned char *p, *body;
    unsigned char *q;
    int clen;
    int vfd = -1;
    XFontStruct *afinfo = NULL;
    int bsize, cbytes;
    unsigned char *buf, *cbuf;
    int ctype;
    char afontname[XP_FONTNAME_LEN];


    int dist;
    int tlen;

    int ret;

    int x, y, b;
    int px, py;
    register char u;

    unsigned char tmpline[LINELEN];
    unsigned char d;



    ret = -1;
    dist = 0;

#if 1
    Trace("_GetStringNchar: csize %d, qx %d. line '%s'\n",
        csize, qx, line);
#endif

    if(line==NULL || *line=='\0') {
#if 0
        return -1;
#endif
        return 0;
    }

    sprintf(afontname, _afontname, csize);
    Trace(" _GetStringNchar: sz %3d\n\
      mask   _a %s\n\
      actual  a %s\n\
      actual  k %s\n", csize, _afontname, afontname, kfontname);



    if((afinfo = XLoadQueryFont(dpy, afontname)) == NULL) {
        Trace("_GetStringNchar: fail XLoadQueryFont '%s'\n", afontname);
        goto out;
    }


    q = line;
    while(*q) {
        ctype = get_chunk(q, tmpline, &clen);
#if 0
        Trace("  len %d, value '%s'\n", clen, tmpline);
#endif

        switch(ctype) {
        case KANJI_CHUNK:

            p = tmpline;
            tlen = 0;

            while(*p) {
                if(*p>0x7f) {
                    code = (*p & 0x7f) * 256 + (*(p+1) & 0x7f);

                    p++;
                }
                else {
                    code = *p;
                }

                p++;

                tlen += csize;
#if 1
    Trace("    *dist %d tlen %d -> %d : %d\n", dist, tlen, dist+tlen, qx);
#endif

#if 1
                if(dist+tlen>qx) {
                    ret = (q-line)+(p-tmpline)-2;
                    goto out;
                }
#endif
#if 0
                if(dist+tlen==qx) {
                    ret = (q-line)+(p-tmpline)-1;
                    goto out;
                }
#endif
            }
            if(dist+tlen==qx) {
                ret = (q-line)+(p-tmpline);
                goto out;
            }

            dist += tlen;

            break;

        case ASCII_CHUNK:
        case EXP_CHUNK:

            tlen = XTextWidth(afinfo, tmpline, clen);
#if 1
    Trace("    -dist %d tlen %d -> %d : %d\n", dist, tlen, dist+tlen, qx);
#endif
            if(dist+tlen>qx) {
                int ww;

                for(ww=1;ww<=clen;ww++) {
                    tlen = XTextWidth(afinfo, tmpline, ww);
#if 1
    Trace("    --dist %d tlen %d -> %d : %d\n", dist, tlen, dist+tlen, qx);
#endif
                    if(dist+tlen>qx) {
                        ret = (q-line)+ww-1;
                        goto out;
                    }
                }
            }
            else
            if(dist+tlen==qx) {
                ret = (q-line)+clen;
                goto out;
            }

            dist += tlen;
            gx += tlen;

            break;

        default:
            break;
        }

#if 1
    Trace("    bit length %d\n", tlen);
#endif


        q += clen;
    }


out:

#if 1
    if(afinfo != NULL) {
    }
#endif


#if 1
    Trace(" ret %d \n", ret);
#endif

    return ret;
}

int
_GetStringXpos(Display *dpy,
    char *_afontname, char *kfontname, int csize, unsigned char *line,
    int exp_expand, int qx)
{
    static int _count_ = 0;

    unsigned long code;
    int i;
    int gx, gy;
    unsigned char *p, *body;
    unsigned char *q;
    int clen;
    int vfd = -1;
    XFontStruct *afinfo = NULL;
    int bsize, cbytes;
    unsigned char *buf, *cbuf;
    int ctype;
    char afontname[XP_FONTNAME_LEN];

    int dist;
    int tlen;

    int ret;

    int x, y, b;
    int px, py;
    int lx;
    register char u;

    unsigned char tmpline[LINELEN];
    unsigned char d;



    ret = -1;
    dist = 0;

#if 1
    Trace("_GetStringXpos: csize %d, qx %d. line '%s'\n",
        csize, qx, line);
#endif

    if(line==NULL || *line=='\0') {
        return -1;
    }

    sprintf(afontname, _afontname, csize);
    Trace(" _GetStringXpos: sz %3d\n\
      mask   _a %s\n\
      actual  a %s\n\
      actual  k %s\n", csize, _afontname, afontname, kfontname);


    if((afinfo = XLoadQueryFont(dpy, afontname)) == NULL) {
        Trace("_GetStringXpos: fail XLoadQueryFont '%s'\n", afontname);
        goto out;
    }


    q = line;
    while(*q) {
        ctype = get_chunk(q, tmpline, &clen);

#if 1
        Trace("  len %d, value '%s'\n", clen, tmpline);
#endif

        switch(ctype) {
        case KANJI_CHUNK:

            p = tmpline;
            tlen = 0;

            while(*p) {
                if(*p>0x7f) {
                    code = (*p & 0x7f) * 256 + (*(p+1) & 0x7f);

                    p++;
                }
                else {
                    code = *p;
                }

                p++;

                tlen += csize;
#if 1
    Trace("    *dist %d tlen %d -> %d : %d\n", dist, tlen, dist+tlen, qx);
#endif

                if(dist+tlen>qx) {
                    ret = dist+tlen-csize;
                    goto out;
                }
            }

            dist += tlen;

            break;

        case ASCII_CHUNK:
        case EXP_CHUNK:

            tlen = XTextWidth(afinfo, tmpline, clen);
#if 1
    Trace("    -dist  %3d tlen %3d -> %3d : %3d\n", dist, tlen, dist+tlen, qx);
#endif
            if(dist+tlen>qx) {
                int ww;

                lx = dist;
                for(ww=1;ww<=clen;ww++) {
                    tlen = XTextWidth(afinfo, tmpline, ww);
#if 1
    Trace("    --dist %3d tlen %3d -> %3d : %3d\n", dist, tlen, dist+tlen, qx);
#endif
                    if(dist+tlen>qx) {
                        ret = lx;
                        goto out;
                    }
                    lx = dist+tlen;
                }
            }
            else
            if(dist+tlen==qx) {
#if 1
    Trace("    -*dist %3d tlen %3d -> %3d : %3d\n", dist, tlen, dist+tlen, qx);
#endif
                ret = dist+tlen;
                goto out;
            }

            dist += tlen;
            gx += tlen;

            break;

        default:
            break;
        }

#if 1
    Trace("    bit length %d\n", tlen);
#endif


        q += clen;
    }


out:

#if 1
    if(afinfo != NULL) {
        XFreeFontInfo(NULL, afinfo, 1);
    }
#endif


#if 1
    Trace(" ret %d \n", ret);
#endif

    return ret;
}



/***
 ***
 *** Complex functions
 ***
 ***/


/*
 * DrawSimpleFoldString - Wrapper for DrawString.
 *
 * Because DrawString not support folding.
 * This routine check length and fold if necessary.
 *
 * You have to check return value rx and ry.
 * They mean drawing area size by this routines.
 */
int
DrawSimpleFoldString(Display *dpy, Window win, GC xgc,
    int sx, int sy, Pixel fg, Pixel bg, int depth,
    int wx, int wy, int *rx, int *ry,
    xp_font_t *ft, char *origline,
    double iratio, double lratio, int opt)
{
    int c;
    int w, cw;
    char pool[STRLEN];
    char tmp[STRLEN];
    char *cur, *rest;
    char *p;
    int  pos;
    int  indent, yskip;
    int  px, py;

    /*
    TracePush;
    */

    Trace("DrawSimpleFoldString: start (%d,%d) width (%d,%d)\n",
            sx, sy, wx, wy);
    Trace(" orig '%s'\n", origline);

    strcpy(pool, origline);
    cur = pool;


    indent = (int)(ft->csize * iratio);
    yskip  = (int)(ft->csize * lratio);

#if 0
    TracePush;
    Trace("csize %5.2f, lratio %5.3f -> yskip %4d\n",
        ft->csize, lratio, yskip);
    TracePop;
#endif

    *rx = 0;
    *ry = 0;
    w = wx - sx;
    c = 0;

    px = sx;
    py = sy;

    while(1) {
        if(c==0) {
            px = sx;
            w = wx;
        }
        else {
            px = sx + indent;
            w = wx - indent;
        }
        cw = GetStringLength(dpy, ft, cur, 1);

    Trace(" %2d: %4d/%4d '%s'\n", c, cw, w, cur);

        if(cw<0) {
            break;
        }

        if(cw<w) {
            Trace(" short string\n");
            rest = NULL;
        }
        else {
            Trace(" long string\n");

            pos = GetStringNchar(dpy, ft, cur, 1, w);
            if(pos>0) {
                rest = &cur[pos];
                strncpy(tmp, cur, pos);
                cur = tmp;
                cur[pos] = '\0';
            }
            else {
                rest = NULL;
            }

            Trace(" splited in %dth char as '%s' and '%s'\n",
                pos, cur, rest);

            cw = GetStringLength(dpy, ft, cur, 1);
        }

draw:
#if 0
        if(c>0) {
            /* plot mark for foliding */
            DrawString(dpy, win, xgc, sx, py, bg, fg, depth, ft, ">");
        }
#endif

        DrawString(dpy, win, xgc, px, py, fg, bg, depth, ft, cur, 1);

        if(cw>*rx) {
            *rx = cw;
        }
#if 0
        TracePush;
        Trace("py %4d, csize %5.2f, yskip %4d -> %4d\n",
            py, ft->csize, yskip, py + (int)ft->csize + yskip);
        TracePop;
#endif
        *ry += (int)ft->csize;
        py += (int)ft->csize;


        if(rest==NULL) {
            break;
        }

        *ry += yskip;
        py += yskip;

        cur = rest;
        c++;
    }

out:
    Trace("  result (%d,%d)\n", *rx, *ry);
    /*
    TracePop;
    */

    return 0;
}


/*
 *
 */
static
int
split_word(char *line, char *word, int *len)
{
    unsigned char *p, *q;
    int i, c;

    p = (unsigned char*)line;
    q = (unsigned char*)word;
    c = 0;
    if(p==NULL) {
        return -1;
    }

    if(*p=='\0') {
        return -1;
    }

    /* delimitor */
    if(*p==' '|| *p=='\t') {
        *q++ = *p;
        *q = '\0';
        *len = 1;
        return 0;
    }

    /* single byte charactors */
    if(*p>0x20&&*p<0x7f) {
        do {
            *q++ = *p++;
            c++;
        } while(*p && *p>0x20 && *p<0x7f);
        *q = '\0';
        *len = c;
    }
    /* double byte charactors */
    else
    if(*p>=0x80) {
        *q++ = *p++;
        *q++ = *p++;
        *q = '\0';
        *len = 2;
    }
    else {
        *q++ = *p;
        *q = '\0';
        *len = 1;
    }

#if 0
Trace("split_word: '%s' %d in '%s'\n", word, *len, line);
#endif

    return 0;
}


/*
 * DrawFancyFoldString - Wrapper for DrawString.
 *
 * Because DrawString not support folding.
 * This routine check length with word splitting and fold if necessary.
 *
 * You have to check return value rx and ry.
 * They mean drawing area size by this routines.
 */
int
DrawFancyFoldString(Display *dpy, Window win, GC xgc,
    int sx, int sy, Pixel fg, Pixel bg, int depth,
    int wx, int wy, int *rx, int *ry,
    xp_font_t *ft, char *origline,
    double iratio, double lratio, int opt)
{
    int c;
    int w, cw;
    unsigned char pool[STRLEN];
    unsigned char tmp[STRLEN];
    unsigned char *cur, *rest;
    unsigned char *p;
    int  pos;
    int  indent, yskip;
    int  px, py;

    /*
    TracePush;
    */

    Trace("DrawFancyFoldString: start (%d,%d) width (%d,%d)\n",
            sx, sy, wx, wy);
    Trace(" orig '%s'\n", origline);

    strcpy(pool, origline);
    cur = pool;


    indent = (int)(ft->csize * iratio);
    yskip  = (int)(ft->csize * lratio);

#if 0
    TracePush;
    Trace("csize %5.2f, lratio %5.3f -> yskip %4d\n",
        ft->csize, lratio, yskip);
    TracePop;
#endif

    *rx = 0;
    *ry = 0;
    w = wx - sx;
    c = 0;

    px = sx;
    py = sy;

    while(1) {
        if(c==0) {
            px = sx;
            w = wx;
        }
        else {
            px = sx + indent;
            w = wx - indent;
        }

        if(cur==NULL || cur[0]=='\0') {
            break;
        }
        else
        if(cur[0]==' ' || cur[0]=='\t') {
            cur++;
        }
        else
        if(cur[0]==0xa1 && cur[1]==0xa1) {
            cur += 2;
        }

        cw = GetStringLength(dpy, ft, cur, 1);

    Trace(" %2d: %4d/%4d '%s'\n", c, cw, w, cur);

        if(cw<0) {
            break;
        }

        if(cw<w) {
            Trace(" short string\n");
            rest = NULL;
        }
        else {
            Trace(" long string\n");

            {
                int hx, hy;
                int hw;
                unsigned char *hstr;
                int ctype;
                int clen;

                hx = 0;
                hstr = cur;
                while(*hstr) {
                    ctype = split_word(hstr, tmp, &clen);

                    /* blank at top of line is skipped */
                    if(hx==0 && (tmp[0]==' ' || tmp[0]=='\t'
                        || (tmp[0]==0xa1 && tmp[1]==0xa1) )) {
                        goto next_chunk;
                    }
                    hw = GetStringLength(dpy, ft, tmp, 1);

#if 0
TracePush;
Trace("type %d [%s] %4d\n", ctype, tmp, clen);
Trace("hx %4d, hw %4d -> current %4d limit %d\n", hx, hw, hx+hw, w);
TracePop;
#endif

                    if(hx+hw>=w) {
#if 0
TracePush;
Trace("\t***over run. break\n");
TracePop;
#endif
                        break;
                    }
next_chunk:
                    hx += hw;
                    hstr += clen;
                }

                /* string length is shorter than draw area width */
                if(*hstr=='\0') {
                    rest = NULL;
                }
                else {
                    pos = hstr - cur;
                    strncpy(tmp, cur, pos);
                    cur = tmp;
                    cur[pos] = '\0';
                    rest = hstr;
                }
            }


            Trace(" splited in %dth char as '%s' and '%s'\n",
                pos, cur, rest);

            cw = GetStringLength(dpy, ft, cur, 1);
        }

draw:
#if 0
        if(c>0) {
            /* plot mark for foliding */
            DrawString(dpy, win, xgc, sx, py, bg, fg, depth, ft, ">");
        }
#endif

        DrawString(dpy, win, xgc, px, py, fg, bg, depth, ft, cur, 1);

        if(cw>*rx) {
            *rx = cw;
        }
#if 0
        TracePush;
        Trace("py %4d, csize %5.2f, yskip %4d -> %4d\n",
            py, ft->csize, yskip, py + (int)ft->csize + yskip);
        TracePop;
#endif
        *ry += (int)ft->csize;
        py += (int)ft->csize;


        if(rest==NULL) {
            break;
        }

        *ry += yskip;
        py += yskip;

        cur = rest;
        c++;
    }


out:
    Trace("  result (%d,%d)\n", *rx, *ry);
    /*
    TracePop;
    */

    return 0;
}

