/*
 * $Id$
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include "alist.h"
#include "ds_misc.h"

#ifndef STRLEN
#define STRLEN 1024
#endif

#define Trace printf


apair_t chunktype_alist[] = {
    {"ASCII",   ASCII_CHUNK},
    {"KANJI",   KANJI_CHUNK},
    {"EXP",     EXP_CHUNK},
    {NULL,      -1},
};


#if 0
void dump_string(char *prefix, char *line)
{
    unsigned char *p;

    Trace("%s[", prefix);
    p = (unsigned char*) line;
    if(*p) {
        while(*p&&*p!='\r'&&*p!='\n') {
            if(*p<' ' || *p>=0x7f) {
                Trace("<%02x>",*p);
            }
            else {
                Trace("%c",*p);
            }
            p++;
        }
        if(*p=='\0') {
            Trace("<NULL>");
        }
        if(*p=='\r') {
            Trace("<CR>");
            p++;
        }
        if(*p=='\n')
            Trace("<LF>");
        Trace("]\n");
    }
    else {
        Trace("<EMPTY>]\n");
    }
}
#endif


void dump_jstring(char *prefix, char *line)
{
    unsigned char *p;

    Trace("%s[", prefix);
    p = (unsigned char*) line;
    if(*p) {
        while(*p&&*p!='\r'&&*p!='\n') {
            if(*p<' ') {
                Trace("<%02x>",*p);
            }
            else
            if(*p>0x7f) {
                Trace("%c%c",*p,*(p+1));
                p++;
            }
            else
            if(*p=='\x7f') {
                Trace("<DEL>");
            }
            else {
                Trace("%c",*p);
            }
            p++;
        }
        if(*p=='\0')
            Trace("<NULL>");

        if(*p=='\r') {
            Trace("<CR>");
            p++;
        }
        if(*p=='\n')
            Trace("<LF>");
        Trace("]\n");
    }
    else {
        Trace("<EMPTY>]\n");
    }
}


/*
 *  return
 *      -1  error
 *       0  control sequence
 *       1  ascii
 *       2  japanese EUC
 *       3  expression for TeX
 *     else undef
 *
 */

int
get_chunk(unsigned char *src, unsigned char *dest, int *len)
{
    register unsigned char *p, *q;
    int ret = -1;

    p = src;
    q = dest;

    if(*p=='\0') {
        return -1;
    }

    if(*p>=0x20 && *p<0x7f) {
        if(*p=='$')
            goto seek_exp_end;
        goto seek_ascii_end;
    }
    if(*p & 0x80) {
        goto seek_kanji_end;
    }
    if(*p==0x1b) {
        p++;
        if(*p=='$') {
            p++;
            if(*p==')') {
                p++;
                if(*p=='B') {
                    p++;

                    goto seek_kanji_end;
                }
            }
        }
    }

    *len = 1;
    ret = 0;

    *q++ = *p++;
    *q = '\0';

    goto out;


seek_ascii_end:
    ret = ASCII_CHUNK;

    while(*p && *p>=0x20 && *p<0x7f) {
        if(*p=='$') {
            break;
        }
#if 0
        if(*p=='\\') {
            p++;
        }
#endif
        *q++ = *p++;
    }
    *q = '\0';

    *len = p - src;

    goto out;

seek_exp_end:
    ret = EXP_CHUNK;

    *q++ = *p++;

    while(*p && *p>=0x20 && *p<0x7f) {
        if(*p=='$') {
            *q++ = *p++;
            break;
        }
        *q++ = *p++;
    }
    *q = '\0';

    *len = p - src;

    goto out;


seek_kanji_end:
    ret = KANJI_CHUNK;

    while(*p && (*p & 0x80)) {
        *q++ = *p++;
    }
    *q = '\0';

    *len = p - src;

    goto out;

out:

    return ret;


}






