/*
 * $Id$
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>

#include <errno.h>

#include "hook.h"

#include "alist.h"
#include "fig.h"



apair_t figtype_alist[] = {
    {"COLOR",       FIG_COLOR},
    {"ELLIPSE",     FIG_ELLIPSE},
    {"POLYLINE",    FIG_POLYLINE},
    {"SPLINE",      FIG_SPLINE},
    {"TEXT",        FIG_TEXT},
    {"ARC",         FIG_ARC},
    {"COMPOUND",    FIG_COMPOUND},
    {NULL, -1},
};

char *_fig_obj_type_str[] = {
    "Color", "Ellipse", "PolyLine", "Spline", "Text", "Arc", "Compound",
    "Undef", ""
};

/*
 *
 */
char *Chop_CRLF(char *line)
{
    char *p;

    if(line==NULL) {
        return NULL;
    }

    p = line;
    while(*p) {
        if(*p=='\r' && *(p+1)=='\n' && *(p+2)=='\0') {
            *p = '\0';
            break;
        }
        if(*p=='\n' && *(p+1)=='\0') {
            *p = '\0';
            break;
        }
        p++;
    }

    return line;
}


char*
read_fig_line(FILE *fp, char *buf, int len)
{
    char *chk;

try:
    chk = fgets(buf, len, fp);
    if(chk) {
        if(buf[0]=='#') {
            if(buf[1]=='F' && buf[2]=='I' && buf[3]=='G' &&
                (buf[4]==' ' || buf[4]=='\t') ) {
                goto out;
            }
            goto try;
        }
        if(buf[0]==' ' || buf[0]=='\t') {
            /* skip strage line, it would be rest of data line
                so, when data enter this block, it is error. */
            goto try;
        }
    }

out:
    return chk;
}

char*
read_fig_linerest(FILE *fp, char *buf, int len)
{
    char *chk;

try:
    chk = fgets(buf, len, fp);
    if(chk) {
        if(buf[0]=='#') {
            goto try;
        }
    }

out:
    return chk;
}


void
show_fig_object(void *target0)
{
    int dtype;
    fig_color_t *t_color;
    fig_ellipse_t *t_ell;
    fig_line_t *t_line;
    fig_spline_t *t_spline;
    fig_text_t *t_text;
    int i;

    if(target0==NULL) {
        return;
    }

    dtype=FIG_OBJTYPE(target0);

    switch(dtype) {
    case FIG_COLOR:
        t_color = (fig_color_t*) target0;
        fprintf(stdout, "  COLOR\n");
        fprintf(stdout, "    type   = %d\n", t_color->objtype);
        fprintf(stdout, "    c#     = %d\n", t_color->cnum);
        fprintf(stdout, "    value  = %08x\n", t_color->value);
        break;

    case FIG_ARC:
        fprintf(stdout, "ARC\n");
        break;

    case FIG_ELLIPSE:
        t_ell = (fig_ellipse_t*) target0;

        fprintf(stdout, "  ELLIPSE\n");
        fprintf(stdout, "    type   = %d\n", t_ell->objtype);
        fprintf(stdout, "    subtype= %d\n", t_ell->subtype);
        fprintf(stdout, "    depth  = %d\n", t_ell->depth);
        fprintf(stdout, "    pen c  = %d\n", t_ell->pen_color);
        fprintf(stdout, "    fill c = %d\n", t_ell->fill_color);
        fprintf(stdout, "    fill a = %d\n", t_ell->area_fill);
        fprintf(stdout, "    center = (%d,%d)\n", t_ell->center.x, t_ell->center.y);
        fprintf(stdout, "    radius = (%d,%d)\n", t_ell->radiuses.x, t_ell->radiuses.y);
        fprintf(stdout, "    start  = (%d,%d)\n", t_ell->start.x, t_ell->start.y);
        fprintf(stdout, "    end    = (%d,%d)\n", t_ell->end.x, t_ell->end.y);

        break;

    case FIG_POLYLINE:
        t_line = (fig_line_t*) target0;

        fprintf(stdout, "  POLYLINE\n");
        fprintf(stdout, "    type   = %d\n", t_line->objtype);
        fprintf(stdout, "    subtype= %d\n", t_line->subtype);
        fprintf(stdout, "    depth  = %d\n", t_line->depth);
        fprintf(stdout, "    radius = %d\n", t_line->radius);
        fprintf(stdout, "    pen c  = %d\n", t_line->pen_color);
        fprintf(stdout, "    fill c = %d\n", t_line->fill_color);
        fprintf(stdout, "    fill a = %d\n", t_line->area_fill);

        fprintf(stdout, "    #of points = %d\n", t_line->npoints);

        if(t_line->for_arrow) {
            fprintf(stdout, "    farrow = <%d %d %f %f %f>\n",
                t_line->for_arrow->type, t_line->for_arrow->style,
                t_line->for_arrow->thickness , t_line->for_arrow->width,
                t_line->for_arrow->height);
        }
        if(t_line->back_arrow) {
            fprintf(stdout, "    barrow = <%d %d %f %f %f>\n",
                t_line->back_arrow->type, t_line->back_arrow->style,
                t_line->back_arrow->thickness , t_line->back_arrow->width,
                t_line->back_arrow->height);
        }

        for(i=0;i<t_line->npoints;i++) {
            fprintf(stdout, "    point[%d] = %d,%d\n",
                    i, t_line->points[i].x, t_line->points[i].y);
        }

        break;

    case FIG_SPLINE:
        t_spline = (fig_spline_t*) target0;

        fprintf(stdout, "  SPLINE\n");
        fprintf(stdout, "    type   = %d\n", t_spline->objtype);
        fprintf(stdout, "    subtype= %d\n", t_spline->subtype);
        fprintf(stdout, "    depth  = %d\n", t_spline->depth);
        fprintf(stdout, "    radius = %d\n", t_spline->radius);

        fprintf(stdout, "    pen c  = %d\n", t_spline->pen_color);
        fprintf(stdout, "    fill c = %d\n", t_spline->fill_color);
        fprintf(stdout, "    fill a = %d\n", t_spline->area_fill);

        fprintf(stdout, "    #of points = %d\n", t_spline->npoints);

        if(t_spline->for_arrow) {
            fprintf(stdout, "    farrow = <%d %d %f %f %f>\n",
                t_spline->for_arrow->type, t_spline->for_arrow->style,
                t_spline->for_arrow->thickness , t_spline->for_arrow->width,
                t_spline->for_arrow->height);
        }
        if(t_spline->back_arrow) {
            fprintf(stdout, "    barrow = <%d %d %f %f %f>\n",
                t_spline->back_arrow->type, t_spline->back_arrow->style,
                t_spline->back_arrow->thickness , t_spline->back_arrow->width,
                t_spline->back_arrow->height);
        }

        for(i=0;i<t_spline->npoints;i++) {
            fprintf(stdout, "    point[%d] = %d,%d\n",
                    i, t_spline->points[i].x, t_spline->points[i].y);
        }

        break;

    case FIG_TEXT:
        t_text = (fig_text_t*)target0;

        fprintf(stdout, "  TEXT\n");
        fprintf(stdout, "    type   = %d\n", t_text->objtype);
        fprintf(stdout, "    subtype= %d (%s)\n", t_text->subtype,
            t_text->subtype==0 ? "Left" : (t_text->subtype==1 ? "Center" : "Right"));
        fprintf(stdout, "    depth  = %d\n", t_text->depth);
        fprintf(stdout, "    color  = %d\n", t_text->color);
        fprintf(stdout, "    font   = %d\n", t_text->font);
        fprintf(stdout, "    fflags = %d %x\n",
            t_text->font_flags, t_text->font_flags);
        fprintf(stdout, "    angle  = %f\n", t_text->angle);
        fprintf(stdout, "    fsize  = %f\n", t_text->font_size);
        fprintf(stdout, "    height = %f\n", t_text->height);
        fprintf(stdout, "    length = %f\n", t_text->length);
        fprintf(stdout, "    pos    = %d,%d\n", t_text->pos.x, t_text->pos.y);
        fprintf(stdout, "    value  = '%s'\n", t_text->value);

        break;


    case FIG_COMPOUND:
    case FIG_UNKNOW:
    default:
        fprintf(stdout, "unsupport object type\n");
    }

}

show_fig_list(fig_list_t *target)
{
    fig_list_t *cur;
    int i;

    fflush(stderr);
    fprintf(stdout, "===\n");


    i = 0;
    cur = target;
    while(cur) {
/*
        fprintf(stdout, "%3d: %p %d - %d\n", i, cur, cur, FIG_OBJTYPE(cur->cont));
*/
        fprintf(stdout, "%3d: %p %4d - %d\n",
            i, cur, cur->len, FIG_OBJTYPE(cur->cont));
        show_fig_object(cur->cont);
        cur = cur->next;
        i++;
    }

    fflush(stdout);
}


static int  fig_version = -1;
fig_list_t *first_obj;
fig_list_t *last_obj;
fig_fpos    scaling;
fig_pos     bb_ul, bb_lr;   /* bounding box: unid left, bottom right */

fig_cnbook_t *cnbook;


static
parse_fig_version(char *line)
{
    register char *p;

    p = line;
    while(*p && !(*p==' ' || *p=='\t')) {
        p++;
    }
    if(*p==' ' || *p=='\t') {
        p++;
        if(isdigit(*p)) {
            if(strncmp(p, "3.1", 3)==0) {
                fig_version = FIG_VERSION_3_1;
            }
            else
            if(strncmp(p, "3.2", 3)==0) {
                fig_version = FIG_VERSION_3_2;
            }
            else {
                Error("unknown version is found\n");
                fig_version = -1;
            }

            if(fig_version>0) {
                Trace("version is [%s]\n", p);
            }
        }
    }
}

parse_fig_options(char *line)
{
#if 0
    fprintf(stdout, "option>>%s\n", line);
#endif
}

char *
parse_fig_textstring(char *line)
{
    unsigned char tmp[2048];
    register unsigned char *p, *q;
    int r, c;

    p = line;
    q = tmp;
    while(*p) {
        if(*p=='\\') {
            p++;
            r = 0;
            c = 0;
            while(c<3 && *p>='0' && *p<='7') {
                r += r*8+(*p-'0');
                p++;
                c++;
            }
            if(r==1) {
                break;
            }
            *q++ = r;
        }
        else {
            *q++ = *p++;
        }
    }
    *q = '\0';

    return strdup(tmp);
}


#define READ_INT(x) { \
    while(*p && (*p==' '|| *p=='\t' || *p=='\r' || *p=='\n')) { \
        p++; \
    } \
    while(*p=='\0') { \
        Trace("**next line**\n"); \
        read_fig_linerest(fp, lbuf, 1024); \
        p = lbuf; \
        while(*p && (*p==' '|| *p=='\t' || *p=='\r' || *p=='\n')) { \
            p++; \
        } \
    } \
    x=atoi(p); \
    while(*p && (isdigit(*p)||*p=='-')) { \
        p++; \
    } \
    while(*p && (*p==' '|| *p=='\t' || *p=='\r' || *p=='\n')) { \
        p++; \
    } \
}

#define READ_HEXINT(x) { \
    while(*p && (*p==' '|| *p=='\t' || *p=='\r' || *p=='\n')) { \
        p++; \
    } \
    while(*p=='\0') { \
        Trace("**next line**\n"); \
        read_fig_linerest(fp, lbuf, 1024); \
        p = lbuf; \
        while(*p && (*p==' '|| *p=='\t' || *p=='\r' || *p=='\n')) { \
            p++; \
        } \
    } \
    x=strtol(p, (char**)NULL, 16); \
    if(errno) {x=0;} \
    while(*p && (isxdigit(*p)||*p=='-')) { \
        p++; \
    } \
    while(*p && (*p==' '|| *p=='\t' || *p=='\r' || *p=='\n')) { \
        p++; \
    } \
}

#define READ_FLOAT(x)   { \
    while(*p && (*p==' '|| *p=='\t' || *p=='\r' || *p=='\n')) { \
        p++; \
    } \
    while(*p=='\0') { \
        Trace("**next line**\n"); \
        read_fig_linerest(fp, lbuf, 1024); \
        p = lbuf; \
        while(*p && (*p==' '|| *p=='\t' || *p=='\r' || *p=='\n')) { \
            p++; \
        } \
    } \
    x=atof(p); \
    while(*p && (isdigit(*p)||*p=='.'||*p=='-')) { \
        p++; \
    } \
    while(*p && (*p==' '|| *p=='\t' || *p=='\r' || *p=='\n')) { \
        p++; \
    } \
}

#define APPEND_FIG_OBJ(x) { \
    fig_list_t *tlist; \
    void *tobj; \
    tlist = (fig_list_t*)MALLOC(sizeof(fig_list_t)); \
    tobj = MALLOC(sizeof(x)); \
    memcpy(tobj, &x, sizeof(x)); \
    tlist->cont = tobj; \
    tlist->len  = sizeof(x); \
    tlist->next = NULL; \
    \
    if(last_obj!=NULL) { \
        last_obj->next = tlist; \
    } \
    last_obj = tlist; \
    if(first_obj==NULL) { \
        first_obj= tlist; \
    } \
}

#define APPEND_COLOR(x, v) { \
    Trace("APPEND_COLOR %d %08x\n", x, v); \
    if(cnbook && (x)<cnbook->max) { \
        cnbook->book[(x)] = (v); \
        if((x)+1>cnbook->len) \
            cnbook->len = (x)+1; \
    } \
}

/*
 *  fprintf(stdout, "<%7d,%7d> bbox <%7d,%7d><%7d,%7d>\n", \
 *      (t).x, (t).y, bb_ul.x, bb_ul.y, bb_lr.x, bb_lr.y); \
 */

#define CHECK_BBOX(t) { \
    if((t).x<bb_ul.x)   { bb_ul.x = (t).x; } \
    if((t).y<bb_ul.y)   { bb_ul.y = (t).y; } \
    if(bb_lr.x<(t).x)   { bb_lr.x = (t).x; } \
    if(bb_lr.y<(t).y)   { bb_lr.y = (t).y; } \
}

fig_list_t *
parse_fig_data(FILE *fp, char *lbuf)
{
    int dtype;
    fig_color_t t_color;
    fig_ellipse_t t_ell;
    fig_line_t t_line;
    fig_spline_t t_spline;
    fig_text_t t_text;
    int t_int;
    float t_float;
    int i;
    unsigned char *p;
    int nitems;
    int have_for_arrow, have_back_arrow;

    fig_pos tmp;


#if 0
    fprintf(stdout, "data> %s\n", lbuf);
#endif

    p = (unsigned char*) lbuf;
    dtype=atoi(lbuf);

    switch(dtype) {
    case FIG_COLOR:
        READ_INT(t_int);
        t_color.objtype = t_int;
        READ_INT(t_int);
        t_color.cnum = t_int;
        READ_HEXINT(t_int);
        t_color.value = t_int;

        APPEND_COLOR(t_color.cnum, t_color.value);
        break;

    case FIG_ARC:
        Trace("ARC\n");
        break;


    case FIG_ELLIPSE:
        READ_INT(t_int);
        t_ell.objtype = t_int;
        READ_INT(t_int);
        t_ell.subtype = t_int;
        READ_INT(t_int);
        t_ell.style = t_int;
        READ_INT(t_int);
        t_ell.thickness = t_int;
        READ_INT(t_int);
        t_ell.pen_color = t_int;
        READ_INT(t_int);
        t_ell.fill_color = t_int;
        READ_INT(t_int);
        t_ell.depth = t_int;
        READ_INT(t_int);
        READ_INT(t_int);
        t_ell.area_fill = t_int;
        READ_FLOAT(t_float);
        READ_INT(t_int);
        READ_FLOAT(t_float);

        READ_INT(t_int);
        t_ell.center.x = t_int;
        READ_INT(t_int);
        t_ell.center.y = t_int;
        CHECK_BBOX(t_ell.center);

        READ_INT(t_int);
        t_ell.radiuses.x = t_int;
        READ_INT(t_int);
        t_ell.radiuses.y = t_int;

        READ_INT(t_int);
        t_ell.start.x = t_int;
        READ_INT(t_int);
        t_ell.start.y = t_int;
        CHECK_BBOX(t_ell.start);

        READ_INT(t_int);
        t_ell.end.x = t_int;
        READ_INT(t_int);
        t_ell.end.y = t_int;
        CHECK_BBOX(t_ell.end);

        Trace("ELLIPSE\n");
        Trace("     type   = %d\n", t_ell.objtype);
        Trace("     depth  = %d\n", t_ell.depth);
        Trace("     center = (%d,%d)\n", t_ell.center.x, t_ell.center.y);
        Trace("     radius = (%d,%d)\n", t_ell.radiuses.x, t_ell.radiuses.y);
        Trace("     start  = (%d,%d)\n", t_ell.start.x, t_ell.start.y);
        Trace("     end    = (%d,%d)\n", t_ell.end.x, t_ell.end.y);

        APPEND_FIG_OBJ(t_ell);
        break;


    case FIG_POLYLINE:
        READ_INT(t_int);
        t_line.objtype = t_int;
        READ_INT(t_int);
        t_line.subtype = t_int;
        READ_INT(t_int);
        t_line.style = t_int;
        READ_INT(t_int);
        t_line.thickness = t_int;
        READ_INT(t_int);
        t_line.pen_color = t_int;
        READ_INT(t_int);
        t_line.fill_color = t_int;
        READ_INT(t_int);
        t_line.depth = t_int;
        READ_INT(t_int);
        READ_INT(t_int);
        t_line.area_fill = t_int;
        READ_FLOAT(t_float);
        READ_INT(t_int);
        READ_INT(t_int);
        READ_INT(t_int);
        t_line.radius = t_int;
        READ_INT(t_int);
        have_for_arrow = t_int;
        READ_INT(t_int);
        have_back_arrow = t_int;
        READ_INT(t_int);
        nitems = t_line.npoints = t_int;

        Trace("POLYLINE\n");
        Trace("     type   = %d\n", t_line.objtype);
        Trace("     subtype= %d\n", t_line.subtype);
        Trace("     depth  = %d\n", t_line.depth);
        Trace("     radius = %d\n", t_line.radius);
        Trace("     #of points = %d\n", t_line.npoints);

        if(have_for_arrow>0) {
            t_line.for_arrow =
                (fig_arrow*)MALLOC(sizeof(fig_arrow));
            if(t_line.for_arrow) {
                READ_INT(t_int);
                t_line.for_arrow->type = t_int;
                READ_INT(t_int);
                t_line.for_arrow->style = t_int;
                READ_FLOAT(t_float);
                t_line.for_arrow->thickness = t_float;
                READ_FLOAT(t_float);
                t_line.for_arrow->width = t_float;
                READ_FLOAT(t_float);
                t_line.for_arrow->height = t_float;
            }
        }
        else {
            t_line.for_arrow = NULL;
        }

        if(have_back_arrow>0) {
            t_line.back_arrow =
                (fig_arrow*)MALLOC(sizeof(fig_arrow));
            if(t_line.back_arrow) {
                READ_INT(t_int);
                t_line.back_arrow->type = t_int;
                READ_INT(t_int);
                t_line.back_arrow->style = t_int;
                READ_FLOAT(t_float);
                t_line.back_arrow->thickness = t_float;
                READ_FLOAT(t_float);
                t_line.back_arrow->width = t_float;
                READ_FLOAT(t_float);
                t_line.back_arrow->height = t_float;
            }
        }
        else {
            t_line.back_arrow = NULL;
        }

        t_line.npoints = nitems;
        t_line.points = (fig_pos*)MALLOC(sizeof(fig_pos)*(nitems+1));
        if(t_line.points==NULL) {
            Error("no memory\n");
            break;
        }
        for(i=0;i<t_line.npoints;i++) {
            READ_INT(t_int);
            t_line.points[i].x = t_int;
            READ_INT(t_int);
            t_line.points[i].y = t_int;
        Trace("    point[%d] = %d,%d\n",
                i, t_line.points[i].x, t_line.points[i].y);
            CHECK_BBOX(t_line.points[i]);
        }

        APPEND_FIG_OBJ(t_line);
        break;



    case FIG_SPLINE:
        READ_INT(t_int);
        t_spline.objtype = t_int;
        READ_INT(t_int);
        t_spline.subtype = t_int;
        READ_INT(t_int);
        t_spline.style = t_int;
        READ_INT(t_int);
        t_spline.thickness = t_int;
        READ_INT(t_int);
        t_spline.pen_color = t_int;
        READ_INT(t_int);
        t_spline.fill_color = t_int;
        READ_INT(t_int);
        t_spline.depth = t_int;
        READ_INT(t_int);
        READ_INT(t_int);
        t_spline.area_fill = t_int;
        READ_FLOAT(t_float);
        READ_INT(t_int);
        READ_INT(t_int);
        have_for_arrow = t_int;
        READ_INT(t_int);
        have_back_arrow = t_int;
        READ_INT(t_int);
        nitems = t_spline.npoints = t_int;

        Trace("SPLINE\n");
        Trace("     type   = %d\n", t_spline.objtype);
        Trace("     subtype= %d\n", t_spline.subtype);
        Trace("     depth  = %d\n", t_spline.depth);
        Trace("     radius = %d\n", t_spline.radius);
        Trace("     #of points = %d\n", t_spline.npoints);

        if(have_for_arrow>0) {
            t_spline.for_arrow =
                (fig_arrow*)MALLOC(sizeof(fig_arrow));
            if(t_spline.for_arrow) {
                READ_INT(t_int);
                t_spline.for_arrow->type = t_int;
                READ_INT(t_int);
                t_spline.for_arrow->style = t_int;
                READ_FLOAT(t_float);
                t_spline.for_arrow->thickness = t_float;
                READ_FLOAT(t_float);
                t_spline.for_arrow->width = t_float;
                READ_FLOAT(t_float);
                t_spline.for_arrow->height = t_float;
        Trace("     farrow = <%d %d %f %f %f>\n",
            t_spline.for_arrow->type, t_spline.for_arrow->style,
            t_spline.for_arrow->thickness , t_spline.for_arrow->width,
            t_spline.for_arrow->height);
            }
        }
        else {
            t_spline.for_arrow = NULL;
        }

        if(have_back_arrow>0) {
            t_spline.back_arrow =
                (fig_arrow*)MALLOC(sizeof(fig_arrow));
            if(t_spline.back_arrow) {
                READ_INT(t_int);
                t_spline.back_arrow->type = t_int;
                READ_INT(t_int);
                t_spline.back_arrow->style = t_int;
                READ_FLOAT(t_float);
                t_spline.back_arrow->thickness = t_float;
                READ_FLOAT(t_float);
                t_spline.back_arrow->width = t_float;
                READ_FLOAT(t_float);
                t_spline.back_arrow->height = t_float;
        Trace("     barrow = <%d %d %f %f %f>\n",
            t_spline.back_arrow->type, t_spline.back_arrow->style,
            t_spline.back_arrow->thickness , t_spline.back_arrow->width,
            t_spline.back_arrow->height);
            }
        }
        else {
            t_spline.back_arrow = NULL;
        }

        t_spline.npoints = nitems;
        t_spline.points = (fig_pos*)MALLOC(sizeof(fig_pos)*(nitems+1));
        if(t_spline.points==NULL) {
            Error("no memory\n");
            break;
        }
        for(i=0;i<t_spline.npoints;i++) {
            READ_INT(t_int);
            t_spline.points[i].x = t_int;
            READ_INT(t_int);
            t_spline.points[i].y = t_int;
        Trace("    point[%d] = %d,%d\n",
                i, t_spline.points[i].x, t_spline.points[i].y);
            CHECK_BBOX(t_spline.points[i]);
        }

        APPEND_FIG_OBJ(t_spline);
        break;

    case FIG_TEXT:
        READ_INT(t_int);
        t_text.objtype = t_int;
        READ_INT(t_int);
        t_text.subtype = t_int;
        READ_INT(t_int);
        t_text.color = t_int;
        READ_INT(t_int);
        t_text.depth = t_int;
        READ_INT(t_int);
        READ_INT(t_int);
        t_text.font = t_int;
        READ_FLOAT(t_float);
        t_text.font_size = t_float;
        READ_FLOAT(t_float);
        t_text.angle = t_float;
        READ_INT(t_int);
        t_text.font_flags = t_int;
        READ_FLOAT(t_float);
        t_text.height = t_float;
        READ_FLOAT(t_float);
        t_text.length = t_float;
        READ_INT(t_int);
        t_text.pos.x = t_int;
        READ_INT(t_int);
        t_text.pos.y = t_int;
        CHECK_BBOX(t_text.pos);
        t_text.value = parse_fig_textstring(p);

    {
        /* 0 left, 1 center, 2 right */
        switch(t_text.subtype) {
        case 0:
            tmp.x = t_text.pos.x;
            tmp.y = t_text.pos.y - (int) t_text.height - 1;
            CHECK_BBOX(tmp);

            tmp.x = t_text.pos.x + (int) t_text.length + 1;
            tmp.y = t_text.pos.y;
            CHECK_BBOX(tmp);

            break;

        case 1:
            tmp.x = t_text.pos.x - (int) t_text.length/2 - 1;
            tmp.y = t_text.pos.y - (int) t_text.height - 1;
            CHECK_BBOX(tmp);

            tmp.x = t_text.pos.x + (int) t_text.length/2 + 1;
            tmp.y = t_text.pos.y;
            CHECK_BBOX(tmp);

            break;

        case 2:
            tmp.x = t_text.pos.x - (int) t_text.length - 1;
            tmp.y = t_text.pos.y - (int) t_text.height - 1;
            CHECK_BBOX(tmp);

            tmp.x = t_text.pos.x;
            tmp.y = t_text.pos.y;
            CHECK_BBOX(tmp);

            break;
        }

    }

        Trace("TEXT\n");
        Trace("     type   = %d\n", t_text.objtype);
        Trace("     subtype= %d\n", t_text.subtype);
        Trace("     depth  = %d\n", t_text.depth);
        Trace("     font   = %d\n", t_text.font);
        Trace("     fflags = %d %x\n", t_text.font_flags, t_text.font_flags);
        Trace("     fsize  = %f\n", t_text.font_size);
        Trace("     angle  = %f\n", t_text.angle);
        Trace("     height = %f\n", t_text.height);
        Trace("     length = %f\n", t_text.length);
        Trace("     pos    = %d,%d\n", t_text.pos.x, t_text.pos.y);
        Trace("     value  = '%s'\n", t_text.value);


        APPEND_FIG_OBJ(t_text);

        break;


    case FIG_COMPOUND:
    case FIG_UNKNOW:
    default:
        fprintf(stdout, "unsupport object type\n");
    }

}

#undef READ_INT
#undef READ_HEXINT
#undef READ_FLOAT
#undef APPEND_FIG_OBJ
#undef CHECK_BBOX


int
release_fig_list(fig_list_t *list)
{
    fig_list_t *p, *o;

    Trace("release_fig_list: %p\n", list);

    if(list==NULL) {
        return -1;
    }

    o = NULL;
    p = list;
    while(p) {
        o = p->next;
        if(p->cont)
            FREE(p->cont);
        FREE(p);
        p = o;
    }

    return 0;
}

int
release_fig_list_link(fig_list_t *list)
{
    fig_list_t *p, *o;

    Trace("release_fig_list_link: %p\n", list);

    if(list==NULL) {
        return -1;
    }

    o = NULL;
    p = list;
    while(p) {
        o = p->next;
        FREE(p);
        p = o;
    }

    return 0;
}

release_fig(xp_fig_elem_t *target)
{
    fig_list_t *p, *o;

    Trace("release_fig: %p\n", target);

    if(target==NULL) {
        return -1;
    }

    if(target->lsbook)
        FREE(target->lsbook);
    if(target->cnbook && target->cnbook->book)
        FREE(target->cnbook->book);

    release_fig_list(target->list);

    return 0;
}

static
int
comp_fig_depth(void const *xa, void const *xb)
{
    fig_list_t *a, *b;

    a = (fig_list_t*)xa;
    b = (fig_list_t*)xb;

    if(FIG_OBJTYPE(a->cont)==FIG_COLOR) {
        if(FIG_OBJTYPE(b->cont)==FIG_COLOR) {
            return (((fig_color_t*)b->cont)->cnum)
                    - (((fig_color_t*)a->cont)->cnum);
        }
        else {
            return 1;
        }
    }
    if(FIG_OBJTYPE(b->cont)==FIG_COLOR) {
        if(FIG_OBJTYPE(a->cont)==FIG_COLOR) {
        }
        else {
            return -1;
        }
    }

    return (((fig_arc_t*)b->cont)->depth)
                - (((fig_arc_t*)a->cont)->depth);
}

xp_fig_elem_t*
sort_depth_fig(xp_fig_elem_t *target)
{
    int boxlen;
    fig_list_t *box;
    int i;
    fig_list_t *p;


    Trace("sort_depth_fig: %p list %p\n", target, target->list);

    if(target->list==NULL) {
        return target;
    }

    i = 0;
    p = target->list;
    while(p) {
#if 0
        Trace(" b object# %03d addr %p, %-10s %4d , next %p\n",
            i, p, FIG_OBJTYPESTR(FIG_OBJTYPE(p->cont)),
            FIG_OBJTYPE(p->cont)==FIG_COLOR ?
                ((fig_color_t*)p->cont)->cnum :
                ((fig_arc_t*)p->cont)->depth,
            p->next);
#endif
        i++;
        p = p->next;
    }

    boxlen = i;
    Trace("boxlen %d\n", boxlen);
    if(boxlen<=0) {
        return target;
    }

    box = (fig_list_t*)MALLOC(sizeof(fig_list_t)*boxlen);

    p = target->list;
    for(i=0;i<boxlen;i++) {
        memcpy(&box[i], p, sizeof(fig_list_t));
        p = p->next;
    }

#if 0
    for(i=0;i<boxlen;i++) {
        Trace(" d object# %03d addr %p, %-10s %4d , next %p\n",
            i, &box[i], FIG_OBJTYPESTR(FIG_OBJTYPE(box[i].cont)),
            FIG_OBJTYPE(box[i].cont)==FIG_COLOR ?
                ((fig_color_t*)box[i].cont)->cnum :
                ((fig_arc_t*)box[i].cont)->depth,
            box[i].next);
    }
#endif

    qsort(box, boxlen, sizeof(fig_list_t), comp_fig_depth);

    release_fig_list_link(target->list);
    target->list = box;
    p = target->list;
    for(i=0;i<boxlen-1;i++) {
        box[i].next = &box[i+1];
    }
    box[i].next = NULL;


#if 0
    for(i=0;i<boxlen;i++) {
        Trace(" a object# %03d addr %p, %-10s %4d , next %p\n",
            i, &box[i], FIG_OBJTYPESTR(FIG_OBJTYPE(box[i].cont)),
            FIG_OBJTYPE(&box[i].cont)==FIG_COLOR ?
                ((fig_color_t*)box[i].cont)->cnum :
                ((fig_arc_t*)box[i].cont)->depth,
            box[i].next);
    }
#endif

#if 0
    show_fig_list(target->list);
#endif


    Trace("sort_depth_fig: done\n");

    return target;
}


xp_fig_elem_t*
read_fig_fp(FILE *fp)
{
    xp_fig_elem_t *ret;
    char line[1025];
    int opt_count;
    int state;
    /*
     * state
     *    -1 ->  0 -->   1 -->   2
     *  init    format  options data
     *
     */

    ret = (xp_fig_elem_t*)MALLOC(sizeof(xp_fig_elem_t));
    if(ret==NULL) {
        return ret; /* so, it is NULL */
    }
    ret->cnbook = NULL;
    ret->lslen = 0;
    ret->lsbook = NULL;


    first_obj = NULL;
    last_obj = NULL;
    scaling.x = -1;
#ifdef INT_MAX
    bb_ul.x =  INT_MAX;
    bb_ul.y =  INT_MAX;
    bb_lr.x = -INT_MAX;
    bb_lr.y = -INT_MAX;
#else
    bb_ul.x =  1000000L;
    bb_ul.y =  1000000L;
    bb_lr.x = -1000000L;
    bb_lr.y = -1000000L;
#endif

    cnbook = (fig_cnbook_t*)MALLOC(sizeof(fig_cnbook_t));
    if(cnbook==NULL) {
        return NULL;
    }
    cnbook->len = 0;
    cnbook->max = 100;
    cnbook->book = (int*)MALLOC(cnbook->max*sizeof(int));
    if(cnbook->book==NULL) {
        FREE(cnbook);
        cnbook = NULL;
        return NULL;
    }
    else {
        memset(cnbook->book, 0, cnbook->max*sizeof(int));
    }

    /* default color index with RGB */
    APPEND_COLOR(0, 0x000000);      /* black    */
    APPEND_COLOR(1, 0x0000FF);      /* blue     */
    APPEND_COLOR(2, 0x00FF00);      /* green    */
    APPEND_COLOR(3, 0x00FFFF);      /* cyan     */
    APPEND_COLOR(4, 0xFF0000);      /* read     */
    APPEND_COLOR(5, 0xFF00FF);      /* magenta  */
    APPEND_COLOR(6, 0xFFFF00);      /* yellow   */
    APPEND_COLOR(7, 0xFFFFFF);      /* white    */

    APPEND_COLOR( 8, 0x000090);     /* */
    APPEND_COLOR( 9, 0x0000b0);     /* */
    APPEND_COLOR(10, 0x0000d0);     /* */
    APPEND_COLOR(11, 0x87ceff);     /* */

    APPEND_COLOR(12, 0x009000);     /* */
    APPEND_COLOR(13, 0x00b000);     /* */
    APPEND_COLOR(14, 0x00d000);     /* */

    APPEND_COLOR(15, 0x009090);     /* */
    APPEND_COLOR(16, 0x00b0b0);     /* */
    APPEND_COLOR(17, 0x00d0d0);     /* */

    APPEND_COLOR(18, 0x900000);     /* */
    APPEND_COLOR(19, 0xb00000);     /* */
    APPEND_COLOR(20, 0xd00000);     /* */

    APPEND_COLOR(21, 0x900090);     /* */
    APPEND_COLOR(22, 0xb000b0);     /* */
    APPEND_COLOR(23, 0xd000d0);     /* */

    APPEND_COLOR(24, 0x803000);     /* */
    APPEND_COLOR(25, 0xa04000);     /* */
    APPEND_COLOR(26, 0xc06000);     /* */

    APPEND_COLOR(27, 0xff8080);     /* */
    APPEND_COLOR(28, 0xffa0a0);     /* */
    APPEND_COLOR(29, 0xffc0c0);     /* */
    APPEND_COLOR(30, 0xffe0e0);     /* */

    APPEND_COLOR(31, 0xffd700);     /* gold */

#if 0
    if(cnbook) {
        Trace("colors: %p book %p len %d max %d\n",
            cnbook, cnbook->book, cnbook->len, cnbook->max);
    }
#endif


    state = -1;
    opt_count = 0;
    while(read_fig_line(fp, line, 1024)) {
        Chop_CRLF(line);
#if 0
        fprintf(stdout, "read <%s>\n", line);
#endif
        if(state==-1 && strncmp(line, "#FIG", 4)==0) {
            state=0;
            parse_fig_version(line);
            state=1;
            continue;
        }

        if(fig_version<0) {
            Trace("figure version is unknow\n");
            break;
        }

        if(state==1) {
            parse_fig_options(line);
            opt_count++;

            if(fig_version==FIG_VERSION_3_1 && opt_count>=4) {
                state = 2;
            }
            if(fig_version==FIG_VERSION_3_2 && opt_count>=8) {
                state = 2;
            }
            continue;
        }

        parse_fig_data(fp, line);
    }

#if 0
    if(!first_obj) {
        Trace("no object\n");
    }
    else {
        fprintf(stdout, "bbox %d,%d,%d,%d\n",
            bb_ul.x, bb_ul.y, bb_lr.x, bb_lr.y);
        show_fig_list(first_obj);
    }
#endif

    ret->list = first_obj;
    ret->ulx = bb_ul.x;
    ret->uly = bb_ul.y;
    ret->lrx = bb_lr.x;
    ret->lry = bb_lr.y;
    ret->cnbook = cnbook;
    ret->lslen = 0;
    ret->lsbook = NULL;

    return ret;
}

xp_fig_elem_t*
read_fig_file(char *fname)
{
    xp_fig_elem_t *ret;
    FILE *fp;

    if(fname==NULL) {
        return NULL;
    }
    if((fp = fopen(fname, "r"))==NULL) {
        return NULL;
    }
    ret = read_fig_fp(fp);
    fclose(fp);

    sort_depth_fig(ret);

    return ret;
}

int
scan_FIG(char *fname, int *xoff, int *yoff, int *wid, int *ht)
{
    FILE *fp;
    xp_fig_elem_t *cur;
    char line[1025];
    int opt_count;
    int state;
    /*
     * state
     *    -1 ->  0 -->   1 -->   2
     *  init    format  options data
     *
     */

#ifdef MEMCHECK
    {
        TracePush;
        Trace("before FIG file scanning\n");
        show_memoryusage();
        Trace("\n");
        fflush(stdout);
        fflush(stderr);
        TracePop;
    }
#endif


    if((fp=fopen(fname, "r"))==NULL) {
        return -1;
    }


    cur = (xp_fig_elem_t*)MALLOC(sizeof(xp_fig_elem_t));
    if(cur==NULL) {
        return -1; /* so, it is NULL */
    }
    cur->cnbook = NULL;
    cur->lslen = 0;
    cur->lsbook = NULL;


    first_obj = NULL;
    last_obj = NULL;
    scaling.x = -1;
#ifdef INT_MAX
    bb_ul.x =  INT_MAX;
    bb_ul.y =  INT_MAX;
    bb_lr.x = -INT_MAX;
    bb_lr.y = -INT_MAX;
#else
    bb_ul.x =  1000000L;
    bb_ul.y =  1000000L;
    bb_lr.x = -1000000L;
    bb_lr.y = -1000000L;
#endif

    cnbook = (fig_cnbook_t*)MALLOC(sizeof(fig_cnbook_t));
    if(cnbook==NULL) {
        return;
    }
    cnbook->len = 0;
    cnbook->max = 100;
    cnbook->book = (int*)MALLOC(cnbook->max*sizeof(int));
    if(cnbook->book==NULL) {
        FREE(cnbook);
        cnbook = NULL;
        return;
    }
    else {
        memset(cnbook->book, 0, cnbook->max*sizeof(int));
    }

    /* default color index with RGB */
    /* see also resources.c in 'xfig' packege */
    APPEND_COLOR( 0, 0x000000);     /* black    */
    APPEND_COLOR( 1, 0x0000FF);     /* blue     */
    APPEND_COLOR( 2, 0x00FF00);     /* green    */
    APPEND_COLOR( 3, 0x00FFFF);     /* cyan     */
    APPEND_COLOR( 4, 0xFF0000);     /* red      */
    APPEND_COLOR( 5, 0xFF00FF);     /* magenta  */
    APPEND_COLOR( 6, 0xFFFF00);     /* yellow   */
    APPEND_COLOR( 7, 0xFFFFFF);     /* white    */

    APPEND_COLOR( 8, 0x000090);     /* */
    APPEND_COLOR( 9, 0x0000b0);     /* */
    APPEND_COLOR(10, 0x0000d0);     /* */
    APPEND_COLOR(11, 0x87ceff);     /* */

    APPEND_COLOR(12, 0x009000);     /* */
    APPEND_COLOR(13, 0x00b000);     /* */
    APPEND_COLOR(14, 0x00d000);     /* */

    APPEND_COLOR(15, 0x009090);     /* */
    APPEND_COLOR(16, 0x00b0b0);     /* */
    APPEND_COLOR(17, 0x00d0d0);     /* */

    APPEND_COLOR(18, 0x900000);     /* */
    APPEND_COLOR(19, 0xb00000);     /* */
    APPEND_COLOR(20, 0xd00000);     /* */

    APPEND_COLOR(21, 0x900090);     /* */
    APPEND_COLOR(22, 0xb000b0);     /* */
    APPEND_COLOR(23, 0xd000d0);     /* */

    APPEND_COLOR(24, 0x803000);     /* */
    APPEND_COLOR(25, 0xa04000);     /* */
    APPEND_COLOR(26, 0xc06000);     /* */

    APPEND_COLOR(27, 0xff8080);     /* */
    APPEND_COLOR(28, 0xffa0a0);     /* */
    APPEND_COLOR(29, 0xffc0c0);     /* */
    APPEND_COLOR(30, 0xffe0e0);     /* */

    APPEND_COLOR(31, 0xffd700);     /* gold */

#if 0
    if(cnbook) {
        Trace("colors: %p book %p len %d max %d\n",
            cnbook, cnbook->book, cnbook->len, cnbook->max);
    }
#endif


    state = -1;
    opt_count = 0;
    while(read_fig_line(fp, line, 1024)) {
        Chop_CRLF(line);
#if 0
        fprintf(stdout, "read <%s>\n", line);
#endif
        if(state==-1 && strncmp(line, "#FIG", 4)==0) {
            state=0;
            parse_fig_version(line);
            state=1;
            continue;
        }

        if(fig_version<0) {
            Trace("figure version is unknow\n");
            break;
        }

        if(state==1) {
            parse_fig_options(line);
            opt_count++;

            if(fig_version==FIG_VERSION_3_1 && opt_count>=4) {
                state = 2;
            }
            if(fig_version==FIG_VERSION_3_2 && opt_count>=8) {
                state = 2;
            }
            continue;
        }

        parse_fig_data(fp, line);
    }

#if 0
    if(!first_obj) {
        Trace("no object\n");
    }
    else {
        fprintf(stdout, "bbox %d,%d,%d,%d\n",
            bb_ul.x, bb_ul.y, bb_lr.x, bb_lr.y);
        show_fig_list(first_obj);
    }
#endif

    cur->list = first_obj;
    cur->ulx = bb_ul.x;
    cur->uly = bb_ul.y;
    cur->lrx = bb_lr.x;
    cur->lry = bb_lr.y;
    cur->cnbook = cnbook;
    cur->lslen = 0;
    cur->lsbook = NULL;

    *xoff = bb_ul.x;
    *yoff = bb_ul.y;
    *wid  = bb_lr.x - bb_ul.x;
    *ht   = bb_lr.y - bb_ul.y;

    release_fig(cur);

#ifdef MEMCHECK
    {
        TracePush;
        Trace("after  FIG file scanning\n");
        show_memoryusage();
        Trace("\n");
        fflush(stdout);
        fflush(stderr);

        TracePop;
    }
#endif

    return 0;
}


#ifdef TEST_FIG
main(int argc, char **argv)
{
    xp_fig_elem_t *fig;
    fig = read_fig_fp(stdin);
    if(fig!=NULL) {
        sort_depth_fig(fig);
        show_fig_list(fig->list);
    }

}
#endif /* TEST_FIG */
