/*
 * $Id$
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <math.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xproto.h>
#include <X11/cursorfont.h>

#include <X11/Xmu/WinUtil.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include <X11/Shell.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/Scrollbar.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/Simple.h>
#include <X11/Xaw/SimpleMenu.h>

#include "Lww.h"

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#ifdef USE_ALLOCA
#ifdef HAVE_ALLOCA_H
#include <alloca.h>
#endif
#endif

#if 0
#include <png.h>
#endif

#include "hook.h"
#include "alist.h"

#if 0
#include "xpobj.h"
#include "xpoptions.h"
#include "img.h"
#include "reduce.h"
#include "png_file.h"
#include "png_pixmap.h"
#include "color.h"
#endif


#if 0
#include "app.h"
#endif

#include "fig.h"
#if 0
#include "xpfont.h"
#endif

typedef struct {
    double csize;
    char *afont;
    char *kfont;
} xp_font_t;

#include "str_draw.h"

#define XP_OPT_FRAME    (0x10)



/*
 * It is mean x==0.  The 'equal' intraction between integer and float
 * point has litte error.  Then, this condition uses the 'less'
 * interaction.
 */
#define NOWIDTH(x)      ((x)<0.001)

/*
#include "strdraw"
*/

ColorCacheStore(Display *dpy, Colormap cmap)
{
    return;
}

int FindPixel(Display *dpy, Visual *vis, Colormap cmap,
        String name, Pixel *rpixel)
{
    XColor qc, rc;
    int ret;
    int chk;
    unsigned int x;

    ret = -1;
    *rpixel = 0;
    qc.pixel = qc.red = qc.green = qc.blue = 0;

/*
    TracePush;
*/

    Trace("FindPixel: name '%s', cmap %d %#x, visual %p (id %d, %#x)\n",
        name, cmap, cmap, vis, vis->visualid, vis->visualid);

    if(vis==NULL) {
        ret = -2;
        return ret;
    }

    if(vis->class == TrueColor || vis->class == StaticColor 
      || vis->class == DirectColor ) {
        Trace("TrueColor/StaticColor/DirectColor\n");

        if((chk=XParseColor(dpy, cmap, name, &qc))==0) {
            Error("fail XParseColor()\n");
            ret = 1;
            *rpixel = 0x00000; /* maybe black */
        }
        else {
            Trace("\tvis mask %#8x, %#8x, %#8x // %d bits\n",
                vis->red_mask, vis->green_mask, vis->blue_mask,
                vis->bits_per_rgb);
            Trace("\tqc  RGB  %#8x, %#8x, %#8x // %#8x\n",
                    qc.red, qc.green, qc.blue, qc.pixel);

            *rpixel = 0;
    
            x = (qc.red >> (16 - vis->bits_per_rgb));
            x |= (x<<vis->bits_per_rgb) | (x<<(vis->bits_per_rgb*2));
            *rpixel |= (vis->red_mask & x);
    
            x = (qc.green >> (16 - vis->bits_per_rgb));
            x |= (x<<vis->bits_per_rgb) | (x<<(vis->bits_per_rgb*2));
            *rpixel |= (vis->green_mask & x);
    
            x = (qc.blue >> (16 - vis->bits_per_rgb));
            x |= (x<<vis->bits_per_rgb) | (x<<(vis->bits_per_rgb*2));
            *rpixel |= (vis->blue_mask & x);

            ret = 0;
        }
    }
    else {
        Trace("PseudoColor ?\n");

        if(XParseColor(dpy, cmap, name, &qc)==0) {
            Error("fail XParseColor()\n");

            ret = 1;
            *rpixel = BlackPixel(dpy, DefaultScreen(dpy));
        }
        else {
            if(XAllocColor(dpy, cmap, &qc)==0) {
                Error("fail XAllocColor()\n");

                ret = 2;
                *rpixel = BlackPixel(dpy, DefaultScreen(dpy));
            }
            else {
                Trace("success qc %d (%#x)\n", qc.pixel, qc.pixel);
                Trace("\tqc RGB %#4x, %#4x, %#4x\n", qc.red, qc.green, qc.blue);

                ret = 0;
                *rpixel = qc.pixel;
            }
        }
    }

    Trace("return %s (%d) , with pixel %d (%#x)\n",
            (ret==0) ? "SUCCESS" : "ERROR",
            ret, *rpixel, *rpixel);

/*
    TracePop;
*/

    return ret;
}

int FindPixelPrivate(Display *dpy, Visual *vis, Colormap cmap,
        String name, Pixel *rpixel)
{
    FindPixel(dpy, vis, cmap, name, rpixel);
}




/*
 * find pixel value for FIG.
 * if it is fail, try to use black (which is often use as default color).
 */
int FindFigPixel(Display *dpy, Visual *vis, Colormap cmap, int use_private,
        fig_cnbook_t* cnbook, int num, Pixel *ans)
{
    char cname[37];
    int chk;

    if(num > cnbook->len) {
        Error("color number #%d is too large color book(%d).\n",
            num, cnbook->len);

        return -1;
    }

    sprintf(cname, "rgb:%x/%x/%x",
        (cnbook->book[num] & 0xff0000) >> 16,
        (cnbook->book[num] & 0xff00) >> 8,
        (cnbook->book[num] & 0xff ) );

#if 0
    TracePush;
    Trace("Color# %3d : 0x%08x [%s] <%s>\n",
        num, cnbook->book[num], cname, use_private ? "private" : "shared");
    TracePop;
#endif

    if(use_private) {
#if 0
        ColorList(dpy, cmap, stderr);
#endif
        chk  = FindPixelPrivate(dpy, vis, cmap, cname, ans);
        if(chk) {
            chk  = FindPixelPrivate(dpy, vis, cmap, "black", ans);
            if(chk) {
                return -1;
            }
        }
    }
    else {
        chk  = FindPixel(dpy, vis, cmap, cname, ans);
        if(chk) {
            chk  = FindPixel(dpy, vis, cmap, "black", ans);
            if(chk) {
                return -1;
            }
        }
    }

#if 0
    TracePush;
    Trace("success. answer %d %08x\n", *ans, (unsigned int)*ans);
    TracePop;
#endif

    return 0;
}


#define XPOS(xx)    (xgo+(int)((double)((xx)-xof)*xsc))
#define YPOS(yy)    (ygo+(int)((double)((yy)-yof)*ysc))
#define XVAL(zz)    ((int)((double)(zz)*ysc))
#define YVAL(zz)    ((int)((double)(zz)*ysc))
#define RXPOS(ff)    (((double)(ff-xgo))/xsc+xof)
#define RYPOS(ff)    (((double)(ff-ygo))/ysc+yof)
#define THICK(t)    ((int)((xsc>ysc ? ysc : xsc)*15.0*(t)))

#if 0
#define VAL(zz)     ((int)((double)(zz)*sqrt(xsc*xsc+ysc*ysc)))
#define DVAL(zz)    (((double)(zz)*sqrt(xsc*xsc+ysc*ysc)))
#endif


static int ndash = 2;
static float dash[2] = {1., 1.};
static int ndot = 2;
static float dot[2] = {0., 0.};

static int ndash_dot = 4;
static float dash_dot[4] = { 1., 0.5, 0., 0.5 };
static int ndash_2dots = 6;
static float dash_2dots[6] = { 1., 0.45, 0., 0.333, 0., 0.45 };
static int ndash_3dots = 8;
static float dash_3dots[8] = { 1., 0.4, 0., 0.3, 0., 0.3, 0., 0.4 };

#define XSC (15.0)

int
set_line_attr(Display *dpy, GC gc, float sc, int tk, int st)
{
    int nd;
    float *fl;
    int i;
    int z;
    int lw;
    unsigned char dash_list[] = {255,255,255,255,255,255,255,255,255,255};
#if 0
    unsigned char dash_list[] = {
        {255,255},{255,255},{255,255}.{255,255},{255,255},
        {255,255},{255,255},{255,255}.{255,255},{255,255}};
#endif

    if(tk>0) {
        lw = (int)(sc*tk*XSC);
    }
    else {
        lw = 0;
    }

#if 0
    fflush(stdout);
    fprintf(stderr, "line-attr: sc %f, tk %d -> lw %d, st %d\n",
        sc, tk, lw, st);
#endif

#if 0
        XSetLineAttributes(dpy, gc, THICK(t_line->thickness),
            LineSolid, CapButt, JoinMiter);
#endif
    if(st==FIG_SOLID_LINE) {
        XSetLineAttributes(dpy, gc, lw, LineSolid, CapButt, JoinMiter);
    }
    else {
        XSetLineAttributes(dpy, gc, lw, LineOnOffDash, CapButt, JoinMiter);

#if 0 
        if(st==FIG_DASH_LINE) {
        }
        if(st==FIG_DOTTED_LINE) {
        }
        if(st==FIG_DASH_DOT_LINE) {
        }
        if(st==FIG_DASH_2_DOTS_LINE) {
        }
        if(st==FIG_DASH_3_DOTS_LINE) {
        }
        dash_line[0] = dash_line[1] = 10;
#endif

        if(st==FIG_DASH_LINE) {
            fl = dash;
            nd = ndash;
        }
        if(st==FIG_DOTTED_LINE) {
            fl = dot;
            nd = ndot;
        }
        if(st==FIG_DASH_DOT_LINE) {
            fl = dash_dot;
            nd = ndash_dot;
        }
        if(st==FIG_DASH_2_DOTS_LINE) {
            fl = dash_2dots;
            nd = ndash_2dots;
        }
        if(st==FIG_DASH_3_DOTS_LINE) {
            fl = dash_3dots;
            nd = ndash_3dots;
        }

        for(i=0;i<nd;i++) {
            if(fl[i] != 0.) {
                z = (int)(fl[i] * sc * XSC * 10);
                if(z>255)   dash_list[i] = 255;
                else        dash_list[i] = z;
            }
            else {
                dash_list[i] = sc;
            }
            if(dash_list[i]==0)
                dash_list[i] = 1;
        }
#if 0
        fprintf(stderr, "nd %d\n", nd);
        for(i=0;i<nd;i++) {
            fprintf(stderr, "%2d: %5.3f %3d\n", i, fl[i], dash_list[i]);
        }
#endif
#if 0
        fprintf(stderr, "nd %2d { ", nd);
        for(i=0;i<nd;i++) fprintf(stderr, "%.2f ", fl[i]);
        fprintf(stderr, "} { ");
        for(i=0;i<nd;i++) fprintf(stderr, "%d ", dash_list[i]);
        fprintf(stderr, "}\n");
        fflush(stdout);
#endif

        XSetDashes(dpy, gc, 0, dash_list, nd);
    }

}

static
int
dfig_ellipse(Display *dpy, Window w, GC gc,
        Visual *vis, int detph, Colormap cmap,
        int want, int use_private,
        int xgo, int ygo, int xof, int yof, double xsc, double ysc,
        void* cnbook,
        void* tar)
{
    fig_ellipse_t *t_ell;
    int chk;
    Pixel pen, fill;


    Trace("dfig_ellipse: ");
    Trace("xgo %d, ygo %d, xof %d, yof %d, xsc %.4f, ysc %.4f, tar %p\n",
        xgo, ygo, xof, yof, xsc, ysc, tar);

    if(tar==NULL) {
        return -1;
    }

    t_ell = tar;

    Trace("  colors: fill %d, pen %d <%s>\n",
        t_ell->fill_color, t_ell->pen_color,
        use_private ? "private" : "shared");

    /* fill shape */
    if(t_ell->area_fill!=-1) {

        if(t_ell->fill_color>=0 && cnbook) {
            chk = FindFigPixel(dpy, vis, cmap, use_private,
                cnbook, t_ell->fill_color, &fill);
            if(chk) {
                Trace("  Fail FindFigPixel() to %d. Then, use white.\n",
                    t_ell->fill_color);
                XSetForeground(dpy, gc, WhitePixel(dpy, DefaultScreen(dpy)));
            }
            else {
                Trace("  fill color -> %d\n", fill);
                XSetForeground(dpy, gc, fill);
            }
        }

        switch(t_ell->subtype) {
        case 1:
        case 2:
        case 3:
        case 4:
            XFillArc(dpy, w, gc,
                XPOS(t_ell->center.x - t_ell->radiuses.x),
                YPOS(t_ell->center.y - t_ell->radiuses.y),
                XVAL(t_ell->radiuses.x)*2,
                YVAL(t_ell->radiuses.y)*2,
                0*64, 360*64);
            break;
        default:
            break;
        }

    }


    Trace("  thickness %d\n", t_ell->thickness);

    /* draw shape */
    if(t_ell->thickness>0) {
#if 0
        XSetLineAttributes(dpy, gc, THICK(t_ell->thickness),
            LineSolid, CapButt, JoinMiter);
#endif
        set_line_attr(dpy, gc,
            (xsc>ysc ? ysc : xsc), t_ell->thickness, t_ell->style);

        if(t_ell->pen_color>=0 && cnbook) {
            chk = FindFigPixel(dpy, vis, cmap, use_private,
                cnbook, t_ell->pen_color, &pen);
            if(chk) {
                Trace("  Fail FindFigPixel() to %d. Then, use black.\n",
                    t_ell->pen_color);
                XSetForeground(dpy, gc, BlackPixel(dpy, DefaultScreen(dpy)));
            }
            else {
                Trace("  pen color -> %d\n", pen);
                XSetForeground(dpy, gc, pen);
            }
        }

        switch(t_ell->subtype) {
        case 1:
        case 2:
        case 3:
        case 4:
            XDrawArc(dpy, w, gc,
                XPOS(t_ell->center.x - t_ell->radiuses.x),
                YPOS(t_ell->center.y - t_ell->radiuses.y),
                XVAL(t_ell->radiuses.x)*2,
                YVAL(t_ell->radiuses.y)*2,
                0*64, 360*64);
            break;
        default:
            break;
        }

    }


}

static
double
calc_theta(int x1, int y1, int x2, int y2)
{
    double t;

    if(x2-x1 == 0) {
        Trace("*** look ***\n");
        if(y2-y1>0) {
            t = 3.0*M_PI/2.0;
        }
        else {
            t = M_PI/2.0;
        }

        goto solve;
    }

    t = acos((x2-x1)/
            sqrt(((double)x2-x1)*((double)x2-x1)
                    +((double)y2-y1)*((double)y2-y1)));

    Trace("calc_theta: x1 y1 %d,%d x2,y2 %d,%d\n", x1, y1, x2, y2);
    Trace("  |x| %d, |y| %d, T %.3f -> t %.3f rad %.3f deg\n",
            x2-x1, y2-y1,
            sqrt(((double)x2-x1)*((double)x2-x1)
                    +((double)y2-y1)*((double)y2-y1)),
            t, (t*180.0)/M_PI
        );

    if(y2-y1>0) {
        t = - M_PI + t;
    }
    else {
        t = M_PI - t;       /* ok */
    }

solve:
#if 0
    Trace("  ---> t %.3f rad %.3f deg\n", t, (t*180.0)/M_PI);
#endif

    return t;
}

static
draw_arrow(Display *dpy, Window w, GC gc,
/*
        Visual *vis, int depth, Colormap cmap,
        int want, int use_private,
*/
        int gx, int gy,
        double xsc, double ysc,
        double angle, fig_arrow *ar)
{
    double r;
    double cp, sp;
    int cx, cy, dx, dy;
    XPoint pos[5];
    int n;

    Trace("   angle %.2f\n", angle);
    Trace("   type %d, style %d\n",
        ar->type, ar->style);

    r = sqrt((double)ar->height*ar->height + (double)ar->width*ar->width/4);

    cp = ar->height/r;
    sp = (ar->width/2)/r;

    Trace("   orig w,h %.3f,%.3f r %.3f, cp %.3f, sp %.3f\n",
        ar->width, ar->height, r, cp, sp);

    /*
     * cx = gx + cos(angle - p) * r
     * cy = gy + sin(angle - p) * r
     * dx = gx + cos(angle + p) * r
     * dy = gy + sin(angle + p) * r
     */
    pos[0].x = gx;
    pos[0].y = gy;
    pos[2].x = gx + XVAL( r * (  cos(angle)*cp - sin(angle) * sp));
    pos[1].y = gy + YVAL( r * (- cos(angle)*sp + sin(angle) * cp));
    pos[1].x = gx + XVAL( r * (  cos(angle)*cp + sin(angle) * sp));
    pos[2].y = gy + YVAL( r * (  cos(angle)*sp + sin(angle) * cp));
    pos[3].x = gx;
    pos[3].y = gy;
    n = 4;

    Trace("    1 %.4f %.4f\n",
                    (  cos(angle)*cp - sin(angle) * sp),
                    (- cos(angle)*sp + sin(angle) * cp) );
    Trace("    1 %5d %5d\n",
            XVAL( r * (  cos(angle)*cp - sin(angle) * sp)),
            YVAL( r * (- cos(angle)*sp + sin(angle) * cp)) );
    Trace("    2 %.4f %.4f\n",
                    (  cos(angle)*cp + sin(angle) * sp),
                    (  cos(angle)*sp + sin(angle) * cp) );
    Trace("    2 %5d %5d\n",
            XVAL( r * (  cos(angle)*cp + sin(angle) * sp)),
            YVAL( r * (  cos(angle)*sp + sin(angle) * cp)) );

    {
        int i;
        for(i=0;i<n;i++) {
            Trace("   %2d: %7d %7d\n", i, pos[i].x, pos[i].y);
        }
    }

#if 0
    XSetForeground(dpy, gc, 0x0000ff);
    XFillPolygon(dpy, w, gc, pos, n, 0, CoordModeOrigin);
    XSetForeground(dpy, gc, 0x00ff00);
    XDrawLines(dpy, w, gc, pos, n, CoordModeOrigin);
#endif

#if 1
    if(ar->style==0)
        XDrawLines(dpy, w, gc, pos, n, CoordModeOrigin);
    if(ar->style==1)
        XFillPolygon(dpy, w, gc, pos, n, 0, CoordModeOrigin);
#endif
}



static
int
dfig_polyline(Display *dpy, Window w, GC gc,
        Visual *vis, int depth, Colormap cmap,
        int want, int use_private,
        int xgo, int ygo, int xof, int yof, double xsc, double ysc,
        void* cnbook,
        void* tar)
{
    fig_line_t *t_line;
    XPoint *xpoints;
    int i, n;
    Pixel pen, fill;
    int chk;
    double theta;

    Trace("dfig_polyline: ");
    Trace("xgo %d, ygo %d, xof %d, yof %d, xsc %.4f, ysc %.4f, tar %p\n",
        xgo, ygo, xof, yof, xsc, ysc, tar);

    if(tar==NULL) {
        return -1;
    }


    t_line = tar;

#ifdef USE_ALLOCA
    xpoints = (XPoint*) alloca(sizeof(XPoint)*t_line->npoints);
#else
    xpoints = (XPoint*) MALLOC(sizeof(XPoint)*t_line->npoints);
#endif

    for(i=0;i<t_line->npoints;i++) {
        xpoints[i].x = XPOS(t_line->points[i].x);
        xpoints[i].y = YPOS(t_line->points[i].y);
    }

#if 0
    /* for debug */
    for(i=0;i<t_line->npoints;i++) {
        fprintf(stdout, " : %4d: %4d %4d\n",
            i, xpoints[i].x, xpoints[i].y);
    }
    fflush(stdout);
#endif

    if(t_line->area_fill!=-1) {
        if(t_line->fill_color>=0 && cnbook) {
            Trace("fill color is defined as %d\n", t_line->fill_color);

            chk = FindFigPixel(dpy, vis, cmap, use_private,
                cnbook, t_line->fill_color, &fill);
            if(chk) {
                Trace("  Fail FindFigPixel()\n");
            }
            else {
                Trace(" -> %d\n", fill);
            }
            XSetForeground(dpy, gc, fill);
        }
        else {
            /* nothing */
        }

        if(t_line->subtype==4) {
            /* arc-box */
        }
        else {
        }
        XFillPolygon(dpy, w, gc, xpoints, t_line->npoints, 0, CoordModeOrigin);
    }


    Trace("  thickness %d\n", t_line->thickness);
    if(t_line->thickness>0) {
#if 0
        XSetLineAttributes(dpy, gc, THICK(t_line->thickness),
            LineSolid, CapButt, JoinMiter);
#endif
        
#if 0
        {
        int j;
        for(j=0;j<6;j++) {
            set_line_attr(dpy, gc, (xsc>ysc ? ysc : xsc), t_line->thickness, j);
        }
        }
#endif

        set_line_attr(dpy, gc,
            (xsc>ysc ? ysc : xsc), t_line->thickness, t_line->style);

        if(t_line->pen_color>=0 && cnbook) {
            Trace("pen color is defined as %d\n", t_line->pen_color);
            chk = FindFigPixel(dpy, vis, cmap, use_private,
                cnbook, t_line->pen_color, &pen);
            if(chk) {
                Trace("  Fail FindFigPixel()\n");
            }
            else {
                Trace(" -> %d\n", pen);
            }
            XSetForeground(dpy, gc, pen);
        }
        else {
            Trace("pen color is not defined\n");
        }

        if(t_line->subtype==4) {
            /* arc-box */
        }
        else {
        }
        XDrawLines(dpy, w, gc, xpoints, t_line->npoints, CoordModeOrigin);

        if(t_line->back_arrow) {
            theta =
                calc_theta(t_line->points[1].x, t_line->points[1].y,
                    t_line->points[0].x, t_line->points[0].y);
            Trace("theta = %f [ %f deg]\n", theta, (theta*180.0)/M_PI);
            draw_arrow(dpy, w, gc,
                xpoints[0].x, xpoints[0].y, xsc, ysc, theta, t_line->back_arrow);
        }
        if(t_line->for_arrow) {
            theta =
                calc_theta(
                    t_line->points[t_line->npoints-2].x,
                    t_line->points[t_line->npoints-2].y,
                    t_line->points[t_line->npoints-1].x,
                    t_line->points[t_line->npoints-1].y);
            Trace("theta = %f [ %f deg]\n", theta, (theta*180.0)/M_PI);
            draw_arrow(dpy, w, gc,
                xpoints[t_line->npoints-1].x, xpoints[t_line->npoints-1].y,
                xsc, ysc, theta, t_line->for_arrow);
        }

    }

out:
#ifdef USE_ALLOCA
#else
    FREE(xpoints);
#endif

    return;
}

/* 
 *  # the term                    | this implementation
 *    in documents of Xfig        | 
 *  ------------------------------------------------------------
 *  0 opened approximated spline    bezier curve ?
 *  1 closed approximated spline    ?
 *  2 opened interpolated spline    cubic spline ?
 *  3 closed interpolated spline    ?
 *  4 opened x-spline               ?
 *  5 closed x-spline               ?
 */

/* this routine is quoted from the book by H.Okumura */
static
int
mktbl_cubic_spline(int n, XPoint *xy, double *z)
{
    int i;
    double t;
    static double *h, *d;

#ifdef USE_ALLOCA
    h = (double*) alloca(sizeof(double)*n);
    d = (double*) alloca(sizeof(double)*n);
#else
    h = (double*) MALLOC(sizeof(double)*n);
    d = (double*) MALLOC(sizeof(double)*n);
#endif

#if 0 
    z[0] = 0;
    z[n-1] = 0;
    for(i=0;i<n-1;i++) {
        h[i]    = x[i+1] - x[i];
        d[i+1]  = (y[i+1] - y[i])/h[i];
    }
    z[1] = d[2] - d[1] - h[0] * z[0];
    d[1] = 2 * (x[2] - x[0]);
    for(i=1;i<n-2;i++) {
        t = h[i] / d[i];
        z[i+1] = d[i+2] - d[i+1] - z[i] *t;
        d[i+i] = 2* (x[i+2] - x[i]) - h[i] * t;
    }
    z[n-2] -= h[n-2] * z[n-1];
    for(i=n-2;i>0;i--){
        z[i] = (z[i] - h[i] * z[i+1])/ d[i];
    }
#endif

    z[0] = 0;
    z[n-1] = 0;
    for(i=0;i<n-1;i++) {
        h[i]    = xy[i+1].x - xy[i].x;
        d[i+1]  = (xy[i+1].y - xy[i].y)/h[i];
    }
    z[1] = d[2] - d[1] - h[0] * z[0];
    d[1] = 2 * (xy[2].x - xy[0].x);
    for(i=1;i<n-2;i++) {
        t = h[i] / d[i];
        z[i+1] = d[i+2] - d[i+1] - z[i] *t;
        d[i+i] = 2* (xy[i+2].x - xy[i].x) - h[i] * t;
    }
    z[n-2] -= h[n-2] * z[n-1];
    for(i=n-2;i>0;i--){
        z[i] = (z[i] - h[i] * z[i+1])/ d[i];
    }

}

static
double
pos_cubic_spline(double t, int n, XPoint *xy, double *z)
{
    int i, j, k;
    double d, h;

    i=0;
    j=n-1;
    while(i<j) {
        k = (i+j)/2;
        if(xy[k].x<t)
            i = k+1;
        else 
            j = k;
    }
    if(i>0)
        i--;

    h = xy[i+1].x - xy[i].x;
    d = t - xy[i].x;

    return (((z[i+1] - z[i])*d / h + z[i]*3)*d
            + ((xy[i+1].y - xy[i].y)/h - (z[i]*2+z[i+1])*h )) *d + xy[i].y;
}

int
comp_XPoint(const void* za, const void* zb)
{
    XPoint *a, *b;
    a = (XPoint*)za;
    b = (XPoint*)zb;
    if(a->x==b->x) {
        return a->y - b->y;
    }
    else {
        return a->x - b->x;
    }
}

int
rcomp_XPoint(const void* za, const void* zb)
{
    XPoint *a, *b;
    a = (XPoint*)za;
    b = (XPoint*)zb;
    if(a->x==b->x) {
        return b->y - a->y;
    }
    else {
        return b->x - a->x;
    }
}

/*
 * This is dummy.
 * In actually,  this is same routine of dfig_polyline X-)
 */
static
int
dfig_spline(Display *dpy, Window w, GC gc,
        Visual *vis, int depth, Colormap cmap,
        int want, int use_private,
        int xgo, int ygo, int xof, int yof, double xsc, double ysc,
        void* cnbook,
        void* tar)
{
    fig_spline_t *t_line;
    XPoint *rxpoints;   /* row positions for points */
    XPoint *xpoints;    /* justified them */
    double *zpoints;
    int i, j, n;
    Pixel pen, fill;
    int chk;
    double theta;

    TraceOn;


    Trace("dfig_spline: ");
    Trace("xgo %d, ygo %d, xof %d, yof %d, xsc %.4f, ysc %.4f, tar %p\n",
        xgo, ygo, xof, yof, xsc, ysc, tar);

    if(tar==NULL) {
        return -1;
    }


    t_line = tar;

#ifdef USE_ALLOCA
    rxpoints = (XPoint*) alloca(sizeof(XPoint)*t_line->npoints);
    xpoints = (XPoint*) alloca(sizeof(XPoint)*t_line->npoints);
    zpoints = (double*) alloca(sizeof(double)*t_line->npoints);
#else
    rxpoints = (XPoint*) MALLOC(sizeof(XPoint)*t_line->npoints);
    xpoints = (XPoint*) MALLOC(sizeof(XPoint)*t_line->npoints);
    zpoints = (double*) MALLOC(sizeof(double)*t_line->npoints);
#endif
    for(i=0;i<t_line->npoints;i++) {
        rxpoints[i].x = t_line->points[i].x;
        rxpoints[i].y = t_line->points[i].y;
        xpoints[i].x = XPOS(t_line->points[i].x);
        xpoints[i].y = YPOS(t_line->points[i].y);
    }
    qsort(rxpoints, t_line->npoints, sizeof(XPoint), comp_XPoint);

    mktbl_cubic_spline(t_line->npoints, rxpoints, zpoints);

#if 0
    /* for debug */
    Trace("  subtype %d, line_style %d\n",
        t_line->subtype, t_line->style);
    for(i=0;i<t_line->npoints;i++) {
        fprintf(stdout, " : %4d: %4d %4d ; %.3f\n",
            i, rxpoints[i].x, rxpoints[i].y, zpoints[i]);
    }
    for(i=0;i<t_line->npoints;i++) {
        fprintf(stdout, " : %4d: %4d %4d ; %.3f\n",
            i, xpoints[i].x, xpoints[i].y, zpoints[i]);
    }
    fflush(stdout);
#endif

#if 0
    if(t_line->area_fill!=-1) {
        if(t_line->fill_color>=0 && cnbook) {
            Trace("fill color is defined as %d\n", t_line->fill_color);

            chk = FindFigPixel(dpy, vis, cmap, use_private,
                cnbook, t_line->fill_color, &fill);
            if(chk) {
                Trace("  Fail FindFigPixel()\n");
            }
            else {
                Trace(" -> %d\n", fill);
            }
            XSetForeground(dpy, gc, fill);
        }
        else {
            /* nothing */
        }

        if(t_line->subtype==4) {
            /* arc-box */
        }
        else {
        }
        XFillPolygon(dpy, w, gc, xpoints, t_line->npoints, 0, CoordModeOrigin);
    }
#endif


    Trace("  thickness %d\n", t_line->thickness);
    if(t_line->thickness>0) {
#if 0
        XSetLineAttributes(dpy, gc, THICK(t_line->thickness),
            LineSolid, CapButt, JoinMiter);
#endif
        set_line_attr(dpy, gc,
            (xsc>ysc ? ysc : xsc), t_line->thickness, t_line->style);

        if(t_line->pen_color>=0 && cnbook) {
            Trace("pen color is defined as %d\n", t_line->pen_color);
            chk = FindFigPixel(dpy, vis, cmap, use_private,
                cnbook, t_line->pen_color, &pen);
            if(chk) {
                Trace("  Fail FindFigPixel()\n");
            }
            else {
                Trace(" -> %d\n", pen);
            }
            XSetForeground(dpy, gc, pen);
        }
        else {
            Trace("pen color is not defined\n");
        }


#if 0
        XDrawLines(dpy, w, gc, xpoints, t_line->npoints, CoordModeOrigin);
#endif

        {
            XPoint pp[3];
            int rx, ty;
            int wx;

            for(i=0;i<t_line->npoints-1;i++) {
                wx = (int)((double)10.0/xsc);
#if 0
                Trace("  width raw %d scaled %d (%f); wx %d\n",
                        rxpoints[i+1].x - rxpoints[i].x,
                        XPOS(rxpoints[i+1].x) - XPOS(rxpoints[i].x),
                        xsc,
                        wx);
#endif

                    rx = rxpoints[i].x;
                    ty = pos_cubic_spline(rx,
                            t_line->npoints, rxpoints, zpoints);
                    pp[1].x = XPOS(rx);
                    pp[1].y = YPOS(ty);
                    rx += wx;

                while(rx<rxpoints[i+1].x) {
                    pp[0].x = pp[1].x;
                    pp[0].y = pp[1].y;

                    ty = pos_cubic_spline(rx,
                            t_line->npoints, rxpoints, zpoints);
                    pp[1].x = XPOS(rx);
                    pp[1].y = YPOS(ty);
                    rx += wx;

#if 0
                    Trace("  draw#i %4d, %4d - %4d, %4d // rx %d\n",
                        pp[0].x, pp[0].y, pp[1].x, pp[1].y, rx);
#endif
                    XDrawLines(dpy, w, gc, pp, 2, CoordModeOrigin);
                }
                    pp[0].x = pp[1].x;
                    pp[0].y = pp[1].y;

                    rx = rxpoints[i+1].x;
                    ty = pos_cubic_spline(rx,
                            t_line->npoints, rxpoints, zpoints);
                    pp[1].x = XPOS(rx);
                    pp[1].y = YPOS(ty);

#if 0
                    Trace("  draw#n %d, %d - %d, %d\n",
                        pp[0].x, pp[0].y, pp[1].x, pp[1].y);
#endif
                    XDrawLines(dpy, w, gc, pp, 2, CoordModeOrigin);

            }
        }

        if(t_line->back_arrow) {
            theta =
                calc_theta(t_line->points[1].x, t_line->points[1].y,
                    t_line->points[0].x, t_line->points[0].y);
            Trace("theta = %f [ %f deg]\n", theta, (theta*180.0)/M_PI);
            draw_arrow(dpy, w, gc,
                xpoints[0].x, xpoints[0].y, xsc, ysc, theta, t_line->back_arrow);
        }
        if(t_line->for_arrow) {
            theta =
                calc_theta(
                    t_line->points[t_line->npoints-2].x,
                    t_line->points[t_line->npoints-2].y,
                    t_line->points[t_line->npoints-1].x,
                    t_line->points[t_line->npoints-1].y);
            Trace("theta = %f [ %f deg]\n", theta, (theta*180.0)/M_PI);
            draw_arrow(dpy, w, gc,
                xpoints[t_line->npoints-1].x, xpoints[t_line->npoints-1].y,
                xsc, ysc, theta, t_line->for_arrow);
        }

    }

out:
#ifdef USE_ALLOCA
#else
    FREE(rxpoints);
    FREE(xpoints);
    FREE(zpoints);
#endif

    TraceOff;

    return;
}


#ifndef FIG_MINCHO_FONT
#define FIG_MINCHO_FONT \
    "-misc-fixed-medium-r-normal--%d-130-75-75-c-140-jisx0208.1983-0"
#endif
#ifndef FIG_GOTHIC_FONT
#define FIG_GOTHIC_FONT \
    "-misc-fixed-medium-r-normal--%d-130-75-75-c-140-jisx0208.1983-0"
#endif

static
xp_font_t adobe_font_set[] = {
  {
    1.0,
    "-adobe-times-medium-r-normal--%d-*-*-*-p-0-iso8859-1",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-times-medium-i-normal--%d-*-*-*-p-0-iso8859-1",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-times-bold-r-normal--%d-*-*-*-p-0-iso8859-1",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-times-bold-i-normal--%d-*-*-*-p-0-iso8859-1",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-itc avant garde gothic-book-r-normal--%d-*-*-*-p-0-iso8859-1",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-itc avant garde gothic-book-o-normal--%d-*-*-*-p-0-iso8859-1",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-itc avant garde gothic-demi-r-normal--%d-*-*-*-p-0-iso8859-1",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-itc avant garde gothic-demi-o-normal--%d-*-*-*-p-0-iso8859-1",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-itc bookman-light-r-normal--%d-*-*-*-p-0-iso8859-1",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-itc bookman-light-i-normal--%d-*-*-*-p-0-iso8859-1",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-itc bookman-demi-r-normal--%d-*-*-*-p-0-iso8859-1",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-itc bookman-demi-i-normal--%d-*-*-*-p-0-iso8859-1",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-courier-medium-r-normal--%d-*-*-*-m-0-iso8859-1",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-courier-medium-o-normal--%d-*-*-*-m-0-iso8859-1",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-courier-bold-r-normal--%d-*-*-*-m-0-iso8859-1",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-courier-bold-o-normal--%d-*-*-*-m-0-iso8859-1",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-helvetica-medium-r-normal--%d-*-*-*-p-0-iso8859-1",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-helvetica-medium-o-normal--%d-*-*-*-p-0-iso8859-1",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-helvetica-bold-r-normal--%d-*-*-*-p-0-iso8859-1",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-helvetica-bold-o-normal--%d-*-*-*-p-0-iso8859-1",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-helvetica-medium-r-narrow--%d-*-*-*-p-0-iso8859-1",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-helvetica-medium-o-narrow--%d-*-*-*-p-0-iso8859-1",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-helvetica-bold-r-narrow--%d-*-*-*-p-0-iso8859-1",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-helvetica-bold-o-narrow--%d-*-*-*-p-0-iso8859-1",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-new century schoolbook-medium-r-normal--%d-*-*-*-p-0-iso8859-1",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-new century schoolbook-medium-i-normal--%d-*-*-*-p-0-iso8859-1",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-new century schoolbook-bold-r-normal--%d-*-*-*-p-0-iso8859-1",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-new century schoolbook-bold-i-normal--%d-*-*-*-p-0-iso8859-1",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-palatino-medium-r-normal--%d-*-*-*-p-0-iso8859-1",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-palatino-medium-i-normal--%d-*-*-*-p-0-iso8859-1",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-palatino-bold-r-normal--%d-*-*-*-p-0-iso8859-1",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-palatino-bold-i-normal--%d-*-*-*-p-0-iso8859-1",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-symbol-medium-r-normal--%d-*-*-*-p-0-adobe-fontspecific",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-itc zapf chancery-medium-i-normal--%d-*-*-*-p-0-iso8859-1",
    FIG_MINCHO_FONT
  },
  {
    1.0,
    "-adobe-itc zapf dingbats-medium-r-normal--%d-*-*-*-p-0-iso8859-1",
    FIG_MINCHO_FONT
  },
};






static
    xp_font_t default_font = {
        14,
        "-adobe-courier-bold-r-normal--%d-0-75-75-m-*-iso8859-1",
#if 0
        "-misc-fixed-medium-r-normal--%d-130-75-75-c-140-jisx0208.1990-0"
#else
        "-misc-fixed-medium-r-normal--%d-130-75-75-c-140-jisx0208.1983-0"
#endif
    };


static
int
dfig_text(Display *dpy, Window w, GC gc,
        Visual *vis, int depth, Colormap cmap,
        int want, int use_private,
        int xgo, int ygo, int xof, int yof, double xsc, double ysc,
        void* cnbook,
        void* tar)
{
    fig_text_t *t_text;
    XPoint *xpoints;
    int i, n;
    Pixel pen, bg;
    int chk;
    int x, y;
    XGCValues gcval;
    xp_font_t cfont;
    int xlen;

    int fit_box=1;

    Trace("dfig_text: ");
    Trace("xgo %d, ygo %d, xof %d, yof %d, xsc %.4f, ysc %.4f, tar %p\n",
        xgo, ygo, xof, yof, xsc, ysc, tar);

    if(tar==NULL) {
        return -1;
    }
#if 0
    show_fig_object(tar);
#endif

    t_text = tar;


    if(t_text->color>=0 && cnbook) {
        Trace("pen color is defined as %d\n", t_text->color);
        chk = FindFigPixel(dpy, vis, cmap, use_private,
            cnbook, t_text->color, &pen);
        if(chk) {
            Trace("  Fail FindFigPixel()\n");
        }
        else {
            Trace(" -> %d\n", pen);
        }
        XSetForeground(dpy, gc, pen);
    }
    else {
        Trace("pen color is not defined\n");
        XGetGCValues(dpy, gc, GCForeground |GCBackground, &gcval);
        pen = (Pixel) gcval.foreground;
        bg  = (Pixel) gcval.background;
    }

    /* fit string size to area */
    if(t_text->font>=0 && t_text->font<=34) {
        memcpy(&cfont, &adobe_font_set[t_text->font], sizeof(xp_font_t));
        cfont.csize = YVAL(t_text->height);
    }
    else {
        Trace("  ignore font# %d. Use default font.\n", t_text->font);
        memcpy(&cfont, &default_font, sizeof(xp_font_t));
        cfont.csize = YVAL(t_text->height);
    }

#if 0
    /* debug, draw area as center justified text */
    XDrawRectangle(dpy, w, gc,
        XPOS(t_text->pos.x) - XVAL(t_text->length)/2,
        YPOS(t_text->pos.y) - YVAL(t_text->height),
        XVAL(t_text->length), YVAL(t_text->height));
#endif

    MEMORY_TRAP("before fitting string size\n");

    /* set size */

    if(fit_box && cfont.csize>4.0) {
        double lastsize;

        xlen = GetStringLength(dpy, &cfont, t_text->value, 0);
        Trace("  csize %f, xlen %d : size %d\n",
            cfont.csize, xlen, XVAL(t_text->length));

        while(cfont.csize>4.0 && xlen>XVAL(t_text->length)) {
            lastsize = cfont.csize;
            cfont.csize = rint((cfont.csize)*0.92);
            if(NOWIDTH(lastsize - cfont.csize)) {
                break;
            }
            xlen = GetStringLength(dpy, &cfont, t_text->value, 0);
            Trace("  csize %f, xlen %d : size %d\n",
                cfont.csize, xlen, XVAL(t_text->length));
            if(xlen<=0) {
                xlen = XVAL(t_text->length);
                break;
            }
        }
    }
    else {
        xlen = GetStringLength(dpy, &cfont, t_text->value, 0);
        Trace("                 ***here*** %d\n", xlen);
        if(xlen<=0) {
            xlen = XVAL(t_text->length);
        }
    }

    if(cfont.csize<4.0) {
        xlen = XVAL(t_text->length);
    }


    MEMORY_TRAP("after  fitting\n");


    /* calc location */
    switch(t_text->subtype) {
    case 0:
        x = XPOS(t_text->pos.x);
        y = YPOS(t_text->pos.y);
        break;

    case 1:
        x = XPOS(t_text->pos.x) - xlen/2;
        y = YPOS(t_text->pos.y);
        break;

    case 2:
        x = XPOS(t_text->pos.x) - xlen;
        y = YPOS(t_text->pos.y);
        break;
    }
    Trace("x %d, y %d, height %d, length %d -> %d\n",
        x, y, XVAL(t_text->height), YVAL(t_text->length), xlen);

    if(cfont.csize<4.0) {
        XFillRectangle(dpy, w, gc, x, y,
            XVAL(t_text->length), YVAL(t_text->height));
        goto out;
    }

    /* draw text */
    if(t_text->font>=0 && t_text->font<=34) {
        DrawString(dpy, w, gc, x, y-YVAL(t_text->height),
            pen, bg, depth, &cfont, t_text->value, 0);
    }
    else {
        XDrawString(dpy, w, gc, x, y-YVAL(t_text->height),
            t_text->value, strlen(t_text->value));
    }

out:
    return 0;
}

#undef XPOS
#undef YPOS
#undef XVAL
#undef YVAL


void
DrawFigure(Display *dpy, Window w, GC gc,
    Visual *vis, int depth, Colormap cmap,
    int ox, int oy, int *pwid, int *pht,
    xp_fig_elem_t* target,
    int want, int use_private, int opt)
{
    double xscale=1.0, yscale=1.0;
    int xoff=0, yoff=0;
    fig_list_t *cur;
    int fc;
    GC ngc;
    int i;
    int wid, ht;
    Pixel black_color, white_color;

    XGCValues       gc_values;


    wid = *pwid;
    ht = *pht;


TracePush;

    Trace("called with %dx%d+%d+%d\n", wid, ht, ox, oy);
    Trace("target ulx, uly, lrx, lry = %d,%d,%d,%d\n",
        target->ulx, target->uly, target->lrx, target->lry);
    Trace("  so it is means wid %d, ht %d\n",
        target->lrx - target->ulx, target->lry - target->uly);

TracePop;

    /*
     * if width == 0, xscale is setted value of yscale.
     * if height == 0, yscale is setted value of xscale.
     * otherwise, xscale and yscale setted independent.
     */
    if(wid<=0 && ht<=0) {
        xscale = 1.0;
        yscale = 1.0;
    }
    else {
        if(ht<=0) {
            Trace("fit to width\n");
            xscale = ((double)wid)/(target->lrx - target->ulx + 1);
            yscale = xscale;
            *pht = (int)((double)(target->lry - target->uly)*yscale+1);
        }
        else
        if(wid<=0) {
            Trace("fit to height\n");
            yscale = ((double)ht)/(target->lry - target->uly + 1);
            xscale = yscale;
            *pwid = (int)((double)(target->lrx - target->ulx)*xscale+1);
        }
        else {
            Trace("fit to width and height\n");
            xscale = ((double)wid)/(target->lrx - target->ulx + 1);
            yscale = ((double)ht)/(target->lry - target->uly + 1);
        }
    }


    ngc = XCreateGC(dpy, w, 0, NULL);
    XCopyGC(dpy, gc, (unsigned long)(-1L), ngc);

    /* use black as default color */
    black_color = BlackPixel(dpy, DefaultScreen(dpy));

#if 0
    white_color = WhitePixel(XtDisplay(imgview_shell),
                            DefaultScreen(XtDisplay(imgview_shell)));
    XSetForeground(dpy, gc, white_color);

    if(opt & XP_OPT_FRAME) {
        XSetForeground(dpy, gc, 0);
        XDrawRectangle(dpy, w, gc, ox, oy, *pwid, *pht);
    }
#endif

    xoff = target->ulx;
    yoff = target->uly;

    Trace("then, scale is %.4f,%.4f // offset %d, %d\n",
        xscale, yscale, xoff, yoff);

    if(!target->cnbook) {
        Trace("color number book is not defined\n");
    }
    else {
        Trace("%d color(s) are defined in color number book (max %d).\n",
            target->cnbook->len, target->cnbook->max);
        for(i=0;i<target->cnbook->len;i++) {
            Trace(" %3d: %08x\n", i, target->cnbook->book[i]);
        }
    }

    ColorCacheStore(dpy, cmap);


    cur = target->list;
    fc = 0;


    while(cur) {
        Trace("\nfig# %d // %s [%d]\n",
            fc, FIG_OBJTYPESTR(FIG_OBJTYPE(cur->cont)), FIG_OBJTYPE(cur->cont));

        MEMORY_TRAP("before appling figure routines\n");

        XSetForeground(dpy, ngc, black_color);

        switch(FIG_OBJTYPE(cur->cont)) {
        case FIG_COLOR:
            Trace("skip color definition\n");
            break;

        case FIG_ELLIPSE:
            dfig_ellipse(dpy, w, ngc, vis, depth, cmap,
                want, use_private,
                ox, oy, xoff, yoff, xscale, yscale,
                target->cnbook, (void*)cur->cont);
            break;


        case FIG_SPLINE:
            dfig_spline(dpy, w, ngc, vis, depth, cmap,
                want, use_private,
                ox, oy, xoff, yoff, xscale, yscale,
                target->cnbook, (void*)cur->cont);
            break;

        case FIG_POLYLINE:
            dfig_polyline(dpy, w, ngc, vis, depth, cmap,
                want, use_private,
                ox, oy, xoff, yoff, xscale, yscale,
                target->cnbook, (void*)cur->cont);
            break;

        case FIG_TEXT:
            dfig_text(dpy, w, ngc, vis, depth, cmap,
                want, use_private,
                ox, oy, xoff, yoff, xscale, yscale,
                target->cnbook, (void*)cur->cont);
            break;

        default:
            Trace("not support such figure type [%d]\n",
                FIG_OBJTYPE(cur->cont));
            break;
        }

        MEMORY_TRAP("after  appling figure routines\n");

        cur = cur->next;
        fc++;
    }

    Trace("there are %d figures(s)\n", fc);

#if 1
    if(opt & XP_OPT_FRAME) {
        white_color = WhitePixel(dpy,
                                DefaultScreen(dpy));
        gc_values.function = GXxor;
        XChangeGC(dpy, ngc, GCFunction, &gc_values);
        XSetForeground(dpy, ngc, white_color);
        XDrawRectangle(dpy, w, ngc, ox, oy, *pwid, *pht);
    }
#endif

    XFreeGC(dpy, ngc);

}




Pixmap
Load_FIG(Display *dpy, Window win,
    GC gc, Visual *visual, int depth, Colormap cmap,
    char *fname, int *width, int *height, int want, int use_private, int opt)
{
    Pixmap ground;
    xp_fig_elem_t *afig;
    int chk;
    Pixel fc;
    extern xp_fig_elem_t* sort_depth_fig(xp_fig_elem_t *target);

    ground = None;

    Trace("Load_FIG: path '%s'\n", fname);
    Trace("  width %d height %d want %d private %d\n",
            *width, *height, want, use_private);

    MEMORY_TRAP("begging of Load_FIG\n");

    afig = read_fig_file(fname);
    if(afig==NULL) {
        return ground;
    }

#if 0
    show_fig_list(afig->list);
#endif

    MEMORY_TRAP("middle of Load_FIG, sorted\n");

    ground = XCreatePixmap(dpy, win, *width, *height, depth);
    if(ground==None) {
        return ground;
    }

    /*
     * fill background
     */
    if(use_private) {
#if 0
        ColorList(dpy, cmap, stderr);
#endif
        chk  = FindPixelPrivate(dpy, visual, cmap, "white", &fc);
    }
    else {
        chk  = FindPixel(dpy, visual, cmap, "white", &fc);
    }
    XSetForeground(dpy, gc, fc);

    XFillRectangle(dpy, ground, gc, 0, 0, *width, *height);

    if(use_private) {
#if 0
        ColorList(dpy, cmap, stderr);
#endif
        chk  = FindPixelPrivate(dpy, visual, cmap, "black", &fc);
    }
    else {
        chk  = FindPixel(dpy, visual, cmap, "black", &fc);
    }
    XSetForeground(dpy, gc, fc);


    MEMORY_TRAP("middle of Load_FIG, background is ready\n");


    DrawFigure(dpy, ground, gc, visual, depth, cmap,
        0, 0, width, height,
        afig, want, use_private, opt);

    MEMORY_TRAP("ending of Load_FIG\n");

    return ground;
}

Pixmap
Overwrite_FIG(Display *dpy, Window win, int offx, int offy,
    GC gc, Visual *visual, int depth, Colormap cmap,
    char *fname, int *width, int *height, int want, int use_private, int opt)
{
    Pixmap ground;
    xp_fig_elem_t *afig;
    int chk;
    Pixel fc;
    extern xp_fig_elem_t* sort_depth_fig(xp_fig_elem_t *target);

    ground = win;

    Trace("Overwrite_FIG: path '%s'\n", fname);
    Trace("  width %d height %d want %d private %d\n",
            *width, *height, want, use_private);

    afig = read_fig_file(fname);
    if(afig==NULL) {
        return ground;
    }

#if 0
    show_fig_list(afig->list);
#endif

    if(use_private) {
#if 0
        ColorList(dpy, cmap, stderr);
#endif
        chk  = FindPixelPrivate(dpy, visual, cmap, "black", &fc);
    }
    else {
        chk  = FindPixel(dpy, visual, cmap, "black", &fc);
    }
    XSetForeground(dpy, gc, fc);

    DrawFigure(dpy, win, gc, visual, depth, cmap,
        offx, offy, width, height,
        afig, want, use_private, opt);

    return ground;
}

