/* 
 * fsave.c - file saving routines
 *
 * $Id$
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <syslog.h>
#ifdef TIME_WITH_SYS_TIME
 #include <sys/time.h>
 #include <time.h>
#else
 #ifdef HAVE_SYS_TIME_H
  #include <sys/time.h>
 #else
  #include <time.h>
 #endif
#endif
#include <errno.h>

#include "alist.h"
#include "sentry.h"
#include "img.h"
#include "pool.h"
/*
#include "client.h"
*/

/*
 * NOTE:
 *      1 day = 86400sec = 1440min
 *
 *  frame rate      sum
 *  (frame/sec)     (frame/day)
 *     0.1             8640
 *       1            86400
 *      30 (max)    2592000
 *      
 */

/*
    type
    0   prefix/yy-mm-dd/HH:MM".sis"
    1   prefix/w/HH/MM".sis"
    >1  prefix/n/HH/MM".sis"
 */


char*
conv_filepath(char *dst, char *prefix, int type, time_t reftime)
{
    char tmp[BUFSIZ];
    struct tm *tmbuf;

    tmbuf = localtime(&reftime);
    
    if(type==0) {
        sprintf(tmp, "%04d-%02d-%02d/%02d:%02d.sis",
            tmbuf->tm_year+1900, tmbuf->tm_mon+1, tmbuf->tm_mday,
            tmbuf->tm_hour, tmbuf->tm_min);
    }
    else
    if(type==1) {
        sprintf(tmp, "%d/%02d/%02d.sis",
            tmbuf->tm_wday, tmbuf->tm_hour, tmbuf->tm_min);
    }
    else {
        sprintf(tmp, "%d/%02d/%02d.sis",
            (reftime/86400)%type, tmbuf->tm_hour, tmbuf->tm_min);
    }

    strcpy(dst, prefix);
    strcat(dst, "/");
    strcat(dst, tmp);

    if(debug>=1000) {
        fprintf(stderr, "target path '%s'\n", dst);
    }

    return dst;
}

int
check_path(char *path)
{
    char tmppath[BUFSIZ];
    register char *q, *p;
    int chk;

    q = tmppath;
    p = path;

    fprintf(stderr, "check_path: '%s'\n", path);

    while(*p) {
        while(*p && *p!='/') {
/*
            fprintf(stderr, "  %2d - '%c' \"%s\"\n", p-path, *p, p);
*/
            *q++ = *p++;
        }
        *q = '\0';

/*
            fprintf(stderr, "  %2d ? '%c' \"%s\"\n", p-path, *p, p);
*/
        if(*p!='/') {
            break;
        }


        if(tmppath[0]) {

        chk = mkdir(tmppath, 0775);
        if(chk<0) {
            if(errno==EEXIST) {
            }
            else {
                fprintf(stderr, "mkdir: path '%s', chk %d, errno %d\n",
                    tmppath, chk, errno);

#if 0
                return -1;
#endif
            }
        }

        fprintf(stderr, "  mkdir '%s' %d\n", tmppath, chk);

        }

        if(*p=='/') {
            *q++ = *p++;
        }
    }

    return 0;
}



static char last_name[BUFSIZ]="";
static int last_time = 0;
static int wfd = -1;

int
write_out(char *prefix, int pathtype, sentry_imgrec *himg)
{
    static int magic_len=0;
    char *magic="SENTRY";
    int ret=0;
    char curpath[BUFSIZ];
    unsigned char header_padding[SENTRY_SHM_HEADER_SIZE];
    sentry_imgrec *nimg; /* network oriented */
    int slen;
    int nw, tnw;
    int tv, tl;
    int chk;
    struct stat sb;

    memset(header_padding, 0, SENTRY_SHM_HEADER_SIZE);
    nimg = (sentry_imgrec*) header_padding;

    if(magic_len<=0) {
        magic_len = strlen(magic);
    }

    if(debug>=100) {
        show_imgattr(stderr, himg);
    }


    nw = 0;

    hton_img(nimg, himg);
    nimg->body_len = 0;
    nimg->body     = NULL;

#if 0
    /* unreadble data (because it is not host order) */
    if(debug>=100) {
        show_imgattr(stderr, nimg);
    }
#endif
    
    conv_filepath(curpath, prefix, pathtype, himg->recv_sec);

    if(debug>=10) {
        fprintf(stderr, "path '%s'\n", curpath);
    }


    chk = strcmp(curpath, last_name);
    fprintf(stderr, "path '%s' <last '%s'> same? %d\n",
        curpath, last_name, chk);

    if(chk!=0) {
        check_path(curpath);

        if(wfd>=0) {
            close(wfd);
            wfd = -1;
        }

        chk = stat(curpath, &sb);
        if(chk<0) {
            if(errno==ENOENT) {
                goto pass_stat;
            }
            fprintf(stderr, "fail stat() '%s' (%d)\n", curpath, errno);
        }
        else {
            if(sb.st_mtime < himg->recv_sec - 60) {
                if(debug>=10) {
                    fprintf(stderr,
                        "delete old file '%s' modified %d sec before.\n",
                        curpath,  himg->recv_sec - sb.st_mtime ); 
                }
                chk = unlink(curpath);
                if(chk) {
                    fprintf(stderr,
                        "fail unlink() '%s' (%d)\n", curpath, errno);
                }
            }
        }
pass_stat:

        wfd = open(curpath, O_WRONLY|O_CREAT, 0644);

        if(wfd<0) {
            fprintf(stderr, "write_out: fail to open '%s' (%d)\n", 
                curpath, errno);
        }
    }

    chk = lseek(wfd, 0, SEEK_END);

#if 0
    slen = length(magic) + sizeof(int32_t) + sizeof(int32_t)
                + sizeof(sentry_imgrec) + himg->length;
#endif
#if 0
    slen = sizeof(sentry_imgrec) + himg->length;
#endif
    slen = SENTRY_SHM_HEADER_SIZE + himg->length;

try_write1:
    tnw = write(wfd, magic, magic_len);
    if(tnw<0) {
        fprintf(stderr, "write_out: fail to write magic for '%s' (%d)\n",
            curpath, errno);
        if(errno==EINTR) {
            goto try_write1;
        }
        goto fail;
    }
    nw += tnw;

    tv = htonl(SENTRY_SHM_VERSION);

try_write2:
    tnw = write(wfd, &tv, sizeof(int32_t));
    if(tnw<0) {
        fprintf(stderr, "write_out: fail to write version for '%s' (%d)\n",
            curpath, errno);
        if(errno==EINTR) {
            goto try_write2;
        }
        goto fail;
    }
    nw += tnw;

    tl = htonl(slen);
    fprintf(stderr, "tl %d %x, slen %d %x\n", tl, tl, slen, slen);
    fflush(stderr);

try_write3:
    tnw = write(wfd, &tl, sizeof(int32_t));
    if(tnw<0) {
        fprintf(stderr, "write_out: fail to write length for '%s' (%d)\n",
            curpath, errno);
        if(errno==EINTR) {
            goto try_write3;
        }
        goto fail;
    }
    nw += tnw;

try_write4:
#if 0
    tnw = write(wfd, nimg, sizeof(sentry_imgrec));
#endif
    tnw = write(wfd, nimg, SENTRY_SHM_HEADER_SIZE);
    if(tnw<0) {
        fprintf(stderr, "write_out: fail to write header for '%s' (%d)\n",
            curpath, errno);
        if(errno==EINTR) {
            goto try_write4;
        }
        goto fail;
    }
    nw += tnw;

try_write5:
    tnw = write(wfd, himg->body, himg->length);
    if(tnw<0) {
        fprintf(stderr, "write_out: fail to write body for '%s' (%d)\n",
            curpath, errno);
        if(errno==EINTR) {
            goto try_write5;
        }
        goto fail;
    }
    nw += tnw;


    goto success;

fail:
    


success:
    strcpy(last_name, curpath); 

    
out:
    return ret;
}



