/*
 *
 * $Id$
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#ifdef TIME_WITH_SYS_TIME
 #include <sys/time.h>
 #include <time.h>
#else
 #ifdef HAVE_SYS_TIME_H
  #include <sys/time.h>
 #else
  #include <time.h>
 #endif
#endif
#include <errno.h>

#include "alist.h"
#include "img.h"
#include "sentry.h"
/*
#include "client.h"
*/

apair_t matching_type_alist[] = {
    {"none",    SEN_M_NONE},
    {"tattr",   SEN_M_TATTR},
    {"time",    SEN_M_TATTR},
    {"iattr",   SEN_M_IATTR},
    {"1point",  SEN_M_1POINT},
    {"1",       SEN_M_1POINT},
    {"center",  SEN_M_1POINT},
    {"4points", SEN_M_4POINTS},
    {"4",       SEN_M_4POINTS},
    {"9points", SEN_M_9POINTS},
    {"9",       SEN_M_9POINTS},
    {"all",     SEN_M_ALL},
    {NULL,      -1},
    };


int
show_imgattr(FILE* fp, sentry_imgrec *ximg)
{
    fprintf(fp, "addr %p, time i %d.%06d r %d.%06d %s %s %d/%d %dx%d %p\n",
        ximg, ximg->send_sec, ximg->send_usec,
        ximg->recv_sec, ximg->recv_usec,
        assoc_v(format_alist, ximg->format),
        assoc_v(color_alist, ximg->corder),
        ximg->length, ximg->body_len, ximg->width, ximg->height,
        ximg->body);
    fflush(fp);
}

int
ntoh_img(sentry_imgrec *himg, sentry_imgrec *nimg)
{
    himg->send_sec   = ntohl(nimg->send_sec);
    himg->send_usec = ntohl(nimg->send_usec);
    himg->recv_sec   = ntohl(nimg->recv_sec);
    himg->recv_usec = ntohl(nimg->recv_usec);
    himg->length    = ntohl(nimg->length);
    himg->format    = ntohl(nimg->format);
    himg->corder    = ntohl(nimg->corder);
    himg->width     = ntohl(nimg->width);
    himg->height    = ntohl(nimg->height);
/*
    himg->body_len  = ntohl(0);
    himg->body      = (uint8_t*)0;
*/
}

int
hton_img(sentry_imgrec *nimg, sentry_imgrec *himg)
{
    nimg->send_sec   = htonl(himg->send_sec);
    nimg->send_usec = htonl(himg->send_usec);
    nimg->recv_sec   = htonl(himg->recv_sec);
    nimg->recv_usec = htonl(himg->recv_usec);
    nimg->length    = htonl(himg->length);
    nimg->format    = htonl(himg->format);
    nimg->corder    = htonl(himg->corder);
    nimg->width     = htonl(himg->width);
    nimg->height    = htonl(himg->height);
/*
    nimg->body_len  = htonl(0);
    nimg->body      = (uint8_t*)0;
*/
}

int
copyattr_img(sentry_imgrec *himg, sentry_imgrec *nimg)
{
    himg->send_sec  = (nimg->send_sec);
    himg->send_usec = (nimg->send_usec);
    himg->recv_sec  = (nimg->recv_sec);
    himg->recv_usec = (nimg->recv_usec);
    himg->length    = (nimg->length);
    himg->format    = (nimg->format);
    himg->corder    = (nimg->corder);
    himg->width     = (nimg->width);
    himg->height    = (nimg->height);
}


#define CMP(x,y,f)  if(x->f!=y->f) {return x->f - y->f;}

int
img_diff_iattr(sentry_imgrec *a, sentry_imgrec *b)
{
    CMP(a,b,format);
    if(a->format==SEN_F_RGB) {
        CMP(a,b,corder);
    }
    CMP(a, b, width);
    CMP(a, b, height);
    CMP(a, b, length);

    return 0;
}

static
int
pixel_diff(sentry_imgrec *a, sentry_imgrec *b, int pos, int nb, int range)
{
    int av, bv;
    int i;
    int omit;

    omit = 0;

    for(i = pos;i<pos+nb;i++) {
        av = a->body[i];
        bv = b->body[i];

        if(av-bv>range || bv-av>range) {
            omit = 1;
        }   
        if(debug>=1000) {
            fprintf(stderr, "av %x, bv %x, diff %d, omit %d\n", 
                av, bv, av-bv, omit);
        }
        if(omit)
            break;
    }

    if(debug>=500) {
        fprintf(stderr, "  pixel_diff: pos %d, nb %d, range %d => omit %d\n",
            pos, nb, range, omit);
    }

    return omit;
}


int
img_diff_1pixel(sentry_imgrec *a, sentry_imgrec *b, int range)
{
    int pos;
    int nb=0;
    int omit;

    CMP(a,b,format);
    if(a->format==SEN_F_RGB) {
        CMP(a,b,corder);
    }
    CMP(a, b, width);
    CMP(a, b, height);
    CMP(a, b, length);

    if(a->format==SEN_C_RGB || a->format==SEN_C_BGR) {
        nb = 3;
    }
    else {
        nb = 4;
    }

    pos = (a->width*(a->height/2) + a->width/2)* nb;
    omit = pixel_diff(a, b, pos, nb, range);

    return omit;
}

int
img_diff_4pixel(sentry_imgrec *a, sentry_imgrec *b, int range)
{
    int pos;
    int nb=0;
    int omit;

    CMP(a,b,format);
    if(a->format==SEN_F_RGB) {
        CMP(a,b,corder);
    }
    CMP(a, b, width);
    CMP(a, b, height);
    CMP(a, b, length);

    if(a->format==SEN_C_RGB || a->format==SEN_C_BGR) {
        nb = 3;
    }
    else {
        nb = 4;
    }

    omit = 0;

    pos = (a->width*(a->height/3) + a->width/3)* nb;
    omit += pixel_diff(a, b, pos, nb, range);
    pos = (a->width*(a->height/3) + a->width*2/3)* nb;
    omit += pixel_diff(a, b, pos, nb, range);
    pos = (a->width*(a->height*2/3) + a->width/3)* nb;
    omit += pixel_diff(a, b, pos, nb, range);
    pos = (a->width*(a->height*2/3) + a->width*2/3)* nb;
    omit += pixel_diff(a, b, pos, nb, range);

    return omit;
}

int
img_diff_9pixel(sentry_imgrec *a, sentry_imgrec *b, int range)
{
    int pos;
    int nb=0;
    int omit;

    CMP(a,b,format);
    if(a->format==SEN_F_RGB) {
        CMP(a,b,corder);
    }
    CMP(a, b, width);
    CMP(a, b, height);
    CMP(a, b, length);

    if(a->format==SEN_C_RGB || a->format==SEN_C_BGR) {
        nb = 3;
    }
    else {
        nb = 4;
    }

    omit = 0;

    pos = (a->width*(a->height*1/4) + a->width*1/4)* nb;
    omit += pixel_diff(a, b, pos, nb, range);
    pos = (a->width*(a->height*1/4) + a->width*2/4)* nb;
    omit += pixel_diff(a, b, pos, nb, range);
    pos = (a->width*(a->height*1/4) + a->width*3/4)* nb;
    omit += pixel_diff(a, b, pos, nb, range);

    pos = (a->width*(a->height*2/4) + a->width*1/4)* nb;
    omit += pixel_diff(a, b, pos, nb, range);
    pos = (a->width*(a->height*2/4) + a->width*2/4)* nb;
    omit += pixel_diff(a, b, pos, nb, range);
    pos = (a->width*(a->height*2/4) + a->width*3/4)* nb;
    omit += pixel_diff(a, b, pos, nb, range);

    pos = (a->width*(a->height*3/4) + a->width*1/4)* nb;
    omit += pixel_diff(a, b, pos, nb, range);
    pos = (a->width*(a->height*3/4) + a->width*2/4)* nb;
    omit += pixel_diff(a, b, pos, nb, range);
    pos = (a->width*(a->height*3/4) + a->width*3/4)* nb;
    omit += pixel_diff(a, b, pos, nb, range);

    return omit;
}

typedef struct {
    int x;
    int y;
} pos_xy_t;

int
img_diff_npos(sentry_imgrec *a, sentry_imgrec *b, int npos, pos_xy_t *geo, int range)
{
    int pos;
    int nb=0;
    int omit;
    int i, c;

    CMP(a,b,format);
    if(a->format==SEN_F_RGB) {
        CMP(a,b,corder);
    }
    CMP(a, b, width);
    CMP(a, b, height);
    CMP(a, b, length);

    if(a->format==SEN_C_RGB || a->format==SEN_C_BGR) {
        nb = 3;
    }
    else {
        nb = 4;
    }

    omit = 0;

    for(i=0;i<npos;i++) {
        pos = (a->width*geo[i].y + geo[i].x) * nb;
        if(pos>a->length) {
            fprintf(stderr, "out of range; %d in %d\n", pos, a->length);
        }
        else {
            omit += pixel_diff(a, b, pos, nb, range);
        }
    }

    return omit;
}


/*
 * compare image like strcmp().
 * If 0 is returned, it means same or very simulare, except SEN_M_NONE.
 * When SEN_M_NONE, this routine always returns 1.
 */

int
img_cmp(sentry_imgrec *a, sentry_imgrec *b, int type, void* opt)
{
    int ret;

    if(type==SEN_M_NONE) {
        ret = 1;
    }
    else {
        switch(type) {
        case SEN_M_IATTR:
            ret = img_diff_iattr(a, b);
            break;
        case SEN_M_1POINT:
            ret = img_diff_1pixel(a, b, *(int*)opt);
            break;
        case SEN_M_4POINTS:
            ret = img_diff_4pixel(a, b, *(int*)opt);
            break;
        case SEN_M_9POINTS:
            ret = img_diff_9pixel(a, b, *(int*)opt);
            break;
        }

#if 0
        if(ret && debug>=10) {
            fprintf(stderr, "img_cmp: diffrent image.\a\a\a\n");
        }
#else
        if(ret && debug>=10) {
            fprintf(stderr, "img_cmp: diffrent image.\n");
        }
#endif
    }

    return ret;
}
