/*==
 ==*/
/*
 * jm.c - job manager
 *                  by k-chinen@is.aist-nara.ac.jp 2003.
 *
 * these routines are wriiten from scratch.
 * however, the idea of them is given from Mr. Arikawa.
 *  
 * $Id$
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <poll.h>
#include <netinet/in.h>
#include <netdb.h>
#include <signal.h>
#ifdef __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#include <errno.h>


#include "smisc.h"
#include "alist.h"
#include "img.h"
#include "pool.h"
#include "client.h"
#include "nv.h"
#include "sf.h"
#include "jm.h"

int jm_errno=JM_ENONE;



#if 0
int
chomp(char *line)
{
    char *p;
    if(line==NULL)
        return -1;
    p = line;
    while(*p) {
        if(*p=='\r'||*p=='\n') {
            *p='\0';
            return 0;
        }
        p++;
    }
    return 1;
}
int
parse_cand_list(char *cl)
{
    int c;
    char *p;

    c = 0;
    p = cl;

    while(*p) {
        while(*p && (*p==' ' || *p=='\t' || *p=='\r' || *p=='\n')) {
            p++;
        }
        if(*p && !(*p==' ' || *p=='\t' || *p=='\r' || *p=='\n')) {
            c++;
            while(*p && !(*p==' ' || *p=='\t' || *p=='\r' || *p=='\n')) {
                p++;
            }
        }
    }

    return c;
}
#endif


#ifdef __STDC__
int fdprintf(const int fd, const char *fmt, ...);
#else
int fdprintf(va_alist);
#endif

extern int shape_factor;

extern char *rm_host;
extern int rm_port;
extern int rm_fd;

#ifdef __STDC__
int send_cmd(const int fd, const char *fmt, ...)
#else
int send_cmd(va_alist)
va_dcl
#endif
{
    va_list args;
#ifndef __STDC__
    int fd;
    char *fmt;
#endif
    char buffer[BUFSIZ];
    int nwrite;

#ifdef __STDC__
    va_start(args,fmt);
#else
    va_start(args);
    fd  = va_arg(args, int);
    fmt = va_arg(args, char *);
#endif

#ifdef HAVE_VSNPRINTF
    vsnprintf(buffer, BUFSIZ, fmt, args);
#else
    vsprintf(buffer, fmt, args);
#endif
#if 1
    fprintf(stderr, "<<<%s", buffer);
#endif
try:
    nwrite = write(fd, buffer, strlen(buffer));
    if(nwrite<0) {
        if(errno==EINTR) {
            goto try;
        }
        else {
            fprintf(stderr, "send_cmd: fail write() fd# %d (%d)\n", fd, errno);
        }
    }

    va_end(args);

    return nwrite;
}

/* XXX */
int
recv_response(int fd, int *rcode, char *resbuf, int reslen)
{
    char linebuf[BUFSIZ];
    char numbuf[5];
    int nread;
    char *p, *q;
    int c;

    *rcode = -1;

        memset(linebuf, 0, BUFSIZ);
    nread = read(fd, linebuf, BUFSIZ-1);
    if(nread<0) {
        fprintf(stderr, "fail read fd# %d (%d)\n", fd, errno);
        return -1;
    }
    if(nread==0) {
#if 1
        fprintf(stderr, "reach end fd# %d\n", fd);
#endif
        return 0;
    }
    linebuf[nread] = '\0';
#if 1
    fprintf(stderr, ">>>%s", linebuf);
#endif

    c = 0;
    p = linebuf;
    q = numbuf;
    while(c<4 && *p>='0' && *p<='9') {
        *q++ = *p++;
        c++;
    }
    *q = '\0';
    if(*p==' ') {
        p++;
    }

    c = 0;
    q = resbuf;
    while(c<reslen && *p) {
        *q++ = *p++;
        c++;
    }
    *q = '\0';

#if 1
    fprintf(stderr, "res code '%s'\n", numbuf);
    fprintf(stderr, "    line '%s'\n", resbuf);
#endif

    *rcode = atoi(numbuf);
    
    return nread;
}


#if 0
apair_t filter_name_al[] = {
    {"thru",    FILTER_THRU},
    {"a3",      FILTER_A3},
    {"gray",    FILTER_GRAY},
    {"flop",    FILTER_FLOP},
    {"flip",    FILTER_FLIP},
    {"diff",    FILTER_HIST},
    {"edge",    FILTER_EDGE},
    {"a1",      FILTER_A1},
    {"a2",      FILTER_A2},
    {NULL,      -1},
};
#endif


int
rotate_filtermode(int *rcm)
{
    int cm;
    apair_t *cp, *pp;
    int i, nc, ne, nn;

    cm = *rcm;

    i = 0;
    cp = NULL;
    pp = filter_name_al;
    while(pp->name) {
        if(pp->value==cm) {
            cp = pp;
            nc = i;
        }
        i++;
        pp++;
    }
    ne = i;

    if(!cp) {
        return -1;
    }

    if(nc+1==ne) {
        nn = 0;
    }
    else {  
        nn = nc+1;
    }
    *rcm = filter_name_al[nn].value;

    return 0;
}


int
isfree_node(node_t *this)
{
    int i;
    int cu, cl;

    cu = cl = 0;
    for(i=0;i<MAX_LOWERNODE;i++) {
        if(this->unid[i]) {
            cu++;
        }
        if(this->lnid[i]) {
            cl++;
        }
#if 0
        if(this->Hup[i]) {
            cu++;
        }
        if(this->Hlo[i]) {
            cl++;
        }
#endif
    }

    if(cu==0 && cl==0) {
        return 1;
    }
    else {
        return 0;
    }
}



count_nodes(node_t *nl)
{
    node_t *p;
    int c;

    p = nl;
    c = 0;
    while (p) {
        c++;
        p = p->next;
    }

    return c;
}

int
count_undef_nodes(node_t *nl)
{
    node_t *p;
    int c;

    p = nl;
    c = 0;
    while (p) {
        if(!p->host || p->host[0]=='\0') {
            c++;
        }
        p = p->next;
    }

    return c;
}

int
count_need_nodes(node_t *nl)
{
    node_t *p;
    int c;

    p = nl;
    c = 0;
    while (p) {
        if(isfree_node(p)) {
            /* nothing */
        }
        else {
            c++;
        }
        p = p->next;
    }

    return c;
}

int
count_needundef_nodes(node_t *nl)
{
    node_t *p;
    int c;

    p = nl;
    c = 0;
    while (p) {
        if(isfree_node(p)) {
            /* nothing */
        }
        else
        if(!p->host || p->host[0]=='\0') {
            c++;
        }
        p = p->next;
    }

    return c;
}

int
store_cid(char **dest, char *src)
{
    char tmp[BUFSIZ];
    char *p, *q;

    p = src;
    q = tmp;
    while(*p && *p!=' ' && *p!='\t' && *p!='\r' && *p!='\n') {
        *q++ = *p++;
    }
    *q = '\0';

    *dest = strdup(tmp);
    
    return 0;
}


int 
query_nodes(char *req_host, int req_port, int need_node, char *tmpnode)
{
    int i;
    int sock, len, n, ret;
    char buffer[BUFSIZ];
    char buf_num[4];
    char *sep=" \n";
    int chk;

    sock = connect_host(req_host, req_port, SOCK_STREAM);
    if(sock<0) {
        fprintf(stderr, "fail connect to '%s':%d\n", req_host, req_port);
        jm_errno = JM_ENORM;
        return -1;
    }

    n = send_cmd(sock, "GET %d\r\n", need_node);
    if (n < 0) {
        fprintf(stderr, "Error: do not write, GET\n");
        return -1;
    }
    fflush(stdout);

    chk = recv_response(sock, &n, buffer, BUFSIZ);

    switch (n) {
    case 201:
        strcpy(tmpnode, buffer);
        len = strlen(buffer);
        break;
    case 400:
        /* XXX */
        jm_errno = JM_ETOOMANYNEED;
    default:
        len = -1;
        break;
    }

    close(sock);

    return len;
}

int
jm_filter_wake_and_open(node_t *this)
{
    int xfd;
    int rcode;
    char resbuf[BUFSIZ];
    int n;
    int chk;

    fprintf(stderr,
        "jm_filter_wake_and_open: start '%s' '%s'\n",
        this->name, this->host);

    fprintf(stderr,
        "  @ id# %d, c/filter wake + open\n", this->id, this->type);

    if(this->sitter_fd>=0) {
        xfd = this->sitter_fd;
        goto skip_connect_wake;
    }

    xfd = -1;

    xfd = connect_host(this->host, this->sitter_port, SOCK_STREAM);
    if(xfd<0) {
        fprintf(stderr, "cannot connect id# %d, name '%s', port %d\n",
            this->id, this->host, this->sitter_port);
        this->sitter_fd = -1;
        return -1;
    }


    chk = isreadable_via_select(xfd, 0, 200000);
    if (chk>=0) {
        chk = recv_response(xfd, &rcode, resbuf, BUFSIZ);
        if(chk<=0) {
            fprintf(stderr, "error or closed\n");
            goto close_out;
        }
        if(chk>0 && rcode!=100) {
            fprintf(stderr, "unexpected response %d '%s'\n",
                rcode, resbuf);
            goto close_out;
        }
    }

    this->sitter_fd = xfd;

#if 1

    n = send_cmd(xfd, "DEBUG\r\n");
    if (n < 0) {
        /* Error */
        fprintf(stderr, "fail cmd WAKE\n");
        goto close_out;
    }

    chk = recv_response(xfd, &rcode, resbuf, BUFSIZ);
    if(chk<=0) {
        fprintf(stderr, "error or closed\n");
        goto close_out;
    }
    if(chk>0 && rcode!=200) {
        fprintf(stderr, "unexpected response %d '%s'\n",
            rcode, resbuf);
#if 0
        goto close_out;
#endif
    }
    if(rcode==200) {
        this->dlcid = NULL;
    }

#endif


    n = send_cmd(xfd, "WAKE %s\r\n", this->prog_name);
    if (n < 0) {
        /* Error */
        fprintf(stderr, "fail cmd WAKE\n");
        goto close_out;
    }

    chk = isreadable_via_select(xfd, 0, 200000);
    if (chk>=0) {
        chk = recv_response(xfd, &rcode, resbuf, BUFSIZ);
        if(chk<=0) {
            fprintf(stderr, "error or closed\n");
            goto close_out;
        }
        if(chk>0 && rcode!=100) {
            fprintf(stderr, "unexpected response %d '%s'\n",
                rcode, resbuf);
            goto close_out;
        }
    }

skip_connect_wake:
    if(this->dlcid) {
        goto skip_open;

#if 0
    n = send_cmd(xfd, "CLOSE %s\r\n", this->dlcid);
    if (n < 0) {
        /* Error */
        fprintf(stderr, "fail cmd OPEN\n");
        goto close_out;
    }

    chk = recv_response(xfd, &rcode, resbuf, BUFSIZ);
    if(chk<=0) {
        fprintf(stderr, "error or closed\n");
        goto close_out;
    }
    if(chk>0 && rcode!=200) {
        fprintf(stderr, "unexpected response %d '%s'\n",
            rcode, resbuf);
        goto close_out;
    }
    if(rcode==200) {
        this->dlcid = NULL;
    }

#endif

    }

    n = send_cmd(xfd, "SHAPE %d\r\n", shape_factor);
    if (n < 0) {
        /* Error */
        fprintf(stderr, "fail cmd SHAPE\n");
        goto close_out;
    }

    chk = recv_response(xfd, &rcode, resbuf, BUFSIZ);
    if(chk<=0) {
        fprintf(stderr, "error or closed\n");
        goto close_out;
    }
    if(chk>0 && rcode!=200) {
        fprintf(stderr, "unexpected response %d '%s'\n",
            rcode, resbuf);
    }


    n = send_cmd(xfd, "OPEN %s %d\r\n", this->host, this->down_listen_port);
    if (n < 0) {
        /* Error */
        fprintf(stderr, "fail cmd OPEN\n");
        goto close_out;
    }

    chk = recv_response(xfd, &rcode, resbuf, BUFSIZ);
    if(chk<=0) {
        fprintf(stderr, "error or closed\n");
        goto close_out;
    }
    if(chk>0 && rcode!=201) {
        fprintf(stderr, "unexpected response %d '%s'\n",
            rcode, resbuf);
        goto close_out;
    }
    if(rcode==201) {
        store_cid(&this->dlcid, resbuf);
    }

skip_open:
    

#if 0
    n = send_cmd(xfd, "LIST\r\n", this->host, this->down_listen_port);
    if (n < 0) {
        /* Error */
        fprintf(stderr, "fail cmd OPEN\n");
        goto close_out;
    }

    chk = recv_response(xfd, &rcode, resbuf, BUFSIZ);
    if(chk<=0) {
        fprintf(stderr, "error or closed\n");
        goto close_out;
    }
#if 0
    if(chk>0 && rcode!=100) {
        fprintf(stderr, "unexpected response %d '%s'\n",
            rcode, resbuf);
        goto close_out;
    }
#endif
#endif

    goto out;

close_out:
    close(this->sitter_fd);
    this->sitter_fd = -1;

out:

    return 0;
}

int
jm_camera_wake_and_open(node_t *this)
{
    return jm_filter_wake_and_open(this);
}

int
jm_camera_kick(node_t *this)
{
    fprintf(stderr,
        "jm_camera_kick: start '%s' '%s'\n",
        this->name, this->host);

    fprintf(stderr,
        "  @ id# %d, camera *skip*\n", this->id);

    return 0;
}

#if 0
int
jm_viewer_kick(node_t *this)
{
    int i;
    node_t *cur;

    fprintf(stderr,
        "jm_viewer_kick: start '%s' '%s'\n",
        this->name, this->host);

    for(i=0;i<MAX_LOWERNODE;i++) {
        if(!this->unid[i]) {
            continue;
        }
        cur = this->unid[i];
        fprintf(stderr,
            "  @ id# %d, viewer connect to id# %d\n", this->id, cur->id);
    }

    return 0;
}
#endif

int
jm_filter_kick(node_t *this)
{
    int rcode;
    char resbuf[BUFSIZ];
    int n;
    int chk;
    int i;
    node_t *cur;
    int c;
    int ctrlfd;

    fprintf(stderr,
        "jm_filter_kick: start '%s' '%s'\n",
        this->name, this->host);

#if 0
    if(this->host==NULL || this->host[0]=='\0') {
        return;
    }
#endif

    if(this->type==NV_T_FILTER||this->type==NV_T_SCANNER) {
        fprintf(stderr,
            "  @ id# %d, v/filter filter -\n", this->id);

        if(this->sitter_fd<0) {
            fprintf(stderr, " sitter is not connected <%d>\n",
                this->sitter_fd);
            goto out;
        }

        chk = isreadable_via_select(this->sitter_fd, 0, 200000);
        if(chk) {
            chk = recv_response(this->sitter_fd, &rcode, resbuf, BUFSIZ);
            if(chk<=0) {
                fprintf(stderr, "error or closed\n");
                goto out;
            }
        }

        chk = send_cmd(this->sitter_fd, "FILTER %s\r\n",
                assoc_v(filter_name_al, this->filtermode));

        chk = recv_response(this->sitter_fd, &rcode, resbuf, BUFSIZ);
        if(chk<=0) {
            fprintf(stderr, "error or closed\n");
            goto out;
        }   

    }

    for(i=0;i<MAX_LOWERNODE;i++) {
#if 0
        fflush(stdout);
        fprintf(stderr, " ? %d %p\n", i, this->unid[i]);
        fflush(stderr);
#endif
        if(!this->unid[i]) {
            continue;
        }
        cur = this->unid[i];
        fprintf(stderr,
            "  @ id# %d, v/filter connect to id# %d\n", this->id, cur->id);

        if(this->type==NV_T_VIEWER) {
            continue;
        }

   
        chk = send_cmd(this->sitter_fd, "CONNECT %s %d\r\n",
                cur->host, cur->down_listen_port);
   
        chk = recv_response(this->sitter_fd, &rcode, resbuf, BUFSIZ);
        if(chk<=0) {
            fprintf(stderr, "error or closed\n");
            continue;
        }  
        if(rcode==201) {
            store_cid(&this->ucid[i], resbuf);
        }

    }

#if 1
    if(this->type==NV_T_VIEWER) {
        /* support single upper node */
        cur = NULL;
        c = 0;
        for(i=0;i<MAX_LOWERNODE;i++) {
            if(this->unid[i]) {
                cur = this->unid[i];
                c++;
            }
        }

        if(c>=1 && cur) {
            ctrlfd = build_ctrl_port(cur->host,
                        cur->down_listen_port);
            if (ctrlfd < 0) {
                fprintf(stderr,
                    "fail setup control fd# %d with port# %d\n",
                    ctrlfd, this->down_listen_port);
            }

            v_start_request(this, ctrlfd);
        }
    }
#endif

out:
    return 0;
}

int
jm_viewer_kick(node_t *this)
{
    return jm_filter_kick(this);
}

jm_viewer_stop(node_t *this)
{
    node_t *upp;
    int i, j;

    if(!this) {
        return -1;
    }

    fprintf(stderr, "jm_viewer_stop: v id# %d\n", this->id);

    v_stop_request(this);

#if 0
    for(i=0;i<MAX_LOWERNODE;i++) {
        if(this->unid[i]) {
            fprintf(stderr, " <%d: %p %p %s>",
                i, this->unid[i], this->Hup[i], this->ucid[i]);
        }
    }
    fprintf(stderr, "\n");
#endif

    for(i=0;i<MAX_LOWERNODE;i++) {
        if(this->unid[i]) {
            this->Hup[i] = NULL;
            if(this->ucid[i]) {
                free(this->ucid[i]);
                this->ucid[i] = NULL;
            }

#if 0
            fprintf(stderr, "  id# %d = ", this->unid[i]->id);
#endif

            for(j=0;j<MAX_LOWERNODE;j++) {
                if(this->unid[i]->lnid[j]) {
#if 0
                    fprintf(stderr, " [%d: %p %p %s]",
                        i, this->unid[i]->lnid[j],
                        this->unid[i]->Hlo[j], this->unid[i]->lcid[j]);
#endif
                    if(this->unid[i]->lnid[j]==this) {
                        this->unid[i]->Hlo[j] = NULL;
                    }
                }
            }
#if 0
            fprintf(stderr, "\n");
#endif
        }
    }
}

int
jm_camera_stop(node_t *this)
{
    int n;
    int rcode;
    char resbuf[BUFSIZ];
    int chk;
    int i, j;

    if(!this) {
        return -1;
    }
    if(this->sitter_fd<0) {
        return -1;
    }

    fprintf(stderr, "jm_camera_stop: c/f id# %d, dlcid '%s'\n",
        this->id, this->dlcid);
    if(this->dlcid) {

        n = send_cmd(this->sitter_fd, "CLOSE %s\r\n", this->dlcid);
        if(n<0) {
            fprintf(stderr, "error on CLOSE to camera\n");
            goto close_out;
        }
        
        chk = recv_response(this->sitter_fd, &rcode, resbuf, BUFSIZ);
        if(chk<=0) {
            fprintf(stderr, "error or closed\n");
            goto close_out;
        }
        if(chk>0 && rcode!=200) {
            fprintf(stderr, "unexpected response %d '%s'\n",
                rcode, resbuf);
            goto close_out;
        }

        for(i=0;i<MAX_LOWERNODE;i++) {
            if(this->lnid[i]) {
                for(j=0;j<MAX_LOWERNODE;j++) {
                    if(this->lnid[i]->unid[j]==this) {
                        this->lnid[i]->Hup[j] = NULL;
                    }
                }
            }
        }

        free(this->dlcid);
        this->dlcid = NULL;
    }

 if(this->type==NV_T_FILTER) {
    fprintf(stderr, "jm_camera_stop: f id# %d, dlcid '%s'\n",
        this->id, this->dlcid);

    for(i=0;i<MAX_LOWERNODE;i++) {
        if(!this->Hup[i]) {
            continue;
        }
        n = send_cmd(this->sitter_fd,
                "DISCONNECT %s\r\n", this->ucid[i]);
        if(n<0) {
            fprintf(stderr, "error on DISCONNECT\n");
            goto close_out;
        }
        
        chk = recv_response(this->sitter_fd, &rcode, resbuf, BUFSIZ);
        if(chk<=0) {
            fprintf(stderr, "error or closed\n");
            goto close_out;
        }
        if(chk>0 && rcode!=200) {
            fprintf(stderr, "unexpected response %d '%s'\n",
                rcode, resbuf);
            goto close_out;
        }

        for(j=0;j<MAX_LOWERNODE;j++) {
            if(this->unid[i]->lnid[j]==this) {
                this->unid[i]->Hlo[j] = NULL;
            }
        }
        this->Hup[i] = NULL;

        free(this->ucid[i]);
        this->ucid[i] = NULL;
    }

 }

    goto out;
    
close_out:
    close(this->sitter_fd);
    this->sitter_fd = -1;
out:

}


int
jm_filter_stop(node_t *this)
{
    if(this) {
        fprintf(stderr, "jm_filter_stop: filter id# %d, dlcid '%s'\n",
            this->id, this->dlcid);
    }
    return jm_camera_stop(this);
}


int
jm_camera_stat(node_t *this)
{
    int rcode;
    char resbuf[BUFSIZ];
    int n;
    int chk;

    send_cmd(this->sitter_fd, "LIST\r\n");

    chk = recv_response(this->sitter_fd, &rcode, resbuf, BUFSIZ);
    if(chk<=0) {
        fprintf(stderr, "error or closed\n");
        goto close_out;
    }
#if 0
    if(chk>0 && rcode!=100) {
        fprintf(stderr, "unexpected response %d '%s'\n",
            rcode, resbuf);
        goto close_out;
    }
#endif
close_out:
}

int
jm_filter_change(node_t *this)
{
    int rcode;
    char resbuf[BUFSIZ];
    int n;
    int chk;
    int i;
    node_t *cur;
    int c;

    fprintf(stderr,
        "jm_filter_change: start '%s' '%s' fd# %d, mode %s\n",
        this->name, this->host, this->sitter_fd,
        assoc_v(filter_name_al, this->filtermode));

    if((this->type!=NV_T_FILTER)&&(this->type!=NV_T_SCANNER)) {
        goto out;
    }

#if 0
    rotate_filtermode(&this->filtermode);
#endif

    if(this->sitter_fd<0) {
        goto out;
    }

    chk = send_cmd(this->sitter_fd, "FILTER %s\r\n",
            assoc_v(filter_name_al, this->filtermode));

    chk = recv_response(this->sitter_fd, &rcode, resbuf, BUFSIZ);
    if(chk<=0) {
        fprintf(stderr, "error or closed\n");
        goto out;
    }   


out:
    return 0;
}

int
jm_tobefree_node(node_t *onode)
{
    node_t *that;
    node_t *here;
    int i, j;
    int f;
    int rcode;
    char resbuf[BUFSIZ];
    int chk;


    fprintf(stderr, "===\n");
    print_nodelist(stderr);
    fflush(stderr);

    
    /*
     * close connections from lower nodes
     */
    for(i=0;i<MAX_LOWERNODE;i++) {
        if(onode->lnid[i]) {
            that = onode->lnid[i];
            f = -1;
            for(j=0;j<MAX_LOWERNODE;j++) {
                if(that->unid[j]==onode) {
                    f = j;
                    fprintf(stderr,
                        "conn node#%d l[%d] = node#%d u[%d]\n",
                        onode->id, i, that->id, j
                        );
                    
                    fprintf(stderr, "disconnect id# %d with fd# %d\n",
                        that->id, that->sitter_fd);

                    if(that->type==NV_T_VIEWER) {
                        v_stop_request(that);
                        that->unid[j] = NULL;
                        free(that->ucid[j]);
                        that->ucid[j] = NULL;
                    }
                    else {
                        if(that->sitter_fd<0) {
                            continue;
                        }

                        chk = send_cmd(that->sitter_fd, "DISCONNECT %s\r\n",
                            that->ucid[j]);

                        chk = recv_response(that->sitter_fd,
                            &rcode, resbuf, BUFSIZ);
                        if(chk<=0) {
                            fprintf(stderr, "error or closed\n");
                            continue;
                        }  
                        if(rcode==200) {
                            that->unid[j] = NULL;
                            that->ucid[j] = NULL;
                        }

                    }
                }
            }

            onode->lnid[i] = NULL;
            onode->lcid[i] = NULL;
        }
    }

#if 0
    fprintf(stderr, "===\n");
    print_nodelist(stderr);
    fflush(stderr);
#endif


    /*
     * close connections from upper nodes
     */
    for(i=0;i<MAX_LOWERNODE;i++) {
        if(onode->unid[i]) {
            that = onode->unid[i];
            f = -1;
            for(j=0;j<MAX_LOWERNODE;j++) {
                if(that->lnid[j]==onode) {
                    f = j;
                    fprintf(stderr,
                        "conn node#%d u[%d] = node#%d l[%d]\n",
                        onode->id, i, that->id, j
                        );
                    
                    that->lnid[j] = NULL;
                    that->lcid[j] = NULL;
                }
            }

            chk = send_cmd(onode->sitter_fd, "DISCONNECT %s\r\n",
                onode->ucid[i]);

            chk = recv_response(onode->sitter_fd,
                &rcode, resbuf, BUFSIZ);
            if(chk<=0) {
                fprintf(stderr, "error or closed\n");
                continue;
            }  
            if(rcode==200) {
                onode->unid[i] = NULL;
                onode->ucid[i] = NULL;
            }

#if 0
            onode->unid[i] = NULL;
            onode->ucid[i] = NULL;
#endif
        }
    }




    fprintf(stderr, "===\n");
    print_nodelist(stderr);
    fflush(stderr);

}

int
jm_release_node(node_t *onode)
{
    jm_tobefree_node(onode);
}

jm_copy_links(node_t *nnode, node_t *onode)
{
    int i, k;
    node_t *that;
    int hv;

    for(i=0;i<MAX_LOWERNODE;i++) {
        if(!onode->unid[i]) {
            continue;
        }
        nnode->unid[i] = onode->unid[i];

        that = nnode->unid[i];
        hv = -1;
        for(k=0;k<MAX_LOWERNODE;k++) {
            if(that->lnid[k]==nnode) {
                hv = i;
            }
        }
        if(hv<0) {
            for(k=0;k<MAX_LOWERNODE;k++) {
                if(that->lnid[k]==NULL) {
                    that->lnid[k] = nnode;
                    hv = i;
                    fprintf(stderr, "bind up node#%d -> nnode# %d\n",
                        that->id, nnode->id);
                    break;
                }
            }
        }
        if(hv<0) {
            fprintf(stderr, "fail bind up node#%d -> nnode# %d\n",
                that->id, nnode->id);
                
        }
    }


    for(i=0;i<MAX_LOWERNODE;i++) {
        if(!onode->lnid[i]) {
            continue;
        }
        nnode->lnid[i] = onode->lnid[i];

        that = nnode->lnid[i];
        hv = -1;
        for(k=0;k<MAX_LOWERNODE;k++) {
            if(that->unid[k]==nnode) {
                hv = i;
            }
        }
        if(hv<0) {
            for(k=0;k<MAX_LOWERNODE;k++) {
                if(that->unid[k]==NULL) {
                    that->unid[k] = nnode;
                    hv = i;
                    fprintf(stderr, "bind down node#%d -> nnode# %d\n",
                        that->id, nnode->id);
                    break;
                }
            }
        }
        if(hv<0) {
            fprintf(stderr, "fail bind down node#%d -> nnode# %d\n",
                nnode->id, that->id);
        }
    }

}


int
jm_replace_node(node_t *onode, node_t *nnode)
{
#if 0
    int i;
#endif
    node_t *that;
    node_t *here;
    int i, j;
    int f;
    int rcode;
    char resbuf[BUFSIZ];
    int chk;
    char node_cand_list[BUFSIZ];    /* XXX */
    char *p, *q;
    int ncand;

    fprintf(stderr, "jm_replace_node: new id# %d, old id# %d\n",
        nnode->id, onode->id);

#if 0
    for(i=0;i<MAX_LOWERNODE;i++) {
        nnode->unid[i] = onode->unid[i];
        nnode->lnid[i] = onode->unid[i];
    }
#endif
    print_nodelist(stderr);
    jm_copy_links(nnode, onode);
    print_nodelist(stderr);

#if 0
    /* dummy */
    onode->filtermode = FILTER_THRU;
#endif
    nnode->filtermode = onode->filtermode;
 
    jm_tobefree_node(onode);

    if(!nnode->host) {
        chk = query_nodes(rm_host, rm_port, 1, node_cand_list);

        fprintf(stderr, "chk %d\n", chk);
        if(chk<0) {
            fprintf(stderr, "cannot get candidate(s)\n");
#if 0
            jm_errno = JM_ETOOMANYNEED;
#endif
            return -1;
        }
        chomp(node_cand_list);
#if 0
        ncand = parse_cand_list(node_cand_list);
#endif
        ncand = count_whitesplitted_list(node_cand_list);
        fprintf(stderr, "ncand %d '%s' vs %d\n",
            ncand, node_cand_list, 1);
        if(ncand==1) {
            nnode->host = strdup(node_cand_list);
        }
    }

    fprintf(stderr, "wake+open+kick self id# %d\n", nnode->id);
    jm_filter_wake_and_open(nnode);


    for(i=0;i<MAX_LOWERNODE;i++) {
        if(nnode->lnid[i]) {
            that = nnode->lnid[i];
            fprintf(stderr, "wake+open+kick lower id# %d\n", that->id);
#if 0
            if(that->type==NV_T_FILTER||that->type==NV_T_SCANNER) {
                jm_filter_wake_and_open(that);
                jm_filter_kick(that);
            }
            else
            if(that->type==NV_T_VIEWER) {
                v_start_request(that);
            }
#endif
            if(that->type!=NV_T_VIEWER) {
                jm_filter_wake_and_open(that);
            }
            jm_filter_kick(that);
        }
    }
    
    jm_filter_kick(nnode);


}



int 
jm_probe_node_state(char *req_host, int req_port, node_t *xn)
{
    int i;
    int sock, len, n, ret;
    char buffer[BUFSIZ];
    char buf_num[4];
    char *sep=" \n";
    int chk;
    int xnidle;

    xnidle = -1;

    sock = connect_host(req_host, req_port, SOCK_STREAM);
    if(sock<0) {
        fprintf(stderr, "fail connect to '%s':%d\n", req_host, req_port);
        return -1;
    }

    n = send_cmd(sock, "STATE %s\r\n", xn->host);
    if (n < 0) {
        fprintf(stderr, "Error: do not write, STATE\n");
        return -1;
    }
    fflush(stdout);

    chk = recv_response(sock, &n, buffer, BUFSIZ);

    switch (n) {
    case 201:
        xnidle = atoi(buffer);
        break;
    case 400:
    default:
        break;
    }

    if(xnidle>=0) {
        xn->hidle = xnidle ;
    }

#if 1
    close(sock);
#endif

    return 0;
}

void
jm_probe_state()
{
#if 1
    node_t *cur;
        
    cur = nodelist;
    while(cur) {
        if(cur->type!=NV_T_VIEWER && cur->type!=NV_T_CAMERA &&
                cur->host!= NULL && cur->host[0]!='\0') {
            jm_probe_node_state(rm_host, rm_port, cur);
        }
        cur = cur->next;
    }
#endif
}


void
dup_links()
{
    node_t *cur;
    int i;

    cur = nodelist;
    while(cur) {
        for(i=0;i<MAX_LOWERNODE;i++) {
            cur->Hup[i] = cur->unid[i];
            cur->Hlo[i] = cur->lnid[i];
        }
        cur = cur->next;
    }
}

node_t*
find_leaf()
{
    node_t *cur;
    int i;
    int cu, cl;

#if 0
    print_nodelist_Hlow(stderr, 0);
#endif

    cur = nodelist;
    while(cur) {
#if 0
        fprintf(stderr, "  cur %p, id# %d\n", cur, cur->id);
#endif
#if 0
        print_node(stderr, cur);
#endif

        if(cur->host==NULL || cur->host[0]=='\0') {
            /* nothing */
        }
        else {
            cu = cl = 0;
            for(i=0;i<MAX_LOWERNODE;i++) {
                if(cur->Hup[i]) {
#if 0
                    fprintf(stderr, "    Hup id# %d\n", cur->Hup[i]->id);
#endif
                    cu++;
                }
                if(cur->Hlo[i]) {
#if 0
                    fprintf(stderr, "    Hlo id# %d\n", cur->Hlo[i]->id);
#endif
                    cl++;
                }
            }
            if(cu>0 && cl==0) {
#if 0
                fprintf(stderr, "    *found* cu %d cl %d\n", cu, cl);
#endif
                return cur;
            }
        }
        cur = cur->next;
    }

    return NULL;
}


int
start_allnodes(node_t *dummy)
{
    int all_nodes;
    int undef_nodes;
    int need_nodes;
    int needundef_nodes;
    node_t *cur;
    int chk;
    char node_cand_list[BUFSIZ];    /* XXX */
    char nhost[BUFSIZ];
    char *p, *q;
    int c;
    int i;
    node_t *atn;
    int j;
    int k;
    int kf;
    int ncand;

    dup_links();

    all_nodes = count_nodes(nodelist);
    undef_nodes = count_undef_nodes(nodelist);
    need_nodes = count_need_nodes(nodelist);
    needundef_nodes = count_needundef_nodes(nodelist);
    fprintf(stderr,
        "nodes: all %d, undef %d, need  %d, needundef %d\n",
        all_nodes, undef_nodes, need_nodes, needundef_nodes);

#if 1
    if(need_nodes<1) {
        fprintf(stderr, "start_allnodes: no target\n");
        goto done;
    }
#endif

    jm_errno = JM_ENONE;

    memset(node_cand_list, 0, sizeof(node_cand_list));
    if(needundef_nodes>0) {
        chk = query_nodes(rm_host, rm_port, needundef_nodes, node_cand_list);
        fprintf(stderr, "chk %d\n", chk);
        if(chk<0) {
            fprintf(stderr, "cannot get candidate(s)\n");
#if 0
            jm_errno = JM_ETOOMANYNEED;
#endif
            return -1;
        }
        chomp(node_cand_list);
#if 0
        ncand = parse_cand_list(node_cand_list);
#endif
        ncand = count_whitesplitted_list(node_cand_list);
        fprintf(stderr, "ncand %d '%s' vs %d\n",
            ncand, node_cand_list, needundef_nodes);
    }
    fprintf(stderr, "candidate list '%s'\n", node_cand_list);
    p = node_cand_list;

    i = 0;
    cur = nodelist;
    while(cur) {
        if(cur->type==NV_T_CAMERA || cur->type==NV_T_VIEWER) {
        }
        else
        if(cur->host==NULL || cur->host[0]=='\0') {
            fprintf(stderr, "jdk id# %d, ptr %p, host %p '%s' %d\n",
                cur->id, cur, cur->host, 
                cur->host ? cur->host : "-",
                cur->host ? cur->host[0] : -1);
            /* cut out new host */
            q = nhost;
            c = 0;
            while(c<BUFSIZ && *p &&
                    (*p!=' ' && *p!='\t'&& *p!='\r'&& *p!='\n')) {
#if 0
                fprintf(stderr,"  in word  '%c'\n", *p);
#endif
                *q++ = *p++;
                c++;
            }
            *q = '\0';
            while( *p && (*p==' '||*p=='\t'||*p=='\r'||*p=='\n')) {
#if 0
                fprintf(stderr,"  out word '%c'\n", *p);
#endif
                p++;
            }
            fprintf(stderr, "* new hostname '%s'\n", nhost);
            cur->host = strdup(nhost);
        }

        if(cur->host==NULL || cur->host[0]=='\0') {
            /* nothing */
        }
        else {

        fprintf(stderr, "fire %3d id# %d, '%s' : '%s' %d '%s'\n",
            i, cur->id,
            cur->name, cur->host, cur->down_listen_port, cur->prog_name);
        switch(cur->type) {
        case NV_T_VIEWER:
            fprintf(stderr, "  @ id# %d, *skip*\n", cur->id);
            break;
        case NV_T_CAMERA:
            jm_camera_wake_and_open(cur);
            break;
        case NV_T_FILTER:
        case NV_T_SCANNER:
            jm_filter_wake_and_open(cur);
            break;
        default:
            break;
        }

        }

        i++;
        cur = cur->next;
    }

    fflush(stdout);
    do {
        fprintf(stderr, "===\n");
        print_nodelist_low(stderr, 0);
        fprintf(stderr, "-\n");
        print_nodelist_Hlow(stderr, 0);

        atn = find_leaf();
        if(atn==NULL) {
            fprintf(stderr, "atn %p\n", atn);
            break;
        }
        else {
            fprintf(stderr, "atn %p, id# %d\n", atn, atn->id);
        }

        switch(atn->type) {
        case NV_T_VIEWER:
            jm_viewer_kick(atn);
            break;
        case NV_T_FILTER:
        case NV_T_SCANNER:
            jm_filter_kick(atn);
            break;
        case NV_T_CAMERA:
            jm_camera_kick(atn);
            break;
        }

        for(j=0;j<MAX_LOWERNODE;j++) {
            if(atn->Hup[j]) {
#if 0
                fprintf(stderr, "  @ id# %d, connect to id# %d\n",
                    atn->id, atn->Hup[j]->id);
#endif


                kf = -1;
                for(k=0;k<MAX_LOWERNODE;k++) {
                    if(atn->Hup[j]->Hlo[k]==atn) {
                        atn->Hup[j]->Hlo[k] = NULL;
                    }
                }
                atn->Hup[j] = NULL;
            }
        }
        
#if 0
        print_node(stderr, atn);
#endif
    } while(1);

    print_nodelist_Hlow(stderr, 0);

    cur = nodelist;
    while(cur) {
        if(cur->host==NULL || cur->host[0]=='\0') {
        }
        else
        if(cur->type==NV_T_CAMERA) {
#if 0
            jm_camera_stat(cur);
#endif
        }
        cur = cur->next;
    }
    

done:
    fprintf(stderr, "start_allnodes: done\n");
    fflush(stderr);
    
    return 0;
}



int
stop_allnodes()
{
    node_t *cur;
    int i;

    dup_links();

    print_nodelistH(stderr);

    do {
        fprintf(stderr, "===\n");
        print_nodelist_low(stderr, 0);
        fprintf(stderr, "-\n");
        print_nodelist_Hlow(stderr, 0);

        cur = find_leaf();
        if(cur==NULL) {
            fprintf(stderr, "leaf %p\n", cur);
            break;
        }
        else {
            fprintf(stderr, "leaf %p, id# %d\n", cur, cur->id);
        }

        switch(cur->type) {
        case NV_T_VIEWER:
            jm_viewer_stop(cur);

#if 0
            for(i=0;i<MAX_LOWERNODE;i++) {
                if(cur->Hup[i]) {
                    cur->Hup[i] = NULL;
                }
            }
#endif

            break;
        case NV_T_FILTER:
        case NV_T_SCANNER:
            jm_filter_stop(cur);
#if 0
            for(i=0;i<MAX_LOWERNODE;i++) {
                if(cur->Hup[i]) {
                    cur->Hup[i] = NULL;
                }
            }

            for(i=0;i<MAX_LOWERNODE;i++) {
                if(cur->Hup[i]) {
                    cur->Hup[i] = NULL;
                }
            }
#endif
            break;
        }
#if 0
        print_node(stderr, cur);
#endif
    } while(1);

    print_nodelist_Hlow(stderr, 0);

    cur = nodelist;
    while(cur) {
        if(cur->type==NV_T_CAMERA) {
            jm_camera_stop(cur);

            for(i=0;i<MAX_LOWERNODE;i++) {
                if(cur->Hlo[i]) {
                    cur->Hlo[i] = NULL;
                }
            }
        }
        cur = cur->next;
    }

    print_nodelist_Hlow(stderr, 0);
}




