/*
 * Sentry Video System
 *   is video server with QoS control from clients' requests.
 *
 *      original by akira-y@is.aist-nara.ac.jp, 1994,1995
 *      modified by k-chinen@is.aist-nara.ac.jp, 1996
 *      modified by satosi-f@is.aist-nara.ac.jp, 1997,2000
 *  
 *  $Id: $
 */

/*
 * manage_client.c - manage (register, release, etc) clients.
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#ifdef TIME_WITH_SYS_TIME
 #include <sys/time.h>
 #include <time.h>
#else
 #ifdef HAVE_SYS_TIME_H
  #include <sys/time.h>
 #else
  #include <time.h>
 #endif
#endif
#include <errno.h>

#include "alist.h"
#include "sentry.h"

#define MAX_CLIENTS     64
struct q_client {
  int id;
#define INVALID        (-1)
#define UDP_UNICAST    0
#define UDP_MULTICAST  1
#define TCP            2
  int flag;                     /* flag */
  int datafd;                   /* data socket fd */
  struct sockaddr_in addr;      /* family/address/port */
  int rate;                     /* request rate */
  struct q_client *next;
} client_pool[MAX_CLIENTS];

int max_client_id;
int multicast_clients;
int unicast_clients;

struct q_client *freelist;

#if 0
struct q_client *clients_udp_uni;
struct q_client *clients_udp_multi;
struct q_client *clients_tcp;
#endif

#ifdef SUPPORT_MULTIRATE
struct {
  int rate_interval;
  int rate_count;
} client_rate[MAX_CLIENTS];
#endif

extern int debug;
extern int verb;

extern int     maxreadfds;
extern fd_set  readmask;
extern int     maxwritefds;
extern fd_set  writemask;


int
add_client(u_char type, struct sockaddr_in *addr, int fd) 
{
  int id;
  extern void clear_client();

  /* client address check */
  if (addr->sin_family != AF_INET) {
    fprintf(stderr, "sentryd: address family is invalid.\n");
    return -1;
  }

  if ((id = find_client(addr)) < 0) {
    /* new client */
    if ((id = malloc_client()) < 0)
      return -1;

  } else if (client_pool[id].flag != type) {
    clear_client(id);
    if ((id = malloc_client()) < 0) 
      return -1;

  } else { /* already registerd */
    return -2; 
  }

  switch(type) {
  case OPEN_UDP_UNICAST:
    client_pool[id].flag = UDP_UNICAST;
    unicast_clients ++;
    break;
  case OPEN_UDP_MULTICAST:
    client_pool[id].flag = UDP_MULTICAST;
    multicast_clients ++;
    break;
  case OPEN_TCP:
    client_pool[id].flag = OPEN_TCP;
    unicast_clients ++;
  default:
    return -1;
  }
  
  client_pool[id].datafd = fd;
  bcopy((char *)addr, (char *)&(client_pool[id].addr), sizeof(struct sockaddr_in));

  
  if (debug) {
    int i;
    extern void dump_client();

    fprintf(stderr, "add_client: %s:%d, id# %d, flag# %x, datafd# %d\n", 
        inet_ntoa(client_pool[id].addr.sin_addr),
#if 0
        inet_ntoa((struct in_addr *)ntohl(client_pool[id].addr.sin_addr.s_addr)),
#endif
        ntohs(client_pool[id].addr.sin_port),
        id, client_pool[id].flag, client_pool[id].datafd);

    for(i=0; i<=max_client_id; i++)
      dump_client(i);
  }


  return id;
}

int
show_client(FILE *fp, char *msg)
{
    int id;
    int c;

    if(msg)
        fprintf(fp, "clients %s\n", msg);
    c = 0;
    for(id=0; id<MAX_CLIENTS; id++) {
        if (client_pool[id].flag == INVALID) continue;
        
        fprintf(fp, "%3d: %3d %s:%d\n",
            id, client_pool[id].datafd,
            inet_ntoa(client_pool[id].addr.sin_addr), 
            ntohs(client_pool[id].addr.sin_port));
        c++;
    }

    return c;
}


int
find_client(struct sockaddr_in *addr) 
{
  int id;

  if (debug) 
    fprintf(stderr, "find_client: %s:%d\n",
        inet_ntoa(addr->sin_addr), ntohs(addr->sin_port));

  for(id=0; id<MAX_CLIENTS; id++) {
    if (client_pool[id].flag == INVALID) continue;
    if (!bcmp((char *)addr, (char *)&(client_pool[id].addr), sizeof(struct sockaddr_in)))
      return id;
  }
  return -1;
}

int 
find_client_by_datafd(int datafd)
{
  int id;

  for(id=0; id<MAX_CLIENTS; id++) {
    if (client_pool[id].flag == INVALID) continue;
    if (client_pool[id].datafd == datafd)
      return id;
  }
  return -1;
}

int
malloc_client()
{
  int id;
  for(id=0; id<MAX_CLIENTS; id++) {
    if (client_pool[id].flag == INVALID) {
      client_pool[id].datafd = -1;
      if (max_client_id < id) max_client_id = id;
      return id;
    }
  }
  return -1;
}

void
dump_client(int id)
{
  struct q_client *cli = &client_pool[id];

  fprintf(stderr, "=== client[%d] ===\n", id);
  fprintf(stderr, "  flag#   %d\n", cli->flag);
  fprintf(stderr, "  datafd# %d\n", cli->datafd);
  fprintf(stderr, "  addr    %s\n", inet_ntoa(cli->addr.sin_addr));
  fprintf(stderr, "  port    %d\n\n", ntohs(cli->addr.sin_port));
}


void
clear_client(int id)
{
  int flag;
  int i;

#ifdef SUPPORT_MULTIRATE
  int cid, maxcid = -1;
  int maxrate = 0;
  extern void set_qos(), set_client_rate();
#endif

  if (id < 0 || id > MAX_CLIENTS) return;

  /* set client flag off */
  switch(client_pool[id].flag) {
  case UDP_MULTICAST:
    /* cli_t = clients_udp_multi; */
    if (--multicast_clients < 0) multicast_clients = 0;
    break;
  case UDP_UNICAST:
  case TCP:
    if (--unicast_clients < 0) unicast_clients = 0;
  default:
    break;
  }

#if 0
  for(cli=cli_t; cli; cli=cli->next) {
      if (cli->id == id)
    if (cli == cli_t)
      cli_t = cli->next;
    else
      cli_prev->next = cli->next;
      cli_prev = cli;
  }
#endif

  FD_CLR(client_pool[id].datafd, &writemask);
  if (maxwritefds == client_pool[id].datafd) {
    for(i=maxwritefds-1; i>=0; i--) 
      if (FD_ISSET(i, &writemask)) {
    maxwritefds = i;
    break;
      }
  }
  if (client_pool[id].flag == TCP) {
    FD_CLR(client_pool[id].datafd, &readmask);
    if (maxreadfds == client_pool[id].datafd)
      for(i=maxreadfds-1; i>=0; i--)
    if (FD_ISSET(i, &readmask)) {
      maxreadfds = i;
      break;
    }
  }
  if (client_pool[id].datafd > 0) {
    shutdown(client_pool[id].datafd, 2);
    close(client_pool[id].datafd);
    client_pool[id].datafd = -1;
  }

  if (debug)
    fprintf(stderr, "clear_client: maxwritefds# %d, maxreadfds# %d\n", 
        maxwritefds, maxreadfds);

  if (max_client_id == id) {
    for(i=max_client_id-1; i>=0; i--) 
      if (client_pool[id].flag != INVALID) {
    max_client_id = i;
    break;
      }
  }
  client_pool[id].flag = INVALID;

  /* update free list */
  client_pool[id].next = freelist;
  freelist = &client_pool[id];
  
  if(debug>=100)
    fprintf(stderr, "client[%d] is clear.\n", id);
}

int
clear_all_client()
{
    int c;
    int id;

    c = 0;
    for(id=0; id<MAX_CLIENTS; id++) {
        if (client_pool[id].flag == INVALID) continue;
        clear_client(id);
        c++;
    }

    return c;
}


void 
init_client(void)
{
  int i;

  for(i=0; i<MAX_CLIENTS; i++) {
    client_pool[i].id = i;
    client_pool[i].flag = INVALID;
    client_pool[i].datafd = -1;
    client_pool[i].next = NULL;

    if (i<MAX_CLIENTS-1) 
      client_pool[i].next = &client_pool[i+1];
    else 
      client_pool[i].next = NULL;
  }    

  max_client_id = -1;
  multicast_clients = 0;
  unicast_clients = 0;

  freelist = client_pool;
#if 0  
  clients_udp_uni = NULL;
  clients_udp_multi = NULL;
  clients_tcp = NULL;
#endif

  if(debug>=100) {
    struct q_client *cli;
    int id = 0;

    for(cli=freelist; cli; cli=cli->next)
      fprintf(stderr, "init_client: free[%d]: flag# %d\n", 
          id++, cli->flag);
  }
}
