/*
 * sentry.c - image viewer for sentry protocol
 *      k-chinen@is.aist-nara.ac.jp, 1995-2001, 2003
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <signal.h>
#include <syslog.h>
#ifdef TIME_WITH_SYS_TIME
 #include <sys/time.h>
 #include <time.h>
#else
 #ifdef HAVE_SYS_TIME_H
  #include <sys/time.h>
 #else
  #include <time.h>
 #endif
#endif
#include <errno.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xproto.h>
#include <X11/cursorfont.h>

#include <X11/Xmu/WinUtil.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/Scrollbar.h>

/* private TrueColor Widget */
#include "Lww.h"

#include "alist.h"
#include "sentry.h"
#include "img.h"
#include "pool.h"
#include "client.h"

#include "rtr.h"

#ifndef MIN
#define MIN(x,y)    ((x)<(y) ? (x):(y))
#endif

#ifndef MAX_IMG_LEN
#define MAX_IMG_LEN     (640*480*4)
#endif
#ifndef TMP_IMGBUF_LEN
#define TMP_IMGBUF_LEN  (320*240*4)
#endif

extern apair_t buflen_xrgb_alist;

extern  void request_call();


typedef struct {
    /* id */
    char *name;
    char *serv_host;
    int   serv_port;
    int   recv_port;
    char *desc;

    struct sockaddr_in addr;
    int   cntl_fd;
    int   data_fd;
    subrecord_t *recbuf;

    /* request */
    int   type;
    int   size;
    int   color;
    int   format;
    int   rate;
    int   zoom;

    /* presentation */
    int   x;
    int   y;
    int   anchor;

    Widget Wlabel;
    Widget Wimg;
    Widget Wmsg;
    sentry_imgrec inimg;
    sentry_imgrec outimg;

} cnode_t;

show_cnode(FILE *fp, char *msg, cnode_t *cn)
{
    fprintf(fp, "cnode %p : %s\n", cn, msg);
    fprintf(fp, "  name '%s'\n", cn->name);
    fprintf(fp, "  server '%s' port %d\n", cn->serv_host, cn->serv_port);
    fprintf(fp, "  receive port %d\n", cn->recv_port);
    fprintf(fp, "  description '%s'\n", cn->desc);
    fprintf(fp, "  addr -\n");
    fprintf(fp, "  fd cntl %d data %d\n", cn->cntl_fd, cn->data_fd);
    fprintf(fp, "  req. type %s(%d) size %s(%d) color %s(%d) format %s(%d)\n",
        assoc_v(req_type_alist, cn->type),
        cn->type,
        assoc_v(size_alist, cn->size),
        cn->size,
        assoc_v(color_alist, cn->color),
        cn->color,
        assoc_v(format_alist, cn->format),
        cn->format);
    fprintf(fp, "    rate %d, zoom %d\n",
        cn->rate,
        cn->zoom);
    fprintf(fp, "  position %d,%d, anchor %s(%d)\n",    
        cn->x, cn->y, assoc_v(anchor_alist, cn->anchor), cn->anchor);

    fflush(fp);
}

int
clear_cnode(cnode_t *cn)
{
    if(!cn) {
        return -1;
    }

    cn->name = NULL;
    cn->serv_host = NULL;
    cn->serv_port = -1;
    cn->recv_port = -1;
    cn->desc = NULL;
    memset(&cn->addr, 0, sizeof(cn->addr));
    cn->cntl_fd = -1;
    cn->data_fd = -1;
    cn->type    = -1;
    cn->size    = -1;
    cn->color   = -1;
    cn->format  = -1;
    cn->zoom    = -1;
    cn->rate    = -1;
    cn->x       = -1;
    cn->y       = -1;
    cn->anchor  = -1;

    cn->recbuf  = NULL;

    cn->inimg.body_len = 0;
    cn->outimg.body_len = 0;
}

int
setdefault_cnode(cnode_t *cn)
{
    int buflen;

    if(!cn) {
        return -1;
    }

    cn->name = NULL;
    cn->serv_host = NULL;
    cn->serv_port = SENTRY_CTRL_PORT;
    cn->recv_port = SENTRY_DATA_PORT;
    cn->desc = NULL;
    memset(&cn->addr, 0, sizeof(cn->addr));
    cn->cntl_fd = -1;
    cn->data_fd = -1;
    cn->type    = OPEN_UDP_UNICAST;
    cn->size    = SIZE_160x120;
    cn->color   = SEN_C_NONE;
    cn->format  = SEN_F_JPEG;
    cn->zoom    = 1;
    cn->rate    = 1000;
    cn->x       = 0;
    cn->y       = 0;
    cn->anchor  = SEN_A_LT;

    cn->recbuf  = NULL;

    buflen = 160*120*4;

    cn->inimg.body      = malloc(buflen);
    cn->inimg.body_len  = buflen;
    cn->outimg.body     = malloc(buflen);
    cn->outimg.body_len = buflen;
}



extern u_char client_type; 
extern int server_zoom;
extern int server_size;
extern int server_color;
extern int server_format; 
extern int server_rate;

static int show_help=0;

static int beep_mute=0;
static int ghost_mode=0;


typedef struct {
    String server_name;
    String server_port;
    String data_port;
    String req_rate;
    String req_type;
    String req_size;
    String req_zoom;
    String req_color;
    String req_format;
    String hivision_flag;
    String voffset;
    /* */
    String shm_key;
    /* */
    String compare_func;
    /* */
    String debug_level;
    String show_help;
    String img_history;
} sentry_resource_t;

sentry_resource_t sentry_resource;

#define Offset(field) (XtOffsetOf(sentry_resource_t, field))
static XtResource resources[] = {
  {"servername", "ServerName",  XtRString,  sizeof(String),
                    Offset(server_name), XtRString,  NULL},
  {"serverport", "ServerPort",  XtRString,  sizeof(String),
                    Offset(server_port), XtRString,  NULL},
  {"dataport",   "DataPort",  XtRString,  sizeof(String),
                    Offset(data_port), XtRString,  NULL},
  {"reqsize",   "ReqSize",  XtRString,  sizeof(String),
                    Offset(req_size), XtRString,  NULL},
  {"reqrate",   "ReqRate",  XtRString,  sizeof(String),
                    Offset(req_rate), XtRString,  NULL},
  {"reqformat", "ReqFormat",  XtRString,  sizeof(String),
                    Offset(req_format), XtRString,  NULL},
  {"reqtype",   "ReqType",  XtRString,  sizeof(String),
                    Offset(req_type), XtRString,  NULL},

  {"shmkey",    "ShmKey",  XtRString,  sizeof(String),
                    Offset(shm_key), XtRString,  NULL},

  {"comp",      "comp",  XtRString,  sizeof(String),
                    Offset(compare_func), XtRString,  NULL},

  {"debug",     "Debug",  XtRString,  sizeof(String),
                    Offset(debug_level), XtRString,  NULL},
  {"help",      "Help",  XtRString,  sizeof(String),
                    Offset(show_help), XtRString,  NULL},
  {"history",   "History",  XtRString,  sizeof(String),
                    Offset(img_history), XtRString,  NULL},
};




String fallback_resources[] = {
    "*reqsize:      160x120",
    "*reqrate:      200",
    "*reqformat:    JPEG",
    "*servername:   localhost",
    "*serverport:   7001",
    "*dataport:     7000",

    "*shmkey:       -1",

    "*comp:         none",

    "*debug:        0",

    "*help:         0",

    "*foreground:   black",
    "*background:   white",

    "*history:      16",
    NULL};

static XrmOptionDescRec options[] = {
{"-fg",         "*Foreground",  XrmoptionSepArg,    NULL},
{"-bg",         "*Background",  XrmoptionSepArg,    NULL},
{"-servername", "*servername",  XrmoptionSepArg,    NULL},
{"-serverport", "*serverport",  XrmoptionSepArg,    NULL},
{"-dataport",   "*dataport",    XrmoptionSepArg,    NULL},

{"-shmkey",     "*shmkey",      XrmoptionSepArg,    NULL},
{"-debug",      "*debug",       XrmoptionSepArg,    NULL},
{"-rate",       "*reqrate",     XrmoptionSepArg,    NULL},
{"-format",     "*reqformat",   XrmoptionSepArg,    NULL},
{"-size",       "*reqsize",     XrmoptionSepArg,    NULL},

{"-history",    "*history",     XrmoptionSepArg,    NULL},

{"-comp",       "*comp",        XrmoptionSepArg,    NULL},

{"-help",       "*help",        XrmoptionNoArg,     "1"},
};

Widget   toplevel;
XtAppContext app_context;
XtInputId   iid, ciid;

XtActionsRec main_action_list[] = {
    NULL,
};


Visual   *xp_slide_visual,   *xp_foil_visual;
int      xp_slide_depth,     xp_foil_depth;
Colormap xp_slide_cmap,      xp_foil_cmap;
int      xp_use_special_cmap;

String fvisual, svisual;


Pixel black, white;

Widget pwinlabel;
Widget pwin;
Widget pwinlabel2;
Widget pwin2;
Widget pwinmsg;
Widget pshell;
Widget pbase;
Widget pcpanel;
Widget plabel;
Widget pcmd;
Widget bquit;
Widget bmute, bghost;
GC  pwin_gc;

Pixmap diff_pix;

typedef struct {
    XtIntervalId    id;
    time_t          interval;
} xp_periodical_rec;

xp_periodical_rec win1;


void periodical_update(XtPointer client_data, XtIntervalId *iid);


extern sentry_shmrec   *sentry_shm_chunk;


show_chunk()
{
    int i;
    int pos;

    for(i=0;i<sentry_shm_chunk_radix;i++) {
        pos = i+1;
        printf(" %3d: %3d %12d %4d\n", i, pos,
            sentry_shm_chunk[pos].entry.header.recv_sec,
            sentry_shm_chunk[pos].entry.header.length);
    }
}

int 
find_pos(time_t reftime)
{
    int i;
    int pos;
    int ret;

    ret = -1;

    for(i=0;i<sentry_shm_chunk_radix;i++) {
        pos = i+1;
#if 0
        printf(" %3d: %3d %12d %4d\n", i, pos,
            sentry_shm_chunk[pos].entry.header.recv_sec,
            sentry_shm_chunk[pos].entry.header.length);
#endif
        if(sentry_shm_chunk[pos].entry.header.recv_sec==reftime) {
            if(debug>=10) {
                printf(
" %3d: %3d time issue %d.%06d recv %d.%06d len %4d (%dx%d) pos %p [%d]\n",
                    i, pos,
                    sentry_shm_chunk[pos].entry.header.send_sec,
                    sentry_shm_chunk[pos].entry.header.send_usec,
                    sentry_shm_chunk[pos].entry.header.recv_sec,
                    sentry_shm_chunk[pos].entry.header.recv_usec,
                    sentry_shm_chunk[pos].entry.header.length,
                    sentry_shm_chunk[pos].entry.header.width,
                    sentry_shm_chunk[pos].entry.header.height,
                    sentry_shm_chunk[pos].entry.header.body,
                    sentry_shm_chunk[pos].entry.header.body_len
                    );
            }
            ret = pos;
        }
    }

    return ret;
}


SenDrawImage2(Widget w, sentry_imgrec *inimg, sentry_imgrec *outimg)
{
    int wid, ht;
    int xwid, xht;
    unsigned char *raw_data;
#if 0
    unsigned char *data_rgb=NULL, *data_bgrx=NULL;
    int buflen=0;
#endif
#if 1
    static unsigned char *data_rgb=NULL, *data_bgrx=NULL;
    static int buflen=0;
#endif
    int raw_len;
    int raw_format, raw_color;
    int chk;
    int cur_format=SEN_F_NONE;
    int cur_corder=SEN_C_NONE;
    int success=0;
    int outlen=-1;
    GC cur_gc;

    if(inimg==NULL)
        return;

#if 0
    if(debug>=10) {
        printf("SenDrawImage: output format '%s', color order '%s'\n",
            assoc_v(format_alist, output_format),   
            assoc_v(color_alist, output_corder));
    }
#endif


    raw_len     = inimg->length;
    raw_data    = (unsigned char*)inimg->body;
    raw_format  = inimg->format;
    raw_color   = inimg->corder;
    wid = inimg->width;
    ht  = inimg->height;

    if(!data_rgb || !data_bgrx) {
        buflen = (int)assoc_p(&buflen_xrgb_alist, MAX_DISPLAY_SIZE);
        if(buflen<=0) {
            buflen = TMP_IMGBUF_LEN;
        }
fprintf(stderr, "#buflen %d\n", buflen);

        data_bgrx = (unsigned char*)malloc(buflen);
        memset(data_bgrx, 0xff, buflen);

        data_rgb = (unsigned char*)malloc(buflen);
        memset(data_rgb, 0xff, buflen);
    }

#if 0
    if(debug>=10) {
        printf("SenDrawImage: %p %d '%s''%s' %dx%d, 3B/p=%d 4B/p=%d\n",
            raw_data, raw_len, assoc_v(format_alist, raw_format),
            assoc_v(color_alist, raw_color), wid, ht, wid*ht*3, wid*ht*4);
    }
#endif
    if(debug>=10) {
        printf(
"draw: got format %s, color %s, ptr %p, size %dx%d, len %d\n\
      out format %s, color %s\n",
        assoc_v(format_alist, raw_format), assoc_v(color_alist, raw_color), 
        raw_data, wid, ht, raw_len,
        assoc_v(format_alist, output_format),
        assoc_v(color_alist, output_corder));
    }

    switch(raw_format) {
    case SEN_F_JPEG:
        chk = decode_jpeg(raw_data, raw_len, data_rgb, &xwid, &xht);
        if(xwid!=wid && xht!=ht) {
            fprintf(stderr,
                "SenDrawImage: wid %d != xwid %d, ht %d != xht %d\n",
                wid, xwid, ht, xht);
        }
        cur_format = SEN_F_RGB;
        cur_corder = SEN_C_RGB;
        if(debug>=20){
            printf("  convert to '%s' (wid %d, ht %d)\n",
                assoc_v(color_alist, cur_corder), xwid, xht);
        }
        break;

    case SEN_F_NONE: /* XXX */
        memcpy(data_rgb, raw_data, raw_len);
        break;

    case SEN_F_RGB: /* XXX */
        memcpy(data_rgb, raw_data, raw_len);
        cur_format = raw_format; /* SEN_F_RGB */
        cur_corder = raw_color;
        break;
    }

    if(debug>=10) {
        printf(
"draw: cur format %s, color %s\n\
      out format %s, color %s\n",
        assoc_v(format_alist, cur_format), assoc_v(color_alist, cur_corder),
        assoc_v(format_alist, output_format),
        assoc_v(color_alist, output_corder)
        );
    }
    if(cur_format != output_format) {
        if(debug>=1) {
            fprintf(stderr, "image format miss-match cur=%s out=%s\n",
                assoc_v(format_alist, cur_format),
                assoc_v(format_alist, output_format));
        }
        goto put_fail;
    }

    if(output_format==SEN_F_RGB) {
        cur_gc = XtGetGC(w, NULL, NULL);
        if(output_corder==SEN_C_XBGR) {
            convert_RGB2XBGR(data_bgrx, data_rgb, wid * ht);

            outlen = wid*ht*4;
            put_image(XtDisplay(w), xp_foil_visual,
                xp_foil_depth, cur_gc, XtWindow(w), wid, ht, data_bgrx);

            success = 1;

            copyattr_img(outimg, inimg);
            outimg->format = output_format;
            outimg->corder = output_corder;
            outimg->length = outlen;
            memcpy(outimg->body, data_bgrx, outlen);
        }
        else
        if(output_corder==SEN_C_XRGB) {
/*
            convert_RGB2XRGB(data_bgrx, data_rgb, wid * ht);
*/
            convert_RGB2BGRX(data_bgrx, data_rgb, wid * ht);

            outlen = wid*ht*4;
            put_image(XtDisplay(w), xp_foil_visual,
                xp_foil_depth, cur_gc, XtWindow(w), wid, ht, data_bgrx);
            success = 1;

            copyattr_img(outimg, inimg);
            outimg->format = output_format;
            outimg->corder = output_corder;
            outimg->length = outlen;
            memcpy(outimg->body, data_bgrx, outlen);
        }
        else
        if(output_corder==SEN_C_RGB) {
            outlen = wid*ht*3;
            put_image(XtDisplay(w), xp_foil_visual,
                xp_foil_depth, cur_gc, XtWindow(w), wid, ht, data_rgb);
            success = 1;

            copyattr_img(outimg, inimg);
            outimg->format = output_format;
            outimg->corder = output_corder;
            outimg->length = outlen;
            memcpy(outimg->body, data_rgb, outlen);
        }
        XtReleaseGC(w, cur_gc);
    }

put_fail:
    if(!success) {
        if(debug>=1) {
            fprintf(stderr, "fail some reason, format cur=%s out=%s\n",
                assoc_v(format_alist, cur_format),
                assoc_v(format_alist, output_format));
        }
    }
    else {
#if 0
        copyattr_img(outimg, inimg);
        memcpy(outimg->body, inimg->body, outlen);
#endif
    }

out:
#if 0
    if(data_rgb!=NULL)
        free(data_rgb);
    if(data_bgrx!=NULL) 
        free(data_bgrx);
#endif
}


static int call_count=0, lost_count=0, got_count=0, c_got_count=0;
static int refresh_flag = 1;


RefreshWins(sentry_imgrec *bank)
{
    Dimension tw, th, tx, ty;
    Dimension bw, bh, bx, by;
    Dimension nw, nh;

    XtVaGetValues(plabel,
         XtNx, &tx, XtNy, &ty, XtNwidth, &tw, XtNheight, &th, NULL);
    XtVaGetValues(pwin,
         XtNx, &bx, XtNy, &by, XtNwidth, &bw, XtNheight, &bh, NULL);
#if 0
    printf("1 %ux%u+%u+%u %ux%u+%u+%u\n",
        tw, th, tx, ty, bw, bh, bx, by);
#endif

    XtVaSetValues(plabel,
        XtNwidth, bank->width,
        XtNlabel, server_host,
        NULL);

    XtVaGetValues(plabel, XtNheight, &th, NULL);
    XtVaGetValues(pwin, XtNheight, &bh, NULL);
    XtVaGetValues(plabel, XtNy, &ty, NULL);
    XtVaGetValues(pwin, XtNy, &by, NULL);
#if 0
    printf("2 th %d, bh %d, ty %d, by %d\n", th, bh, ty, by);
#endif


    XtVaSetValues(pwin,
        XtNwidth, bank->width,
        XtNheight, bank->height,
        NULL);

    nw = bank->width+10;
/*
    nh = bank->height+th+10;
*/
    nh = MAX(by,24) + bank->height+10;

#if 0
    printf("nw %d, nh %d\n", nw, nh);
#endif

#if 0
    XtVaSetValues(pshell,
        XtNwidth, nw, XtNheight, nh,
/*
*/
        NULL);
#endif

    XtVaSetValues(pbase,
/*
        XtNwidth, bank->width,
        XtNheight, bank->height,
*/
/*
        XtNwidth, nw, XtNheight, nh,
*/
        NULL);

    XtVaGetValues(plabel, XtNheight, &th, NULL);
    XtVaGetValues(pwin, XtNheight, &bh, NULL);
#if 0
    printf("3 th %d, bh %d\n", th, bh);
#endif

    fflush(stdout);
    fflush(stderr);

    refresh_flag = 0;
}





static sentry_shmrec bk_info;
static sentry_imgrec bk_inframe;
static unsigned char *inimgbuf=NULL;
static sentry_imgrec bk_drframe[2];
static unsigned char *drimgbuf0=NULL;
static unsigned char *drimgbuf1=NULL;





alloc_inimgbufs()
{
    int buflen;

    buflen = (int)assoc_p(&buflen_xrgb_alist, MAX_DISPLAY_SIZE);
 if(debug) {
    printf("buflen %d\n", buflen);
 }
    if(buflen==0) {
        buflen = TMP_IMGBUF_LEN;
    }

    inimgbuf = (unsigned char*)malloc(buflen);
    bk_inframe.body = inimgbuf;
    bk_inframe.body_len = buflen;
}

alloc_drimgbufs()
{
    int buflen;

    if(output_format!=SEN_F_RGB) {
        fprintf(stderr, "no idea for no RGB device\n");
        exit(1);
    }

    buflen = (int)assoc_p(&buflen_xrgb_alist, MAX_DISPLAY_SIZE);
 if(debug) {
    printf("buflen %d\n", buflen);
 }
    if(buflen==0) {
        buflen = TMP_IMGBUF_LEN;
    }

    drimgbuf0 = (unsigned char*)malloc(buflen);
    bk_drframe[0].body = drimgbuf0;
    bk_drframe[0].body_len = buflen;

    drimgbuf1 = (unsigned char*)malloc(buflen);
    bk_drframe[1].body = drimgbuf1;
    bk_drframe[1].body_len = buflen;
}




static int img_his_radix=-1;
static sentry_imgrec *img_his=NULL;
static int img_his_pos=0;

void
show_image_history(FILE *fp)
{
    int i, u;

    for(i=0;i<img_his_radix;i++) {
        fprintf(fp, "%2d: %p %6d/%6d %3dx%3d\n",
            i,
            &img_his[i], img_his[i].length, img_his[i].body_len,
            img_his[i].width, img_his[i].height);
    }
#if 0
    for(u=img_his_pos;u<img_his_pos+img_his_radix;u++) {
        i = u%img_his_radix;
        fprintf(fp, "%2d: %c %2d : %p %6d/%6d %3dx%3d\n",
            u, (i==img_his_pos ? '*': '-'),  i,
            &img_his[i], img_his[i].length, img_his[i].body_len,
            img_his[i].width, img_his[i].height);
    }
#endif
}

init_image_history(int n)
{
    int i;
    int w;

    img_his_radix = n;
    img_his = (sentry_imgrec*) malloc(sizeof(sentry_imgrec)*img_his_radix);

    for(i=0;i<img_his_radix;i++) {
        w = (int)assoc_p(&buflen_xrgb_alist, MAX_DISPLAY_SIZE);
        img_his[i].body = (uint8_t*)malloc(w);
        img_his[i].body_len = w;
        img_his[i].length = -1;
    }

    if(debug) {
        fprintf(stderr, "img_his_radix %d\n", img_his_radix);
    }
}


int
RefreshHist(int n)
{
    sentry_imgrec *cur;
    int o;
    sentry_imgrec *last;
    int wid, ht;
    char msg[BUFSIZ];
    struct tm htm;
    char data_body[MAX_IMG_LEN];
    GC cur_gc;

    void DoDiffExpose(Widget w, XEvent *ev, String *s, Cardinal *c);

    if(n<0 || n>=img_his_radix) {
        return -1;
    }

    cur = &img_his[n];
    if(!cur) {
        return -1;
    }

    o = (img_his_radix + n - 1) % img_his_radix;
    last = &img_his[o];
    if(!last) {
        return -1;
    }
#if 0
    fprintf(stderr, "n %d, o %d\n", n, o);
#endif
    
    wid = cur->width;
    ht  = cur->height;

    if(cur->length>0) {

        localtime_r(&cur->recv_sec, &htm);
        sprintf(msg, "%02d %2d-%02d %02d:%02d:%02d.%02d",
            n, htm.tm_mon+1, htm.tm_mday,
            htm.tm_hour, htm.tm_min, htm.tm_sec, cur->recv_usec/10000);
        RefreshMsg(msg);

        cur_gc = XtGetGC(pwin2, NULL, NULL);

        if(ghost_mode && last->length>0) {
#if 0
fprintf(stderr,  "#corder %d output %d SEN_C_XRGB %d\n",
    cur->corder, output_corder, SEN_C_XRGB);
#endif
            if(output_corder=SEN_C_XRGB) {
                subghost_bgrx(cur->body, last->body, data_body, wid, ht);
            }
            else
            if(output_corder=SEN_C_RGB) {
                subghost_rgb(cur->body, last->body, data_body, wid, ht);
            }

            put_image(XtDisplay(pwin2), xp_foil_visual,
                xp_foil_depth, cur_gc, diff_pix, wid, ht, data_body);
        }
        else {
            put_image(XtDisplay(pwin2), xp_foil_visual,
                xp_foil_depth, cur_gc, diff_pix, wid, ht, cur->body);
        }
        XtReleaseGC(pwin2, cur_gc);

        DoDiffExpose(NULL, NULL, NULL, NULL);
    }
    else {
        fprintf(stderr, "#show n %d, length %d\n", n, cur->length);
    }
}


int
store_image_history(sentry_imgrec *cur, sentry_imgrec *last)
{
    char data_bgrx[MAX_IMG_LEN];
    char data_rgb[MAX_IMG_LEN];
    int wid, ht;
    int cnum;
    void DoDiffExpose(Widget w, XEvent *ev, String *s, Cardinal *c);
    
    wid = cur->width;
    ht  = cur->height;

    if(debug)
        show_image_history(stderr);

    cnum = img_his_pos;

    if(debug)
        fprintf(stderr, "#store to %d\n", cnum);

    copyattr_img(&img_his[img_his_pos], cur);
    memcpy(img_his[img_his_pos].body, cur->body, cur->length);
    img_his[img_his_pos].length = cur->length;

    img_his_pos = (img_his_pos+1) % img_his_radix;

    if(debug)
        show_image_history(stderr);

    RefreshHist(cnum);
}


/*
int diff_capacity = 0x20;
*/
int diff_capacity = 0x30;
int cmp_type = SEN_M_NONE;

void
eventical_update(caddr_t client_data, int *fd, XtInputId *id)
{
    static struct timeval lasttime={-1,0};
    struct timeval curtime;
    time_t now;
    int chk;
    sentry_shmrec *info;
    sentry_imgrec *inframe;
    static sentry_imgrec *cur=NULL, *last=NULL;
    char *tmp_buff;
    int pos;
    int diff;
    double rate;
    int  drfno;
    char *drframe;
    int recv_image(int datafd, char **buff,
        sentry_shmrec *info, sentry_imgrec *dest);

    gettimeofday(&curtime, NULL);
    now = curtime.tv_sec;

    if(use_shm_flag) {
        info = &sentry_shm_chunk[0];
        pos  = 1+now%sentry_shm_chunk_radix;
        inframe = &(sentry_shm_chunk[pos].entry.header);
    }
    else {
        info = &bk_info;

        if(c_got_count==0) {
            alloc_inimgbufs();
            alloc_drimgbufs();
        }

        inframe = &bk_inframe;
    }

 if(debug>=100) {
    fprintf(stderr, "inframe %p\n", inframe);
 }

    chk = recv_image(*fd, &tmp_buff, info, inframe);
    if(chk) {
        return;
    }

    if(got_count%report_interval==0) {
        syslog(LOG_NOTICE, "sentryc key=%d; got %d data\n",
            sentry_shmkey, got_count);
    }
    if(refresh_flag){
        RefreshWins(inframe);
    }

    drfno = c_got_count%2;
    cur = &bk_drframe[drfno];

 if(debug>=100) {
    fprintf(stderr, "drfno %d, bk_drframe[%d] %p, cur %p\n",
        drfno, drfno, &bk_drframe[drfno], cur);
 }

    SenDrawImage2(pwin, inframe, cur);

    if(last) {
        if(debug) {
            show_imgattr(stderr, cur);
            show_imgattr(stderr, last);
        }

        chk = img_cmp(cur, last, cmp_type, (void*)&diff_capacity);
#if 0 
        if(debug) {
            fprintf(stderr, "same ? %d\n", chk);
        }
#endif
        if(chk) {
            if(!beep_mute) {
                fprintf(stderr, "DIFF %d\a\a\n", chk);
            }
            else {
                fprintf(stderr, "DIFF %d\n", chk);
            }

            store_image_history(cur, last);

        }
        else {
            if(debug) {
                fprintf(stderr, "SAME\n");
            }
        }
    }


    got_count++;
    c_got_count++;

    if(curtime.tv_sec-lasttime.tv_sec>report_interval) {
        if(lasttime.tv_sec<=0) {
        }
        else {
            diff = SUB_TIME(curtime, lasttime);
            rate = (double)got_count*1000.0/diff;
            fprintf(stderr, "##### rate %5.2f\n", rate);
        }
        lasttime = curtime;
        got_count = 0;
    }

    last = cur;

    return;
}


void periodical_update(XtPointer client_data, XtIntervalId *iid)
{
    struct timeval now;
    xp_periodical_rec *prof;
    int chk;
    int pos;
    sentry_imgrec *inframe;
    int  drfno;
    char *drframe;
    static sentry_imgrec *cur=NULL, *last=NULL;

    gettimeofday(&now, NULL);
    if(debug>=10) {
        printf("periodical_update: %d.%06d %p %p\n",
            now.tv_sec, now.tv_usec, client_data, iid);
    }
    prof = (xp_periodical_rec*) client_data;

/*
    show_chunk();
*/
    pos = find_pos(now.tv_sec);
    if(debug>=20) {
        printf("pos %d\n", pos);
    }
    if(pos>=0) {
#if 0
        if(got_count<=0) {
            RefreshWins(&sentry_shm_chunk[pos]);
        }
#endif
        if(c_got_count==0) {
            alloc_drimgbufs();
        }
        inframe = &sentry_shm_chunk[pos];

        if(refresh_flag){
            RefreshWins(&sentry_shm_chunk[pos].entry.header);
        }

        drfno = c_got_count%2;
        cur = &bk_drframe[drfno];

        SenDrawImage2(pwin, inframe, cur);

        if(last) {
            if(debug) {
                show_imgattr(stderr, cur);
                show_imgattr(stderr, last);
            }

            chk = img_cmp(cur, last, cmp_type, (void*)&diff_capacity);
            if(debug) {   
                fprintf(stderr, "same ? %d\n", chk);
            }
            if(chk) {
                if(!beep_mute) {
                    fprintf(stderr, "DIFF %d\a\a\n", chk);
                }
                else {
                    fprintf(stderr, "DIFF %d\n", chk);
                }
            }

        }

        got_count++;

        last = cur;
    }
    else {
        lost_count++;
    }

    if(debug<10) {
        printf(".");
        fflush(stdout);
    }


    prof->id = XtAppAddTimeOut(app_context,
        prof->interval, periodical_update, prof);

    call_count++;

    if(call_count%30==0) {
        fprintf(stderr, "##### total %d, got %d, lost %d, err%% %.2f\n",
            call_count, got_count, lost_count,
            (double)lost_count*100.0/call_count);
        fflush(stderr);
    }


    return;
}



void
DoLiveExpose(Widget w, XEvent *ev, String *s, Cardinal *c)
{
/*
*/
    printf("DoLiveExpose\n"); fflush(stdout);
}

void
DoDiffExpose(Widget w, XEvent *ev, String *s, Cardinal *c)
{
    Dimension wwid, wht;
    Dimension dwid, dht;
    GC cur_gc;
/*
*/
    fflush(stdout);
    fprintf(stderr, "DoDiffExpose\n"); 
    fflush(stderr);

    XtVaGetValues(pwin2, 
        XtNwidth,   &wwid,
        XtNheight,  &wht,
        NULL);

    dwid = MIN(640, wwid);
    dht  = MIN(480, wht);

    cur_gc = XtGetGC(pwin2, NULL, NULL);

    XCopyArea(XtDisplay(pwin2),
        diff_pix, XtWindow(pwin2), cur_gc, 0, 0, dwid, dht, 0, 0);

    XtReleaseGC(pwin2, cur_gc);

    fflush(stdout);
    fprintf(stderr, "done\n");
    fflush(stderr);
}

RefreshMsg(char *msg)
{
    Dimension wid;

    XtVaGetValues(pwinlabel, XtNwidth, &wid, NULL);
    XtVaSetValues(pwinmsg,
        XtNlabel, msg,
        XtNwidth, wid,
        NULL);
}

static int img_his_view = 0;

void
DoDiffRotateUp(Widget w, XEvent *ev, String *s, Cardinal *c)
{
    char msg[BUFSIZ];
    int n;

    n = (img_his_view + img_his_radix + 1) % img_his_radix;

#if 0 
    sprintf(msg, "up %3d %3d %3d", n, img_his_view, img_his_radix);
    RefreshMsg(msg);
#endif
    RefreshHist(n);

    img_his_view = n;
}
void
DoDiffRotateDown(Widget w, XEvent *ev, String *s, Cardinal *c)
{
    char msg[BUFSIZ];
    int n;

    n = (img_his_view + img_his_radix - 1) % img_his_radix;

#if 0
    sprintf(msg, "down %3d %3d %3d", n, img_his_view, img_his_radix);
    RefreshMsg(msg);
#endif
    RefreshHist(n);

    img_his_view = n;
}

void
DoDiffRotate(Widget w, XEvent *ev, String *s, Cardinal *c)
{
    char msg[BUFSIZ];

    if(ev && ev->type==KeyPress) {
        sprintf(msg, "key %d", ev->xkey.keycode);
        RefreshMsg(msg);
    }
    else {
        sprintf(msg, "press any key");
        RefreshMsg(msg);
    }
}

void
DoResize(Widget w, XEvent *ev, String *s, Cardinal *c)
{
    refresh_flag = 1;
}

#if 0
static String pwin_trans="#override\n\
    <Expose>:           live-expose()\n\
    <Key>:              live-expose()\n\
";
#endif

static String pwin_trans="#override\n\
";

static String pwin2_trans="#override\n\
    <Expose>:           diff-expose()\n\
    <Key>Up:            diff-rotate-up()\n\
    <Key>Down:          diff-rotate-down()\n\
    <Key>Left:          diff-rotate-down()\n\
    <Key>Right:         diff-rotate-up()\n\
";

static String common_trans="#override\n\
    <Configure>:        live-resize()\n\
";

static XtActionsRec pwin_action_list[] = {
    {"live-expose",     DoLiveExpose},
    {"live-resize",     DoResize},
    {"diff-expose",     DoDiffExpose},
    {"diff-rotate-up",  DoDiffRotateUp},
    {"diff-rotate-down",DoDiffRotateDown},
};


void
quit_sentry(Widget w, XtPointer client_data, XtPointer call_data)
{
    exit(0);
}

void
toggle_mute(Widget w, XtPointer client_data, XtPointer call_data)
{
    beep_mute = 1 - beep_mute;
    fprintf(stderr, "beep_mute %d\n", beep_mute);
    fflush(stderr);

    if(beep_mute) {
        XtVaSetValues(bmute, XtNforeground, black);
        XtVaSetValues(bmute, XtNbackground, white);
    }
    else {
        XtVaSetValues(bmute, XtNforeground, white);
        XtVaSetValues(bmute, XtNbackground, black);
    }
}

void
toggle_ghost(Widget w, XtPointer client_data, XtPointer call_data)
{
    ghost_mode = 1 - ghost_mode;
    fprintf(stderr, "ghost_mode %d\n", ghost_mode);
    fflush(stderr);

    if(ghost_mode) {
        XtVaSetValues(bghost, XtNforeground, black);
        XtVaSetValues(bghost, XtNbackground, white);
    }
    else {
        XtVaSetValues(bghost, XtNforeground, white);
        XtVaSetValues(bghost, XtNbackground, black);
    }
}




void
usage()
{
        printf("sentry - image viewer of sentry system\n\
usage: sentry [options]\n\
\n\
option:\n\
    -help\n\
    -servername host    default host is localhost\n\
    -serverport port    \n\
    -dataport port      \n\
    -reqrate num        frame rate within per 100sec\n\
    -format format      frame format (e.g., NONE, JPEG, PNG, etc.)\n\
    -shmkey key         \n\
    -debug num          debug leve\n\
\n\
note:\n\
  * you should change dataport when you start multipul program\n\
  * frame rate does not changed by this program, if you access\n\
    with shared memory\n\
\n\
example:\n\
  %% sentry -severname camera1 -dataport 723\n\
  %% sentry -shmkey 129\n\
");
}


store_color(Pixel *sa, int r, int g, int b)
{
    if(output_corder==SEN_C_XRGB) {
        *sa  = r*0x010000;
        *sa += g*0x000100;
        *sa += b*0x000001;
    }
    else 
    if(output_corder==SEN_C_XBGR) {
        *sa  = b*0x010000;
        *sa += g*0x000100;
        *sa += r*0x000001;
    }
    
}


int
split_hostport(char *hn, char *ps, char *hp)
{
    char *p, *q;
    int c;

    p = hp;
    q = hn;
    c = 0;
    while(*p && *p!=':' && c<BUFSIZ) {
        *q++ = *p++;
        c++;
    }
    *q = '\0';

    if(!*p) {
        return 1;
    }

    p++; /* skip ':' */
    
    q = ps;
    while(*p && c<BUFSIZ) {
        *q++ = *p++;
        c++;
    }
    *q = '\0';
    
}

int
recv_image_multi(int datafd, subrecord_t **recvlist, char **buff,
    sentry_shmrec *info, sentry_imgrec *dest);

void
cwin_update(caddr_t client_data, int *fd, XtInputId *id)
{
    cnode_t *cn;
    char msg[BUFSIZ];
    int nread;
    sentry_shmrec info;
    int chk;
    char *dummy;
    struct tm htm;

    cn = (cnode_t*)client_data;

#if 0
    fprintf(stderr, "cn %p: host %s, id %p %d\n",
        cn, cn->serv_host, id, *id);
#endif
    
    chk = recv_image_multi(*fd, &(cn->recbuf), &dummy, &info, &cn->inimg);
    SenDrawImage2(cn->Wimg, &cn->inimg, &cn->outimg);

#if 0
    localtime_r(&(cn->inimg.recv_sec), &htm);
    sprintf(msg, "%2d-%02d %02d:%02d:%02d.%02d",
        htm.tm_mon+1, htm.tm_mday,
        htm.tm_hour, htm.tm_min, htm.tm_sec, cn->inimg.recv_usec/10000);

    XtVaSetValues(cn->Wmsg,
        XtNlabel, msg,
        NULL);
#endif


}

int
make_cwin(char *tmppeer)
{
    char host[BUFSIZ];
    char portstr[BUFSIZ];
    int port;
    cnode_t ncn;
    int i;
    int panelwid;
    int c;

    int cnum;
    WidgetList clist;
    Widget nbaselast;
    Widget nbase;
    Widget nlabel;
    Widget nimg;
    Widget nmsg;

    host[0] = '\0';
    portstr[0] = '\0';
    split_hostport(host, portstr, tmppeer);

    if(*portstr) {
        port = atoi(portstr);
    }
    else {
        port = SENTRY_CTRL_PORT;
    }

    fprintf(stderr, "make_cwin: '%s' -> '%s' '%s' %d\n",
        tmppeer, host, portstr, port);

    setdefault_cnode(&ncn);
    show_cnode(stderr, "before", &ncn);

    ncn.name      = strdup(tmppeer);
    ncn.serv_host = strdup(host);
    ncn.serv_port = port;

    c = 0;
    do {
        ncn.data_fd = build_data_port(ncn.recv_port);
fprintf(stderr, "    port %d -> fd %d\n", ncn.recv_port, ncn.data_fd);
        if(ncn.data_fd<0)
            ncn.recv_port++;
        c++;
    } while(ncn.data_fd<0 && c<10);

    ncn.cntl_fd = build_ctrl_port(ncn.serv_host, ncn.serv_port);
    if (ncn.cntl_fd < 0) {
        printf("fail setup control fd with port# %d (%d)\n",
            port, errno);
    }

    show_cnode(stderr, "wake", &ncn);


    /*
     *
     */

    XtVaGetValues(pcpanel,
        XtNnumChildren, &cnum, 
        XtNchildren, &clist, NULL);

    nbaselast = NULL;
    fprintf(stderr, "cnum %d: ", cnum);
    for(i=0;i<cnum;i++) {
        fprintf(stderr, " %p", clist[i]);
        nbaselast = clist[i];
    }
    fprintf(stderr, "\n");

    panelwid = (cnum+1) * 170 + (cnum+2) * 20;
    fprintf(stderr, "panelwidth %d\n", panelwid);

    if(nbaselast) {
        fprintf(stderr, " right on %p\n", nbaselast);

    nbase = XtVaCreateManagedWidget(
        "nbase", formWidgetClass, pcpanel,
        XtNvisual,  xp_foil_visual,
        XtNdepth,   xp_foil_depth,
        XtNcolormap,xp_foil_cmap,
        XtNfromHoriz, nbaselast,
        XtNwidth, 170,
        XtNheight, 170,
/*
        XtNbackground, 0x00cccc,
*/
        XtNresizable,   TRUE,
        NULL);

    }
    else {

    nbase = XtVaCreateManagedWidget(
        "nbase", formWidgetClass, pcpanel,
        XtNvisual,  xp_foil_visual,
        XtNdepth,   xp_foil_depth,
        XtNcolormap,xp_foil_cmap,
        XtNwidth, 170,
        XtNheight, 170,
/*
        XtNbackground, 0xcccc00,
*/
        XtNresizable,   TRUE,
        NULL);

    }

fprintf(stderr, "nbase %d %p\n", nbase, nbase);

    nlabel = XtVaCreateManagedWidget(
        "label", labelWidgetClass, nbase,
        XtNlabel, "----------------------",
        XtNjustify, XtJustifyCenter,
        XtNborderWidth, 0,
        XtNresize,  True,
        NULL);

    nimg = XtVaCreateManagedWidget(
        "live", lwwWidgetClass, nbase,
        XtNfromVert, nlabel,
        XtNvisual,  xp_foil_visual,
        XtNdepth,   xp_foil_depth,
        XtNcolormap,xp_foil_cmap,
/*
        XtNwidth,  320, XtNheight, 240,
*/
        XtNwidth,  160, XtNheight, 120,
        NULL);

    nmsg = XtVaCreateManagedWidget(
        "difflabel", labelWidgetClass, nbase,
        XtNlabel, "no topic",
        XtNfromVert, nimg,
        XtNborderWidth, 0,
        XtNwidth,       160,
        NULL);

    XtVaSetValues(nlabel,
        XtNlabel, tmppeer,
        NULL);

    XtVaSetValues(pcpanel,
        XtNwidth, panelwid,
        XtNheight, 170,
        NULL);

    XtVaSetValues(pbase,
        XtNwidth, panelwid,
        XtNheight, 300,
        NULL);

    XtVaSetValues(pshell,
        XtNwidth, panelwid,
        XtNheight, 400,
        NULL);



    XawFormDoLayout(pcpanel, TRUE);

    ncn.Wimg = nimg;
    ncn.Wlabel = nlabel;
    ncn.Wmsg = nmsg;


 {
    cnode_t *tmp;
    sentry_protocol3_request_hdr req;
    int nwrite;
    int tmpiid;

    tmp = (cnode_t*)malloc(sizeof(cnode_t));
    *tmp = ncn;

/*
    tmpiid = XtAppAddInput(app_context, tmp->data_fd,
                (caddr_t)XtInputReadMask, eventical_update, (void*)tmp);
*/
    tmpiid = XtAppAddInput(app_context, tmp->data_fd,
                (caddr_t)XtInputReadMask, cwin_update, (void*)tmp);

    fprintf(stderr, "tmpiid %d with fd %d\n", tmpiid, tmp->data_fd);

    req.hd_ver  = SENTRY_PROTOCOL_VERSION;
    req.type    = OPEN_UDP_UNICAST;
    req.port    = htons(tmp->recv_port);
    req.size    = htonl(tmp->size);
    req.zoom    = htonl(tmp->zoom);
    req.color   = htonl(tmp->color);
    req.format  = htonl(tmp->format);
    req.rate    = htonl(tmp->rate);

    nwrite = write(tmp->cntl_fd, (void*)&req, sizeof(req));

 }


goto out;

    close(ncn.cntl_fd);
    ncn.cntl_fd = -1;

    close(ncn.data_fd);
    ncn.data_fd = -1;
    

out:
    return 0;
}


#if 0
fd_set  InputFdMask;
#endif
subrecord_t *recvlist=NULL;


/* MAIN */

int
main(int argc, char **argv)
{
    int skey;
    XGCValues values;
    int valuemask;
    int tv;
    int nhistory;

    rtr_initmem();

    toplevel = XtVaAppInitialize(
        &app_context,                   /* Application context */
        "Xp",                           /* Application class */
        options, XtNumber(options),     /* option list */
        &argc, argv,                    /* command line args */
        fallback_resources,             /* for missing app-defaults file */
        XtNtitle,       "XP",
        XtNiconName,    "XP",
        NULL);                          /* terminate varargs list */

    XtVaGetApplicationResources(toplevel,
        (XtPointer) &sentry_resource, resources, XtNumber(resources), NULL);

    XtAppAddActions(app_context, main_action_list, XtNumber(main_action_list));

#if 0
    nhistory = atoi(sentry_resource.img_history);
    init_image_history(nhistory);
#endif

    tv = assoc_n(matching_type_alist, sentry_resource.compare_func);
    if(debug) {
        fprintf(stderr, "compare type %d %x '%s' <= '%s'\n",
            tv, tv, assoc_v(matching_type_alist, tv),
            sentry_resource.compare_func);
    }
    cmp_type = tv;

    if(debug) {
        fprintf(stderr, "severname '%s'\n", sentry_resource.server_name);
    }

    tv = assoc_n(format_alist, sentry_resource.req_format);
    server_format = tv;

    tv = assoc_n(size_alist, sentry_resource.req_size);
    server_size = tv;

    if(debug>=10) {
    fprintf(stderr, "format %d, '%s' => '%s'\n",    
        tv, sentry_resource.req_format, assoc_v(format_alist, tv));
    fprintf(stderr, "size %d, '%s' => '%s'\n",    
        tv, sentry_resource.req_size, assoc_v(size_alist, tv));
    }

#if 0
    rtr_set_global_param(RTR_P_VERBOSE, NULL);
    if(debug) {
        rtr_set_global_param(RTR_P_DEBUG, NULL);
    }
#endif

    if(show_help) {
        usage();
        exit(0);
    }

    fflush(stdout);
    fflush(stderr);
    SetupVisual(XtDisplay(toplevel), fvisual, svisual);
    fflush(stdout);
    fflush(stderr);

    output_format = SEN_F_RGB;
#if 0
    /* corder is solved by SetupVisual() */
    output_corder = SEN_F_RGB;
#endif

    store_color(&black, 0x00, 0x00, 0x00);
    store_color(&white, 0xff, 0xff, 0xff);

    diff_pix =  XCreatePixmap(XtDisplay(toplevel),
                DefaultRootWindow(XtDisplay(toplevel)),
                640, 480, xp_foil_depth);

    pshell = XtVaAppCreateShell(
        "sentry", "Sentry",
        topLevelShellWidgetClass, XtDisplay(toplevel),
        XtNvisual,  xp_foil_visual,
        XtNdepth,   xp_foil_depth,
        XtNcolormap, xp_foil_cmap,
        NULL);

    pbase = XtVaCreateManagedWidget(
        "base", formWidgetClass, pshell,
        XtNvisual,  xp_foil_visual,
        XtNdepth,   xp_foil_depth,
        XtNcolormap, xp_foil_cmap,
        XtNresizable, TRUE,
        NULL);

    plabel = XtVaCreateManagedWidget(
        "label", labelWidgetClass, pbase,
        XtNlabel, "multiple cameras",
        XtNjustify, XtJustifyCenter,
        XtNborderWidth, 0,
        XtNresize,  True,
        NULL);

    pcmd = XtVaCreateManagedWidget(
        "cmd", formWidgetClass, pbase,
        XtNfromVert, plabel,
        XtNresizable, FALSE,
        NULL);

    bquit = XtVaCreateManagedWidget(
        "quit", commandWidgetClass, pcmd,
        NULL);
    XtAddCallback(bquit, XtNcallback, quit_sentry, NULL);

    bmute = XtVaCreateManagedWidget(
        "mute", commandWidgetClass, pcmd,
        XtNfromHoriz, bquit,
        NULL);
    XtAddCallback(bmute, XtNcallback, toggle_mute, NULL);

    bghost = XtVaCreateManagedWidget(
        "ghost", commandWidgetClass, pcmd,
        XtNfromHoriz, bmute,
        NULL);
    XtAddCallback(bghost, XtNcallback, toggle_ghost, NULL);

    pcpanel = XtVaCreateManagedWidget(
#if 0
        "cpanel", formWidgetClass, pbase,
#endif
        "cpanel", boxWidgetClass, pbase,
        XtNorientation, XtEhorizontal,
        XtNfromVert, pcmd,
/*
        XtNforeground, 0x00ff00,
        XtNbackground, 0xff00ff,
*/
        XtNvisual,  xp_foil_visual,
        XtNdepth,   xp_foil_depth,
        XtNcolormap,xp_foil_cmap,
#if 0
#endif
        XtNwidth,  150, XtNheight, 80,
        XtNresizable,   TRUE,
        NULL);


#if 0 

    pwinlabel = XtVaCreateManagedWidget(
        "livelabel", labelWidgetClass, pbase,
        XtNlabel, "live image",
        XtNfromVert, pcmd,
        XtNborderWidth, 0,
        NULL);

    pwin = XtVaCreateManagedWidget(
        "live", lwwWidgetClass, pbase,
        XtNfromVert, pwinlabel,
        XtNvisual,  xp_foil_visual,
        XtNdepth,   xp_foil_depth,
        XtNcolormap,xp_foil_cmap,
/*
        XtNwidth,  320, XtNheight, 240,
*/
        XtNwidth,  160, XtNheight, 120,
        NULL);

    pwinlabel2 = XtVaCreateManagedWidget(
        "difflabel", labelWidgetClass, pbase,
        XtNlabel, "diff. image",
        XtNfromVert, pwin,
        XtNborderWidth, 0,
        NULL);

    pwin2 = XtVaCreateManagedWidget(
        "diff", lwwWidgetClass, pbase,
        XtNfromVert, pwinlabel2,
        XtNvisual,  xp_foil_visual,
        XtNdepth,   xp_foil_depth,
        XtNcolormap,xp_foil_cmap,
/*
        XtNwidth,  320, XtNheight, 240,
*/
        XtNwidth,  160, XtNheight, 120,
        NULL);

    pwinmsg = XtVaCreateManagedWidget(
        "difflabel", labelWidgetClass, pbase,
        XtNlabel, "no topic",
        XtNfromVert, pwin2,
        XtNborderWidth, 0,
        XtNwidth,       160,
        XtNresize,      1,
        NULL);

#endif

    pwinmsg = XtVaCreateManagedWidget(
        "difflabel", labelWidgetClass, pbase,
        XtNlabel, "",
        XtNfromVert, pcpanel,
        XtNborderWidth, 0,
        XtNwidth,       160,
        XtNresize,      1,
        NULL);

    XtRealizeWidget(pshell);

#if 0
    values.background =
        values.foreground = WhitePixel(XtDisplay(pwin),
                                DefaultScreen(XtDisplay(pwin)));
    valuemask = GCForeground | GCBackground;
    pwin_gc = XCreateGC(XtDisplay(pwin), XtWindow(pwin), valuemask, &values);
#endif
#if 0
    pwin_gc = XtGetGC(pwin, NULL, NULL);
#endif

    XtAppAddActions(app_context,
        pwin_action_list, XtNumber(pwin_action_list));
#if 0
    XtOverrideTranslations(pwin,    XtParseTranslationTable(pwin_trans));
    XtOverrideTranslations(pwin2,   XtParseTranslationTable(pwin2_trans));
#endif
#if 0
    XtOverrideTranslations(pbase,   XtParseTranslationTable(pwin_trans));
#endif
    XtOverrideTranslations(pshell,  XtParseTranslationTable(common_trans));


#if 0


PASS;
    {
        ctrlfd = build_ctrl_port(server_host, server_port);
        if (ctrlfd < 0) {
            printf("fail setup control fd with port# %d (%d)\n",
                server_port, errno);
            exit (1);
        }

PASS;
        datafd = build_data_port(data_port);
        if (datafd < 0) {
            printf("fail setup data fd with port# %d (%d)\n",
                data_port, errno);
            exit (1);
        }
    }
PASS;

    {
        fprintf(stderr, "sentry image viewer with direct network access\n");

#if 0
        init_ibuf();
#endif

        init_request();
        request_call();
        signal(SIGALRM, request_call);

PASS;

        FD_ZERO(&InputFdMask);
        FD_SET(ctrlfd, &InputFdMask);
        FD_SET(datafd, &InputFdMask);

        iid = XtAppAddInput(app_context, datafd,
                (caddr_t)XtInputReadMask, eventical_update, NULL);
#if 0
        ciid = XtAppAddInput(app_context, ctrlfd,
                (caddr_t)XtInputReadMask, retry_connect, NULL);
#endif

    }

#endif


#if 1

    XtRealizeWidget(pshell);

 {
    int i;
    for(i=1;i<argc;i++) {
        fprintf(stderr, "%4d: '%s'\n", i, argv[i]);
        fflush(stderr);
        
        make_cwin(argv[i]);
    }
 }
#endif
    

PASS;

    XtAppMainLoop(app_context);

}


