/*==
 ==*/

/*
 * nv - node viewer
 * $Id$
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <time.h>
#include <sys/time.h>
#include <signal.h>
#include <sys/wait.h>
#include <math.h>
#ifdef __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#include <errno.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xproto.h>
#include <X11/cursorfont.h>

#include <X11/Xmu/WinUtil.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/Scrollbar.h>
#include <X11/Xaw/SimpleMenu.h>

#include "Lww.h"

#include "alist.h"
#include "sentry.h"
#include "img.h"
#include "pool.h"
#include "client.h"

#include "rtr.h"
#include "sf.h"
#include "nv.h"
#include "jm.h"


#ifndef MIN
#define MIN(x,y)    ((x)<(y) ? (x) : (y))
#endif
#ifndef MAX
#define MAX(x,y)    ((x)>(y) ? (x) : (y))
#endif

#ifndef MAX_IMG_LEN   
#define MAX_IMG_LEN     (640*480*4)
#endif
#ifndef TMP_IMGBUF_LEN
#define TMP_IMGBUF_LEN  (320*240*4)
#endif

#if 0
#define PASS    fprintf(stderr,"PASS %s:%d\n", __FILE__, __LINE__)
#else
#define PASS    ((void)1)
#endif


static int img_his_radix=-1;
static sentry_imgrec *img_his=NULL;
static int img_his_pos  = 0;
static int img_his_view = 0;

#if 0
static int call_count=0, lost_count=0, got_count=0, c_got_count=0;
#endif
static int c_got_count=0;
static int refresh_flag = 1;


Visual   *xp_slide_visual,   *xp_foil_visual;
int      xp_slide_depth,     xp_foil_depth;
Colormap xp_slide_cmap,      xp_foil_cmap;
int      xp_use_special_cmap;

String fvisual, svisual;


subrecord_t *recvlist=NULL;

apair_t node_type_alist[] = {
    {"camera",  NV_T_CAMERA},
    {"viewer",  NV_T_VIEWER},
    {"filter",  NV_T_FILTER},
    {"scanner", NV_T_SCANNER},
    {NULL,      -1},
};

apair_t node_type_c_alist[] = {
    {".",  NV_T_NONE},
    {"C",  NV_T_CAMERA},
    {"V",  NV_T_VIEWER},
    {"F",  NV_T_FILTER},
    {"S",  NV_T_SCANNER},
    {NULL,      -1},
};


int run_mode=0;
int employ_limit=0;

Widget   toplevel;
XtAppContext app_context;

static XrmOptionDescRec options[] = {
{"-fg",         "*Foreground",  XrmoptionSepArg,    NULL},
{"-foreground", "*Foreground",  XrmoptionSepArg,    NULL},
{"-bg",         "*Background",  XrmoptionSepArg,    NULL},
{"-background", "*Background",  XrmoptionSepArg,    NULL},
{"-font",       "*Font",        XrmoptionSepArg,    NULL},

{"-camerahost", "*camerahost",  XrmoptionSepArg,    NULL},
{"-rmhost",     "*rmhost",      XrmoptionSepArg,    NULL},
#if 0
{"-cameraport", "*cameraport",  XrmoptionSepArg,    NULL},
#endif
{"-shape",      "*shape",       XrmoptionSepArg,    NULL},

{"-help",       "*help",        XrmoptionNoArg,     "1"},
};

typedef struct {
    String fvisual;
    String svisual;

    /* story */
    String text_color;
    String mark_color;
    String warn_color;
    String fatal_color;
    String cpos_color;

    String select_color;
    String unselect_color;

    String first_filename;

    String shape;

#if 0
    String camera_host;
    String camera_port;
#endif

    String camera_servpair_list;
    String rm_servpair;

    String show_help;
} nv_resource_t;


nv_resource_t nv_resource;

#define Offset(topofield) (XtOffsetOf(nv_resource_t, topofield))
static XtResource resources[] = {
  {"help",  "Help",  XtRString,  sizeof(String),
                    Offset(show_help), XtRString,  NULL},

  {"svisual", "SVisual",  XtRString,  sizeof(String),
                    Offset(svisual), XtRString,  NULL},
  {"fvisual", "FVisual",  XtRString,  sizeof(String),
                    Offset(fvisual), XtRString,  NULL},

  {"shape", "Shape",  XtRString,  sizeof(String),
                    Offset(shape), XtRString,  NULL},

  {"camerahost", "CameraHost",  XtRString,  sizeof(String),
                    Offset(camera_servpair_list), XtRString,  NULL},
  {"rmhost",    "RMHost",  XtRString,  sizeof(String),
                    Offset(rm_servpair), XtRString,  NULL},
};
#undef Offset


String fallback_resources[] = {
    "*sBase.story.cursor:       xterm",
    "*Title:  NV - Node Viewer",
    "*title:  NV - node viewer",
    "*help:     0",

    "*camerahost: localhost:7001",
    "*rmhost: mint040:2236",
    "*nppanel*justify: left",

    "*shape: 6",
    NULL};


#if 0
void
DoDiffRotateUp(Widget w, XEvent *ev, String *s, Cardinal *c)
{
    char msg[BUFSIZ];
    int n;

    n = (img_his_view + img_his_radix + 1) % img_his_radix;

#if 0 
    sprintf(msg, "up %3d %3d %3d", n, img_his_view, img_his_radix);
    RefreshMsg(msg);
#endif
    RefreshHist(n);

    img_his_view = n;
}

void
DoDiffRotateDown(Widget w, XEvent *ev, String *s, Cardinal *c)
{
    char msg[BUFSIZ];
    int n;

    n = (img_his_view + img_his_radix - 1) % img_his_radix;

#if 0
    sprintf(msg, "down %3d %3d %3d", n, img_his_view, img_his_radix);
    RefreshMsg(msg);
#endif
    RefreshHist(n);

    img_his_view = n;
}


void
DoDiffRotate(Widget w, XEvent *ev, String *s, Cardinal *c)
{
    char msg[BUFSIZ];

    if(ev && ev->type==KeyPress) {
        sprintf(msg, "key %d", ev->xkey.keycode);
        RefreshMsg(msg);
    }
    else {
        sprintf(msg, "press any key");
        RefreshMsg(msg);
    }
}


void
DoResize(Widget w, XEvent *ev, String *s, Cardinal *c)
{
    refresh_flag = 1;
}

void
DoLiveExpose(Widget w, XEvent *ev, String *s, Cardinal *c)
{
/*
*/
    printf("DoLiveExpose\n"); fflush(stdout);
}
#endif

#if 0
void
DoDiffExpose(Widget w, XEvent *ev, String *s, Cardinal *c)
{
    Dimension wwid, wht;
    Dimension dwid, dht;
/*
*/
#if 0
    fflush(stdout);
    fprintf(stderr, "DoDiffExpose\n"); 
    fflush(stderr);
#endif

    XtVaGetValues(pwin2, 
        XtNwidth,   &wwid,
        XtNheight,  &wht,
        NULL);

    dwid = MIN(640, wwid);
    dht  = MIN(480, wht);

    XCopyArea(XtDisplay(pwin2),
        diff_pix, XtWindow(pwin2), pwin_gc, 0, 0, dwid, dht, 0, 0);

#if 0
    fflush(stdout);
    fprintf(stderr, "done\n");
    fflush(stderr);
#endif
}
#endif


#if 0
RefreshMsg(char *msg)
{
    Dimension wid;

    XtVaGetValues(pwinlabel, XtNwidth, &wid, NULL);
    XtVaSetValues(pwinmsg,
        XtNlabel, msg,
        XtNwidth, wid,
        NULL);
}
#endif

static String pwin_trans="#override\n\
";

static String pwin2_trans="#override\n\
    <Expose>:           diff-expose()\n\
    <Key>Up:            diff-rotate-up()\n\
    <Key>Down:          diff-rotate-down()\n\
    <Key>Left:          diff-rotate-down()\n\
    <Key>Right:         diff-rotate-up()\n\
";

static String common_trans="#override\n\
    <Configure>:        live-resize()\n\
";

#if 0
static XtActionsRec pwin_action_list[] = {
    {"live-expose",     DoLiveExpose},
    {"live-resize",     DoResize},
    {"diff-expose",     DoDiffExpose},
    {"diff-rotate-up",  DoDiffRotateUp},
    {"diff-rotate-down",DoDiffRotateDown},
};
#endif

node_t *nodelist=NULL;
int node_wid = 40;
int node_ht = 30;
int node_range_near = 625;
    /* (node_wid/2*node_wid/2 + node_ht/2*node_ht/2) */
int node_range_middle = 2500;
    /* (node_wid/2*node_wid/2 + node_ht/2*node_ht/2) */

/* XXX */
long    black   = 0x000000;
long    white   = 0xffffff;
long    red     = 0xff0000;
long    blue    = 0x0000ff;

long    normc = 0x000000;
long    highc = 0xff0000;


int
print_node(FILE *ofp, node_t *this)
{
    int i;

    if(!this) {
        fprintf(ofp, "node %p\n", this);
        return -1;
    }
    
    fprintf(ofp, "node %p, id# %d name '%s', host '%s'\n",
        this, this->id, this->name, this->host);

    fprintf(ofp, "  unid ");
    for(i=0;i<MAX_LOWERNODE;i++) {
        if(this->unid[i])
            fprintf(ofp, "id#%d ", this->unid[i]->id);
        else {
            fprintf(ofp, "%p ", this->unid[i]);
        }
    }
    fprintf(ofp, "\n");

    fprintf(ofp, "  lnid ");
    for(i=0;i<MAX_LOWERNODE;i++) {
        if(this->lnid[i])
            fprintf(ofp, "id#%d ", this->lnid[i]->id);
        else {
            fprintf(ofp, "%p ", this->lnid[i]);
        }
    }
    fprintf(ofp, "\n");

    fprintf(ofp, "  Hup  ");
    for(i=0;i<MAX_LOWERNODE;i++) {
        fprintf(ofp, "%p ", this->Hup[i]);
    }
    fprintf(ofp, "\n");

    fprintf(ofp, "  Hlo  ");
    for(i=0;i<MAX_LOWERNODE;i++) {
        fprintf(ofp, "%p ", this->Hlo[i]);
    }
    fprintf(ofp, "\n");


}

node_t*
new_node(void)
{
    node_t *r;
    static int seq=0;
    int i;

    r = malloc(sizeof(node_t));
    if(r==NULL) {
        fprintf(stderr, "fail malloc\n");
    }
    if(r) {
        r->id   = seq++;
        r->type = NV_T_NONE;
        r->name = NULL;
        r->ctime = time(NULL);
        r->host = NULL;
        r->sitter_port = COMM_SITTER_PORT;
        r->prog_name[0] = '\0';
        r->filtermode = FILTER_THRU;
        r->down_listen_port = -1;
        r->dlcid = NULL;
        r->up_recv_port = -1;
        r->prop = NULL;
        r->desc = NULL;
        r->sitter_fd = -1;
        r->x    = -1;
        r->y    = -1;
        r->hidle = -1;
        for(i=0;i<MAX_LOWERNODE;i++) r->unid[i] = NULL;
        for(i=0;i<MAX_LOWERNODE;i++) r->lnid[i] = NULL;
        for(i=0;i<MAX_LOWERNODE;i++) r->ucid[i] = NULL;
        for(i=0;i<MAX_LOWERNODE;i++) r->lcid[i] = NULL;
        r->next = NULL;
    }
    return r;
}

int
add_node(node_t *n)
{
    if(nodelist)
        n->next = nodelist;
    nodelist = n;

    return 0;
}

void
print_nodelist_low(FILE *fp, int df)
{
    int i, c;
    node_t *cur;
    char *fmt;
    char *fmt_df   =
        "%3d 0x%08x : %3d %s %-8s : %-9s %5d %5d %-8s : %3d %3d %3d : 0x%08x";
    char *fmt_norm = "%3d: %3d %s %-8s : %-9s %5d %5d %-8s : %3d %3d %3d";

    if(df) {
        fmt = fmt_df;
    }
    else {
        fmt = fmt_norm;
    }

    c = 0;
    cur = nodelist;
    while(cur) {    
        if(df) {

            fprintf(fp, fmt_df,
                c, cur, cur->id, 
                assoc_v(node_type_c_alist, cur->type),
                cur->name ? cur->name : "-",
                cur->host ? cur->host : "-",
                cur->down_listen_port,
                cur->up_recv_port,
                cur->prog_name[0] ? cur->prog_name : "-", 
                cur->sitter_fd,
                cur->x, cur->y,
                cur->next);

        }
        else {

            fprintf(fp, fmt_norm,
                c, cur->id, 
                assoc_v(node_type_c_alist, cur->type),
                cur->name ? cur->name : "-",
                cur->host ? cur->host : "-",
                cur->down_listen_port,
                cur->up_recv_port,
                cur->prog_name[0] ? cur->prog_name : "-", 
                cur->sitter_fd,
                cur->x, cur->y);

        }

        if(df)
            fprintf(fp, " :unid");
        else
            fprintf(fp, " :up");

        for(i=0;i<MAX_LOWERNODE;i++) {
            if(df) {
                if(cur->unid[i])
                    fprintf(fp, " %d-0x%08x", i, cur->unid[i]);
            }
            else {
                if(cur->unid[i])
                    fprintf(fp, " #%d", cur->unid[i]->id);
            }
        }

        if(df)
            fprintf(fp, " :lnid");
        else
            fprintf(fp, " :lo");
        for(i=0;i<MAX_LOWERNODE;i++) {
            if(df) {
                if(cur->lnid[i])
                    fprintf(fp, " %d-0x%08x", i, cur->lnid[i]);
            }
            else {
                if(cur->lnid[i])
                    fprintf(fp, " #%d", cur->lnid[i]->id);
            }
        }

        fprintf(fp, "\n");

        cur = cur->next;
        c++;
    }
    fflush(fp);
}

void
print_nodelist_Hlow(FILE *fp, int df)
{
    int i, c;
    node_t *cur;
    char *fmt;
    char *fmt_df   =
        "%3d 0x%08x : %3d %s %-8s : %-9s %5d %-8s : %3d %3d %3d : 0x%08x";
    char *fmt_norm = "%3d: %3d %s %-8s : %-9s %5d %-8s : %3d %3d %3d";

    if(df) {
        fmt = fmt_df;
    }
    else {
        fmt = fmt_norm;
    }

    c = 0;
    cur = nodelist;
    while(cur) {    
        if(df) {

            fprintf(fp, fmt_df,
                c, cur, cur->id, 
                assoc_v(node_type_c_alist, cur->type),
                cur->name ? cur->name : "-",
                cur->host ? cur->host : "-",
                cur->down_listen_port,
                cur->prog_name[0] ? cur->prog_name : "-", 
                cur->sitter_fd,
                cur->x, cur->y,
                cur->next);

        }
        else {

            fprintf(fp, fmt_norm,
                c, cur->id, 
                assoc_v(node_type_c_alist, cur->type),
                cur->name ? cur->name : "-",
                cur->host ? cur->host : "-",
                cur->down_listen_port,
                cur->prog_name[0] ? cur->prog_name : "-", 
                cur->sitter_fd,
                cur->x, cur->y);

        }

        if(df)
            fprintf(fp, " :Hup");
        else
            fprintf(fp, " :Hu");

        for(i=0;i<MAX_LOWERNODE;i++) {
            if(df) {
                if(cur->Hup[i])
                    fprintf(fp, " %d-0x%08x", i, cur->Hup[i]);
            }
            else {
                if(cur->Hup[i])
                    fprintf(fp, " #%d", cur->Hup[i]->id);
            }
        }

        if(df)
            fprintf(fp, " :Hlo");
        else
            fprintf(fp, " :Hl");
        for(i=0;i<MAX_LOWERNODE;i++) {
            if(df) {
                if(cur->Hlo[i])
                    fprintf(fp, " %d-0x%08x", i, cur->Hlo[i]);
            }
            else {
                if(cur->Hlo[i])
                    fprintf(fp, " #%d", cur->Hlo[i]->id);
            }
        }

        fprintf(fp, "\n");

        cur = cur->next;
        c++;
    }
    fflush(fp);
}

int
print_nodelist(FILE *fp)
{
    print_nodelist_low(fp, 1);
    print_nodelist_low(fp, 0);
}

int
print_nodelistH(FILE *fp)
{
    print_nodelist_Hlow(fp, 1);
    print_nodelist_Hlow(fp, 0);
}



int
entry_upper(node_t *cnode, node_t *u)
{
    int i;
    int p, q;
    int s, t;

    if(!cnode)
        return -1;

    p = -1;
    for(i=0;i<MAX_LOWERNODE;i++) {
        if(u->lnid[i]==cnode) {
            p = i;
            break;
        }
    }
    s = -1;
    if(p==-1) {
        for(i=0;i<MAX_LOWERNODE;i++) {
            if(u->lnid[i]==NULL) {
                s = i;
                u->lnid[i] = cnode;
                break;
            }
        }
    }

    q = -1;
    for(i=0;i<MAX_LOWERNODE;i++) {
        if(cnode->unid[i]==u) {
            q = i;
            break;
        }
    }
    t = -1;
    if(q==-1) {
        for(i=0;i<MAX_LOWERNODE;i++) {
            if(cnode->unid[i]==NULL) {
                t = i;
                cnode->unid[i] = u;
                break;
            }
        }
    }

printf("p %d, q %d ; s %d t %d\n", p, q, s, t);

    if(q>=0) {
        return 0;
    }

    return 1;
}

int
entry_lower(node_t *cnode, node_t *l)
{
    int i;

    if(!cnode)
        return -1;

    for(i=0;i<MAX_LOWERNODE;i++) {
        if(cnode->lnid[i]==NULL) {
            cnode->lnid[i] = l;
            return 0;
        }
    }

    return 1;
}

int
draw_link(Display *dpy, Window win, GC gc, node_t *fr, node_t *to)
{
    int sx, sy;
    int ex, ey;
    int cx, cy;
    int dx, dy;
    double a, c;
    double dd;
    int r1, r2;
    XPoint p[3];

    r1 = 5;
    r2 = 10;
    
    sx = fr->x + node_wid/4;    sy = fr->y;
    ex = to->x - node_wid/4;    ey = to->y;
    cx = (sx+ex)/2;             cy = (sy+ey)/2;
    dx = ex-sx;                 dy = ey-sy;

    a = ((double)dy)/dx;
    c = ((double)dx)/dy;
    dd = sqrt(dx*dx+dy*dy);

#if 0
    fprintf(stderr, "s %d,%d ; e %d,%d ; c %d,%d\n",
        sx, sy, ex, ey, cx, cy);
#endif

    XFillRectangle(dpy, win, gc, sx-r1/2, sy-r1/2, r1, r1);
    XFillRectangle(dpy, win, gc, ex-r1/2, ey-r1/2, r1, r1);
    XDrawLine(dpy, win, gc, sx, sy, ex, ey);

    if(dx==0) {
        p[0].x = cx-r1; p[0].y = cy;
        p[1].x = cx+r1; p[1].y = cy;
        p[2].x = cx;
        if(dy<0) {      p[2].y  = cy-r2;
        } else {        p[2].y  = cy+r2; }
        goto draw_arrow;
    }
    else
    if(dy==0) {
        p[0].x = cx;    p[0].y = cy-5;
        p[1].x = cx;    p[1].y = cy+5;
        if(dx>0) {      p[2].x = cx+r2; 
        } else {        p[2].x = cx-r2; }
        p[2].y  = cy;
        goto draw_arrow;
    }

#if 0
    fprintf(stderr, "dx %d, dy %d => p1 %d,%d ; p2 %d,%d ; p3 %d,%d\n",
        dx, dy, p1x, p1y, p2x, p2y, p3x, p3y);
#endif

    p[0].x = cx + (int)((double)dy*r1/dd);
    p[0].y = cy - (int)((double)dx*r1/dd);
    p[1].x = cx - (int)((double)dy*r1/dd);
    p[1].y = cy + (int)((double)dx*r1/dd);
    p[2].x = cx + (int)((double)dx*r2/dd);
    p[2].y = cy + (int)((double)dy*r2/dd);

draw_arrow:
    XFillPolygon(dpy, win, gc, p, 3, Convex, CoordModeOrigin);

}

#if 0
char *_nlf_name = "6x13bold";
#endif
char *_nlf_name = "a14";
XFontStruct *_nlf_st = NULL;

int
load_node_label_prop(Display *dpy)
{
    if(_nlf_st==NULL) {
        _nlf_st = XLoadQueryFont(dpy, _nlf_name);
        if(_nlf_st==NULL) {
            fprintf(stderr, "fail XLoadQueryFont\n");
            return -1;
        }
#if 1
        fprintf(stderr, "_nlf name '%s', st %p\n",
                _nlf_name, _nlf_st);
        if(_nlf_st)
            fprintf(stderr, "\tnum %d\n", _nlf_st->fid);
#endif
    }
    return 0;
}

void
draw_node_label(Display *dpy, Window win, GC gc,
    int x, int y, char *dt, int jh, int jv)
{
    int dtwid, dtht, dtdep;
    int gx, gy;
    int dtlen;
    Font fid;
    XGCValues gcv;

    dtlen = strlen(dt);

    if(_nlf_st==NULL) {
        load_node_label_prop(dpy);
    }
#if 0
    if(_nlf_st==NULL) {
        _nlf_st = XLoadQueryFont(dpy, _nlf_name);
        if(_nlf_st==NULL) {
            fprintf(stderr, "fail XLoadQueryFont\n");
            gx = x;
            gy = y;
            goto rroll;
        }
#if 1
        fprintf(stderr, "_nlf name '%s', st %p\n",
                _nlf_name, _nlf_st);
        if(_nlf_st)
            fprintf(stderr, "\tnum %d\n", _nlf_st->fid);
#endif
    }
#endif

    dtht  = _nlf_st->ascent;
    dtdep = _nlf_st->descent;

    XSetFont(dpy, gc, _nlf_st->fid);
    dtwid = XTextWidth(_nlf_st, dt, dtlen);
    if(dtwid==0) {
        dtwid = dtlen*7;        /* XXX */
    }

    switch(jh) {
    case 0:     gx = x;         break;
    case 1:     gx = x-dtwid/2; break;
    case 2:     gx = x-dtwid;   break;
    }
    switch(jv) {
    case 0:     gy = y;         break;
    case 1:     gy = y+(dtht)/2;    break;
    case 2:     gy = y+dtht;    break;
    }

#if 0
    fprintf(stderr,
    "just %d,%d; x %3d wid %3d - %3d; y %3d ht %2d dep %2d - %3d; %2d '%s'\n",
        jh, jv, x, dtwid, gx, y, dtht, dtdep, gy, dtlen, dt);
#endif
    
rroll:
#if 0
    XDrawLine(dpy, win, gc, x-5, y-5, x+5, y+5);
    XDrawLine(dpy, win, gc, x+5, y-5, x-5, y+5);
    XDrawLine(dpy, win, gc, gx-5, gy, gx+5, gy);
    XDrawLine(dpy, win, gc, gx, gy-5, gx, gy+5);
#endif
    XDrawString(dpy, win, gc, gx, gy, dt, dtlen);
}

int
draw_node_shape(Display *dpy, Window win, GC gc, node_t *nn)
{
    switch(nn->type) {
    case NV_T_VIEWER:
        XDrawRectangle(dpy, win, gc,
            nn->x-node_wid/2, nn->y-node_ht/2, node_wid, node_ht);
        XDrawRectangle(dpy, win, gc,
            nn->x-node_wid*3/8, nn->y-node_ht*3/8, node_wid*3/4, node_ht*3/4);
        break;
    case NV_T_CAMERA:
        XDrawRectangle(dpy, win, gc,
            nn->x-node_wid/3, nn->y-node_ht/2, node_wid*5/6, node_ht);
        XDrawLine(dpy, win, gc,
            nn->x-node_wid/3, nn->y-node_ht/4, 
            nn->x-node_wid/2, nn->y-node_ht/3);
        XDrawLine(dpy, win, gc,
            nn->x-node_wid/2, nn->y-node_ht/3,
            nn->x-node_wid/2, nn->y+node_ht/3);
        XDrawLine(dpy, win, gc,
            nn->x-node_wid/3, nn->y+node_ht/4, 
            nn->x-node_wid/2, nn->y+node_ht/3);
        break;
#if 0
        XDrawLine(dpy, win, gc,
            nn->x-node_wid/2, nn->y, nn->x, nn->y-node_ht/2);
        XDrawLine(dpy, win, gc,
            nn->x+node_wid/2, nn->y, nn->x, nn->y-node_ht/2);
        XDrawLine(dpy, win, gc,
            nn->x-node_wid/2, nn->y, nn->x, nn->y+node_ht/2);
        XDrawLine(dpy, win, gc,
            nn->x+node_wid/2, nn->y, nn->x, nn->y+node_ht/2);
        break;
#endif

    case NV_T_FILTER:
        /* 3box */
        XDrawRectangle(dpy, win, gc,
            nn->x-node_wid*2/8, nn->y-node_ht*4/8, node_wid*2/4, node_ht*4/4);
        XDrawRectangle(dpy, win, gc,
            nn->x-node_wid*4/8, nn->y-node_ht*2/8, node_wid*1/4, node_ht*2/4);
        XDrawRectangle(dpy, win, gc,
            nn->x+node_wid*2/8, nn->y-node_ht*2/8, node_wid*1/4, node_ht*2/4);
        break;

    case NV_T_SCANNER:
        /* 3box + cross*/
        XDrawRectangle(dpy, win, gc,
            nn->x-node_wid*2/8, nn->y-node_ht*4/8, node_wid*2/4, node_ht*4/4);
        XDrawRectangle(dpy, win, gc,
            nn->x-node_wid*4/8, nn->y-node_ht*2/8, node_wid*1/4, node_ht*2/4);
        XDrawRectangle(dpy, win, gc,
            nn->x+node_wid*2/8, nn->y-node_ht*2/8, node_wid*1/4, node_ht*2/4);
        XDrawLine(dpy, win, gc,
            nn->x+node_wid*2/8, nn->y-node_ht*2/8,
            nn->x+node_wid*4/8, nn->y+node_ht*2/8);
        XDrawLine(dpy, win, gc,
            nn->x+node_wid*4/8, nn->y-node_ht*2/8,
            nn->x+node_wid*2/8, nn->y+node_ht*2/8);
        break;

    default:
#if 0
        /* grid */
        XDrawRectangle(dpy, win, gc,
            nn->x-node_wid/2, nn->y-node_ht/2, node_wid, node_ht);
        XDrawRectangle(dpy, win, gc,
            nn->x-node_wid*3/8, nn->y-node_ht*3/8, node_wid*3/4, node_ht*3/4);
        XDrawLine(dpy, win, gc,
            nn->x-node_wid*3/8, nn->y-node_ht*2/8,
            nn->x+node_wid*3/8, nn->y-node_ht*2/8);
        XDrawLine(dpy, win, gc,
            nn->x-node_wid*3/8, nn->y-node_ht*1/8,
            nn->x+node_wid*3/8, nn->y-node_ht*1/8);
        XDrawLine(dpy, win, gc,
            nn->x-node_wid*3/8, nn->y-node_ht*0/8,
            nn->x+node_wid*3/8, nn->y-node_ht*0/8);
        XDrawLine(dpy, win, gc,
            nn->x-node_wid*3/8, nn->y+node_ht*1/8,
            nn->x+node_wid*3/8, nn->y+node_ht*1/8);
        XDrawLine(dpy, win, gc,
            nn->x-node_wid*3/8, nn->y+node_ht*2/8,
            nn->x+node_wid*3/8, nn->y+node_ht*2/8);
        XDrawLine(dpy, win, gc,
            nn->x-node_wid*2/8, nn->y-node_ht*3/8,
            nn->x-node_wid*2/8, nn->y+node_ht*3/8);
        XDrawLine(dpy, win, gc,
            nn->x-node_wid*1/8, nn->y-node_ht*3/8,
            nn->x-node_wid*1/8, nn->y+node_ht*3/8);
        XDrawLine(dpy, win, gc,
            nn->x-node_wid*0/8, nn->y-node_ht*3/8,
            nn->x-node_wid*0/8, nn->y+node_ht*3/8);
        XDrawLine(dpy, win, gc,
            nn->x+node_wid*1/8, nn->y-node_ht*3/8,
            nn->x+node_wid*1/8, nn->y+node_ht*3/8);
        XDrawLine(dpy, win, gc,
            nn->x+node_wid*2/8, nn->y-node_ht*3/8,
            nn->x+node_wid*2/8, nn->y+node_ht*3/8);
#endif

        /* black box */
        XFillRectangle(dpy, win, gc,
            nn->x-node_wid/2, nn->y-node_ht/2, node_wid, node_ht);
        break;
    }
}

int
draw_node(Display *dpy, Window win, GC gc, node_t *nn, int opt)
{
    node_t *u;
    int i;
    char tmp[BUFSIZ];

    if(!nn) {
        return -1;
    }

#if 1
    opt = 0;
#endif

    draw_node_shape(dpy, win, gc, nn);

    if(_nlf_st==NULL) {
        load_node_label_prop(dpy);
    }
    if(!_nlf_st) {
        goto skip_labels;
    }

    /*
     * labels
     */
    if(nn->type!=NV_T_VIEWER) {
        draw_node_label(dpy, win, gc,
            nn->x, nn->y-node_ht/2, nn->host ? nn->host : "*undef*", 1, 0);
#if 0
        if(nn->prog_name) {
                draw_node_label(dpy, win, gc,
                    nn->x, nn->y+node_ht/2, nn->prog_name, 1, 2);
        }
#endif
    }
    if(nn->name) {
        draw_node_label(dpy, win, gc,
            nn->x, nn->y-node_ht/2-(_nlf_st->ascent+_nlf_st->descent),
            nn->name, 1, 0);
    }

#if 0
    draw_node_label(dpy, win, gc,
        nn->x, nn->y+node_ht/2+(_nlf_st->ascent+_nlf_st->descent),
        nn->desc ? nn->desc : "-", 1, 2);
#endif

    sprintf(tmp, "%d", nn->hidle);
    draw_node_label(dpy, win, gc,
        nn->x, nn->y+node_ht/2, tmp, 1, 2);

    if(nn->type==NV_T_FILTER||nn->type==NV_T_SCANNER) {
        sprintf(tmp, "%s", assoc_v(filter_name_al, nn->filtermode));
        draw_node_label(dpy, win, gc,
            nn->x, nn->y+node_ht/2+(_nlf_st->ascent+_nlf_st->descent),
            tmp, 1, 2);
    }

skip_labels:


    for(i=0;i<MAX_LOWERNODE;i++) {
        if(nn->unid[i]) {
            draw_link(dpy, win, gc, nn->unid[i], nn);
        }
    }
    for(i=0;i<MAX_LOWERNODE;i++) {
        if(nn->lnid[i]) {
            draw_link(dpy, win, gc, nn, nn->lnid[i]);
        }
    }
}

typedef struct {
    Widget  Vw;
    Widget  Lw;
    Widget  Rw;
    int     req_sendfd;
    int     res_recvfd;
    node_t  *node;
    XtIntervalId tid;
    int     running;

    subrecord_t *pque;

    /* for stat. */
    int     got_count;
    struct timeval lasttime;
} viewer_rec;

int max_viewer=3;
viewer_rec vslot[MAX_VIEWER];

Widget ground;
Widget topofield;
Widget pvpanel;
Widget pvlabel[MAX_VIEWER];
Widget pview[MAX_VIEWER];
#if 0
Widget pvtime[MAX_VIEWER];
#endif
Widget pvrate[MAX_VIEWER];

Widget tcpanel;
Widget mcpanel;
Widget bcpanel;
Widget tcpanelh;
Widget mcpanelh;
Widget bcpanelh;
Widget bquit, bhelp;
#if 0
Widget bncamera, bnfilter, bredraw, bcheck;
#endif
Widget bnfilter, bredraw, bcheck;
Widget bstart, bstop, bswapper;
#if 0
Widget bstart, bstop, bredraw, bcheck, bswapper;
#endif

Widget smsg;
Widget tmsg;

node_t *nprop_target_node=NULL;
int node_prop_tryfiltermode=FILTER_THRU;
Widget node_prop_shell;
Widget node_prop_base;
Widget node_prop_spanel;
Widget node_prop_panel;
Widget node_prop_unb; Widget node_prop_unbh;
Widget node_prop_lnb; Widget node_prop_lnbh;
Widget node_prop_ucb; Widget node_prop_ucbh;
Widget node_prop_lcb; Widget node_prop_lcbh;
Widget node_prop_label;
Widget node_prop_id;
Widget node_prop_type;
Widget node_prop_name;
Widget node_prop_host;
Widget node_prop_sitter_port;
Widget node_prop_sitter_fd;
Widget node_prop_prog_name;
Widget node_prop_down_listen_port;
Widget node_prop_dlcid;
Widget node_prop_up_recv_port;
Widget node_prop_filtermode;
Widget node_prop_unid[MAX_LOWERNODE];
Widget node_prop_lnid[MAX_LOWERNODE];
Widget node_prop_ucid[MAX_LOWERNODE];
Widget node_prop_lcid[MAX_LOWERNODE];
Widget node_prop_ok;
Widget node_prop_cancel;

GC  pwin_gc;

static int show_help=0;

static int beep_mute=0;
static int ghost_mode=0;
static int stick_mode=1;





String topofield_trans="#override\n\
    Shift<Btn1Down>:    link-node-start()\n\
    Shift<Btn1Motion>:  link-node()\n\
    Shift<Btn1Up>:      link-node-end()\n\
    <Btn1Down>:         drag-node-start()\n\
    <Btn1Motion>:       drag-node()\n\
    <Btn1Up>:           drag-node-end()\n\
    Shift<Btn3Down>:    replace-node()\n\
    <Btn3Down>:         release-node()\n\
    <Btn4Down>:         pick-node()\n\
    Shift<Btn2Down>:    toggle-filtermode()\n\
    <Btn2Down>:         wake-nprop()\n\
    <Motion>:           report-place()\n\
    <Expose>:           topofield-redraw()\n\
";

#if 0
"\
    <Btn3Down>:         mk-newnode()\n\
"
#endif

/*
void Keyin(Widget w,XEvent *eve,String *parms,Cardinal *num_parm);
*/

#define sdist(x1,y1,x2,y2)  (((x2)-(x1))*((x2)-(x1))+((y2)-(y1))*((y2)-(y1)))

node_t*
near_node(int x, int y)
{
    node_t *nn;

    if(!nodelist) 
        return NULL;

    nn = nodelist;
    while(nn) {
        if(sdist(nn->x, nn->y, x, y)<=node_range_near)
            break;
        nn = nn->next;
    }
     
    return nn;
}

node_t*
middle_node(int x, int y)
{
    node_t *nn;

    if(!nodelist) 
        return nn;

    nn = nodelist;
    while(nn) {
        if(sdist(nn->x, nn->y, x, y)<=node_range_middle) 
            break;
        nn = nn->next;
    }
     
    return nn;
}


void
pick_node(Widget w,XEvent *eve,String *parms,Cardinal *num_parm)
{
    node_t *nn;
    GC tgc;
    XGCValues gc_v;

    printf("pick_node: x,y %d,%d\n", eve->xbutton.x, eve->xbutton.y);

    nn = near_node(eve->xbutton.x, eve->xbutton.y);
    if(!nn) {
        fprintf(stderr, "no found near node\n");
        return;
    }

    gc_v.function = GXcopy;
    tgc = XtGetGC(w, GCFunction, &gc_v);
    gc_v.foreground = highc;
    XChangeGC(XtDisplay(w), tgc, GCForeground, &gc_v);
    XDrawRectangle(XtDisplay(w), XtWindow(w), tgc,
        nn->x-node_wid/2, nn->y-node_ht/2, node_wid, node_ht);
}


#ifdef __STDC__
int print_msg(const char *fmt, ...)
#else
int print_msg(va_alist)
va_dcl
#endif
{
    va_list args;
#ifndef __STDC__
    char *fmt;
#endif
    char buffer[BUFSIZ];
    int nwrite;

#ifdef __STDC__
    va_start(args,fmt);
#else
    va_start(args);
    fmt = va_arg(args, char *);
#endif


#ifdef HAVE_VSNPRINTF
    vsnprintf(buffer, BUFSIZ, fmt, args);
#else
    vsprintf(buffer, fmt, args);
#endif
    if(XtIsRealized(smsg)) {
        XtVaSetValues(smsg, XtNlabel, buffer, NULL);
        XFlush(XtDisplay(smsg));
    }
    else {
        fprintf(stderr, "session msg [%s]\n", buffer);
        fflush(stderr);
    }

    va_end(args);

    return nwrite;
}


int
PopNporpMenu(Widget w, node_t *nn)
{

#define CpStr(x,s)\
    if(nn->##x) {\
        sprintf(tmp, "%s %s", s, nn->##x);\
        XtVaSetValues(node_prop_##x, XtNwidth, ww, XtNlabel, tmp, NULL);\
    } else {\
        sprintf(tmp, "%s -", s);\
        XtVaSetValues(node_prop_##x, XtNwidth, ww, XtNlabel, tmp, NULL);\
    }

#define CpNum(x,s)\
    {\
        sprintf(tmp, "%s %d", s, nn->##x);\
        XtVaSetValues(node_prop_##x, XtNwidth, ww, XtNlabel, tmp, NULL);\
    }

    char tmp[BUFSIZ];
    int k;
    char *np;
    int ww;
    Dimension wwid;

    if(nn==NULL)
        return;


    XtVaGetValues(node_prop_spanel, XtNwidth, &wwid, NULL);
    XtVaSetValues(node_prop_label, XtNwidth, wwid, NULL);

    XtVaGetValues(node_prop_panel, XtNwidth, &wwid, NULL);
    ww = MAX(160, (int)wwid);
    fprintf(stderr, "ww %d, wwid %d\n", ww, wwid);

    nprop_target_node = nn;


    CpNum(id,"node id# ");
    np = assoc_v(node_type_alist, nn->type);
    if(np)  { XtVaSetValues(node_prop_type,
                    XtNwidth, ww, XtNlabel, np, NULL);    }
    else    { XtVaSetValues(node_prop_type,
                    XtNwidth, ww, XtNlabel, "-", NULL);   }
    CpStr(name,"name:");
    CpStr(host,"host:");
    CpStr(prog_name, "program:");
    if(nn->type==NV_T_FILTER) {
        np = assoc_v(filter_name_al, nn->filtermode);
        sprintf(tmp, "filtermode: %s", np ? np : "-");
        XtVaSetValues(node_prop_filtermode,
            XtNsensitive, TRUE,
            XtNwidth, ww*2,
            XtNbackground, 0xffff00,
            XtNlabel, tmp, NULL);   
    }
    else {
        XtVaSetValues(node_prop_filtermode,
            XtNsensitive, FALSE,
            XtNlabel, "not filter", NULL);
    }
    CpNum(down_listen_port, "down listen port#");
    CpStr(dlcid, "down conn id#");
    CpNum(up_recv_port, "up receive port#");
    CpNum(sitter_port, "sitter port#");
    CpNum(sitter_fd, "sitter fd#");

    ww = 80;
    for(k=0;k<MAX_LOWERNODE;k++) {
        if(nn->unid[k]) {
            sprintf(tmp, "id# %d", nn->unid[k]->id);
            XtVaSetValues(node_prop_unid[k],    
                XtNwidth, ww, XtNlabel, tmp, NULL);
        }
        else {
            XtVaSetValues(node_prop_unid[k],
                XtNwidth, ww, XtNlabel, "-", NULL);
        }
        CpStr(ucid[k],"cid#");
        if(nn->lnid[k]) {
            sprintf(tmp, "id# %d", nn->lnid[k]->id);
            XtVaSetValues(node_prop_lnid[k],
                XtNwidth, ww, XtNlabel, tmp, NULL);
        }
        else {
            XtVaSetValues(node_prop_lnid[k],
                XtNwidth, ww, XtNlabel, "-", NULL);
        }
        CpStr(lcid[k],"cid");
    }

    XtPopup(w, XtGrabExclusive);
}

#undef CpStr

void
WakeNprop(Widget w,XEvent *eve,String *parms,Cardinal *num_parm)
{
    node_t *nn;
    GC tgc;
    XGCValues gc_v;
    int cx, cy;

    if(eve->type==ButtonPress||eve->type==ButtonRelease
        ||eve->type==MotionNotify) {
        cx = eve->xbutton.x_root;
        cy = eve->xbutton.y_root;
        XtMoveWidget(node_prop_shell, cx, cy);
    
        nn = near_node(eve->xbutton.x, eve->xbutton.y);

        PopNporpMenu(node_prop_shell, nn);
    }

    return;
}

void
toggle_filtermode(Widget w,XEvent *eve,String *parms,Cardinal *num_parm)
{
    node_t *nn;
    GC tgc;
    XGCValues gc_v;
    int cx, cy;


    printf("toggle_filtermode: x,y %d,%d\n",
        eve->xbutton.x, eve->xbutton.y);

    nn = near_node(eve->xbutton.x, eve->xbutton.y);
    if(!nn) {
        fprintf(stderr, "no found near node\n");
        return;
    }
    if(nn->type!=NV_T_FILTER && nn->type!=NV_T_SCANNER) {
        fprintf(stderr, "not filter\n");
        return;
    }

    rotate_filtermode(&nn->filtermode);
    jm_filter_change(nn);

    redraw_topofield(topofield, NULL, NULL, NULL);
}

void
release_node(Widget w,XEvent *eve,String *parms,Cardinal *num_parm)
{
    node_t *nn;
    GC tgc;
    XGCValues gc_v;

    printf("release_node: x,y %d,%d\n", eve->xbutton.x, eve->xbutton.y);

    nn = near_node(eve->xbutton.x, eve->xbutton.y);
    if(!nn) {
        fprintf(stderr, "no found near node\n");
        return;
    }
    fprintf(stderr, "node %p id# %d\n", nn, nn->id);
    if(nn->type==NV_T_VIEWER) {
    }   
    else {
        jm_release_node(nn);
    }

    redraw_topofield(topofield, NULL, NULL, NULL);
}

#if 0
void
replace_node(Widget w,XEvent *eve,String *parms,Cardinal *num_parm)
#endif
void
_replace_node(node_t *on)
{
#if 0
    node_t *on;
#endif
    node_t *nn;
    char tname[BUFSIZ];
    Dimension wwid, wht;
    int nx, ny;
    int i;

#if 0
    printf("replace_node: x,y %d,%d\n", eve->xbutton.x, eve->xbutton.y);
    on = near_node(eve->xbutton.x, eve->xbutton.y);
#endif
    fprintf(stderr, "_replace_node: on %p\n", on);

    if(!on) {
        fprintf(stderr, "no found near node\n");
        return;
    }
    fprintf(stderr, "old node %p id# %d\n", on, on->id);
    if(on->type==NV_T_VIEWER ||on->type==NV_T_CAMERA) {
        return;
    }   

    XtVaGetValues(topofield,
        XtNwidth, &wwid, XtNheight, &wht, NULL);

    nx = wwid/2;
    ny = wht/2;
    nn = near_node(nx, ny);
    if(!nn) {
        goto get_loc;
    }

    do {
        nx = rand()%wwid;
        ny = rand()%wht;
        nn = middle_node(nx, ny);
        fprintf(stderr, "nx %d, ny %d, nn %p\n", nx, ny, nn);
    } while(nn);

get_loc:
    fprintf(stderr, "make new node at %d,%d\n", nx, ny);

    nn = new_node();
    sprintf(tname, "node# %d", nn->id); 
    nn->type = NV_T_FILTER;
    nn->name = strdup(tname);
    nn->x = nx;
    nn->y = ny;
    strcpy(nn->prog_name, "ssf");
    nn->down_listen_port = 10000+nn->id;
    add_node(nn);

    redraw_topofield(topofield, NULL, NULL, NULL);

#if 1
    for(i=0;i<MAX_LOWERNODE;i++) {
        nn->unid[i] = on->unid[i];
        nn->lnid[i] = on->lnid[i];
    }
    redraw_topofield(topofield, NULL, NULL, NULL);
    XFlush(XtDisplay(topofield));
    sleep(1);
#endif

    for(i=0;i<MAX_LOWERNODE;i++) {
        nn->unid[i] = NULL;
        nn->lnid[i] = NULL;
    }

    jm_replace_node(on, nn);

    redraw_topofield(topofield, NULL, NULL, NULL);

}

void
replace_node(Widget w,XEvent *eve,String *parms,Cardinal *num_parm)
{
    node_t *on;

    on = near_node(eve->xbutton.x, eve->xbutton.y);
    _replace_node(on);
}


node_t *thenode=NULL;
int thenode_tic=0;

void
drag_node_start(Widget w,XEvent *eve,String *parms,Cardinal *num_parm)
{
    node_t *nn;
    GC tgc;
    XGCValues gc_v;

#if 0
    printf("drag_node_start: x,y %d,%d\n", eve->xbutton.x, eve->xbutton.y);
#endif

    nn = near_node(eve->xbutton.x, eve->xbutton.y);
    if(!nn) {
        fprintf(stderr, "no found near node\n");
        return;
    }

    thenode = nn;
    thenode_tic = 0;

    gc_v.function = GXcopy;
    tgc = XtGetGC(w, GCFunction, &gc_v);
    gc_v.foreground = highc;
    XChangeGC(XtDisplay(w), tgc, GCForeground, &gc_v);
/*
    XDrawRectangle(XtDisplay(w), XtWindow(w), tgc,
        thenode->x-node_wid/2, thenode->y-node_ht/2, node_wid, node_ht);
*/
    draw_node(XtDisplay(w), XtWindow(w), tgc, thenode, 1);
}

void
drag_node(Widget w,XEvent *eve,String *parms,Cardinal *num_parm)
{
    node_t *nn;
    GC tgc;
    XGCValues gc_v;

/*
    printf("drag_node: x,y %d,%d\n", eve->xbutton.x, eve->xbutton.y);
*/

    if(!thenode) {
#if 0
        fprintf(stderr, "no dragged node\n");
#endif
        return;
    }

#if 0
    printf("drag_node: tic %d, x,y %d,%d\n", 
        thenode_tic, eve->xbutton.x, eve->xbutton.y);
#endif

    if(thenode_tic==0) {
        gc_v.function = GXcopy;
        tgc = XtGetGC(w, GCFunction, &gc_v);
        gc_v.foreground = white;
        XChangeGC(XtDisplay(w), tgc, GCForeground, &gc_v);
/*
        XDrawRectangle(XtDisplay(w), XtWindow(w), tgc,
            thenode->x-node_wid/2, thenode->y-node_ht/2, node_wid, node_ht);
*/
    draw_node(XtDisplay(w), XtWindow(w), tgc, thenode, 1);

        gc_v.function = GXxor;
        XChangeGC(XtDisplay(w), tgc, GCFunction, &gc_v);
    }
    else {
        gc_v.function = GXxor;
        tgc = XtGetGC(w, GCFunction, &gc_v);
        gc_v.foreground = highc;
        XChangeGC(XtDisplay(w), tgc, GCForeground, &gc_v);
/*
        XDrawRectangle(XtDisplay(w), XtWindow(w), tgc,
            thenode->x-node_wid/2, thenode->y-node_ht/2, node_wid, node_ht);
*/
    draw_node(XtDisplay(w), XtWindow(w), tgc, thenode, 1);
    }

    thenode->x = eve->xbutton.x;
    thenode->y = eve->xbutton.y;
    thenode_tic++;

#if 0
    gc_v.function = GXxor;
    tgc = XtGetGC(w, GCFunction, &gc_v);
#endif
    gc_v.foreground = highc;
    XChangeGC(XtDisplay(w), tgc, GCForeground, &gc_v);
/*
    XDrawRectangle(XtDisplay(w), XtWindow(w), tgc,
        thenode->x-node_wid/2, thenode->y-node_ht/2, node_wid, node_ht);
*/
    draw_node(XtDisplay(w), XtWindow(w), tgc, thenode, 1);

}

void
drag_node_end(Widget w,XEvent *eve,String *parms,Cardinal *num_parm)
{
    node_t *nn;
    GC tgc;
    XGCValues gc_v;

#if 0
    printf("drag_node_end: x,y %d,%d\n", eve->xbutton.x, eve->xbutton.y);
#endif

    if(!thenode) {
        fprintf(stderr, "no dragged node\n");
        return;
    }

    gc_v.function = GXcopy;
    tgc = XtGetGC(w, GCFunction, &gc_v);
    gc_v.foreground = normc;
    XChangeGC(XtDisplay(w), tgc, GCForeground, &gc_v);
/*
    XDrawRectangle(XtDisplay(w), XtWindow(w), tgc,
        thenode->x-node_wid/2, thenode->y-node_ht/2, node_wid, node_ht);
*/
    draw_node(XtDisplay(w), XtWindow(w), tgc, thenode, 1);

    thenode = NULL;
    thenode_tic = 0;
}


void
link_node_start(Widget w,XEvent *eve,String *parms,Cardinal *num_parm)
{
    node_t *nn;
    GC tgc;
    XGCValues gc_v;

#if 0
    printf("link_node_start: x,y %d,%d\n", eve->xbutton.x, eve->xbutton.y);
#endif

    nn = near_node(eve->xbutton.x, eve->xbutton.y);

    if(!nn) {
        fprintf(stderr, "no found near node\n");
        return;
    }

    thenode = nn;
    thenode_tic = 0;

    gc_v.function = GXcopy;
    tgc = XtGetGC(w, GCFunction, &gc_v);
    gc_v.foreground = highc;
    XChangeGC(XtDisplay(w), tgc, GCForeground, &gc_v);
/*
    XDrawRectangle(XtDisplay(w), XtWindow(w), tgc,
        thenode->x-node_wid/2, thenode->y-node_ht/2, node_wid, node_ht);
*/
    draw_node(XtDisplay(w), XtWindow(w), tgc, thenode, 1);
}

void
link_node(Widget w,XEvent *eve,String *parms,Cardinal *num_parm)
{
    node_t *nn;
    GC tgc;
    XGCValues gc_v;

/*
    printf("link_node: x,y %d,%d\n", eve->xbutton.x, eve->xbutton.y);
*/

    if(!thenode) {
        fprintf(stderr, "no linkged node\n");
        return;
    }

#if 0
    printf("link_node: tic %d, x,y %d,%d\n", 
        thenode_tic, eve->xbutton.x, eve->xbutton.y);
#endif

#if 0 
        gc_v.function = GXxor;
        tgc = XtGetGC(w, GCFunction, &gc_v);
        gc_v.foreground = highc;
        XChangeGC(XtDisplay(w), tgc, GCForeground, &gc_v);
/*
        XDrawRectangle(XtDisplay(w), XtWindow(w), tgc,
            thenode->x-node_wid/2, thenode->y-node_ht/2, node_wid, node_ht);
*/
    draw_node(XtDisplay(w), XtWindow(w), tgc, thenode, 1);
#endif

    nn = near_node(eve->xbutton.x, eve->xbutton.y);

    if(nn) {
        gc_v.function = GXxor;
        tgc = XtGetGC(w, GCFunction, &gc_v);
        gc_v.foreground = highc;
        XChangeGC(XtDisplay(w), tgc, GCForeground, &gc_v);
    /*
        XDrawRectangle(XtDisplay(w), XtWindow(w), tgc,
            thenode->x-node_wid/2, thenode->y-node_ht/2, node_wid, node_ht);
    */
        draw_node(XtDisplay(w), XtWindow(w), tgc, nn, 1);

    }
}

void
link_node_end(Widget w,XEvent *eve,String *parms,Cardinal *num_parm)
{
    node_t *nn;
    GC tgc;
    XGCValues gc_v;

#if 0
    printf("link_node_end: x,y %d,%d\n", eve->xbutton.x, eve->xbutton.y);
#endif

    if(!thenode) {
        fprintf(stderr, "no linkged node\n");
        return;
    }

    nn = near_node(eve->xbutton.x, eve->xbutton.y);
    if(!nn) {
        fprintf(stderr, "no linkged node\n");
        return;
    }

    if(nn==thenode) {
        fprintf(stderr, "loopback is not allowed\n");
        return;
    }

#if 0
    gc_v.function = GXcopy;
    tgc = XtGetGC(w, GCFunction, &gc_v);
    gc_v.foreground = normc;
    XChangeGC(XtDisplay(w), tgc, GCForeground, &gc_v);
    draw_node(XtDisplay(w), XtWindow(w), tgc, nn, 1);
#endif

    gc_v.function = GXcopy;
    tgc = XtGetGC(w, GCFunction, &gc_v);
    gc_v.foreground = normc;
    XChangeGC(XtDisplay(w), tgc, GCForeground, &gc_v);

    entry_upper(nn, thenode);
#if 0
    entry_upper(thenode, nn);
    entry_lower(nn, thenode);
#endif

    draw_node(XtDisplay(w), XtWindow(w), tgc, nn, 1);
    draw_node(XtDisplay(w), XtWindow(w), tgc, thenode, 1);

    thenode = NULL;
    thenode_tic = 0;
}

void
report_place(Widget w,XEvent *eve,String *parms,Cardinal *num_parm)
{
    node_t *nn;
    char tmp[BUFSIZ];

#if 0
    fprintf(stderr, "report_trace: type %d\n", eve->type);
#endif

    nn = near_node(eve->xmotion.x, eve->xmotion.y);
    if(nn) {
        sprintf(tmp, "%s node, id# %d",
            assoc_v(node_type_alist, nn->type), nn->id);
        XtVaSetValues(tmsg, XtNlabel, tmp, NULL);
    }
    else {
        XtVaSetValues(tmsg, XtNlabel, "", NULL);
    }
}

void
Nnode(Widget w, XtPointer client_data, XtPointer call_data)
{
    node_t *nn;
    GC tgc;
    XGCValues gc_v;
    char tname[BUFSIZ];
    Dimension wwid, wht;
    int nx, ny;


    XtVaGetValues(topofield,
        XtNwidth, &wwid, XtNheight, &wht, NULL);

    nx = wwid/2;
    ny = wht/2;
    nn = near_node(nx, ny);
    if(!nn) {
        goto get_loc;
    }

    do {
        nx = rand()%wwid;
        ny = rand()%wht;
        nn = middle_node(nx, ny);
        fprintf(stderr, "nx %d, ny %d, nn %p\n", nx, ny, nn);
    } while(nn);

get_loc:
    fprintf(stderr, "make new node at %d,%d\n", nx, ny);

    nn = new_node();
    sprintf(tname, "node# %d", nn->id); 
    nn->type = NV_T_FILTER;
    nn->name = strdup(tname);
    nn->x = nx;
    nn->y = ny;
    strcpy(nn->prog_name, "ssf");
    nn->down_listen_port = 10000+nn->id;
    add_node(nn);

#if 0
    gc_v.function = GXcopy;
    tgc = XtGetGC(w, GCFunction, &gc_v);
    gc_v.foreground = normc;
    XChangeGC(XtDisplay(w), tgc, GCForeground, &gc_v);

    draw_node(XtDisplay(w), XtWindow(w), tgc, nn, 1);
#endif

    redraw_topofield(topofield, NULL, NULL, NULL);
}


void
mk_newnode(Widget w,XEvent *eve,String *parms,Cardinal *num_parm)
{
    node_t *nn;
    GC tgc;
    XGCValues gc_v;
    char tname[BUFSIZ];

#if 0
    printf("mk_newnode: x,y %d,%d\n", eve->xbutton.x, eve->xbutton.y);
#endif
#if 0
    print_nodelist(stderr);
#endif

    nn = new_node();
    sprintf(tname, "node# %d", nn->id); 
    nn->type = NV_T_FILTER;
    nn->name = strdup(tname);
    nn->x = eve->xbutton.x;
    nn->y = eve->xbutton.y;
#if 1
    strcpy(nn->prog_name, "ssf");
    nn->down_listen_port = 10000+nn->id;
#endif
    add_node(nn);

#if 0
    nn = new_node();
    sprintf(tname, "node# %d", nn->id); 
    nn->type = NV_T_SCANNER;
    nn->name = strdup(tname);
    nn->x = eve->xbutton.x;
    nn->y = eve->xbutton.y;
#if 1
    strcpy(nn->prog_name, "ssf");
    nn->down_listen_port = 10000+nn->id;
#endif
    add_node(nn);
#endif

#if 0
    print_nodelist(stderr);
#endif

    gc_v.function = GXcopy;
    tgc = XtGetGC(w, GCFunction, &gc_v);
    gc_v.foreground = normc;
    XChangeGC(XtDisplay(w), tgc, GCForeground, &gc_v);

    draw_node(XtDisplay(w), XtWindow(w), tgc, nn, 1);
}

/*
void
*/
int
redraw_topofield(Widget w,XEvent *eve,String *parms,Cardinal *num_parm)
{
    node_t *nn;
    GC tgc;
    XGCValues gc_v;
    Dimension wwid, wht;

#if 0
    fprintf(stderr, "redraw_topofield:\n");
#endif

    if(!nodelist) {
        fprintf(stderr, "no node\n");
        return;
    }

    gc_v.function = GXcopy;
    tgc = XtGetGC(w, GCFunction, &gc_v);

    XtVaGetValues(w, XtNwidth, &wwid, XtNheight, &wht, NULL);
#if 0
    fprintf(stderr, "widget wid %d, ht %d\n", wwid, wht);
#endif

    gc_v.foreground = white;
    XChangeGC(XtDisplay(w), tgc, GCForeground, &gc_v);
    XFillRectangle(XtDisplay(w), XtWindow(w), tgc, 0, 0, wwid, wht);

#if 0
    print_nodelist(stderr);
#endif

    gc_v.foreground = normc;
    XChangeGC(XtDisplay(w), tgc, GCForeground, &gc_v);

    nn = nodelist;
    while(nn) {
        draw_node(XtDisplay(w), XtWindow(w), tgc, nn, 1);
        nn = nn->next;
    }
}


void
Quit(Widget w, XtPointer client_data, XtPointer call_data)
{
    exit(0);
}

void
HCheck(Widget w, XtPointer client_data, XtPointer call_data)
{
    print_nodelist(stderr);
    print_nodelistH(stderr);
}

void
NpropTfilter(Widget w, XtPointer client_data, XtPointer call_data)
{
    char *np;
    char tmp[BUFSIZ];

    rotate_filtermode(&node_prop_tryfiltermode);
    
#if 0
    np = assoc_v(filter_name_al, node_prop_tryfiltermode);
    if(np)  { XtVaSetValues(node_prop_filtermode, XtNlabel, np, NULL);    }
    else    { XtVaSetValues(node_prop_filtermode, XtNlabel, "-", NULL);   }
#endif

    np = assoc_v(filter_name_al, node_prop_tryfiltermode);
    sprintf(tmp, "filtermode: %s", np ? np : "-");
    XtVaSetValues(node_prop_filtermode, XtNlabel, tmp, NULL);

}


int
NodeSwapper()
{
    node_t *cur;
    int isfree_node(node_t*);
    int chk;

    fprintf(stderr, "NodeSwapper: employ_limit %d\n", employ_limit);

    cur = nodelist;
    while(cur) {
#if 0
        fprintf(stderr, "node id# %d\n", cur->id);
#endif
        chk = isfree_node(cur);
        if(chk||
            cur->type==NV_T_VIEWER||cur->type==NV_T_CAMERA) {
#if 0
                fprintf(stderr, "  It is swappable. sorry <%d>\n", chk);
#endif
        }
        else {
            if(cur->hidle < employ_limit) {
                fprintf(stderr, "  too busy id# %d. swap !\n", cur->id);
                print_msg("swap node #%d to new one", cur->id);
                _replace_node(cur);
            }
        }

        cur = cur->next;
    }
}


XtIntervalId probe_tid;
int probe_interval = 30*1000;

void  
ProbeNodeState(XtPointer client_data, XtIntervalId *iid)
{
    node_t *cur;   
    XtIntervalId tid;
    char tmp[BUFSIZ];
    time_t now;
    struct tm *tmbuf;
    char *tbuf;

    now = time(NULL);
    tmbuf = localtime(&now);
    tbuf = asctime(tmbuf);
    strcpy(tmp, tbuf);
    chomp(tmp);

    fprintf(stderr, "ProbeNodeState: |%s|\n", tmp);
    fflush(stderr);

    jm_probe_state();  

    redraw_topofield(topofield, NULL, NULL, NULL);

    if(employ_limit>0) {
        NodeSwapper();
    }

    v_print_slots();

    probe_tid = XtAppAddTimeOut(app_context, probe_interval,
            ProbeNodeState, NULL);

}

void
Redraw(Widget w, XtPointer client_data, XtPointer call_data)
{
    redraw_topofield(topofield, NULL, NULL, NULL);
}

void
Start(Widget w, XtPointer client_data, XtPointer call_data)
{
    if(run_mode) {
        fprintf(stderr, "already running...\n");
        return;
    }
    run_mode = 1;

    fprintf(stderr, "C Start begin\n");
    
    start_allnodes(nodelist);

    if(!jm_errno) {
        print_msg("no error");
    }
    else {
        print_msg("error code = %d", jm_errno);
    }

#if 1
    redraw_topofield(topofield, NULL, NULL, NULL);
#endif
    fprintf(stderr, "C Start end\n");
}

void
Stop(Widget w, XtPointer client_data, XtPointer call_data)
{
    if(!run_mode) {
        fprintf(stderr, "Why do you think to stop ?\n");
        return;
    }

    run_mode = 0;

    fprintf(stderr, "C Stop begin\n");

    stop_allnodes();

#if 1
    redraw_topofield(topofield, NULL, NULL, NULL);
#endif
    fprintf(stderr, "C Stop end\n");
}

void
ToggleSwapper(Widget w, XtPointer client_data, XtPointer call_data)
{
    static int k=0;

#if 0
    if(employ_limit<=0) {
        employ_limit = 3000;
    }
    else {
        employ_limit = 0;
    }
    fprintf(stderr, "employ_limit %d\n", employ_limit);
#endif

#if 1
    k++;
    employ_limit = (k%6)*1000;
    fprintf(stderr, "employ_limit %d <%d>\n", employ_limit, k);
#endif
}

void
NpropOk(Widget w, XtPointer client_data, XtPointer call_data)
{
    XtPopdown(node_prop_shell);

    if(nprop_target_node) {
        nprop_target_node->filtermode = node_prop_tryfiltermode;
#if 0
        rotate_filtermode(&nprop_target_node->filtermode);
#endif
        jm_filter_change(nprop_target_node);
    }

    nprop_target_node = NULL;
}

void
NpropCancel(Widget w, XtPointer client_data, XtPointer call_data)
{
    XtPopdown(node_prop_shell);

    nprop_target_node = NULL;
}
    
    

#ifndef RETSIGTYPE
#define RETSIGTYPE  void
#endif

/*
 * TrapChildDie - Trap child's die. Check child is what.
 *
 */
RETSIGTYPE TrapChildDie()
{
    pid_t pid;
    int status;
    volatile int eno;

    do {
try_waitpid:
        eno = 0;
        pid = waitpid((pid_t)-1, &status, WNOHANG | WUNTRACED);
        eno = errno;
        if(pid<0) {
            if(eno==EINTR) {
                goto try_waitpid;
            }
        }
        if(pid>0) {
            fprintf(stderr, "die pid %d\n", pid);
        }
    } while(pid>0);
}


static sentry_shmrec bk_info;
static sentry_imgrec bk_inframe;
static unsigned char *inimgbuf=NULL;
static sentry_imgrec bk_drframe[2];
static unsigned char *drimgbuf0=NULL;
static unsigned char *drimgbuf1=NULL;

extern apair_t buflen_xrgb_alist;
extern void request_call();



typedef struct {
    int x;
    int y;
} pos_xy_t;

void
put_mark(Widget w, int n, pos_xy_t *ge, Pixel col)
{
    int i;

    XSetForeground(XtDisplay(w), pwin_gc, col);
    for(i=0;i<n;i++) {
        XDrawLine(XtDisplay(w), XtWindow(w),
            pwin_gc, ge[i].x-2, ge[i].y, ge[i].x+2, ge[i].y);
        XDrawLine(XtDisplay(w), XtWindow(w),
            pwin_gc, ge[i].x, ge[i].y-2, ge[i].x, ge[i].y+2);
    }
}

int diff_capacity = 0x30;
int cmp_type = SEN_M_NONE;


extern int
    img_diff_npos(sentry_imgrec*, sentry_imgrec*, int, pos_xy_t*, int);



int sanpos_npos=21;
pos_xy_t sanpos[100] = {
    {  20,  40},
    {  40,  40},
    {  60,  40},
    {  80,  40},
    { 100,  40},
    { 120,  40},
    { 140,  40},
    {  20,  80},
    {  40,  80},
    {  60,  80},
    {  80,  80},
    { 100,  80},
    { 120,  80},
    { 140,  80},
    {  20,  60},
    {  40,  60},
    {  60,  60},
    {  80,  60},
    { 100,  60},
    { 120,  60},
    { 140,  60},
    {   0,   0},
};


int
find_pos(time_t reftime)
{
    int i;
    int pos;
    int ret;

    ret = -1;

    for(i=0;i<sentry_shm_chunk_radix;i++) {
        pos = i+1;
#if 0
        printf(" %3d: %3d %12d %4d\n", i, pos,
            sentry_shm_chunk[pos].entry.header.recv_sec,
            sentry_shm_chunk[pos].entry.header.length);
#endif
        if(sentry_shm_chunk[pos].entry.header.recv_sec==reftime) {
            if(debug>=10) {
                printf(
" %3d: %3d time issue %d.%06d recv %d.%06d len %4d (%dx%d) pos %p [%d]\n",
                    i, pos,
                    sentry_shm_chunk[pos].entry.header.send_sec,
                    sentry_shm_chunk[pos].entry.header.send_usec,
                    sentry_shm_chunk[pos].entry.header.recv_sec,
                    sentry_shm_chunk[pos].entry.header.recv_usec,
                    sentry_shm_chunk[pos].entry.header.length,
                    sentry_shm_chunk[pos].entry.header.width,
                    sentry_shm_chunk[pos].entry.header.height,
                    sentry_shm_chunk[pos].entry.header.body,
                    sentry_shm_chunk[pos].entry.header.body_len
                    );
            }
            ret = pos;
        }
    }

    return ret;
}


static volatile int djbusy=0;

SenDrawImage2(Widget w, sentry_imgrec *inimg, sentry_imgrec *outimg)
{
    int wid, ht;
    int xwid, xht;
    unsigned char *raw_data;
#if 0
    unsigned char *data_rgb=NULL, *data_bgrx=NULL;
    int buflen=0;
#endif
#if 1
    static unsigned char *data_rgb=NULL, *data_bgrx=NULL;
    static int buflen=0;
#endif
    int raw_len;
    int raw_format, raw_color;
    int chk;
    int cur_format=SEN_F_NONE;
    int cur_corder=SEN_C_NONE;
    int success=0;
    int outlen=-1;


    PASS;

    if(inimg==NULL)
        return;

#if 0
    if(debug>=10) {
        printf("SenDrawImage: output format '%s', color order '%s'\n",
            assoc_v(format_alist, output_format),
            assoc_v(color_alist, output_corder));
    }
#endif

 {
 int i;
    while(djbusy) {
        fprintf(stderr, "$"); fflush(stderr);
        for(i=0;i<(1<<31);i++) {
            chk = (int)(sin(i)*10);
        }
    }
    djbusy++;
 }


    raw_len     = inimg->length;
    raw_data    = (unsigned char*)inimg->body;
    raw_format  = inimg->format;
    raw_color   = inimg->corder;
    wid = inimg->width;
    ht  = inimg->height;

    if(!data_rgb || !data_bgrx) {
        buflen = (int)assoc_p(&buflen_xrgb_alist, MAX_DISPLAY_SIZE);
        if(buflen<=0) {
            buflen = TMP_IMGBUF_LEN;
        }
fprintf(stderr, "#buflen %d\n", buflen);

        data_bgrx = (unsigned char*)malloc(buflen);
        memset(data_bgrx, 0xff, buflen);

        data_rgb = (unsigned char*)malloc(buflen);
        memset(data_rgb, 0xff, buflen);
    }

#if 0
    if(debug>=10) {
        printf("SenDrawImage: %p %d '%s''%s' %dx%d, 3B/p=%d 4B/p=%d\n",
            raw_data, raw_len, assoc_v(format_alist, raw_format),
            assoc_v(color_alist, raw_color), wid, ht, wid*ht*3, wid*ht*4);
    }
#endif
    if(debug>=10) {
        printf(
"draw: got format %s, color %s, ptr %p, size %dx%d, len %d\n\
      out format %s, color %s\n",
        assoc_v(format_alist, raw_format), assoc_v(color_alist, raw_color),
        raw_data, wid, ht, raw_len,
        assoc_v(format_alist, output_format),
        assoc_v(color_alist, output_corder));
    }

    switch(raw_format) {
    case SEN_F_JPEG:

#if 1
{
    static int ct=0;
    static char *fname[] = {"/tmp/nv0.jpg","/tmp/nv1.jpg"};

    dump_to_file(fname[ct%2], raw_data, raw_len);
    ct++;
}
#endif

        chk = decode_jpeg(raw_data, raw_len, data_rgb, buflen, &xwid, &xht);
        if(chk) {
            fprintf(stderr, "ERROR: JPEG decoding\n");
            fflush(stderr);
            goto out;
        }
        if(xwid!=wid && xht!=ht) {
            fprintf(stderr,
                "SenDrawImage: wid %d != xwid %d, ht %d != xht %d\n",
                wid, xwid, ht, xht);
        }
        cur_format = SEN_F_RGB;
        cur_corder = SEN_C_RGB;
        if(debug>=20){
            printf("  convert to '%s' (wid %d, ht %d)\n",
                assoc_v(color_alist, cur_corder), xwid, xht);
        }
        break;

    case SEN_F_NONE: /* XXX */
        memcpy(data_rgb, raw_data, raw_len);
        break;

    case SEN_F_RGB: /* XXX */
        memcpy(data_rgb, raw_data, raw_len);
        cur_format = raw_format; /* SEN_F_RGB */
        cur_corder = raw_color;
        break;
    }


#if 1
    {
    char tmpstr[BUFSIZ];
    struct tm htm;

    localtime_r(&inimg->recv_sec, &htm);
    sprintf(tmpstr, "%02d-%02d %02d:%02d:%02d.%02d",
        htm.tm_mon+1, htm.tm_mday,
        htm.tm_hour, htm.tm_min, htm.tm_sec, inimg->recv_usec/10000);

    put_letter_rgb(data_rgb, wid, ht, 3, 0, ht-8, tmpstr);
    }
#endif
#if 1
    {
    char tmpstr[BUFSIZ];
    struct tm htm;

    localtime_r(&inimg->send_sec, &htm);
    sprintf(tmpstr, "%02d-%02d %02d:%02d:%02d.%02d",
        htm.tm_mon+1, htm.tm_mday,
        htm.tm_hour, htm.tm_min, htm.tm_sec, inimg->send_usec/10000);

    put_letter_rgb(data_rgb, wid, ht, 3, 0, ht-16, tmpstr);
    }
#endif


    if(debug>=10) {
        printf(
"draw: cur format %s, color %s\n\
      out format %s, color %s\n",
        assoc_v(format_alist, cur_format), assoc_v(color_alist, cur_corder),
        assoc_v(format_alist, output_format),
       assoc_v(color_alist, output_corder)
        );
    }
    if(cur_format != output_format) {
        if(debug>=1) {
            fprintf(stderr, "image format miss-match cur=%s out=%s\n",
                assoc_v(format_alist, cur_format),
                assoc_v(format_alist, output_format));
        }
        goto put_fail;
    }

    if(output_format==SEN_F_RGB) {
        if(output_corder==SEN_C_XBGR) {
            convert_RGB2XBGR(data_bgrx, data_rgb, wid * ht);

            PASS;
            outlen = wid*ht*4;
            put_image(XtDisplay(w), xp_foil_visual,
                xp_foil_depth, pwin_gc, XtWindow(w), wid, ht, data_bgrx);
            success = 1;
            PASS;

            copyattr_img(outimg, inimg);
            outimg->format = output_format;
            outimg->corder = output_corder;
            outimg->length = outlen;
            memcpy(outimg->body, data_bgrx, outlen);
        }
        else
        if(output_corder==SEN_C_XRGB) {
/*
            convert_RGB2XRGB(data_bgrx, data_rgb, wid * ht);
*/
            convert_RGB2BGRX(data_bgrx, data_rgb, wid * ht);

            outlen = wid*ht*4;

            PASS;
            put_image(XtDisplay(w), xp_foil_visual,
                xp_foil_depth, pwin_gc, XtWindow(w), wid, ht, data_bgrx);
            success = 1;
            PASS;

            copyattr_img(outimg, inimg);
            outimg->format = output_format;
            outimg->corder = output_corder;
            outimg->length = outlen;
            memcpy(outimg->body, data_bgrx, outlen);
        }
        else
        if(output_corder==SEN_C_RGB) {
            outlen = wid*ht*3;

            PASS;
            put_image(XtDisplay(w), xp_foil_visual,
                xp_foil_depth, pwin_gc, XtWindow(w), wid, ht, data_rgb);
            success = 1;
            PASS;

            copyattr_img(outimg, inimg);
            outimg->format = output_format;
            outimg->corder = output_corder;
            outimg->length = outlen;
            memcpy(outimg->body, data_rgb, outlen);
        }
    }

put_fail:
    if(!success) {
        if(debug>=1) {
            fprintf(stderr, "fail some reason, format cur=%s out=%s\n",
                assoc_v(format_alist, cur_format),
                assoc_v(format_alist, output_format));
        }
    }
    else {
#if 0
        copyattr_img(outimg, inimg);
        memcpy(outimg->body, inimg->body, outlen);
#endif
    }

out:
#if 0
    if(data_rgb!=NULL)
        free(data_rgb);
    if(data_bgrx!=NULL)
        free(data_bgrx);
#endif

    PASS;

    djbusy--;

    return 0;
}


RefreshWins(sentry_imgrec *bank)
{
}


alloc_inimgbufs()
{
    int buflen;

    buflen = (int)assoc_p(&buflen_xrgb_alist, MAX_DISPLAY_SIZE);
 if(debug) {
    printf("buflen %d\n", buflen);
 }
    if(buflen==0) {
        buflen = TMP_IMGBUF_LEN;
    }

    inimgbuf = (unsigned char*)malloc(buflen);
    bk_inframe.body = inimgbuf;
    bk_inframe.body_len = buflen;
}

alloc_drimgbufs()
{
    int buflen;

    if(output_format!=SEN_F_RGB) {
        fprintf(stderr, "no idea for no RGB device\n");
        exit(1);
    }

    buflen = (int)assoc_p(&buflen_xrgb_alist, MAX_DISPLAY_SIZE);
 if(debug) {
    printf("buflen %d\n", buflen);
 }
    if(buflen==0) {
        buflen = TMP_IMGBUF_LEN;
    }

    drimgbuf0 = (unsigned char*)malloc(buflen);
    bk_drframe[0].body = drimgbuf0;
    bk_drframe[0].body_len = buflen;

    drimgbuf1 = (unsigned char*)malloc(buflen);
    bk_drframe[1].body = drimgbuf1;
    bk_drframe[1].body_len = buflen;
}



int
RefreshHist(int n)
{
}

void
show_image_history(FILE *fp)
{
    int i, u;

    for(i=0;i<img_his_radix;i++) {
        fprintf(fp, "%2d: %p %6d/%6d %3dx%3d\n",
            i,
            &img_his[i], img_his[i].length, img_his[i].body_len,
            img_his[i].width, img_his[i].height);
    }
#if 0
    for(u=img_his_pos;u<img_his_pos+img_his_radix;u++) {
        i = u%img_his_radix;
        fprintf(fp, "%2d: %c %2d : %p %6d/%6d %3dx%3d\n",
            u, (i==img_his_pos ? '*': '-'),  i,
            &img_his[i], img_his[i].length, img_his[i].body_len,
            img_his[i].width, img_his[i].height);
    }
#endif
}

init_image_history(int n)
{
    int i;
    int w;

    img_his_radix = n;
    img_his = (sentry_imgrec*) malloc(sizeof(sentry_imgrec)*img_his_radix);

    for(i=0;i<img_his_radix;i++) {
        w = (int)assoc_p(&buflen_xrgb_alist, MAX_DISPLAY_SIZE);
        img_his[i].body     = (uint8_t*)malloc(w);
        img_his[i].body_len = w;
        img_his[i].length   = -1;
        img_his[i].width    = 0;
        img_his[i].height   = 0;
    }

    if(debug) {
        fprintf(stderr, "img_his_radix %d\n", img_his_radix);
    }
}



int
store_image_history(sentry_imgrec *cur, sentry_imgrec *last)
{
    char data_bgrx[MAX_IMG_LEN];
    char data_rgb[MAX_IMG_LEN];
    int wid, ht;
    int cnum;
    void DoDiffExpose(Widget w, XEvent *ev, String *s, Cardinal *c);

    wid = cur->width;
    ht  = cur->height;

    if(debug)
        show_image_history(stderr);

    cnum = img_his_pos;

    if(debug)
        fprintf(stderr, "#store to %d\n", cnum);

    copyattr_img(&img_his[img_his_pos], cur);
    memcpy(img_his[img_his_pos].body, cur->body, cur->length);
    img_his[img_his_pos].length = cur->length;

    img_his_pos = (img_his_pos+1) % img_his_radix;

    if(debug) {
        show_image_history(stderr);
        fprintf(stderr, "# %d %d\n", cnum, img_his_pos);
    }

    RefreshHist(cnum);
    if(stick_mode) {
        img_his_view = cnum;
    }
}

/*
 * as viewer
 */


int
v_print_slots()
{
    int i;
    time_t now;

    now = time(NULL);
    for(i=0;i<max_viewer;i++) {
        fprintf(stderr, "%2d: %c %4d %6d node id# %d\n",
            i, vslot[i].running ? 'R' : 'S',
            vslot[i].got_count, now - vslot[i].lasttime.tv_sec,
            vslot[i].node->id);
    }
}

viewer_rec*
v_find_slot(node_t *this)
{
    int k;
    viewer_rec *rc;
    
    rc = NULL;
    for(k=0;k<max_viewer;k++) {
        if(vslot[k].node==this) {
            rc = &vslot[k];
            break;
        }
    }
    return rc;
}

int
v_stop_request(node_t *this)
{
    viewer_rec *rc;

    fprintf(stderr, "v_stop_request: this %p, id# %d\n", this, this->id);
    
    rc = v_find_slot(this);
    if(!rc) {
        return;
    }
    if(!rc->running) {
        fprintf(stderr, "  skip. because it is never running\n");
        return;
    }


    fprintf(stderr, "v_stop_request: slot %p, id# %d; tid %d\n",
        rc, rc->node->id, rc->tid);
    
    XtRemoveTimeOut(rc->tid);
    rc->running = 0;
    
}

int
v_start_request(node_t *this, int rfd)
{
    viewer_rec *rc;
    XtIntervalId req_tid;

    extern XtAppContext app_context;
    void _request_call(int, int);
    void RequestRepeat(XtPointer client_data, XtIntervalId *iid);
    
    rc = v_find_slot(this);
    if(!rc) {
        return;
    }

    fprintf(stderr,"v_start_request: id# %d, send fd# %d; recv port# %d\n",
        this->id, rfd, this->up_recv_port);
    fprintf(stderr,"  req_sendfd %d\n", rc->req_sendfd);
    if(rc->req_sendfd>=0) {
        rfd = rc->req_sendfd;
    }

    _request_call(rfd, this->up_recv_port);

    req_tid = XtAppAddTimeOut(app_context,
                request_interval*1000,
                RequestRepeat, rc);

    rc->req_sendfd  = rfd;
    rc->running     = 1;
    rc->tid         = req_tid;
    
    fprintf(stderr, "v_start_request: slot %p, id# %d; tid %d\n",
        rc, rc->node->id, rc->tid);
}





void
eventical_update(caddr_t client_data, int *fd, XtInputId *id)
{
#if 0
    static struct timeval lasttime={-1,0};
#endif
    struct timeval curtime;
    time_t now;
    int chk;
    sentry_shmrec *info;
    sentry_imgrec *inframe;
    static sentry_imgrec *cur=NULL, *last=NULL;
    char *tmp_buff;
    int pos;
    int diff;
    double rate;
    int  drfno;
    char *drframe;
    int recv_image(int datafd, char **buff,
        sentry_shmrec *info, sentry_imgrec *dest);
    char tmpstr[BUFSIZ];
    viewer_rec *vrec;
    struct tm htm;


    vrec = (viewer_rec*)client_data;
    if(vrec==NULL) {
        return;
    }
#if 0
    fprintf(stderr,
        "eventical_update: vrec %p; Vw %d, Rw %d, node %p, fd %d\n",
        vrec, vrec->Vw, vrec->Rw, vrec->node, vrec->res_recvfd);
#endif

    PASS;

    gettimeofday(&curtime, NULL);
    now = curtime.tv_sec;
    if(use_shm_flag) {
        info = &sentry_shm_chunk[0];
        pos  = 1+now%sentry_shm_chunk_radix;
        inframe = &(sentry_shm_chunk[pos].entry.header);
    }
    else {
        info = &bk_info;

        if(c_got_count==0) {
            alloc_inimgbufs();
            alloc_drimgbufs();
        }

        inframe = &bk_inframe;
    }

 if(debug>=100) {
   fprintf(stderr, "inframe %p\n", inframe);
 }

    PASS;

#if 0
    chk = recv_image(*fd, &tmp_buff, info, inframe);
#endif
    chk = recv_image_multi(*fd, &vrec->pque, &tmp_buff, info, inframe);
    if(chk) {
#if 0
        fprintf(stderr, "WARN: recv_image id# %d, fd# %d, <%d>\n",
            vrec->node ? vrec->node->id : -1, *fd, chk);
#endif

        if(chk==RTR_E_EOF) {
            fprintf(stderr, "ERROR: recv_image id# %d, fd# %d, <%d>\n",
                vrec->node ? vrec->node->id : -1, *fd, chk);
            v_stop_request(vrec->node);
        }
        return;
    }


    if(refresh_flag){
        RefreshWins(inframe);
    }

    drfno = c_got_count%2;
    cur = &bk_drframe[drfno];

 if(debug>=100) {
    fprintf(stderr, "drfno %d, bk_drframe[%d] %p, cur %p\n",
        drfno, drfno, &bk_drframe[drfno], cur);
}

    PASS;

    SenDrawImage2(vrec->Vw, inframe, cur);
    
    PASS;

#if 0
    put_mark(vrec->Rw, sanpos_npos, sanpos, 0xff0000);
#endif

    PASS;

#if 0
    if(last) {
        if(debug) {
            show_imgattr(stderr, cur);
            show_imgattr(stderr, last);
        }

#if 0
        chk = img_cmp(cur, last, cmp_type, (void*)&diff_capacity);
#endif
        chk = img_diff_npos(cur, last, sanpos_npos, sanpos, diff_capacity);
#if 0
        if(debug) {
            fprintf(stderr, "same ? %d\n", chk);
        }
#endif
        if(chk) {
            if(!beep_mute) {
                fprintf(stderr, "DIFF %d\a\a\n", chk);
            }
            else {
                fprintf(stderr, "DIFF %d\n", chk);
            }

            store_image_history(cur, last);

        }
        else {
            if(debug) {
                fprintf(stderr, "SAME\n");
            }
        }
    }
#endif

    PASS;


    vrec->got_count++;
    c_got_count++;

#if 0
    sprintf(tmpstr, "Date %d.%06d", curtime.tv_sec, curtime.tv_usec);
#endif

#if 0
    localtime_r(&cur->recv_sec, &htm);
    sprintf(tmpstr, "%2d-%02d %02d:%02d:%02d.%02d",
        htm.tm_mon+1, htm.tm_mday,
        htm.tm_hour, htm.tm_min, htm.tm_sec, cur->recv_usec/10000);
    XtVaSetValues(vrec->Tw, XtNlabel, tmpstr, NULL);
#endif


#if 0
    fprintf(stderr, "rateview_interval %d\n", rateview_interval);
#endif
    if(curtime.tv_sec-vrec->lasttime.tv_sec>rateview_interval) {
        if(vrec->lasttime.tv_sec<=0) {
        }
        else {
            diff = SUB_TIME(curtime, vrec->lasttime);
            rate = (double)vrec->got_count*1000.0/diff;
            fprintf(stderr, "##### id# %d, rate %5.2f, count %d time %d\n",
                vrec->node->id,
                rate, vrec->got_count, diff);
            sprintf(tmpstr, "Rate %5.1f f/s", rate);

            XtVaSetValues(vrec->Rw, XtNlabel, tmpstr, NULL);
        }

#if 0
        v_print_slots();
#endif

        vrec->lasttime = curtime;
        vrec->got_count = 0;

    }

    last = cur;

    PASS;

    return;
}


typedef struct {
    XtIntervalId    id;
    time_t          interval;
} xp_periodical_rec;

xp_periodical_rec win1;


Pixmap __bgpix;
Pixmap g_bgpix;
Pixmap bp_bgpix;
#define d_width 16
#define d_height 16
static unsigned char d_bits[] = {
   0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x10, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x10, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
#define d2_width 8
#define d2_height 8
static unsigned char d2_bits[] = {
   0x00, 0xfe, 0x82, 0xba, 0xa2, 0xbe, 0x80, 0xff};
#define d3_width 16
#define d3_height 16
static unsigned char d3_bits[] = {
   0xe3, 0xe3, 0x77, 0x77, 0x3e, 0x3e, 0x1c, 0x1c, 0x3e, 0x3e, 0x77, 0x77,
   0xe3, 0xe3, 0xc1, 0xc1, 0xe3, 0xe3, 0x77, 0x77, 0x3e, 0x3e, 0x1c, 0x1c,
   0x3e, 0x3e, 0x77, 0x77, 0xe3, 0xe3, 0xc1, 0xc1};
#define dslash_width 16
#define dslash_height 16
static unsigned char dslash_bits[] = {
   0x44, 0x44, 0x22, 0x22, 0x11, 0x11, 0x88, 0x88, 0x44, 0x44, 0x22, 0x22,
   0x11, 0x11, 0x88, 0x88, 0x44, 0x44, 0x22, 0x22, 0x11, 0x11, 0x88, 0x88,
   0x44, 0x44, 0x22, 0x22, 0x11, 0x11, 0x88, 0x88};
#define dbslash_width 16
#define dbslash_height 16
static unsigned char dbslash_bits[] = {
   0x22, 0x22, 0x44, 0x44, 0x88, 0x88, 0x11, 0x11, 0x22, 0x22, 0x44, 0x44,
   0x88, 0x88, 0x11, 0x11, 0x22, 0x22, 0x44, 0x44, 0x88, 0x88, 0x11, 0x11,
   0x22, 0x22, 0x44, 0x44, 0x88, 0x88, 0x11, 0x11};
#define d4_width 32
#define d4_height 8
static unsigned char d4_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0x11, 0x00, 0x00, 0x00,
   0x0a, 0x00, 0x04, 0x00, 0x04, 0x00, 0x0a, 0x00, 0x00, 0x00, 0x11, 0x00,
   0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00};







WakeWins()
{
    int k;
    char tmp[BUFSIZ];
    int ww;
    int nw;

    ground = XtVaCreateManagedWidget(
        "ground", formWidgetClass, toplevel,
        XtNbackgroundPixmap, g_bgpix,
        NULL);

    tcpanel = XtVaCreateManagedWidget(
        "tcpanel", boxWidgetClass, ground,
        XtNorientation, XtorientHorizontal,
        XtNbackgroundPixmap, bp_bgpix,
        XtNborderWidth, 0,
        NULL);

    topofield = XtVaCreateManagedWidget(
        "topofield", boxWidgetClass, ground,
        XtNfromVert, tcpanel,
        XtNwidth, 500,
        XtNheight, 360,
        NULL);

    tmsg = XtVaCreateManagedWidget(
        "tmsg", labelWidgetClass, ground,
        XtNfromVert, topofield,
        XtNbackgroundPixmap, bp_bgpix,
        XtNborderWidth, 0,
        XtNwidth, 500,
        XtNlabel, "",
        NULL);

    mcpanel = XtVaCreateManagedWidget(
        "mcpanel", boxWidgetClass, ground,
        XtNorientation, XtorientHorizontal,
#if 1
        XtNfromVert, tmsg,
#else
        XtNfromVert, topofield,
#endif
        XtNbackgroundPixmap, bp_bgpix,
        XtNborderWidth, 0,
        NULL);

    bcpanel = XtVaCreateManagedWidget(
        "tcpanel", boxWidgetClass, ground,
        XtNorientation, XtorientHorizontal,
        XtNfromVert, mcpanel,
        XtNbackgroundPixmap, bp_bgpix,
        XtNborderWidth, 0,
        NULL);

    tcpanelh = XtVaCreateManagedWidget(
        "panelhead", labelWidgetClass, tcpanel,
        XtNlabel, "Program",
        XtNbackgroundPixmap, bp_bgpix,
        XtNborderWidth, 0,
        NULL);

    bquit  = XtVaCreateManagedWidget(
        "quit", commandWidgetClass, tcpanel,
        NULL);
    XtAddCallback(bquit, XtNcallback, Quit, NULL);

    bhelp  = XtVaCreateManagedWidget(
        "help", commandWidgetClass, tcpanel,
        XtNfromHoriz, bquit,
        XtNsensitive, False,
        NULL);

    mcpanelh = XtVaCreateManagedWidget(
        "panelhead", labelWidgetClass, mcpanel,
        XtNlabel, "Topology",
        XtNbackgroundPixmap, bp_bgpix,
        XtNborderWidth, 0,
        NULL);

    bnfilter = XtVaCreateManagedWidget(
        "nnode", commandWidgetClass, mcpanel,   
        XtNlabel, "new node",
        NULL);
    bredraw = XtVaCreateManagedWidget(
        "redraw", commandWidgetClass, mcpanel, NULL);
    XtAddCallback(bredraw, XtNcallback, Redraw, NULL);
    XtAddCallback(bnfilter, XtNcallback, Nnode, NULL);

    bcpanelh = XtVaCreateManagedWidget(
        "panelhead", labelWidgetClass, bcpanel,
        XtNlabel, "Session",
        XtNbackgroundPixmap, bp_bgpix,
        XtNborderWidth, 0,
        NULL);

    bstart = XtVaCreateManagedWidget(
        "start", commandWidgetClass, bcpanel, NULL);
    bstop = XtVaCreateManagedWidget(
        "stop", commandWidgetClass, bcpanel, NULL);
    bswapper = XtVaCreateManagedWidget(
        "swapper", commandWidgetClass, bcpanel, NULL);
    bcheck = XtVaCreateManagedWidget(
        "check", commandWidgetClass, bcpanel, NULL);
    XtAddCallback(bstart, XtNcallback, Start, NULL);
    XtAddCallback(bstop, XtNcallback, Stop, NULL);
    XtAddCallback(bswapper, XtNcallback, ToggleSwapper, NULL);
    XtAddCallback(bcheck, XtNcallback, HCheck, NULL);

    pvpanel = XtVaCreateManagedWidget(
        "pvpanel", formWidgetClass, ground,
        XtNfromVert, bcpanel,
        NULL);

    for (k=0;k<max_viewer;k++) {
        sprintf(tmp, "viewer#%d", k);
        pvlabel[k] = XtVaCreateManagedWidget(
            "label", labelWidgetClass, pvpanel,
            XtNlabel, tmp,
            XtNwidth, 160,
            XtNborderWidth, 0,
            NULL);
        if(k>0) {
            XtVaSetValues(pvlabel[k],
                XtNfromHoriz,  pvlabel[k-1],
                NULL);
        }
        pview[k] = XtVaCreateManagedWidget(
            "view", lwwWidgetClass, pvpanel,
            XtNvisual,  xp_foil_visual,
            XtNdepth,   xp_foil_depth,
            XtNcolormap,xp_foil_cmap,
            XtNwidth,  160, XtNheight, 120,
            NULL);
        if(k==0) {
            XtVaSetValues(pview[k],
                XtNfromVert,  pvlabel[k],
                NULL);
        }
        else
        if(k>0) {
            XtVaSetValues(pview[k],
                XtNfromVert,  pvlabel[k],
                XtNfromHoriz, pview[k-1],
                NULL);
        }

#if 0
        sprintf(tmp, "date#%d", k);
        pvtime[k] = XtVaCreateManagedWidget(
            "time", labelWidgetClass, pvpanel,
            XtNlabel, tmp,
            XtNwidth, 160,
            XtNborderWidth, 0,
            NULL);
        if(k==0) {
            XtVaSetValues(pvtime[k],
                XtNfromVert,  pview[k],
                NULL);
        }
        else
        if(k>0) {
            XtVaSetValues(pvtime[k],
                XtNfromVert,  pview[k],
                XtNfromHoriz, pvrate[k-1],
                NULL);
        }
#endif

        sprintf(tmp, "rate#%d", k);
        pvrate[k] = XtVaCreateManagedWidget(
            "rate", labelWidgetClass, pvpanel,
            XtNlabel, tmp,
            XtNwidth, 160,
            XtNborderWidth, 0,
            NULL);
        if(k==0) {
            XtVaSetValues(pvrate[k],
                XtNfromVert,  pview[k],
                NULL);
        }
        else
        if(k>0) {
            XtVaSetValues(pvrate[k],
                XtNfromVert,  pview[k],
                XtNfromHoriz, pvrate[k-1],
                NULL);
        }
    }

    smsg = XtVaCreateManagedWidget(
        "smsg", labelWidgetClass, ground,
        XtNfromVert, pvpanel,
        XtNbackgroundPixmap, bp_bgpix,
        XtNborderWidth, 0,
        XtNwidth, 500,
        XtNlabel, "",
        NULL);

    XtOverrideTranslations(topofield, XtParseTranslationTable(topofield_trans));

    XtRealizeWidget(toplevel);

#if 1
    ww = 180;
    nw = 48;

    node_prop_shell =  XtVaCreatePopupShell(
        "npshell", transientShellWidgetClass,
        toplevel, NULL);
#endif
#if 0
    node_prop_shell =  XtVaCreatePopupShell(
        "npshell", simpleMenuWidgetClass,
        toplevel, NULL);
#endif
    node_prop_base =  XtVaCreateManagedWidget(
        "npbase", boxWidgetClass, node_prop_shell,
        NULL);

    node_prop_label =  XtVaCreateManagedWidget(
        "nplabel", labelWidgetClass, node_prop_base,
        XtNborderWidth, 0,
        XtNwidth, ww,
        XtNlabel, "Node Property Window",
        NULL);

    node_prop_spanel =  XtVaCreateManagedWidget(
        "nppanel", boxWidgetClass, node_prop_base,
        XtNorientation, XtorientHorizontal,
        NULL);

    node_prop_panel =  XtVaCreateManagedWidget(
        "nppanel", formWidgetClass, node_prop_spanel,
        XtNborderWidth, 0,
        NULL);


    node_prop_id =  XtVaCreateManagedWidget(
        "npid", labelWidgetClass, node_prop_panel,
        XtNborderWidth, 0,
        XtNwidth, ww,
        NULL);

    node_prop_type =  XtVaCreateManagedWidget(
        "nptype", labelWidgetClass, node_prop_panel,
        XtNfromVert, node_prop_id,
        XtNborderWidth, 0,
        XtNwidth, ww,
        NULL);

    node_prop_name =  XtVaCreateManagedWidget(
        "npname", labelWidgetClass, node_prop_panel,
        XtNfromVert, node_prop_type,
        XtNborderWidth, 0,
        XtNwidth, ww,
        NULL);

    node_prop_host =  XtVaCreateManagedWidget(
        "nphost", labelWidgetClass, node_prop_panel,
        XtNborderWidth, 0,
        XtNwidth, ww,
        XtNfromVert, node_prop_name,
        NULL);
    node_prop_prog_name =  XtVaCreateManagedWidget(
        "npprog", labelWidgetClass, node_prop_panel,
        XtNborderWidth, 0,
        XtNwidth, ww,
        XtNfromVert, node_prop_host,
        NULL);
    node_prop_filtermode =  XtVaCreateManagedWidget(
        "npprog", commandWidgetClass, node_prop_panel,
        XtNborderWidth, 0,
        XtNwidth, ww,
        XtNlabel, "filtermode: it is dummy to maximized size",
        XtNfromVert, node_prop_prog_name,
        NULL);
    XtAddCallback(node_prop_filtermode, XtNcallback, NpropTfilter, NULL);
    node_prop_sitter_port =  XtVaCreateManagedWidget(
        "npsport", labelWidgetClass, node_prop_panel,
        XtNborderWidth, 0,
        XtNwidth, ww,
        XtNfromVert, node_prop_filtermode,
        NULL);
    node_prop_sitter_fd  =  XtVaCreateManagedWidget(
        "npsfd", labelWidgetClass, node_prop_panel,
        XtNborderWidth, 0,
        XtNwidth, ww,
        XtNfromVert, node_prop_sitter_port,
        NULL);
    node_prop_down_listen_port  =  XtVaCreateManagedWidget(
        "npdlport", labelWidgetClass, node_prop_panel,
        XtNborderWidth, 0,
        XtNwidth, ww,
        XtNfromVert, node_prop_sitter_fd,
        NULL);
    node_prop_dlcid  =  XtVaCreateManagedWidget(
        "npdlcid", labelWidgetClass, node_prop_panel,
        XtNborderWidth, 0,
        XtNwidth, ww,
        XtNfromVert, node_prop_down_listen_port,
        NULL);
    node_prop_up_recv_port  =  XtVaCreateManagedWidget(
        "npurport", labelWidgetClass, node_prop_panel,
        XtNborderWidth, 0,
        XtNwidth, ww,
        XtNfromVert, node_prop_down_listen_port,
        NULL);
    node_prop_unb =  XtVaCreateManagedWidget(
        "npunb", boxWidgetClass, node_prop_spanel,
        XtNfromHoriz, node_prop_panel,
        NULL);
    node_prop_unbh =  XtVaCreateManagedWidget(
        "npunbh", labelWidgetClass, node_prop_unb,
        XtNlabel,   "upper nodes",
        XtNbackground, 0xffccff,
        XtNborderWidth, 0,
        NULL);
    node_prop_ucb =  XtVaCreateManagedWidget(
        "npucb", boxWidgetClass, node_prop_spanel,
        XtNfromHoriz, node_prop_unb,
/*
        XtNwidth, nw,
*/
        NULL);
    node_prop_ucbh =  XtVaCreateManagedWidget(
        "npucbh", labelWidgetClass, node_prop_ucb,
        XtNlabel,   "conn. id.s",
        XtNbackground, 0xffccff,
        XtNborderWidth, 0,
        NULL);
    node_prop_lnb =  XtVaCreateManagedWidget(
        "nplnb", boxWidgetClass, node_prop_spanel,
        XtNfromHoriz, node_prop_ucb,
        XtNwidth, nw,
        NULL);
    node_prop_lnbh =  XtVaCreateManagedWidget(
        "nplnbh", labelWidgetClass, node_prop_lnb,
        XtNlabel,   "lower nodes",
        XtNbackground, 0xffccff,
        XtNborderWidth, 0,
        NULL);
    node_prop_lcb =  XtVaCreateManagedWidget(
        "nplcb", boxWidgetClass, node_prop_spanel,
        XtNfromHoriz, node_prop_lnb,
        XtNwidth, nw,
        NULL);
    node_prop_lcbh =  XtVaCreateManagedWidget(
        "nplcbh", labelWidgetClass, node_prop_lcb,
        XtNlabel,   "conn. id.s",
        XtNbackground, 0xffccff,
        XtNborderWidth, 0,
        NULL);
    for(k=0;k<MAX_LOWERNODE;k++) {
        node_prop_unid[k] =  XtVaCreateManagedWidget(
            "npunid", labelWidgetClass, node_prop_unb,
            XtNborderWidth, 0,
            XtNwidth, nw,
            XtNlabel, "nid# ####",
            NULL);
        node_prop_ucid[k] =  XtVaCreateManagedWidget(
            "npucid", labelWidgetClass, node_prop_ucb,
            XtNborderWidth, 0,
            XtNwidth, nw,
            XtNlabel, "cid# ####",
            NULL);
        node_prop_lnid[k] =  XtVaCreateManagedWidget(
            "npunid", labelWidgetClass, node_prop_lnb,
            XtNborderWidth, 0,
            XtNwidth, nw,
            XtNlabel, "nid# ####",
            NULL);
        node_prop_lcid[k] =  XtVaCreateManagedWidget(
            "npucid", labelWidgetClass, node_prop_lcb,
            XtNborderWidth, 0,
            XtNwidth, nw,
            XtNlabel, "cid# ####",
            NULL);
    }

    node_prop_ok =  XtVaCreateManagedWidget(
        "npok", commandWidgetClass, node_prop_base,
        XtNlabel, "ok",
        XtNfromVert, node_prop_unb,
        NULL);
    XtAddCallback(node_prop_ok, XtNcallback, NpropOk, NULL);
    node_prop_cancel =  XtVaCreateManagedWidget(
        "npcancel", commandWidgetClass, node_prop_base,
        XtNlabel, "cancel",
        XtNfromVert, node_prop_unb,
        XtNfromHoriz, node_prop_ok,
        NULL);
    XtAddCallback(node_prop_cancel, XtNcallback, NpropCancel, NULL);

    XtRealizeWidget(node_prop_shell);

}


XtActionsRec main_action_list[] = {
/*
    {"keyin",       Keyin},
*/
    {"topofield-redraw",    redraw_topofield},
    {"mk-newnode",          mk_newnode},
    {"report-place",        report_place},
    {"drag-node",           drag_node},
    {"drag-node-start",     drag_node_start},
    {"drag-node-end",       drag_node_end},
    {"link-node",           link_node},
    {"link-node-start",     link_node_start},
    {"link-node-end",       link_node_end},
    {"pick-node",           pick_node},
    {"release-node",        release_node},
    {"replace-node",        replace_node},
    {"wake-nprop",          WakeNprop},
    {"toggle-filtermode",   toggle_filtermode},
};

void
quit_sentry(Widget w, XtPointer client_data, XtPointer call_data)
{
    exit(0);
}

#if 0
void
toggle_mute(Widget w, XtPointer client_data, XtPointer call_data)
{
    beep_mute = 1 - beep_mute;
    fprintf(stderr, "beep_mute %d\n", beep_mute);
    fflush(stderr);

    if(beep_mute) {
        XtVaSetValues(bmute, XtNforeground, black);
        XtVaSetValues(bmute, XtNbackground, white);
    }
    else {
        XtVaSetValues(bmute, XtNforeground, white);
        XtVaSetValues(bmute, XtNbackground, black);
    }
}
#endif

#if 0
void
toggle_ghost(Widget w, XtPointer client_data, XtPointer call_data)
{
    ghost_mode = 1 - ghost_mode;
    fprintf(stderr, "ghost_mode %d\n", ghost_mode);
    fflush(stderr);

    if(ghost_mode) {
        XtVaSetValues(bghost, XtNforeground, black);
        XtVaSetValues(bghost, XtNbackground, white);
    }
    else {
        XtVaSetValues(bghost, XtNforeground, white);
        XtVaSetValues(bghost, XtNbackground, black);
    }
}
#endif

#if 0
void
toggle_stick(Widget w, XtPointer client_data, XtPointer call_data)
{
    stick_mode = 1 - stick_mode;
    fprintf(stderr, "stick_mode %d\n", stick_mode);
    fflush(stderr);

    if(stick_mode) {
        XtVaSetValues(bstick, XtNforeground, black);
        XtVaSetValues(bstick, XtNbackground, white);
    }
    else {
        XtVaSetValues(bstick, XtNforeground, white);
        XtVaSetValues(bstick, XtNbackground, black);
    }
}
#endif


void
usage()
{
    printf("nv - node viewer, viewer for meta-distributed sentry\n\
\n\
usage: nv [options]\n\
\n\
options:\n\
    -help                   print this message\n\
    -camerahost hosts[:port][,host[:port]]\n\
                            camera host(s)\n\
    -rmhost host[:port]     Resource Manager (RM) host\n\
    -shape factor           rate shaping factor ('2' means half rate)\n\
\n\
examples:\n\
    %% nv -camerahost foo,bar:1357,junk:8179 -rmhost buzz:8129\n\
    %% nv -camerahost foo -shape 2\n\
\n");
}



store_color(Pixel *sa, int r, int g, int b)
{
    if(output_corder==SEN_C_XRGB) {
        *sa  = r*0x010000;
        *sa += g*0x000100;
        *sa += b*0x000001;
    }
    else 
    if(output_corder==SEN_C_XBGR) {
        *sa  = b*0x010000;
        *sa += g*0x000100;
        *sa += r*0x000001;
    }
    
}


XtIntervalId req_tid;

void
RequestRepeat(XtPointer client_data, XtIntervalId *iid)
{
    node_t *cur;
    viewer_rec *rec;

#if 0
    cur = nodelist;
    while(cur) {
        if(cur->type==NV_T_VIEWER) {
            rec = v_find_slot(cur);
            if(rec && rec->req_sendfd>=0) {
#if 0
                _request_call(rec->req_sendfd, data_port);
#endif
                _request_call(rec->req_sendfd, cur->up_recv_port);
            }
        }
        cur = cur->next;
    }
#endif
    rec = (viewer_rec*)client_data;
    if(rec) {
        cur = rec->node;
        if(cur && rec->req_sendfd>=0) {
            _request_call(rec->req_sendfd, cur->up_recv_port);
        }
    }


#if 0
    _request_call(data_port);
    req_tid = XtAppAddTimeOut(app_context, request_interval*1000,
                        RequestRepeat, NULL);
#endif

    req_tid = XtAppAddTimeOut(app_context, request_interval*1000,
                        RequestRepeat, rec);
}

int shape_factor = 6;    

char *rm_host = "mint040"; 
int rm_port = RM_QUERY_PORT;
int rm_fd = -1;


main(int argc, char **argv)
{
    XtInputId   iid, ciid;

    node_t *camera = NULL;
    node_t *viewer = NULL;

    use_shm_flag =0;

#if 0
    SetSignalHandler(SIGCHLD,   TrapChildDie);
    SetSignalHandler(SIGPIPE,   SIG_IGN);
#endif
    signal(SIGCHLD,   TrapChildDie);
    signal(SIGPIPE,   SIG_IGN);


    normc = black;
    highc = red;

    toplevel = XtVaAppInitialize(
        &app_context,                   /* Application context */
        "Nv",                           /* Application class */
        options, XtNumber(options),     /* option list */
        &argc, argv,                    /* command line args */
        fallback_resources,             /* for missing app-defaults file */
        XtNtitle,       "NV",
        XtNiconName,    "NV",
        NULL);                          /* terminate varargs list */

    XtVaGetApplicationResources(toplevel,
        (XtPointer) &nv_resource, resources, XtNumber(resources), NULL);

    XtAppAddActions(app_context, main_action_list, XtNumber(main_action_list));


    SetupVisual(XtDisplay(toplevel), fvisual, svisual);  

    output_format = SEN_F_RGB;
    store_color(&black, 0x00, 0x00, 0x00);
    store_color(&white, 0xff, 0xff, 0xff);

#if 0
    diff_pix =  XCreatePixmap(XtDisplay(toplevel),
                DefaultRootWindow(XtDisplay(toplevel)),
                640, 480, xp_foil_depth);
#endif



    show_help = atoi(nv_resource.show_help);
    if(show_help) {
        usage();
        exit(0);
    }


#if 0
    rtr_set_global_param(RTR_P_VERBOSE, NULL);
    if(debug) {
        rtr_set_global_param(RTR_P_DEBUG, NULL);
    }
#endif
    rtr_set_global_param(RTR_P_DEBUG, NULL);

    rtr_set_global_param(RTR_P_LENFREELIST, 1024);
    rtr_set_global_param(RTR_P_QTIME, 2);

    rtr_initmem();


#if 0
    XtRealizeWidget(toplevel);
#endif

#if 1
 {
    Window root;
    int scr;
    int dep;

    root = DefaultRootWindow(XtDisplay(toplevel));
    scr = DefaultScreen(XtDisplay(toplevel));
    dep = DefaultDepth(XtDisplay(toplevel), scr);
    fprintf(stderr, "dep %d\n", dep);
    fflush(stderr);

    g_bgpix = XCreatePixmap(XtDisplay(toplevel),
                DefaultRootWindow(XtDisplay(toplevel)),
                d_width, d_height, dep);
#if 1
    g_bgpix = XCreatePixmapFromBitmapData(XtDisplay(toplevel),
                g_bgpix,
                dslash_bits, dslash_width, dslash_height,
                0xffccff, 0xcccccc, dep);
#endif
#if 0
    g_bgpix = XCreatePixmapFromBitmapData(XtDisplay(toplevel),
                g_bgpix,
                d4_bits, d4_width, d4_height,
                0xffccff, 0xcccccc, dep);
#endif

#if 0
    bp_bgpix = XCreatePixmapFromBitmapData(XtDisplay(toplevel),
                g_bgpix,
                d_bits, d_width, d_height,
                0xccccff, 0xcccc99, dep);
#endif
#if 1
    bp_bgpix = XCreatePixmapFromBitmapData(XtDisplay(toplevel),
                g_bgpix,
                dbslash_bits, dbslash_width, dbslash_height,
                0xccccff, 0xcccccc, dep);
#endif

 }
#endif

    WakeWins();
#if 0
    XtRealizeWidget(toplevel);
#endif

    shape_factor = atoi(nv_resource.shape);
    if(shape_factor==0) {
        shape_factor = -1;
    }
    fprintf(stderr, "shape %d\n", shape_factor);

    fprintf(stderr, "camera(s): %s\n", nv_resource.camera_servpair_list);
    fprintf(stderr, "RM:        %s\n", nv_resource.rm_servpair);

#if 0
    fprintf(stderr, "camera '%s' %s -> %d\n",
        nv_resource.camera_host,
        nv_resource.camera_port, atoi(nv_resource.camera_port));

    if(!nv_resource.camera_host[0]) {
        fprintf(stderr, "empty host\n");
    }
    if(!nv_resource.camera_port[0]) {
        fprintf(stderr, "empty port\n");
    }
#endif


    {
        char hname[BUFSIZ], pname[BUFSIZ];
        int pno;
        int chk;

        chk = split_colonedpair(nv_resource.rm_servpair, hname, pname);
        if(chk==1) {
            pno = RM_QUERY_PORT;
        }
        else {
            pno = atoi(pname);
        }

        rm_host = strdup(hname);
        rm_port = pno;
    }


#if 0
    camera = new_node();
    camera->type = NV_T_CAMERA;
    camera->host = strdup(nv_resource.camera_host);
    camera->name = strdup("camera");
    strcpy(camera->prog_name, "sentryd");
    camera->down_listen_port = atoi(nv_resource.camera_port);
    camera->x = 120;
    camera->y = 180;
    add_node(camera);
#endif

 {
    int k;
    char tname[BUFSIZ];
    char *p, *q;
    int ncamera;
    char hname[BUFSIZ], pname[BUFSIZ];
    int pno;
    int chk;

    ncamera = count_commasplitted_list(nv_resource.camera_servpair_list);

    fprintf(stderr, "cameralist '%s'-> %d\n",
        nv_resource.camera_servpair_list, ncamera);

    p = nv_resource.camera_servpair_list;
    k = 0;
    while(*p) {
        while(*p && (*p==','||*p==' '||*p=='\t'||*p=='\r'||*p=='\n'))
            p++;
        q = tname;
        while(*p && !(*p==','||*p==' '||*p=='\t'||*p=='\r'||*p=='\n'))
            *q++ = *p++;
        *q = '\0';

        if(tname[0]) {
            chk = split_colonedpair(tname, hname, pname);
            if(chk==1) {
                pno = SENTRY_CTRL_PORT;
            }
            else {
                pno = atoi(pname);
            }
            fprintf(stderr, "camera#%d, '%s' -> <%d> '%s' '%s' (%d)\n",
                k, tname, chk, hname, pname, pno);
            camera = new_node();
            camera->type = NV_T_CAMERA;
            camera->host = strdup(hname);
            camera->name = strdup("camera");
            strcpy(camera->prog_name, "sentryd");
            camera->down_listen_port = pno;
            camera->x = 500/4;
            camera->y = 360*(k+1)/(ncamera+1);
            add_node(camera);

            k++;
        }

    }
 }

 {
    int k;
    int c;
    int tport;
    char tname[BUFSIZ];
    int datafd;

    for(k=0;k<max_viewer;k++) {
        datafd = build_data_port_try(-1, &tport);
        fprintf(stderr, "socket for receive fd# %d, port# %d\n",
            datafd, tport);
        if(datafd<0) {
            continue;
        }

        sprintf(tname, "viewer#%d", k);

        viewer = new_node();
        viewer->type = NV_T_VIEWER;
        viewer->host = strdup("localhost");
        viewer->name = strdup(tname);
        strcpy(viewer->prog_name, "nv");
        viewer->down_listen_port = -1;
        viewer->up_recv_port = tport;
        viewer->x = 360;
#if 0
        viewer->y = 180;
#endif
        viewer->y = 480*(k+1)/(max_viewer+2);
        add_node(viewer);

        vslot[k].Vw = pview[k];
        vslot[k].Lw = pvlabel[k];
#if 0
        vslot[k].Tw = pvtime[k];
#endif
        vslot[k].Rw = pvrate[k];
        vslot[k].req_sendfd = -1;
        vslot[k].res_recvfd = datafd;
        vslot[k].node   = viewer;
        vslot[k].pque   = NULL;
        vslot[k].running = 0;

        if(datafd>=0)
        iid = XtAppAddInput(app_context, datafd,
                (caddr_t)XtInputReadMask, eventical_update, &vslot[k]);

        datafd = -1;

        print_nodelist(stdout);
    }

    for(k=0;k<max_viewer;k++) {
        fprintf(stderr, "%d: %p node %p id# %d, V,R=%d,%d fd# %d\n",
            k, &vslot[k], vslot[k].node, vslot[k].node->id,
            vslot[k].Vw, vslot[k].Rw, vslot[k].res_recvfd);
    }

 }


    pwin_gc = XtGetGC(pview[0], NULL, NULL);

    probe_tid = XtAppAddTimeOut(app_context, probe_interval,
            ProbeNodeState, NULL);

    XtAppMainLoop(app_context);
}


#ifdef TEST_NV
/* dummy */
int
start_allnodes(node_t *nd)
{
    fprintf(stderr, "start_allnodes\n");
}
#endif
