/*
 * get_data.c - get data (image header and body)
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>

#include "hook.h"
#include "alist.h"
#include "sentry.h"
#include "img.h"
#include "pool.h"

#include "rtr.h"
int rtr_async_recv(int sock, subrecord_t **recvlist,
    void *buffer, int size, int *rsize);


#ifdef  HAVE_TYPEDEF_FD_SET
typedef fd_set          fdset_t;
#else
typedef struct fd_set   fdset_t;
#endif  /* HAVE_FDSET_T */



#if 0
/*
 * isreadable - check specified FD is readable or not.
 *      if sec<0        wait with blocking
 *
 *
 * return:
 *      1   readable
 *      0   not readable
 *     -1   error
 *
 */

int
isreadable_via_select(int fd, int sec, int usec)
{
    fdset_t fdset, rfdset;
    struct timeval timeout;
    struct timeval *ptimeout;
    int chk;

    FD_ZERO(&fdset);
    FD_SET(fd, &fdset);

    if(sec<0) {
        ptimeout = NULL;
    }
    else {
        ptimeout = &timeout;

        timeout.tv_sec  =  sec;
        if(usec<0) {
            timeout.tv_usec = 0;
        }
        else {
            timeout.tv_usec = usec;
        }
    }

retry:
    rfdset = fdset;
    chk = select(fd+1, &rfdset, NULL, NULL, ptimeout);

#if 0
    dump_fdset(fd+1, &fdset, &rfdset, NULL, NULL);
#endif
    if(chk<0) {
        if(errno==EINTR) {
            goto retry;
        }
        else {
            Error(" #isreadable:fd# %d, errno %d# ", fd, errno);
            return -1;
        }
    }

    if(FD_ISSET(fd, &rfdset)) {
#if 0
        Trace("readable ");
#endif
        return 1;
    }
    else {
#if 0
        Trace("unreadable ");
#endif
        return 0;
    }
}

#endif


static char hidden_buf[BUFSIZ*10];

#ifndef PASS
#if 0
#define PASS  fprintf(stderr,"PASS %s:%d\n", __FILE__, __LINE__)
#else
#define PASS
#endif
#endif


int
recv_image_multi(int datafd, subrecord_t **recvlist, char **buff,
    sentry_shmrec *info, sentry_imgrec *dest)
{
    struct sockaddr_in from;
    int fromlen;
    char packet[BUFSIZ];
    int nread;
    sentry_protocol3_image_hdr *hdr;
    int chk;
    struct timeval image_tstamp;
    int image_format, image_length, image_width, image_height;
    struct timeval curtime;
    time_t now;
    int recvlen = -1;
    int rlen;
    int ret;
    
    ret = 0;

    fflush(stdout);

    recvlen = rtr_async_recv(datafd, recvlist,
        hidden_buf, sizeof(hidden_buf), &rlen); 
    if(debug>300) {
        fprintf(stderr, "async: recvlen %d, rlen %d\n", recvlen, rlen);
        fflush(stderr);
    }

    if(recvlen<0) {
#if 0
        fprintf(stderr, "rtr_async_recv returns %d\n", recvlen);
#endif
        if(debug>300) {
            fprintf(stderr, "negative length data recieved %d\n", recvlen);
        }

        if(recvlen==RTR_E_AGAIN)
            goto suspend;
#if 1
        fprintf(stderr, "rtr_async_recv returns %d\n", recvlen);
#endif

        goto fail;
    }

    hdr = (sentry_protocol3_image_hdr*)hidden_buf;

    gettimeofday(&curtime, NULL);
    now = curtime.tv_sec;

#if 0
    fprintf(stderr, "fd# %d, recvlen %d, rlen %d vs %d (offset %d; %dx%d)\n",
        datafd,
        recvlen, rlen, ntohl(hdr->size), ntohl(hdr->offset),
        ntohl(hdr->width), ntohl(hdr->height));
    fflush(stderr);
#endif

    if(recvlen - sizeof(sentry_protocol3_image_hdr) < ntohl(hdr->size)) {
        fprintf(stderr,
            "the received data is too short (%d ?= %d+%d)\n", 
            recvlen, sizeof(sentry_protocol3_image_hdr), ntohl(hdr->size));
    }

#if 0 
    dump_hexstringsegment("recv", hidden_buf, 46);
#endif
#if 0
    dump_to_file("/var/tmp/x.jpg", hidden_buf, chk);
#endif
#if 0
    dump_to_file("/var/tmp/y.jpg",
        hidden_buf+sizeof(sentry_protocol3_image_hdr), chk);
#endif

    PASS;

    if(dest->length>0 && recvlen>dest->body_len) {
        fprintf(stderr, "length in data %d is larger than storage %d\n",
            dest->length, dest->body_len);
        goto out;
    }
    if(recvlen>0 && recvlen>dest->body_len) {
        fprintf(stderr,
            "length of received data %d is larger than storage %d\n",
            recvlen, dest->body_len);
        goto out;
    }

    PASS;

#if 0
printf("format: direct %d %x ntoh %d %x\n", 
    hdr->format, hdr->format, ntohl(hdr->format), ntohl(hdr->format));
#endif

    dest->send_sec  = ntohl(hdr->hd_tstamp.tv_sec);
    dest->send_usec = ntohl(hdr->hd_tstamp.tv_usec);
    dest->recv_sec  = curtime.tv_sec;
    dest->recv_usec = curtime.tv_usec;
    dest->format    = ntohl(hdr->format);
    dest->corder    = ntohl(hdr->color);
    dest->length    = ntohl(hdr->size);
    dest->width     = ntohl(hdr->width);
    dest->height    = ntohl(hdr->height);

#if 0
    fprintf(stderr, "network %dx%d -> memory %dx%d\n",
        ntohl(hdr->width), ntohl(hdr->height),
        dest->width, dest->height);
    fflush(stderr);
#endif

    if(debug>=100) {
        printf("  pos %p, length %d/%d, format '%s'(%d;%x) corder '%s'(%d)\n",
            dest->body,
            dest->length, dest->body_len,
            assoc_v(format_alist, dest->format),
            dest->format,
            dest->format,
            assoc_v(color_alist, dest->corder),
            dest->corder
            );
        printf("  hidden_buf size %d\n", sizeof(hidden_buf));
    }

    if(debug>=300) {
        show_imgattr(stderr, dest);
    }

    PASS;

    memcpy(dest->body, hidden_buf, dest->length);

    PASS;

#if 0
    fprintf(stderr, "body %p, buf %p, len %d\n",
        dest->body, hidden_buf+sizeof(sentry_protocol3_image_hdr),
        dest->length);
#endif

    memcpy(dest->body, hidden_buf+sizeof(sentry_protocol3_image_hdr),
        dest->length);
    info->info.last_update = now;

 if(debug>=300) {
    fprintf(stderr,
        "packet %d.%06d; stored issued %d.%06d received %d.%06d\n",
        ntohl(hdr->hd_tstamp.tv_sec), ntohl(hdr->hd_tstamp.tv_usec),
        dest->send_sec, dest->send_usec,
        dest->recv_sec, dest->recv_usec);
 }
  
#if 0
    dump_rgb_to_file("/var/tmp/x.ppm", dest->body,
        dest->width, dest->height);
#endif
#if 0
    dump_to_file("/var/tmp/z.jpg", dest->body,
        dest->length);
#endif

    goto out;

suspend:
#if 0
    fprintf(stderr, "suspend fd# %d\n", datafd);
#endif
    ret = RTR_E_AGAIN;
    goto out;

fail:
    fprintf(stderr, "error or end fd# %d\n", datafd);
    ret = RTR_E_EOF;
    goto out;


out:
    PASS;

    return ret;
}

extern subrecord_t *recvlist;

int
recv_image(int datafd, char **buff,
    sentry_shmrec *info, sentry_imgrec *dest)
{

    return recv_image_multi(datafd, &recvlist, buff, info, dest);
}



