/*
 * sentryc.c - sentry network client
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <signal.h>
#include <syslog.h>
#ifdef TIME_WITH_SYS_TIME
 #include <sys/time.h>
 #include <time.h>
#else
 #ifdef HAVE_SYS_TIME_H
  #include <sys/time.h>
 #else
  #include <time.h>
 #endif
#endif
#include <errno.h>

#include "alist.h"
#include "sentry.h"
#include "img.h"
#include "pool.h"
#include "client.h"

#include "rtr.h"

#ifdef USE_LOCKFILE
char    lockfile[] = "/var/tmp/sentryc.pid";
#endif

int write_out(char *prefix, int pathtype, sentry_imgrec *img);
int isreadable_via_select(int fd, int sec, int usec);


#ifndef DEFAULT_PREFIX
#define DEFAULT_PREFIX "/var/tmp/sen_bank"
#endif
#ifndef DEFAULT_PATHTYPE
#define DEFAULT_PATHTYPE (1)
#endif

int  save_file_flag = 0;
char prefix[BUFSIZ] = DEFAULT_PREFIX;
int  pathtype = DEFAULT_PATHTYPE;



#ifndef TMP_IMGBUF_LEN
#define TMP_IMGBUF_LEN  (100*1024)
#endif


void
init_sentryc()
{
#ifdef USE_LOCKFILE
  int chk;
  FILE *fp;
  char mesg[80];

  if (debug)
    fprintf(stderr, "init_sentryc: lockfile=%s\n", lockfile);

  while (1) {
    chk = access(lockfile, F_OK);
    if (chk < 0) break;
    
    if (errno == EINTR)
      continue;

    sprintf(mesg, "init_sentryc: locckfile(%s) is found", lockfile);
    perror(mesg);
    exit (1);
  }

  if ((fp = fopen(lockfile, "w")) == NULL) {
    sprintf(mesg, "init_sentryc: cannot open lockfile(%s)", lockfile);
    perror(mesg);
    exit (1);
  }

  sprintf(mesg, "%d\n", getpid());
  fprintf(stderr, "pid# %d\n", getpid());
  
  if (fputs(mesg, fp) == EOF) {
    perror("init_sentryc: cannot fputs()");
    exit (1);
  }
  fclose(fp);
#endif /* USE_LOCKFILE */
}
  
void
goodbye(void)
{
    if(debug)
        fprintf(stderr, "goodbye\n");


#ifdef DELETE_IPC
    delete_pool(sentry_shmid);
#endif

#if 0
    syslog(LOG_NOTICE, "sentryc closed key=%d (%#x)\n",
        sentry_shmkey, sentry_shmkey);
#endif
    if(use_shm_flag) {
        syslog(LOG_NOTICE, "sentryc closed key=%d (%#x)",
            sentry_shmkey, sentry_shmkey);
    }
    else {
        syslog(LOG_NOTICE, "sentryc exit");
    }
    closelog();

#ifdef USE_LOCKFILE
    unlink(lockfile);
#endif
}

void
usage(void)
{
  fprintf(stderr, "Usage: sentryc [options] hostname.\n\
\n\
option:\n\
                                     current default\n\
  -h        help\n\
  -d num    debug                    %7d %7d\n\
  -t type   unicast(0)/multicast(1)        - unicast(0)\n\
  -c color  monochrome(0)/color(1)         - color(1)\n\
  -p num    server port number       %7d %7d\n\
  -q num    data receiving port      %7d %7d\n\
  -x type   frame size               %7s %7s\n\
  -r num    frame rate (/100sec)     %7d %7d\n\
  -k num    key number upon shm      %7d %7d\n\
  -s num    shm. size                %7d %7d\n\
  -m        use shared memory        %7d %7d\n\
  -w        write to disk            %7d %7d\n\
  -y path   path for writing         %-7s %-7s\n\
  -z num    pathtype                 %7d %7d\n\
            allday(0)/weekday(1)/n-day(>1)\n\
\n\
example:\n\
  %% sentryc -w -m -z 1 video.junk.com\n\
  %% sentryc -r 1000 video.bar.com\n\
  %% sentryc -k 121 -q 6301 video1.foo.com\n\
  %% sentryc -k 122 -q 6302 video2.foo.com\n\
  %% sentryc -k 781 -q 9237 -p 2319 video4.foo.com\n\
  %% sentryc -k 782 -q 9275 -p 2342 video4.foo.com\n\
\n",
    debug, 0,
    server_port, SENTRY_CTRL_PORT,
    data_port, SENTRY_DATA_PORT,
    assoc_v(size_alist, server_size),
        assoc_v(size_alist, SIZE_160x120),
    server_rate, DEFAULT_RATE,
    sentry_shmkey, SENTRY_SHMKEY,
    shm_size, SHM_SIZE,
    use_shm_flag, 1,
    save_file_flag, 0,
    prefix, DEFAULT_PREFIX,
    pathtype, DEFAULT_PATHTYPE
);

    exit (1);
}

#define TMP_IMGBUF_LEN  (100*1024)

static sentry_shmrec bk_info;
static sentry_imgrec bk_inframe;
static unsigned char *inimgbuf=NULL;
static sentry_imgrec bk_drframe[2];
static unsigned char *drimgbuf0=NULL;
static unsigned char *drimgbuf1=NULL;


extern apair_t buflen_xrgb_alist;


alloc_inimgbufs()
{
    inimgbuf = (unsigned char*)malloc(TMP_IMGBUF_LEN);
    bk_inframe.body = inimgbuf;
    bk_inframe.body_len = TMP_IMGBUF_LEN;
}

alloc_drimgbufs()
{
    int buflen;

#if 0
    if(output_format!=SEN_F_RGB) {
        fprintf(stderr, "no idea for no RGB device\n");
        exit(1);
    }
#endif

#if 0
    if(output_corder==SEN_C_RGB||
        output_corder==SEN_C_BGR) {
        buflen = TMP_IMGBUF_LEN;
    }
    else {
        buflen = TMP_IMGBUF_LEN;
    }
#endif
    buflen = (int)assoc_p(&buflen_xrgb_alist, MAX_DISPLAY_SIZE);
 if(debug) {
    printf("buflen %d\n", buflen);
 }
    if(buflen==0) {
        buflen = TMP_IMGBUF_LEN;
    }

    drimgbuf0 = (unsigned char*)malloc(buflen);
    bk_drframe[0].body = drimgbuf0;
    bk_drframe[0].body_len = buflen;

    drimgbuf1 = (unsigned char*)malloc(buflen);
    bk_drframe[1].body = drimgbuf1;
    bk_drframe[1].body_len = buflen;
}


int cmp_type;
int diff_capacity = 0x20;

void
mainloop(void)
{
#if 0
    sentry_shmrec *bank;
    sentry_shmrec back_bank;
    sentry_shmrec back_info;
    unsigned char *imgbuf=NULL;
#endif
    int got_count;
    int chk;
    time_t now;
    sentry_shmrec *info;
    sentry_imgrec *inframe;
    sentry_imgrec *cur, *last;
    char *tmp_buff;
    int drfno;
    int pos;
    int change;


    init_request();
    request_call();
    signal(SIGALRM, request_call);

    cur = last = NULL;

    got_count = 0;
    while (1) {
        chk = isreadable_via_select(datafd, 10, 0);
        if(chk==1) {
            now = time(NULL);

            if(use_shm_flag) {
                info = &sentry_shm_chunk[0];
                pos = 1 + now%sentry_shm_chunk_radix;
                inframe = &(sentry_shm_chunk[pos].entry.header);
            }
            else {
                info = &bk_info;
                if(got_count==0) {
                    alloc_inimgbufs();
                    alloc_drimgbufs();
                }

                inframe = &bk_inframe;
            }

/*
            chk = get_data2(datafd, &tmp_buff, info, inframe);
*/
            chk = recv_image(datafd, &tmp_buff, info, inframe);
            if(chk) {
                continue;
            }

            drfno = got_count%2;
            cur = &bk_drframe[drfno];

            copyattr_img(cur, inframe);
            if(inframe->format==SEN_F_JPEG) {
                chk = decode_jpeg(inframe->body, inframe->length,
                        cur->body, &cur->width, &cur->height);
                if(chk==0) {
                    cur->length = cur->width * cur->height *3;
                    cur->format = SEN_F_RGB;
                    cur->corder = SEN_C_RGB;

#if 0
                    show_imgattr(stderr, cur);
#endif
                }
            }

            if(last) {
#if 0
                show_imgattr(stderr, last);
#endif
#if 0
                change = img_cmp(cur, last, SEN_M_9POINTS,
                            (void*)&diff_capacity);
#endif
                change = img_cmp(cur, last, cmp_type,
                            (void*)&diff_capacity);

                if(change && save_file_flag) {
#if 0
                    write_out(prefix, pathtype, cur); 
#endif
                    write_out(prefix, pathtype, inframe); 
                }
            }

            if(got_count%report_interval==0) {
                syslog(LOG_NOTICE, "sentryc key=%d; got %d data\n",
                    sentry_shmkey, got_count);
            }

            last = cur;

            got_count++;
        }
    }
}

subrecord_t *recvlist=NULL;

int
main(int argc, char **argv)
{
  int c;
  extern char *optarg;
  extern int optind;
  int skey;
  int clear;
  char *qformat;
  char *qsize;
  char *qcolor;
  int tv;

#if 1
  printf("SENTRY_SHM_VERSION     %6d\n", SENTRY_SHM_VERSION);
  printf("sizeof(sentry_imgrec)  %6d\n", sizeof(sentry_imgrec));
  printf("SENTRY_SHM_HEADER_SIZE %6d\n", SENTRY_SHM_HEADER_SIZE);
  printf("SENTRY_SHM_BLOCK_SIZE  %6d\n", SENTRY_SHM_BLOCK_SIZE);
  if(SENTRY_SHM_HEADER_SIZE>SENTRY_SHM_BLOCK_SIZE) {
    printf("SENTRY_SHM_HEADER_SIZE > SENTRY_SHM_BLOCK_SIZE\n");
  }
  if(sizeof(sentry_imgrec) > SENTRY_SHM_HEADER_SIZE) {
    printf("sizeof(sentry_imgrec) > SENTRY_SHM_HEADER_SIZE\n");
  }
#endif

  rtr_initmem();

  clear = 0;
  data_port = SENTRY_DATA_PORT;
  skey = -1;

  openlog("sentryc", LOG_PID, LOG_DAEMON);
  syslog(LOG_NOTICE, "sentryc starts.\n");

  signal(SIGQUIT, complete);
  signal(SIGINT,  complete);
  signal(SIGTERM, complete);
  signal(SIGALRM, SIG_IGN);
  atexit(goodbye);

  qformat   = NULL;
  qsize     = NULL;
  qcolor    = NULL;
  cmp_type  = SEN_M_9POINTS;

  while ((c = getopt(argc, argv, "hd:vp:q:s:x:k:cr:i:mwy:z:F:")) != -1) {
    switch(c) {
    case 'h': /* help */
    default:
      usage();
      exit (1);
    case 'd':
      debug = atoi(optarg);
      break;
    case 'v': /* verbose mode */
      verbose = 1;
      break;
    case 'p':
      server_port = atoi(optarg);
      break;
    case 'q':
      data_port = atoi(optarg);
      break;
    case 's':
      shm_size = atoi(optarg);
      break;
    case 'x':
/*
      server_size = atoi(optarg);
*/
      qsize     = strdup(optarg);
      break;
    case 'k':
      skey = atol(optarg);
      break;
    case 'r':
      server_rate = atol(optarg);
      break;
    case 'c':
      clear = 1;
      break;
    case 'i':
      report_interval = atoi(optarg);
      break;
    case 'm':
      use_shm_flag = 0;
      break;
    case 'w':
      save_file_flag = 1;
      break;
    case 'y':
      strcpy(prefix, optarg);
      break;
    case 'z':
      pathtype = atoi(optarg);
      break;
    case 'F':
        tv = assoc_n(matching_type_alist, optarg);
        if(debug) {
            fprintf(stderr, "compare type %d %x '%s' <= '%s'\n",
                tv, tv, assoc_v(matching_type_alist, tv), optarg);
        }   
        cmp_type = tv;
        break;
    }      
  }

  if (argc <= 1) 
    usage();

  if(argv[optind]) {
    strcpy(server_host, argv[optind]);
  }
  else {
    strcpy(server_host, "localhost");
  }

  if(qsize)
      server_size = assoc_n(size_alist, qsize);

  if(debug) {
    fprintf(stderr, "size %s %d\n",
        assoc_v(size_alist, server_size), server_size);
  }

  if(verbose) {
    rtr_set_global_param(RTR_P_VERBOSE, NULL);
  }
  if(debug) {
    rtr_set_global_param(RTR_P_DEBUG, NULL);
    rtr_set_global_param(RTR_P_VERBOSE, NULL);
  }


  /*
   * daemonize if debug == 0
   */
  if (debug == 0) { /* future work */ }

  syslog(LOG_NOTICE, "sentryc key=%d (%#x)\n", skey, skey);

#if USE_LOCKFILE
  init_sentryc();
#endif

  if(use_shm_flag) {
    init_shm(skey);
    if(clear)
        clear_shm();
  }

  ctrlfd = build_ctrl_port(server_host, server_port);
  if (ctrlfd < 0) {
    printf("fail setup control fd\n");
    exit (1);
  }

  datafd = build_data_port(data_port);
  if (datafd < 0) {
    printf("fail setup data fd\n");
    exit (1);
  }

  (void)mainloop();
}

