/*
 * Sentry Video System
 *   controls QoS from clients' requests.
 * 
 *        original by akira-y@is.aist-nara.ac.jp, 1994,1995
 *        modified by k-chinen@is.aist-nara.ac.jp, 1996, 2002-2003
 *        modified by satosi-f@is.aist-nara.ac.jp, 1997, 2000
 *        modified by atsush-m@is.aist-nara.ac.jp, 2002
 *
 * $Id: $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/signal.h>
#include <sys/wait.h>
#include <time.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#ifdef USE_SYSLOG
#include <syslog.h>
#endif
#include <errno.h>

#include "alist.h"
#include "sentry.h"
#include "server.h"
#include "video.h"

int debug;
int verb;

apair_t qos_attr_alist[] = {
    {"Size",    QOS_SIZE},
    {"Zoom",    QOS_ZOOM},
    {"Color",   QOS_COLOR},
    {"Format",  QOS_FORMAT},
    {"Rate",    QOS_RATE},
    {NULL,      -1},
};

apair_t qos_desc_alist[] = {
    {size_alist,    QOS_SIZE},
    {color_alist,   QOS_COLOR},
    {format_alist,  QOS_FORMAT},
    {NULL, -1},
};

int capture_format=SEN_C_NONE;
int capture_corder=SEN_C_NONE;

#ifdef LOGTIME
time_t logtime;
struct tm *log_tm;
#endif

int batch_mode = 0;

int conn_no=0;
char conn_id[BUFSIZ];

int data_port = SENTRY_DATA_PORT;
int ctrl_port = SENTRY_CTRL_PORT;

int tcpsock=-1;            /* socket(ctrl/data) for tcp connection */
int udpsock=-1;            /* socket(ctrl only) for udp connection */

#if 0
int unicast_datasock;   /* data socket for udp uniicast */
#endif
int multicast_datasock; /* data socket for udp multicast */

int maxreadfds;
fd_set readmask;

int maxwritefds;
fd_set writemask;

char vserver[80] = "localhost";
int  require_encode_type = SEN_F_NONE;

extern int xl_offset, xr_offset, yt_offset, yb_offset;



void
reaper()
{
  union wait status;
  while(wait3((int *)&status, WNOHANG, (struct rusage *)0) >= 0)
    ;
}

void
finish_sentryd()
{
#ifdef LOGTIME
  time_t logtime;
  struct tm *log_tm;
#endif

  if (tcpsock >= 0) {
    shutdown(tcpsock, 2);
    close(tcpsock);
    tcpsock = -1;
  }
  if (udpsock >=0) {
    shutdown(udpsock, 2);
    close(udpsock);
    udpsock = -1;
  }
  if (multicast_datasock >= 0) {
    shutdown(multicast_datasock, 2);
    close(multicast_datasock);
    multicast_datasock = -1;
  }

  if (verb) {
#ifdef LOGTIME
    logtime = time(NULL);
    log_tm = localtime(&logtime);
    fprintf(stderr, "[%4d/%02d/%02d %02d:%02d:%02d] ",
        log_tm->tm_year + 1900, log_tm->tm_mon+1, log_tm->tm_mday,
        log_tm->tm_hour, log_tm->tm_min, log_tm->tm_sec);
#endif
    fprintf(stderr, "finish sentryd\nBye.\n");
  }

#ifdef USE_SYSLOG
  syslog(LOG_NOTICE, "sentryd is stopped.\n");
  closelog();
#endif

  exit (0);
}

#if 1
#define CMD_QUIT        (1)
#define CMD_HELP        (2)
#define CMD_OPEN        (3)
#define CMD_CLOSE       (4)
#define CMD_CONNECT     (5)
#define CMD_DISCONNECT  (6)
#define CMD_LIST        (7)
#define CMD_SHAPE      (31)
#define CMD_SYSTEM     (50)

apair_t cmdal[] = {
    {"QUIT",        CMD_QUIT},
    {"HELP",        CMD_HELP},
    {"OPEN",        CMD_OPEN},
    {"CLOSE",       CMD_CLOSE},
    {"CONNECT",     CMD_CONNECT},
    {"DISCONNECT",  CMD_DISCONNECT},
    {"LIST",        CMD_LIST},
    {"SHAPE",       CMD_SHAPE},
    {"SYSTEM",      CMD_SYSTEM},
    {NULL,          -1},
};
#endif

int
store_conn_id(char *ids, int idn)
{
    int mid;

    *ids++ = 'C';

    mid = idn;
    while(mid) {
        *ids = 'A' + (mid%26);
        ids++;
        mid -= mid%26;
    }
    *ids = '\0';
}

help()
{
    fprintf(stdout,"100\r\n\
supported commands:\r\n\
  HELP\r\n\
  OPEN <host> <port>\r\n\
  CLOSE <#id>\r\n\
  QUIT\r\n\
\r\n\
not supported:\r\n\
  CONNECT <host> <port>\r\n\
  DISCONNECT <#id>\r\n\
.\r\n");
    fflush(stdout);
}

int
do_cmd(char *cmdbuf)
{
    char xcmd[BUFSIZ];
    char arg1[BUFSIZ], arg2[BUFSIZ], arg3[BUFSIZ], arg4[BUFSIZ];
    int cmd;
    int xport;
    int chk;
    int tmpi;

#if 0
    chk = split_cmd(cmdbuf, xcmd, arg);
#endif
    chk = split_arg4(cmdbuf, xcmd, arg1, arg2, arg3, arg4);

    cmd = assoc_n(cmdal, xcmd);
    fprintf(stderr, "cmd %d [%s] arg [%s] [%s]\n",
        cmd, xcmd, arg1, arg2);
    
    switch(cmd) {
    case CMD_HELP:
            help();
            break;
    case CMD_QUIT:
            fprintf(stdout, "200\r\n");
            fflush(stdout);
            exit(134);
            break;
    case CMD_OPEN:
            if(arg1[0] && arg2[0]) {
                xport = atoi(arg2);
                tcpsock = build_listen_port(xport, SOCK_STREAM);
                udpsock = build_listen_port(xport, SOCK_DGRAM);
                if(tcpsock>=0 && udpsock>=0) {
                    ctrl_port = xport;
                    FD_SET(tcpsock, &readmask);
                    maxreadfds = MAX(maxreadfds, tcpsock);
                    FD_SET(udpsock, &readmask);
                    maxreadfds = MAX(maxreadfds, udpsock);
                    conn_no++;
                    store_conn_id(conn_id, conn_no);
                    fprintf(stdout, "201 %s\r\n", conn_id);
                }
                else {
                    if(tcpsock>=0) {
                        close(tcpsock);
                        tcpsock = -1;
                    }
                    if(udpsock>=0) {
                        close(udpsock);
                        udpsock = -1;
                    }
                    fprintf(stdout, "400\r\n");
                }
            }
            else {
                fprintf(stdout, "400\r\n");
            }
            fflush(stdout);
            break;
    case CMD_CLOSE:
            fprintf(stderr, "'%s' vs '%s'\n", arg1, conn_id);
            if(strcmp(arg1, conn_id)==0) {
#if 1
                show_client(stderr, "all-stars");
#endif
                chk = clear_all_client();

                if(tcpsock>=0) {
                    FD_CLR(tcpsock, &readmask);
                    shutdown(tcpsock, 2);
                    close(tcpsock);
                    tcpsock = -1;
                }

                if(udpsock>=0) {
                    FD_CLR(udpsock, &readmask);
                    shutdown(udpsock, 2);
                    close(udpsock);
                    udpsock = -1;
                }

                fprintf(stdout, "200\r\n");
            }
            else {
                fprintf(stdout, "400\r\n");
            }
            fflush(stdout);
            break;
    case CMD_LIST:
            fprintf(stdout, "100\r\n");
            show_client(stdout, "I have");
            fprintf(stdout, ".\r\n");
            fflush(stdout);
            break;
    case CMD_CONNECT:
    case CMD_DISCONNECT:
            fprintf(stdout, "400\r\n");
            fflush(stdout);
            break;
    case CMD_SHAPE:
            tmpi = atoi(arg1);
            chk = set_send_shape_radix(tmpi);
            if(chk) {
                fprintf(stdout, "400\r\n");
            }
            else {
                fprintf(stdout, "200\r\n");
            }
            break;
    case CMD_SYSTEM:
            fprintf(stdout, "201 %d %s %s\r\n",
                SENTRY_PROTOCOL_VERSION, __DATE__, __TIME__);
            fflush(stdout);
            break;
    default:
            fprintf(stdout, "400\r\n");
            fflush(stdout);
            break;
    }
    fflush(stdout);
}


void
mainloop()
{
  fd_set readfds;
  int nfds;
  char sysmesg[256];
  int newsockfd;
  struct sockaddr_in from;
  int fromlen;
  int nr;
  char buf[8192];
  sentry_protocol3_request_hdr req;
  struct timeval *wtime;
  struct timeval _wtime;

  extern int capture_xsize;
  extern int capture_ysize;
  extern _req_val req_val;
    extern int xsize, ysize;
    
    char cmdbuf[BUFSIZ];
    int  nreadcmd;
    int  chk;
    extern int multicast_clients, unicast_clients;

  extern void init_req_val(), inti_client();

  (void)init_req_val();
  (void)init_client();

  fromlen = sizeof(from);

  fprintf(stderr, "batch_mode? %d\n", batch_mode);
  fflush(stderr);

  if(batch_mode) {
      fprintf(stderr, "listening TCP fd# %d, UDP fd# %d\n",
            tcpsock, udpsock);
    FD_ZERO(&readmask);
      FD_SET(tcpsock, &readmask); 
      maxreadfds = tcpsock;
      FD_SET(udpsock, &readmask);
/*
      if (maxreadfds < udpsock) maxreadfds = udpsock;
*/
      maxreadfds  = MAX(maxreadfds, udpsock);
  }
  else {
        FD_ZERO(&readmask);
        maxreadfds = 0;
        FD_SET(0, &readmask); 

        fprintf(stdout,"100 welcome. this is sentryd.\r\n.\r\n");
        fflush(stdout);
  }

  FD_ZERO(&writemask);
  maxwritefds = 0;

    xsize = capture_xsize = 160;
    ysize = capture_ysize = 120;
    capture_format = SEN_F_RGB;
    capture_format = SEN_F_JPEG;
    SET_QOS(QOS_FORMAT, SEN_F_JPEG);
 
#if 1
    start_capture();
#endif
 
    while (1) {

#if 0
    fprintf(stderr, "<%d>", maxreadfds);
    fflush(stderr);
#endif



        /* no client, no work, long sleep */ 
        if (unicast_clients <= 0 && multicast_clients <= 0) {
            _wtime.tv_sec  = 0;
            _wtime.tv_usec = 100000;        /* 100 ms */
            wtime = &_wtime;
        }
        /* have client */
        else {
            /*
             * by kernel sync., do not sleep like busy loop.
             * do not warrly, in this case the program waits in
             * the capturring.
             * the program should do select rapidly, because
             * it has to enter the capturring.
             */
            if(video_sync_type==SENV_SYNC_BY_KERNEL) {
                _wtime.tv_sec  = 0;
                _wtime.tv_usec = 0;         /* 0 ms */
                wtime = &_wtime;
            }
            /*
             * otherwise, sleep little.
             * because, in this case the program only handle
             * request arrivals from clients.
             */
            else {
                _wtime.tv_sec  = 0;
                _wtime.tv_usec = 50000;     /* 50 ms */
                wtime = &_wtime;
            }
            
        }

        readfds = readmask;

        nfds = select(maxreadfds+1, &readfds,
                (fd_set *)NULL, (fd_set *)NULL, wtime);
        if (nfds <= 0) {
#ifdef USE_SYSLOG
            sprintf(sysmesg, "sentryd: select: %s", sys_errlist[errno]);
            syslog(LOG_ERR, sysmesg);
#endif
            if(video_sync_type==SENV_SYNC_BY_KERNEL) {
                send_image(129);
            }
            continue;
        }

        if(!batch_mode && FD_ISSET(0, &readfds)) {
            chk  = get_cmdline(0, cmdbuf);
            if(chk==0) {
                break;
            }
            fflush(stdout);
            fprintf(stderr, "++++++\n");
            fprintf(stderr, " %4d:%s", nreadcmd, cmdbuf);
            fflush(stderr);
            if(chk>0) {
                do_cmd(cmdbuf);
            }
            continue;
        }

        /* connection request arrived at tcp port */
        if (FD_ISSET(tcpsock, &readfds)) {
            FD_CLR(tcpsock, &readfds);
            newsockfd = accept(tcpsock, (struct sockaddr *)&from, &fromlen);
            if (newsockfd < 0)
                goto udp_check;

            if (verb) 
                fprintf(stderr, "TCP connection is arrived.\n");

            if (debug) {
                fprintf(stderr,
                    "from->sin_addr = %s\n", inet_ntoa(from.sin_addr));
                fprintf(stderr,
                    "from->sin_port = %d\n", 
                    ntohs(from.sin_port));
            }

#if 0
            get_tcp_request(newsockfd);
#endif

            if (add_client(0xf2, &from, newsockfd) < 0) {
                fprintf (stderr, "close connection port# %d.\n", newsockfd);
                shutdown(newsockfd, 2);
                close(newsockfd);
                continue;
            }

#if 0
            FD_SET(newsockfd, &readmask);
            if (maxreadfds < newsockfd) maxreadfds = newsockfd;
            FD_SET(newsockfd, &writemask);
            if (maxwritefds < newsockfd) maxwritefds = newsockfd;
#endif
            FD_SET(newsockfd, &readmask);
            maxreadfds = MAX(newsockfd, maxreadfds);
            FD_SET(newsockfd, &writemask);
            maxwritefds = MAX(newsockfd, maxwritefds);
      
            fprintf(stderr,
                "sentryd: tcp sock# %d is connected from %s:%d.\n", 
                newsockfd,
                inet_ntoa((struct in_addr *)ntohl(from.sin_addr.s_addr)),
                ntohs(from.sin_port));
            continue;
}

udp_check:
        if (FD_ISSET(udpsock, &readfds)) { /* request arrived at udp port */
            if (verb)
                fprintf(stderr, "UDP connection/request is arrived.\n");

            get_udp_request(&from, &fromlen, &req);
            process_request(&from, &fromlen, &req, 1);

            if (debug) {
                u_long addr = ntohl(from.sin_addr.s_addr);

                fprintf(stderr, "addr# %x -> %x, %s\n", 
                    ntohl(from.sin_addr.s_addr), addr,
                    inet_ntoa((struct in_addr *)ntohl(addr)));
            }

            fprintf(stderr,
                "sentryd: request is arrived to udp port# %d from %s:%d\n", 
                udpsock, inet_ntoa(from.sin_addr), ntohs(from.sin_port));

        }
        else {
            /* QoS request arrived at TCP port */
            /*
             * QoS requests are arrived from TCP port.
             *   not implemented yet... :)
             */
            if (verb) 
                fprintf(stderr, "TCP request is arrived.\n");

            fprintf(stderr, "this routine has not implemented yet.\n");
#if 0
            get_tcp_request();
#endif
            continue;
        }

    } /* end-of-while */
}


int
main(int argc, char **argv)
{
  int c;
  extern char *optarg;
  extern void usage(), finish_sentryd();
  extern void send_image();
#ifdef LOGTIME
  time_t logtime;
  struct tm *log_tm;
#endif /* LOGTIME */
#ifdef DEBUG
  debug = 1;
#endif /* DEBUG */
#ifdef VERBOSE
  verb = 1;
#endif /* VERBOSE */

#ifdef USE_SYSLOG
  openlog("sentryd", LOG_PID, LOG_DAEMON);
  syslog(LOG_NOTICE, "sentryd starts.");
#endif

  if (verb) {
#ifdef LOGTIME
    logtime = time(null);
    log_tm = localtime(&logtime);
    fprintf(stderr, "[%4d/%02d/%02d %02d:%02d:%02d] ",
        log_tm->tm_year+1900, log_tm->tm_mon+1, log_tm->tm_mday,
        log_tm->tm_hour, log_tm->tm_min, log_tm->tm_sec);
#endif
    fprintf(stderr, "sentryd starts.\n");
  }

  signal(SIGQUIT, finish_sentryd);
  signal(SIGTERM, finish_sentryd);
  signal(SIGINT, finish_sentryd);

#if 0
    if(video_sync_type==SENV_SYNC_BY_SIGNAL) {
        signal(SIGUSR2, send_image);
    }
#endif

  while ((c = getopt(argc, argv, "hDd:vjp:l:r:t:b:B")) != -1) {
    switch(c) {
    case 'h':
      usage();
      exit (1); /* not reached */
    case 'D':
      debug = 100;
      break;
    case 'd':
      debug = atoi(optarg);
      break;
    case 'v':
      verb = 1;
      break;
    case 'j':
      require_encode_type = SEN_F_JPEG;
      break;
    case 'p':
      data_port = atoi(optarg);
      ctrl_port = data_port + 1;
      break;
#ifndef IRIX  /* Linux */
    case 'l':
      xl_offset = atoi(optarg);
      if(xl_offset < 0) xl_offset = 0;
      break;
    case 'r':
      xr_offset = atoi(optarg);
      if(xr_offset < 0) xr_offset = 0;
      break;
    case 't':
      yt_offset = atoi(optarg);
      if(yt_offset < 0) yt_offset = 0;
      break;
    case 'b':
      yb_offset = atoi(optarg);
      if(yb_offset < 0) yb_offset = 0;
      break;
#else  /* IRIX */
    case 'l':
    case 'r':
      usage();
      exit(1);
#endif
    case 'B':
        batch_mode = 1 - batch_mode;
        break;
    }
  }
  
  /*
   * video setup
   */
  if (check_video(vserver)) {
    fprintf(stderr, "sentryd: sorry, video(%s) is not avaliable.\n",
        "localhost");
    fprintf(stdout, "500 video device is busy\r\n");
    exit (2);
  }

  init_video(vserver);

  if(batch_mode) {
      tcpsock = build_listen_port(ctrl_port, SOCK_STREAM);
      udpsock = build_listen_port(ctrl_port, SOCK_DGRAM);

      multicast_datasock = build_udp_port(0);

      if (tcpsock < 0 || udpsock < 0)
        exit (1);
  }
  
  (void)mainloop();
  exit (0);
}

void
usage() {
  printf("\
sentryd - daemon for sentry protocol\n\
    by sentry developer team, communications laboratory, NAIST, JAPAN\n\
    compiled at [%s %s]\n\
\n\
Sentryd send images form camera according to clients' request.\n\
In addition to, the program supports interactive mode.\n\
\n\
usage: sentryd [options]\n\
\n", __DATE__, __TIME__);

  printf("option:\n");
  printf("       -h       help\n");
  printf("       -D       debug mode\n");
  printf("       -d num   debug level\n");
  printf("       -v       verbose mode\n");
  printf("       -B       batch mode; othewise, interactive mode\n");
  printf("       -p port  data port(default=%d)\n", SENTRY_DATA_PORT);
  printf("       -j       force jpeg encoding (default=disable)\n");
  printf("       -t       top offset(default=0)\n");
  printf("       -b       bottom offset(default=0)\n");
#ifndef IRIX
  printf("       -l       left side offset(default=0)\n");
  printf("       -r       right side offset(default=0)\n");
#endif
  printf("\n");
  printf("example:\n");
  printf("  %% sentryd -B\n");
  printf("  %% sentryd\n");
  printf("\n");

  exit (1);
}
