/*
 * sfview.c - image file viewer for sentry protocol
 *
 * $Id$
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <signal.h>
#include <syslog.h>
#ifdef TIME_WITH_SYS_TIME
 #include <sys/time.h>
 #include <time.h>
#else
 #ifdef HAVE_SYS_TIME_H
  #include <sys/time.h>
 #else
  #include <time.h>
 #endif
#endif
#include <errno.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xproto.h>
#include <X11/cursorfont.h>

#include <X11/Xmu/WinUtil.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/Scrollbar.h>

/* private TrueColor Widget */
#include "Lww.h"

#include "alist.h"
#include "sentry.h"
#include "img.h"
#include "fpair.h"
#include "pool.h"
#include "client.h"

#if 0
#include "rtr.h"
subrecord_t *recvlist=NULL;
#endif



extern  void request_call();

extern u_char client_type; 
extern int server_zoom;
extern int server_size;
extern int server_color;
extern int server_format; 
extern int server_rate;

static int show_help=0;


#define HOUR    (3600)
#define DAY     (86400)
#define WEEK    (DAY*7)

#define MAX_N_PATH  (5)

int npath=0;
apair_t path_alist[MAX_N_PATH];
int scan_range=WEEK;

typedef struct {
    String server_name;
    String server_port;
    String data_port;
    String req_rate;
    String req_type;
    String req_size;
    String req_zoom;
    String req_color;
    String req_format;
    String hivision_flag;
    String voffset;
    /* */
    String shm_key;
    /* */
    String compare_func;
    /* */
    String path;
    String scan_range;
    /* */
    String debug_level;
    String show_help;
} sentry_resource_t;

sentry_resource_t sentry_resource;

#define Offset(field) (XtOffsetOf(sentry_resource_t, field))
static XtResource resources[] = {
  {"servername", "ServerName",  XtRString,  sizeof(String),
                    Offset(server_name), XtRString,  NULL},
  {"serverport", "ServerPort",  XtRString,  sizeof(String),
                    Offset(server_port), XtRString,  NULL},
  {"dataport",   "DataPort",  XtRString,  sizeof(String),
                    Offset(data_port), XtRString,  NULL},
  {"reqsize",   "ReqSize",  XtRString,  sizeof(String),
                    Offset(req_size), XtRString,  NULL},
  {"reqrate",   "ReqRate",  XtRString,  sizeof(String),
                    Offset(req_rate), XtRString,  NULL},
  {"reqformat", "ReqFormat",  XtRString,  sizeof(String),
                    Offset(req_format), XtRString,  NULL},
  {"reqtype",   "ReqType",  XtRString,  sizeof(String),
                    Offset(req_type), XtRString,  NULL},

  {"shmkey",    "ShmKey",  XtRString,  sizeof(String),
                    Offset(shm_key), XtRString,  NULL},

  {"comp",      "comp",  XtRString,  sizeof(String),
                    Offset(compare_func), XtRString,  NULL},

  {"path",      "path",  XtRString,  sizeof(String),
                    Offset(path), XtRString,  NULL},
  {"range",     "range",  XtRString,  sizeof(String),
                    Offset(scan_range), XtRString,  NULL},

  {"debug",     "Debug",  XtRString,  sizeof(String),
                    Offset(debug_level), XtRString,  NULL},
  {"help",      "Help",  XtRString,  sizeof(String),
                    Offset(show_help), XtRString,  NULL},
};




String fallback_resources[] = {
    "*reqsize:      160x120",
    "*reqrate:      200",
    "*reqformat:    JPEG",
    "*servername:   localhost",
    "*serverport:   7001",
    "*dataport:     7000",

    "*shmkey:       -1",

    "*comp:         none",

    "*debug:        0",
    
    "*path:         /var/tmp/sen_bank",
    "*range:        28800",

    "*help:         0",

    "*foreground:   black",
    "*background:   white",

    "*font:         a14",

    NULL};

static XrmOptionDescRec options[] = {
{"-fg",         "*Foreground",  XrmoptionSepArg,    NULL},
{"-bg",         "*Background",  XrmoptionSepArg,    NULL},
{"-servername", "*servername",  XrmoptionSepArg,    NULL},
{"-serverport", "*serverport",  XrmoptionSepArg,    NULL},
{"-dataport",   "*dataport",    XrmoptionSepArg,    NULL},

{"-shmkey",     "*shmkey",      XrmoptionSepArg,    NULL},
{"-debug",      "*debug",       XrmoptionSepArg,    NULL},
{"-rate",       "*reqrate",     XrmoptionSepArg,    NULL},
{"-format",     "*reqformat",   XrmoptionSepArg,    NULL},
{"-size",       "*reqsize",     XrmoptionSepArg,    NULL},

{"-comp",       "*comp",        XrmoptionSepArg,    NULL},
{"-path",       "*path",        XrmoptionSepArg,    NULL},
{"-range",      "*range",        XrmoptionSepArg,    NULL},

{"-help",       "*help",        XrmoptionNoArg,     "1"},

};

Widget   toplevel;
XtAppContext app_context;
XtInputId   iid, ciid;

XtActionsRec main_action_list[] = {
    NULL,
};


Visual   *xp_slide_visual,   *xp_foil_visual;
int      xp_slide_depth,     xp_foil_depth;
Colormap xp_slide_cmap,      xp_foil_cmap;
int      xp_use_special_cmap;

String fvisual, svisual;



Widget pwin_base, pwin_label;
Widget pwin;
Widget pcan_base, pcan_label;
Widget pcan_time;
Widget pcan;
Widget pban_base, pban_label;
Widget pban_dn, pban_d0, pban_d1, pban_d2, pban_d3, pban_d4, pban_d5;
Widget pban;
Widget ptscr;
Widget pshell;
Widget pbase;
Widget plabel;

Widget scbase;
Widget sc_quit;
Widget sc_begin, sc_rev, sc_play, sc_ff, sc_end;
Widget sc_skip;
int frame_skip_flag = 0;

XFontStruct *time_font_info;

GC pwin_gc;

Pixel pwin_black, pwin_red, pwin_grayC;
Pixel pcan_black, pcan_white, pcan_blue, pcan_red, pcan_gray9, pcan_grayC;
Pixel pban_black, pban_white, pban_blue, pban_red, pban_gray9, pban_grayC;
Pixel pban_l0, pban_l1, pban_l2, pban_l3, pban_l4, pban_l5, pban_lerr;


typedef struct {
    XtIntervalId    id;
    time_t          interval;
} xp_periodical_rec;

xp_periodical_rec clock_timer;



int
mk_timestr(char *dst, char *pr, time_t *rt)
{
    char tmp2[BUFSIZ];
    struct tm *tmbuf;
    int len2, len1;

    dst[0] = '\0';
    tmbuf = localtime(rt);
    strftime(tmp2, BUFSIZ, "%Y-%m-%d %H:%M:%S %a", tmbuf);

    if(pr && *pr) {
        len1 = strlen(pr);
        len2 = strlen(tmp2);
        if(len1+len2>=BUFSIZ-2) {
            len1 = BUFSIZ - len2 - 1;
            strncpy(dst, pr, len1);
            dst[len1] = '\0';
            strcat(dst, " ");
        }
        else {
            strcpy(dst, pr);
        }
        strcat(dst, " ");
    }
    strcat(dst, tmp2);

    return 0;
}

int
mk_timestr_short(char *dst, char *pr, time_t *rt)
{
    char tmp2[BUFSIZ];
    struct tm *tmbuf;
    int len2, len1;

    dst[0] = '\0';
    tmbuf = localtime(rt);
    strftime(tmp2, BUFSIZ, "%m-%d %H:%M:%S", tmbuf);

    if(pr && *pr) {
        len1 = strlen(pr);
        len2 = strlen(tmp2);
        if(len1+len2>=BUFSIZ-2) {
            len1 = BUFSIZ - len2 - 1;
            strncpy(dst, pr, len1);
            dst[len1] = '\0';
            strcat(dst, " ");
        }
        else {
            strcpy(dst, pr);
        }
        strcat(dst, " ");
    }
    strcat(dst, tmp2);

    return 0;
}


print_time(char *pr, time_t *rt)
{
    char tmp[BUFSIZ];
    int chk;


    chk = mk_timestr(tmp, pr, rt);
    printf("%s\n", tmp);
}


time_t today_start(time_t ref)
{
    struct tm *tmbuf;
    time_t ot;
    
    ot = ref;
    tmbuf = localtime(&ot);
    ot -= (tmbuf->tm_hour*HOUR + tmbuf->tm_min*60 + tmbuf->tm_sec);

    return ot;
}

time_t today_end(time_t ref)
{
    struct tm *tmbuf;
    time_t ot;
    
    ot = ref;
    tmbuf = localtime(&ot);
    ot -= (tmbuf->tm_hour*HOUR + tmbuf->tm_min*60 + tmbuf->tm_sec);
    ot += DAY - 1;

    return ot;
}

time_t nday_start(time_t ref, int n)
{
    struct tm *tmbuf;
    time_t ot;
    
    ot = ref;
    tmbuf = localtime(&ot);
    ot -= (tmbuf->tm_hour*HOUR + tmbuf->tm_min*60 + tmbuf->tm_sec);
    ot += DAY*n;

#if 0
    print_time("ref ", &ref);
    print_time("ot  ", &ot);
#endif

    return ot;
}

time_t nday_end(time_t ref, int n)
{
    struct tm *tmbuf;
    time_t ot;
    
    ot = ref;
    tmbuf = localtime(&ot);
    ot -= (tmbuf->tm_hour*HOUR + tmbuf->tm_min*60 + tmbuf->tm_sec);
    ot += DAY*n + DAY-1;

    return ot;
}

/*
DrawCrossRectangle(XtDisplay(pwin), XtWindow(pwin), pwin_gc,
    zx + x, y, 160, 120);
*/
void
DrawCrossRectangle(Display *dpy, Drawable dr, GC gc,
    int x, int y, int w, int h)
{
    XDrawLine(dpy, dr, gc, x+0, y+0, x+w-1, y+0);
    XDrawLine(dpy, dr, gc, x+0, y+0, x+w-1, y+h-1);
    XDrawLine(dpy, dr, gc, x+0, y+0, x+0, y+h-1);

    XDrawLine(dpy, dr, gc, x+w-1, y+0, x+w-1, y+h-1);
    XDrawLine(dpy, dr, gc, x+w-1, y+0, x+0, y+h-1);
    XDrawLine(dpy, dr, gc, x+0, y+h-1, x+w, y+h-1);
}



#define D_LT    (1)
#define D_CT    (2)
#define D_RT    (3)
#define D_LC    (4)
#define D_C     (5)
#define D_RC    (6)
#define D_LB    (7)
#define D_CB    (8)
#define D_RB    (9)

DrawStr(Display *dpy, Drawable dr, GC gc, XFontStruct *finfo,
    int jtype, int x, int y, char *str)
{
    int slen;
    int gx, gy;
    int tw, th;

#if 0
printf("DrawStr: finfo %p, type %d, x,y %d,%d str '%s'\n",
    finfo, jtype, x, y, str);
#endif

    slen = strlen(str);
    tw = XTextWidth(finfo, str, slen);
    th = finfo->ascent + finfo->descent;
    
    if(jtype==D_LT || jtype==D_LC || jtype==D_LB) {
        gx = x;
    }
    else
    if(jtype==D_CT || jtype==D_C || jtype==D_CB) {
        gx = x-tw/2;
    }
    else
    if(jtype==D_RT || jtype==D_RC || jtype==D_RB) {
        gx = x-tw;
    }

    if(jtype==D_LT || jtype==D_CT || jtype==D_RT) {
        gy = y + th;
    }
    else
    if(jtype==D_LC || jtype==D_C || jtype==D_RC) {
        gy = y + th/2;
    }
    else
    if(jtype==D_LB || jtype==D_CB || jtype==D_RB) {
        gy = y;
    }

#if 0
printf(" gx,gy %d,%d ; tw,th %d,%d\n", gx, gy, tw, th);
#endif

    XSetFont(dpy, gc, finfo->fid);
    XDrawString(dpy, dr, gc, gx, gy, str, slen);
}




void
put_image(Display *display, Visual *visual, int depth, GC gc, Window window,
    int width, int height, char *data);
void
put_image_xy(Display *display, Visual *visual, int depth, GC gc, Window window,
    int x, int y, int width, int height, char *data);


SenDrawImage3(Widget w, int x, int y,
    sentry_imgrec *inimg, sentry_imgrec *outimg)
{
    int wid, ht;
    int xwid, xht;
    unsigned char *raw_data;
    unsigned char *data_rgb=NULL, *data_bgrx=NULL;
    int raw_len;
    int raw_format, raw_color;
    int chk;
    int cur_format=SEN_F_NONE;
    int cur_corder=SEN_C_NONE;
    int success=0;
    int outlen=-1;

    if(inimg==NULL)
        return;

#if 0
    if(debug>=10) {
        printf("SenDrawImage: output format '%s', color order '%s'\n",
            assoc_v(format_alist, output_format),   
            assoc_v(color_alist, output_corder));
    }
#endif

    raw_len     = inimg->length;
    raw_data    = (unsigned char*)inimg->body;
    raw_format  = inimg->format;
    raw_color   = inimg->corder;
    wid = inimg->width;
    ht  = inimg->height;

    data_rgb = (unsigned char*)malloc(wid*ht*3);
    memset(data_rgb, 0xff, wid*ht*3);
    data_bgrx = (unsigned char*)malloc(wid*ht*4);
    memset(data_bgrx, 0xff, wid*ht*4);

#if 0
    if(debug>=10) {
        printf("SenDrawImage: %p %d '%s''%s' %dx%d, 3B/p=%d 4B/p=%d\n",
            raw_data, raw_len, assoc_v(format_alist, raw_format),
            assoc_v(color_alist, raw_color), wid, ht, wid*ht*3, wid*ht*4);
    }
#endif
    if(debug>=10) {
        printf(
"draw: got format %s, color %s, ptr %p, size %dx%d, len %d\n\
      out format %s, color %s\n",
        assoc_v(format_alist, raw_format), assoc_v(color_alist, raw_color), 
        raw_data, wid, ht, raw_len,
        assoc_v(format_alist, output_format),
        assoc_v(color_alist, output_corder));
    }

    switch(raw_format) {
    case SEN_F_JPEG:
        chk = decode_jpeg(raw_data, raw_len, data_rgb, &xwid, &xht);
        if(xwid!=wid && xht!=ht) {
            fprintf(stderr,
                "SenDrawImage: wid %d != xwid %d, ht %d != xht %d\n",
                wid, xwid, ht, xht);
        }
        cur_format = SEN_F_RGB;
        cur_corder = SEN_C_RGB;
        if(debug>=20){
            printf("  convert to '%s' (wid %d, ht %d\n",
                assoc_v(color_alist, cur_corder), xwid, xht);
        }
        break;

    case SEN_F_NONE: /* XXX */
        memcpy(data_rgb, raw_data, raw_len);
        break;

    case SEN_F_RGB: /* XXX */
        memcpy(data_rgb, raw_data, raw_len);
        cur_format = raw_format; /* SEN_F_RGB */
        cur_corder = raw_color;
        break;
    }

    if(debug>=10) {
        printf(
"draw: cur format %s, color %s\n\
      out format %s, color %s\n",
        assoc_v(format_alist, cur_format), assoc_v(color_alist, cur_corder),
        assoc_v(format_alist, output_format),
        assoc_v(color_alist, output_corder)
        );
    }
    if(cur_format != output_format) {
        if(debug>=1) {
            fprintf(stderr, "image format miss-match cur=%s out=%s\n",
                assoc_v(format_alist, cur_format),
                assoc_v(format_alist, output_format));
        }
        goto put_fail;
    }

    if(output_format==SEN_F_RGB) {
        if(output_corder==SEN_C_XBGR) {
            convert_RGB2XBGR(data_bgrx, data_rgb, wid * ht);

            outlen = wid*ht*4;
            put_image_xy(XtDisplay(w), xp_foil_visual,
                xp_foil_depth, pwin_gc, XtWindow(w),
                x, y, wid, ht, data_bgrx);
            success = 1;

            if(outimg->body_len>outlen) {
                copyattr_img(outimg, inimg);
                outimg->format = output_format;
                outimg->corder = output_corder;
                outimg->length = outlen;
                memcpy(outimg->body, data_bgrx, outlen);
            }
        }
        else
        if(output_corder==SEN_C_XRGB) {
/*
            convert_RGB2XRGB(data_bgrx, data_rgb, wid * ht);
*/
            convert_RGB2BGRX(data_bgrx, data_rgb, wid * ht);

            outlen = wid*ht*4;
            put_image_xy(XtDisplay(w), xp_foil_visual,
                xp_foil_depth, pwin_gc, XtWindow(w), 
                x, y, wid, ht, data_bgrx);
            success = 1;

            if(outimg->body_len>outlen) {
                copyattr_img(outimg, inimg);
                outimg->format = output_format;
                outimg->corder = output_corder;
                outimg->length = outlen;
                memcpy(outimg->body, data_bgrx, outlen);
            }
        }
        else
        if(output_corder==SEN_C_RGB) {
            outlen = wid*ht*3;
            put_image_xy(XtDisplay(w), xp_foil_visual,
                xp_foil_depth, pwin_gc, XtWindow(w),
                x, y, wid, ht, data_rgb);
            success = 1;

            if(outimg->body_len>outlen) {
                copyattr_img(outimg, inimg);
                outimg->format = output_format;
                outimg->corder = output_corder;
                outimg->length = outlen;
                memcpy(outimg->body, data_rgb, outlen);
            }
        }
    }

put_fail:
    if(!success) {
        if(debug>=1) {
            fprintf(stderr, "fail some reason, format cur=%s out=%s\n",
                assoc_v(format_alist, cur_format),
                assoc_v(format_alist, output_format));
        }
    }
    else {
#if 0
        copyattr_img(outimg, inimg);
        memcpy(outimg->body, inimg->body, outlen);
#endif
    }

out:
    if(data_rgb!=NULL)
        free(data_rgb);
    if(data_bgrx!=NULL) 
        free(data_bgrx);
}

SenDrawImage2(Widget w, int x, int y,
    sentry_imgrec *inimg, sentry_imgrec *outimg)
{
    SenDrawImage3(w, 0, 0, inimg, outimg);
}


static int call_count=0, lost_count=0, got_count=0, c_got_count=0;
static int refresh_flag = 1;


RefreshWins(sentry_imgrec *bank)
{
    Dimension tw, th, tx, ty;
    Dimension bw, bh, bx, by;
    Dimension nw, nh;

    XtVaGetValues(plabel,
         XtNx, &tx, XtNy, &ty, XtNwidth, &tw, XtNheight, &th, NULL);
    XtVaGetValues(pwin,
         XtNx, &bx, XtNy, &by, XtNwidth, &bw, XtNheight, &bh, NULL);
    printf("1 %ux%u+%u+%u %ux%u+%u+%u\n",
        tw, th, tx, ty, bw, bh, bx, by);

    XtVaSetValues(plabel,
        XtNwidth, bank->width,
        XtNlabel, server_host,
        NULL);

    XtVaGetValues(plabel, XtNheight, &th, NULL);
    XtVaGetValues(pwin, XtNheight, &bh, NULL);
    XtVaGetValues(plabel, XtNy, &ty, NULL);
    XtVaGetValues(pwin, XtNy, &by, NULL);
    printf("2 th %d, bh %d, ty %d, by %d\n", th, bh, ty, by);


    XtVaSetValues(pwin,
        XtNwidth, bank->width,
        XtNheight, bank->height,
        NULL);

    nw = bank->width+10;
/*
    nh = bank->height+th+10;
*/
    nh = MAX(by,24) + bank->height+10;

    printf("nw %d, nh %d\n", nw, nh);

    XtVaSetValues(pshell,
        XtNwidth, nw, XtNheight, nh,
/*
*/
        NULL);
    XtVaSetValues(pbase,
/*
        XtNwidth, bank->width,
        XtNheight, bank->height,
*/
/*
        XtNwidth, nw, XtNheight, nh,
*/
        NULL);

    XtVaGetValues(plabel, XtNheight, &th, NULL);
    XtVaGetValues(pwin, XtNheight, &bh, NULL);
    printf("3 th %d, bh %d\n", th, bh);

    fflush(stdout);
    fflush(stderr);

    refresh_flag = 0;
}












time_t begin_t, end_t;
int begin_i, end_i;
time_t wbegin_t=-1, wend_t=-1;
int wbegin_i, wend_i;
double tscale=1.0;
time_t focus_t=-1;


static sentry_imgrec zxo;


int
remove_dupimage()
{
    fpair_t **ar;
    int pc, qc;
    fpair_t *p;
    fpair_t **q;
    fpair_t *r;
    int i;
    time_t limit;


    if(img_pool_count<=0) 
        return 0;
    
    if(debug>=10)
        printf("remove_dupimage:\n");

    ar = (fpair_t**)malloc(sizeof(fpair_t*)*img_pool_count);

    if(debug>=10)
        printf("alloc %p as array[%d] of fpair_t\n",
            ar, img_pool_count);

#if 0
    for(i=0;i<img_pool_count;i++) {
        printf("i %4d: %p %d %d %p :: %p %d , %7d\n",
            i, &fpairs[i], fpairs[i].etime.tv_sec,
                fpairs[i].chan, fpairs[i].imgpos,
            &img_pool[i], img_pool[i].recv_sec, img_pool[i].length);
    }
    fflush(stdout);
#endif

    limit = time(NULL) - WEEK;

    pc = 0;
    qc = 0;
    p = fpairs;
    q = ar;
    r = NULL;
    do {
#if 0
        printf("p %4d: %p %d vs q %4d: %p %d\n",
            pc, p, p->etime.tv_sec,
            qc, *q, (*q)->etime.tv_sec);
#endif

#if 0
        if(r)
            printf("p %4d: %p %d : r %p %d : %7d\n",
                pc, p, p->etime.tv_sec, r, r->etime.tv_sec, SUB_TIME(p->etime, r->etime));
        else
            printf("p %4d: %p %d\n", pc, p, p->etime.tv_sec);
        fflush(stdout);
#endif

        *q = p;

#if 0
        printf("q %4d: %p %p %d\n", qc, q, *q, (*q)->etime.tv_sec);
        fflush(stdout);
#endif

        if(r && r->etime.tv_sec<limit) {
            /* nothing */
        }
        else
        if(r && (r!=p) && ((r->etime.tv_sec == p->etime.tv_sec) 
                    && (r->etime.tv_usec == p->etime.tv_usec)) ) {
            /* nothing */
        }
        else {
            qc++;
            q++;
        }
        
        r = p;
        pc++;
        p++;
    } while(p && pc<img_pool_count);

#if 0
    for(i=0;i<img_pool_count;i++) {
        printf("fpair %4d/%d: %p %d %d %p\n",
            i, pc, &fpairs[i], fpairs[i].etime.tv_sec,
            fpairs[i].chan, fpairs[i].imgpos);
    }
    for(i=0;i<qc;i++) {
        printf("ar    %4d/%d: %p %d %d %p\n",
            i, qc, ar[i], ar[i]->etime.tv_sec,
            ar[i]->chan, ar[i]->imgpos);
    }
#endif

    for(i=0;i<qc;i++) {
        if(i<qc) {
            fpairs[i] = *ar[i];
        }
        else {
            free((fpairs[i].imgpos)->body);
            (fpairs[i].imgpos)->body_len = 0;
            (fpairs[i].imgpos)->recv_sec = 0;

            img_heap_use -= (fpairs[i].imgpos)->length;
        }
    }
    img_pool_count = qc;

#if 0
    for(i=0;i<img_pool_count;i++) {
        printf("fpair %4d/%d: %p %d %d %p\n",
            i, pc, &fpairs[i], fpairs[i].etime.tv_sec,
            fpairs[i].chan, fpairs[i].imgpos);
    }
#endif


    free(ar);

    if(debug>=10) {
        printf("\t%d\n", img_pool_count);
        fflush(stdout);
    }

    return 0;
}

void
Update_pwin()
{
    static sentry_imgrec *xo=NULL;
    int pos;
    int max_i, min_i;
    volatile int limg, rimg;
    volatile int limgp, rimgp;
    int i, t;
    int x, y;
    int zx;
    Dimension vwid, vht;
    int ypit;
    struct tm *tmbuf;
    char ts1[BUFSIZ], ts2[BUFSIZ];
    char tbuf[BUFSIZ];
    sentry_imgrec *cimg;

    
    if(xo==NULL) {
        xo = &zxo;
        xo->body = (unsigned char*)malloc(128*1024);
        xo->body_len = 128*1024;
    }

    XtVaGetValues(pwin, XtNwidth, &vwid, XtNheight, &vht, NULL);

    XSetForeground(XtDisplay(pwin), pwin_gc, pwin_grayC);
    XFillRectangle(XtDisplay(pwin), XtWindow(pwin), pwin_gc,
        0,  0, vwid, vht);

    ypit = vht/npath;

    max_i = end_i;
    min_i = -1;

    limg = rimg = 0;
    limgp = rimgp = -1;

    for(i=0;i<img_pool_count;i++) {
        t = fpairs[i].etime.tv_sec;
        if(t>=wbegin_t && t<=wend_t) {
#if 0
printf("  %4d: t %d\n", i, t);
#endif
            if(t<focus_t) {
                limg++;
                limgp = i;
            }
            if(t>=focus_t) {
                rimg++;
                if(rimgp<0) {
                    rimgp = i;
                }
            }
        }
    }

    if(debug>=10) {
        fflush(stderr);
        printf("images left %d (..%d), right %d (%d..)\n",
            limg, limgp, rimg, rimgp);
        fflush(stdout);
    }


    if(limg==0) {
        zx = 0;
    }
    else
    if(rimg==0) {
        zx = vwid;
    }
    else {
        zx = ((int)vwid)/2;
    }

    XSetForeground(XtDisplay(pwin), pwin_gc, pwin_red);
    XFillRectangle(XtDisplay(pwin), XtWindow(pwin), pwin_gc,
        zx-1, 0, 2, vht);

    XSetForeground(XtDisplay(pwin), pwin_gc, pwin_black);
    if(time_font_info) {
        XSetFont(XtDisplay(pcan), pwin_gc, time_font_info->fid);
    }

    /* left */
    if(limgp>=0)
        x = 0;
        for(i=limgp;i>=0;i--) {
            y = fpairs[i].chan*ypit;
            cimg = fpairs[i].imgpos;

            if(cimg==NULL) {
#if 0
printf("--left  %4d; ---\n", i);
#endif
/*
                XFillRectangle(XtDisplay(pwin), XtWindow(pwin), pwin_gc,
                    zx + x, y, 160, 120);
*/
                DrawCrossRectangle(XtDisplay(pwin), XtWindow(pwin), pwin_gc,
                    zx + x, y, 160, 120);
                continue;
            }


            x -= cimg->width;
#if 0
            printf(" left  i %4d t %d: x,y %4d(%4d) %3d :pos %d : %p %dx%d\n",
                i, fpairs[i].etime.tv_sec, zx+x, x, y, pos, 
                cimg, cimg->width, cimg->height);
#endif

            if(cimg->body && cimg->body_len>0) {
                SenDrawImage3(pwin, zx + x, y, cimg, xo);
            }
            else {
                XFillRectangle(XtDisplay(pwin), XtWindow(pwin), pwin_gc,
                    zx + x, y, cimg->width, cimg->height);
            }

            tmbuf = localtime(&fpairs[i].etime.tv_sec);
            strftime(ts1, BUFSIZ, "[%T", tmbuf);
            sprintf(ts2, ".%03d]", fpairs[i].etime.tv_usec/1000);
            strcat(ts1, ts2);
            DrawStr(XtDisplay(pwin), XtWindow(pwin), pwin_gc, time_font_info,
                D_LT, zx+x, y + cimg->height, ts1);

            if(x>0) {
                break;
            }
            if(zx+x<0) {
                break;
            }

        }

    /* right */
    if(rimgp>=0)
        x = 0;
        for(i=rimgp;i<img_pool_count;i++) {
            y = fpairs[i].chan*ypit;
            cimg = fpairs[i].imgpos;

            if(cimg==NULL) {
#if 0
printf("--right %4d; ---\n", i);
#endif
/*
                XFillRectangle(XtDisplay(pwin), XtWindow(pwin), pwin_gc,
                    zx + x, y, 160, 120);
*/
                DrawCrossRectangle(XtDisplay(pwin), XtWindow(pwin), pwin_gc,
                    zx + x, y, 160, 120);
                continue;
            }


#if 0 
            printf(" right i %4d t %d: x,y %4d(%4d) %3d :pos %d : %p %dx%d\n",
                i, fpairs[i].etime.tv_sec, zx+x, x, y, pos, 
                cimg, cimg->width, cimg->height);
#endif


            if(cimg->body && cimg->body_len>0) {
                SenDrawImage3(pwin, zx + x, y, cimg, xo);
            }
            else {
                XFillRectangle(XtDisplay(pwin), XtWindow(pwin), pwin_gc,
                    zx + x, y, cimg->width, cimg->height);
            }

            tmbuf = localtime(&fpairs[i].etime.tv_sec);
            strftime(ts1, BUFSIZ, "[%T", tmbuf);
            sprintf(ts2, ".%03d]", fpairs[i].etime.tv_usec/1000);
            strcat(ts1, ts2);
            DrawStr(XtDisplay(pwin), XtWindow(pwin), pwin_gc, time_font_info,
                D_LT, zx+x, y + cimg->height, ts1);

            if(zx+x>vwid) {
                break;
            }
            if(zx+x<0) {
                break;
            }

            x += cimg->width;
        }
}


int pointsize = 4;

time_t thre;

Verify_twindow()
{
    Dimension can_wid, can_ht;
    int np;
    time_t now;

    XtVaGetValues(pcan, XtNwidth, &can_wid, XtNheight, &can_ht, NULL);

    begin_i = 0;
    begin_t = fpairs[0].etime.tv_sec;
    end_i = img_pool_count-1;
    end_t = fpairs[img_pool_count-1].etime.tv_sec;

    np = can_wid / pointsize;

    now = time(NULL);
    thre = nday_start(now, -6);

    if(debug>=10) {
        printf("g  time %d(%d) - %d(%d); width %d ; act window width %d\n",
            begin_t, begin_i, end_t, end_i, end_t - begin_t, can_wid);
        printf("focus_t %d [np %d]\n", focus_t, np);
        printf("wb time %d(%d) - %d(%d); width %d ; act window width %d\n",
            wbegin_t, wbegin_i, wend_t, wend_i, wend_t - wbegin_t, can_wid);

        print_time("now    ", &now);
        print_time("thre   ", &thre);
        print_time("begin  ", &begin_t);
        print_time("end    ", &end_t);
        print_time("wbegin ", &wbegin_t);
        print_time("wend   ", &wend_t);
    }


    if(wbegin_t<0||wend_t<0) {
        if(focus_t<=begin_t) {
            wbegin_t = begin_t;
            wend_t   = begin_t+np;
            focus_t  = begin_t;
        }
        else
        if(focus_t>=end_t) {
            wend_t   = focus_t;
            wbegin_t = wend_t - np;
            focus_t  = end_t;
        }
    }
    else {
        if(focus_t<=begin_t) {
            wbegin_t = begin_t;
            wend_t   = begin_t+np;
        }
        else
        if(focus_t>=end_t) {
            wend_t   = focus_t;
            wbegin_t = wend_t - np;
            focus_t  = end_t;
        }
        else {
            if(focus_t<wbegin_t) {
                wbegin_t = focus_t;
                wend_t   = wbegin_t + np;
            }
            else
            if(focus_t>wend_t) {
                wend_t   = focus_t;
                wbegin_t = wend_t - np;
            }
        }
    }

    if(debug>=10) {
        printf("wa time %d(%d) - %d(%d); width %d ; act window width %d\n",
            wbegin_t, wbegin_i, wend_t, wend_i, wend_t - wbegin_t, can_wid);
        printf("focus_t %d\n", focus_t);
    }
}




Update_pcan()
{
    Dimension can_wid, can_ht;
    GC pcan_gc;
    int ch;
    int y;
    int t;
    int i;
    int ypit, ybase;
    int xbase, x;
    int tbase;
    int twid;
    Font cf;
    struct tm *tmbuf;
    char tbuf[BUFSIZ];
    float scr_pos;
    float scr_len;
    

    pcan_gc = XtGetGC(pcan, NULL, NULL);

    XtVaGetValues(pcan, XtNwidth, &can_wid, XtNheight, &can_ht, NULL);

    XSetForeground(XtDisplay(pcan), pcan_gc, pcan_white);
    XFillRectangle(XtDisplay(pcan), XtWindow(pcan), pcan_gc,
        0, 0, can_wid, can_ht);
    

    Verify_twindow();

    scr_pos = ((double)(focus_t - begin_t)) / (end_t - begin_t);
    scr_len = ((double)(wend_t - wbegin_t)) / (end_t - begin_t);
    
    if(debug>=100) 
        printf("pos %f len %f\n", scr_pos, scr_len);
/*
    XawScrollbarSetThumb(ptscr, scr_pos, scr_len);
*/
    XawScrollbarSetThumb(ptscr, scr_len, scr_pos);


    XSetForeground(XtDisplay(pcan), pcan_gc, pcan_black);

    if(time_font_info) {
        XSetFont(XtDisplay(pcan), pcan_gc, time_font_info->fid);
    }

#if 0
    tmbuf = localtime(&wbegin_t);
    strftime(tbuf, BUFSIZ, "start %y/%m/%d %T", tmbuf);
#endif
    mk_timestr_short(tbuf, "start", &wbegin_t);
    DrawStr(XtDisplay(pcan), XtWindow(pcan), pcan_gc, time_font_info,
        D_LT, 1, 1, tbuf);

#if 0
    tmbuf = localtime(&wend_t);
    strftime(tbuf, BUFSIZ, "end %y/%m/%d %T", tmbuf);
#endif
    mk_timestr_short(tbuf, "end", &wend_t);
    DrawStr(XtDisplay(pcan), XtWindow(pcan), pcan_gc, time_font_info,
        D_RB, can_wid-1, can_ht-1, tbuf);


    /* y (channel) axis */

    XSetForeground(XtDisplay(pcan), pcan_gc, pcan_gray9);

    ypit = can_ht/(npath+2);
    ybase = ypit;

    if(debug>=500)
        printf("ybase %d, ypit = %3d <= can_ht(%d) /( npath(%d)+1)\n",
            ybase, ypit, can_ht, npath);

    y = ybase;
    y += ypit;
    for(ch=0;ch<npath;ch++) {
        if(debug>=500)
            printf(" y %3d\n", y);
        XDrawLine(XtDisplay(pcan), XtWindow(pcan), pcan_gc, 0, y, can_wid, y);
        y += ypit;
    }

    /* x (time) axis */

    for(t=wbegin_t;t<=wend_t;t++) {
        if(t%60!=0)
            continue;

        x = (t-wbegin_t)*pointsize;
        if(debug>=500)
            printf(" * t %d, x %4d\n", t, x);

        if(t%HOUR==0) {
            XSetForeground(XtDisplay(pcan), pcan_gc, pcan_gray9);
        }
        else {
            XSetForeground(XtDisplay(pcan), pcan_gc, pcan_grayC);
        }
        XDrawLine(XtDisplay(pcan), XtWindow(pcan), pcan_gc,
            x, ybase+ypit/2, x, ybase+ypit*npath+ypit/2);
    }


    /* current time position*/

    XSetForeground(XtDisplay(pcan), pcan_gc, pcan_red);
    x = can_wid - (wend_t - focus_t) * pointsize;
    if(debug>=100) 
        printf("wid %d, x %d\n", can_wid, x);
#if 0
    XFillRectangle(XtDisplay(pcan), XtWindow(pcan), pcan_gc, x-1, 0, 2, can_ht);
#endif
    XFillRectangle(XtDisplay(pcan), XtWindow(pcan), pcan_gc,
        x-1, ybase+ypit/2, 2, ypit*npath);

#if 0
    tmbuf = localtime(&focus_t);
    strftime(tbuf, BUFSIZ, "%y/%m/%d %T", tmbuf);
#endif
    mk_timestr_short(tbuf, NULL, &focus_t);
    if(x-1<can_wid/2)
        DrawStr(XtDisplay(pcan), XtWindow(pcan), pcan_gc, time_font_info,
            D_LB, x-1, ypit+ypit/2, tbuf);
    else
        DrawStr(XtDisplay(pcan), XtWindow(pcan), pcan_gc, time_font_info,
            D_RB, x-1, ypit+ypit/2, tbuf);


    twid = wend_t - wbegin_t;
    if(twid*pointsize < can_wid) {
        xbase = can_wid - twid*pointsize;
    }
    else {
        xbase = 0;
    }
    if(debug>=100) { 
        printf("xbase %d (can_wid %d, twid %d)\n", xbase, can_wid, twid);
        printf("wbegin_t %d, wend_t %d, twid %d\n", wbegin_t, wend_t, twid);
    }


    /*
     * plot frame existance
     */

    XSetForeground(XtDisplay(pcan), pcan_gc, pcan_blue);

    for(i=0;i<img_pool_count;i++) {
        t = fpairs[i].etime.tv_sec;
        if(t>=wbegin_t && t<=wend_t) {
            x = xbase + (t-wbegin_t)*pointsize;
            XFillRectangle(XtDisplay(pcan), XtWindow(pcan), pcan_gc,
                x-1, ybase+ypit*(fpairs[i].chan+1)-1, pointsize-1, pointsize-1);
        }
    }

    
}

char *wdname[7] = {"Sun", "Mon", "Tue", "Web", "Thr", "Fri", "Sat"};
int dh_freq[7][24];
int dh_time[7][24];

calc_pban()
{
    int i;
    int d, h;
    time_t t;
    struct tm *tmbuf;
    char tbuf[BUFSIZ];
    int today_zd, today_zt;
    int td;
    time_t xxt, yyt;
    time_t gbegin, gend;

    if(debug>=100) {
        printf("calc_pban:\n");
        printf(" begin_t %d, end_t %d\n", begin_t, end_t);
    }

    for(d=0;d<7;d++) {
        for(h=0;h<24;h++) {
            dh_freq[d][h] = -1;
        }
    }

    tmbuf = localtime(&end_t);
#if 0
    today_zt = end_t - (tmbuf->tm_hour*HOUR+ tmbuf->tm_min*60 + tmbuf->tm_sec);
#endif
    today_zt = today_start(end_t);
    today_zd = tmbuf->tm_wday;

    if(debug>=100) 
        printf("today zt %d, zd %d\n", today_zt, today_zd);

    for(td=0;td<7;td++) {
        d = (today_zd-td+14)%7;
        for(h=0;h<24;h++) {
            dh_time[d][h] = today_zt - td*DAY+ h*HOUR;
        }
    }


    gbegin  = fpairs[0].etime.tv_sec;
    gbegin  = (gbegin/HOUR)*HOUR;
    gend    = fpairs[img_pool_count-1].etime.tv_sec;
    gend    = (gend/HOUR)*HOUR;

    xxt = (end_t/HOUR)*HOUR;
    if(debug>=100) {
        printf("gbegin %d\ngend   %d\nxxt    %d\n", gbegin, gend, xxt);
    }
    if(debug>=100) {
        print_time("gbegin", &gbegin);
        print_time("gend  ", &gend);
        print_time("xxt   ", &xxt);
    }

    if(debug>=100) {
        for(d=0;d<7;d++) {
            printf("%d:", d);
            for(h=0;h<24;h++) {
                printf(" %d%c",
                    dh_time[d][h],
                    (dh_time[d][h]>=gbegin && dh_time[d][h]<=gend) ? '+' : '-');
                if(dh_time[d][h]==xxt) {
                    printf("*");
                }
            }
            printf("\n");
        }
    }

    if(debug>=100) {
        for(d=0;d<7;d++) {
            printf("%d: ", d);
            for(h=0;h<24;h++) {
                printf("%d-%02d ",
                    (dh_time[d][h]-today_zt)/DAY,
                    ((dh_time[d][h]-today_zt)%DAY)/HOUR
                    );
            }
            printf("\n");
        }
    }
    if(debug>=100) {
        printf("%d: ", 0);
        for(h=0;h<24;h++) {
                printf("%4d ", h);
        }
        printf("\n");
    }

#if 0
    for(d=0;d<7;d++) {
        if(debug>=100) {
            printf("%d: ", d);
        }
        for(h=0;h<24;h++) {
            if(dh_time[d][h]>=gbegin && dh_time[d][h]<=gend) {
                if(debug>=100) {
                    printf("%4s ", "*");
                }
                dh_freq[d][h] = 0;
            }
            else {
                if(debug>=100) {
                    printf("%4s ", "-");
                }
            }
        }
        if(debug>=100) {
            printf("\n");
        }
    }
#endif

    for(i=0;i<img_pool_count;i++) {
        t = fpairs[i].etime.tv_sec;
        xxt = (t/HOUR)*HOUR;

        for(d=0;d<7;d++) {
            for(h=0;h<24;h++) {
                if(dh_time[d][h]==xxt) {
#if 0
printf("i %4d t %d, xxt %d, dh_time[%d][%d] %d\n",
    i, t, xxt, d, h, dh_time[d][h]);
#endif
                    if(dh_freq[d][h]<0) {
                        dh_freq[d][h] = 0;
                    }
                    dh_freq[d][h]++;
                    break;
                }
            }
        }
    }

    if(debug>=100) {
        for(d=0;d<7;d++) {
            printf("%d: ", d);
            for(h=0;h<24;h++) {
                printf("%4d ", dh_freq[d][h]);
            }
            printf("\n");
        }
    }

#if 0
    exit(3);
#endif

}

Update_pban()
{
    Dimension ban_wid, ban_ht;
    GC pban_gc;
    static int init_pban=0;
    int d, h;
    double hpit;
    double dpit;
    int x, y;
    char tmp[BUFSIZ];
    int nt;

    pban_gc = XtGetGC(pban, NULL, NULL);
    XtVaGetValues(pban, XtNwidth, &ban_wid, XtNheight, &ban_ht, NULL);

    if(init_pban==0) {
        calc_pban();
        init_pban=1;
    }

    dpit = ban_ht/8;
    hpit = ((double)ban_wid)/25;

    XSetForeground(XtDisplay(pban), pban_gc, pban_black);
    d = 0;
        for(h=0;h<24;h++) {
            sprintf(tmp, "%02d", h);
#if 0
            x = (int)(h+2)*hpit;
            y = dpit*d+dpit/2;
            DrawStr(XtDisplay(pban), XtWindow(pban), pban_gc, time_font_info,
                D_RB, x, y, tmp);
#endif
#if 1
            x = (int)(h+1)*hpit;
            y = dpit*d+dpit;
            DrawStr(XtDisplay(pban), XtWindow(pban), pban_gc, time_font_info,
                D_LB, x, y-1, tmp);
#endif
        }

    nt = (focus_t/HOUR)*HOUR;
    for(d=0;d<7;d++) {
/*
        y = dpit*d+dpit/2;
*/
        y = dpit*d+dpit;
        XSetForeground(XtDisplay(pban), pban_gc, pban_black);
        DrawStr(XtDisplay(pban), XtWindow(pban), pban_gc, time_font_info,
            D_RT, hpit-2, y, wdname[d]);
        for(h=0;h<24;h++) {
            x = (int)((h+1)*hpit);
#if 0
            if(dh_time[d][h]<begin_t || dh_time[d][h]>end_t) {
                    XSetForeground(XtDisplay(pban), pban_gc, pban_grayC);
            }
            else
#endif
            {

#if 0
                if(dh_freq[d][h]>=0) 
                    printf(" %d-%02d %4d\n", d, h, dh_freq[d][h]);
#endif

#if 1
                if(dh_freq[d][h]<0) {
                    XSetForeground(XtDisplay(pban), pban_gc, pban_grayC);
                }
                else 
#endif
                if(dh_freq[d][h]==0) {
                    XSetForeground(XtDisplay(pban), pban_gc, pban_l0);
                }
                else 
                if(dh_freq[d][h]<10) {
                    XSetForeground(XtDisplay(pban), pban_gc, pban_l1);
                }
                else
                if(dh_freq[d][h]<50) {
                    XSetForeground(XtDisplay(pban), pban_gc, pban_l2);
                }
                else 
                if(dh_freq[d][h]<100) {
                    XSetForeground(XtDisplay(pban), pban_gc, pban_l3);
                }
                else 
                if(dh_freq[d][h]<500) {
                    XSetForeground(XtDisplay(pban), pban_gc, pban_l4);
                }
                else
                if(dh_freq[d][h]>=500) {
                    XSetForeground(XtDisplay(pban), pban_gc, pban_l5);
                }
                else {
                    XSetForeground(XtDisplay(pban), pban_gc, pban_lerr);
                }
            }
            XFillRectangle(XtDisplay(pban), XtWindow(pban), pban_gc,
                x, y, (int)hpit, dpit);

            XSetForeground(XtDisplay(pban), pban_gc, pban_black);
            XDrawRectangle(XtDisplay(pban), XtWindow(pban), pban_gc,
                x, y, (int)hpit, dpit);
        }

        for(h=0;h<24;h++) {
            x = (int)((h+1)*hpit);
            if(nt==dh_time[d][h]) {
#if 0
printf("dt %d, dh_time[%d][%d] %d; x %d\n", nt, d, h, dh_time[d][h], x);
#endif
                x += (int)(((double)focus_t-dh_time[d][h])*hpit/HOUR);

#if 0
printf("  focus_t(%d) - dh_time[%d][%d](%d) = %d, hpit %f\n",
        focus_t, d, h, dh_time[d][h], focus_t-dh_time[d][h], hpit);
printf("  new x %d\n", x);
#endif
                XSetForeground(XtDisplay(pban), pban_gc, pban_red);
                XFillRectangle(XtDisplay(pban), XtWindow(pban), pban_gc,
                    x-1, y, 2, dpit);
            }
        }
    }
}


void
Scroll_canvas(Widget w, XtPointer client_data, XtPointer call_data)
{
    int position = (int)call_data;
    sentry_imgrec xo;

    if(debug>=100) 
        printf("Scroll_canvas apply %d to %d\n", position, focus_t);

    if(position>0 && focus_t>begin_t) {
        focus_t--;
    }
    if(position<=0 && focus_t<end_t) {
        focus_t++;
    }

    if(debug>=100) 
        printf("  scroll focus_t %d\n", focus_t);


    Update_pban();
    Update_pcan();
    Update_pwin();
}

void
Jump_canvas(Widget w, XtPointer client_data, XtPointer call_data)
{
    float percent = *(float*) call_data;

    focus_t = (int)(end_t-begin_t)*percent + begin_t;
    if(debug>=100) 
        printf("  jump   focus_t %d\n", focus_t);

    Update_pban();
    Update_pcan();
    Update_pwin();
}

void
ToggleSkip(Widget w, XtPointer client_data, XtPointer call_data)
{
    Pixmap fg, bg;

    frame_skip_flag = 1 - frame_skip_flag;
    if(debug>=100) 
        printf("skip? %d\n", frame_skip_flag); 

    XtVaGetValues(w, XtNforeground, &fg, XtNbackground, &bg, NULL);
    XtVaSetValues(w, XtNforeground, bg, XtNbackground, fg, NULL);

}


Skip(int dir) 
{
    int t;
    int gt;
    int i;
    int p;
    int o_focus_t;

    if(dir>0) {
        p = -1;
        for(i=0;i<img_pool_count;i++) {
            if(focus_t<fpairs[i].etime.tv_sec) {
                p = i;
                break;
            }
        }
        if(p>0) {
            focus_t = fpairs[p].etime.tv_sec;
        }
    }
    else
    if(dir<0) {
        p = -1;
        for(i=0;i<img_pool_count;i++) {
            if(focus_t>fpairs[i].etime.tv_sec) {
                p = i;
            }
        }
        if(p>0) {
            focus_t = fpairs[p].etime.tv_sec;
        }
    }

    if(debug>=100) 
        printf("dir %2d; focus_t %d -> %d\n", dir, o_focus_t, focus_t);
}


void
bar_hand(Widget w, XtPointer client_data, XtPointer call_data)
{
    int x;
    int y;

    x = (int)client_data;
    y = (int)call_data;

    if(debug>=100) 
        printf("x %d, y %d\n", x, y);
    if(frame_skip_flag) {
        switch(x) {
        case -2:    focus_t = begin_t;              break;
        case 2:     focus_t = end_t;                break;
        case -1:    Skip(-1);                       break;
        case 1:     Skip(1);                        break;
        }
    }
    else {
        switch(x) {
        case -2:    focus_t = begin_t;              break;
        case 2:     focus_t = end_t;                break;
        case -1:    if(focus_t>begin_t) focus_t--;  break;
        case 1:     if(focus_t<end_t)   focus_t++;  break;
        }
    }

    Verify_twindow();

    Update_pban();
    Update_pcan();
    Update_pwin();
}

void
PickTime(Widget w, XEvent *ev, String *s, Cardinal *c)
{
    int cx;

    if(debug>=100)
        printf("PickTime: type %d\n", ev->type);

    if(ev->xbutton.window != XtWindow(pcan)) {
        goto out;
    }
    if(debug>=100)
        printf("PickTime: x %d\n", ev->xbutton.x);

    focus_t = wbegin_t + ev->xbutton.x/pointsize;

    if(debug>=100) 
        printf("  button %d,%d => %d\n", ev->xbutton.x, ev->xbutton.y, focus_t);

    Verify_twindow();

    Update_pban();
    Update_pcan();
    Update_pwin();

out:
}

static int last_rollpos=-1;

void
RollTimeStart(Widget w, XEvent *ev, String *s, Cardinal *c)
{
    if(ev->xbutton.window != XtWindow(pcan)) {
        goto out;
    }
    last_rollpos = ev->xbutton.x;
out:
}

void
RollTimeMid(Widget w, XEvent *ev, String *s, Cardinal *c)
{
    int cx, dx;
    time_t dt;

/*
    printf("RollTimeMid\n");
*/
    cx = ev->xbutton.x;
    dx = last_rollpos - cx;
    dt = dx / pointsize;
    if(debug>=100)
        printf("roll cx %d, last %d -> dx %d, dt %d\n",
            cx, last_rollpos, dx, dt);

    wbegin_t += dt;
    wend_t += dt;

    if(focus_t<wbegin_t) {
        focus_t = wbegin_t;
    }
    else 
    if(focus_t>wend_t) {
        focus_t = wend_t;
    }

    last_rollpos = cx;

    Verify_twindow();

    Update_pban();
    Update_pcan();
out:
}


void
RollTimeEnd(Widget w, XEvent *ev, String *s, Cardinal *c)
{
    RollTimeMid(w, ev, s, c);

    Update_pwin();
}


void
PickCell(Widget w, XEvent *ev, String *s, Cardinal *c)
{
    int dx, dy;
    Dimension ban_wid, ban_ht;
    int dpit;
    double hpit;
    int d, h;
    double hp;
    time_t gt;
    struct tm *tmbuf;
    char tbuf[BUFSIZ];

    XtVaGetValues(pban, XtNwidth, &ban_wid, XtNheight, &ban_ht, NULL);
    
    dpit = ban_ht/8;
    hpit = ((double)ban_wid)/25;

    if(debug>=100) 
        printf("PickCell: type %d, dpit %d, hpit %.3f\n",
            ev->type, dpit, hpit);

    if(ev->xbutton.window != XtWindow(pban)) {
        goto out;
    }
/*
    d = (ev->xbutton.y - dpit/2)/dpit;
*/
    d = (ev->xbutton.y - dpit)/dpit;
    h = (ev->xbutton.x - hpit)/hpit;
    hp = ((double)ev->xbutton.x - hpit)/(hpit*24)*DAY;

    if(debug>=100) 
        printf("  button %d,%d => %d, %d\n",
            ev->xbutton.x, ev->xbutton.y,
            d, h);

    gt = dh_time[d][0] + hp;
    if(debug>=100) 
        printf("  d %d + hp %.2f = %d\n", dh_time[d][0], hp, gt);

    if(debug>=100) {
        tmbuf = localtime(&gt);
        strftime(tbuf, BUFSIZ, "%y/%m/%d %T", tmbuf);
        printf("gt %d '%s'\n", gt, tbuf);
    }

    if(gt<begin_t || gt>end_t) {
        printf("%d is out of range %d..%d\a\a\a\n", gt, begin_t, end_t);
        goto out;
    }

    focus_t = gt;

    Verify_twindow();

    Update_pban();
    Update_pcan();
    Update_pwin();

out:
}

void
ScanCell(Widget w, XEvent *ev, String *s, Cardinal *c)
{
    int dx, dy;
    Dimension ban_wid, ban_ht;
    int dpit;
    double hpit;
    int d, h;
    double hp;
    time_t gt;
    time_t gh;
    struct tm *tmbuf;
    char tbuf[BUFSIZ];
    int i, dosort;
    time_t sbegin, send;

    XtVaGetValues(pban, XtNwidth, &ban_wid, XtNheight, &ban_ht, NULL);
    
    dpit = ban_ht/8;
    hpit = ((double)ban_wid)/25;

    if(debug>=100) 
        printf("ScanCell: type %d, dpit %d, hpit %.3f\n",
            ev->type, dpit, hpit);

    if(ev->xbutton.window != XtWindow(pban)) {
        goto out;
    }
/*
    d = (ev->xbutton.y - dpit/2)/dpit;
*/
    d = (ev->xbutton.y - dpit)/dpit;
    h = (ev->xbutton.x - hpit)/hpit;
    hp = ((double)ev->xbutton.x - hpit)/(hpit*24)*DAY;

    if(debug>=100) 
        printf("  button %d,%d => %d, %d\n",
            ev->xbutton.x, ev->xbutton.y,
            d, h);

    gt = dh_time[d][0] + hp;
    if(debug>=100) 
        printf("  d %d + hp %.2f = %d\n", dh_time[d][0], hp, gt);

    if(debug>=100) {
        tmbuf = localtime(&gt);
        strftime(tbuf, BUFSIZ, "%y/%m/%d %T", tmbuf);
        printf("gt %d '%s'\n", gt, tbuf);
    }
#if 0
printf("gt %d, begin_t %d, end_t %d\n  diff %d, %d\n",
    gt, begin_t, end_t, begin_t-gt, end_t-gt);
#endif

    if(gt>=begin_t && gt<=end_t) {
        focus_t = gt;

        Verify_twindow();
    }
    else {
        gh = (gt/HOUR)*HOUR;

#if 0
printf("gt %d, gh %d\n", gt, gh);
printf("b cur %d\n", img_pool_count);
printf("  #0 %d, last #%d %d\n",
    fpairs[0].etime.tv_sec,
    img_pool_count-1, fpairs[img_pool_count-1].etime.tv_sec);
#endif

        if(gt<begin_t) {
            sbegin = gh;
            send = begin_t;
        }
        if(gt>end_t) {
            sbegin = end_t;
            send = gh+HOUR;
        }

        if(debug) {
            printf("%10s %10s %10s %10s %10s %10s\n",
                    "gt", "gh", "sbegin","send","begin","end");
            printf("%10d %10d %10d %10d %10d %10d\n",
                    gt, gh, sbegin, send, begin_t, end_t);
            printf("%10d %10d %10d %10d %10d %10d\n",
                    gt-gh, gh-gh, sbegin-gh, send-gh, begin_t-gh, end_t-gh);
        }

        for(i=0;i<npath;i++) {
#if 0
            scan_frames(path_alist[i].name, path_alist[i].value, send,
                send-sbegin, time(NULL) - WEEK);
#endif
            scan_frames(path_alist[i].name, path_alist[i].value, send,
                send-sbegin, nday_start(time(NULL), -6) );
        }
        qsort(fpairs, img_pool_count, sizeof(fpair_t), fpair_cmp);
        remove_dupimage();

#if 0
printf("a cur %d\n", img_pool_count);
printf("  #0 %d, last #%d %d\n",
    fpairs[0].etime.tv_sec,
    img_pool_count-1, fpairs[img_pool_count-1].etime.tv_sec);
#endif

        focus_t = gt;

        Verify_twindow();
        calc_pban();
    }

    Update_pban();
    Update_pcan();
    Update_pwin();

out:
}


void
ExposePban(Widget w, XEvent *ev, String *s, Cardinal *c)
{
    Update_pban();
}

void
ExposePcan(Widget w, XEvent *ev, String *s, Cardinal *c)
{
    Update_pcan();
}

void
ExposePwin(Widget w, XEvent *ev, String *s, Cardinal *c)
{
    Update_pwin();
}

void
DoResize(Widget w, XEvent *ev, String *s, Cardinal *c)
{
    refresh_flag = 1;
}

static String pwin_trans="#override\n\
    <Expose>:           pwin-expose()\n\
    <Configure>:        pwin-resize()\n\
    <Key>:              pwin-expose()\n\
";
static String pcan_trans="#override\n\
    <Expose>:           pcan-expose()\n\
    <Btn1Up>:           pcan-pick()\n\
    <Btn2Down>:         pcan-rollstart()\n\
    <Btn2Motion>:       pcan-rollmid()\n\
    <Btn2Up>:           pcan-rollend()\n\
";
static String pban_trans="#override\n\
    <Expose>:           pban-expose()\n\
    <Btn1Up>:           pban-pick()\n\
    <Btn2Up>:           pban-check()\n\
";
static String pothers_trans="#override\n\
    <Configure>:        pwin-resize()\n\
";

static XtActionsRec pwin_action_list[] = {
    {"pwin-expose",     ExposePwin},
    {"pwin-resize",     DoResize},
    {"pcan-expose",     ExposePcan},
    {"pcan-pick",       PickTime},
    {"pcan-rollstart",  RollTimeStart},
    {"pcan-rollmid",    RollTimeMid},
    {"pcan-rollend",    RollTimeEnd},

    {"pban-expose",     ExposePban},
    {"pban-pick",       PickCell},
    {"pban-check",      ScanCell},
};

void
usage()
{
        printf("sfview - image viewer from sentry packed file\n\
usage: sfview [options]\n\
\n\
option:\n\
    -help\n\
    -debug num          debug leve\n\
    -path dir           dir\n\
    -range num          scan range (by seconds)\n\
\n\
note:\n\
  * you should change dataport when you start multipul program\n\
  * frame rate does not changed by this program, if you access\n\
    with shared memory\n\
\n\
example:\n\
  %% sfview -path /var/tmp/sen_bank/camera1,/var/tmp/sen_bank/camera2\n\
  %% sfview -path /var/tmp/sen_bank -range 86400\n\
");
}





void
show_paths()
{
    int i;

    for(i=0;i<npath;i++) {
        printf("%4d: %-32s %6d\n",
            i, path_alist[i].name, path_alist[i].value);
    }
}

void
clear_paths()
{
    int i;

    for(i=0;i<MAX_N_PATH;i++) {
        path_alist[i].name  = NULL;
        path_alist[i].value = -1;
    }
    npath = 0;
}

void
expand_paths(char *ls)
{
    int i;
    char tmp[BUFSIZ];
    char *p, *q;

    p = ls;
    while(*p) {
        q = tmp;
        while(*p && *p!=',') {
            *q++ = *p++;
        }
        *q = '\0';


        if(debug>=100)
            printf("tmp '%s'\n", tmp);
        path_alist[npath].name  = strdup(tmp);
        path_alist[npath].value = npath;
        npath++;

        if(*p=='\0')
            break;
        if(*p==',')
            p++;
    }

#if 0
    show_paths();
#endif
}

void
store_images(int len)
{
    time_t now;
    int i;

    now = time(NULL);

    for(i=0;i<npath;i++) {
#if 0
        scan_frames(path_alist[i].name, path_alist[i].value,
            now, len, now - WEEK);
#endif
        scan_frames(path_alist[i].name, path_alist[i].value,
            now, len, nday_start(now, -6) );
    }
    qsort(fpairs, img_pool_count, sizeof(fpair_t), fpair_cmp);
    remove_dupimage();

    focus_t = fpairs[img_pool_count-1].etime.tv_sec;

    Verify_twindow();

    Update_pban();
    Update_pcan();
    Update_pwin();
}


store_color(Pixel *sa, int r, int g, int b)
{
    if(output_corder==SEN_C_XRGB) {
        *sa  = r*0x010000;
        *sa += g*0x000100;
        *sa += b*0x000001;
    }
    else 
    if(output_corder==SEN_C_XBGR) {
        *sa  = b*0x010000;
        *sa += g*0x000100;
        *sa += r*0x000001;
    }
    
}

pick_colors()
{
    Display *dp;
    int     sc;
    Colormap    cmap;
    XColor  qc, rc;

    store_color(&pwin_black,    0x00, 0x00, 0x00);
    store_color(&pwin_grayC,    0xcc, 0xcc, 0xcc);
    store_color(&pwin_red,      0xff, 0x00, 0x00);

    store_color(&pcan_black,    0x00, 0x00, 0x00);
    store_color(&pcan_grayC,    0xcc, 0xcc, 0xcc);
    store_color(&pcan_gray9,    0x99, 0x99, 0x99);
    store_color(&pcan_white,    0xff, 0xff, 0xff);
    store_color(&pcan_red,      0xff, 0x00, 0x00);
    store_color(&pcan_blue,     0x00, 0x00, 0xff);

    store_color(&pban_black,    0x00, 0x00, 0x00);
    store_color(&pban_grayC,    0xcc, 0xcc, 0xcc);
    store_color(&pban_gray9,    0x99, 0x99, 0x99);
    store_color(&pban_white,    0xff, 0xff, 0xff);
    store_color(&pban_red,      0xff, 0x00, 0x00);
    store_color(&pban_blue,     0x00, 0x00, 0xff);

#if 1
    store_color(&pban_l0,       0xff, 0xff, 0xff);
    store_color(&pban_l1,       0xcc, 0xff, 0xff);
    store_color(&pban_l2,       0x99, 0xff, 0xff);
    store_color(&pban_l3,       0x66, 0xff, 0xff);
    store_color(&pban_l4,       0x33, 0xff, 0xff);
    store_color(&pban_l5,       0x00, 0xcc, 0xff);
    store_color(&pban_lerr,     0xff, 0x99, 0xff);
#endif

#if 0
    store_color(&pban_l0,       0xff, 0xff, 0xff);
    store_color(&pban_l1,       0xcc, 0xcc, 0xff);
    store_color(&pban_l2,       0x99, 0x99, 0xff);
    store_color(&pban_l3,       0x66, 0x66, 0xff);
    store_color(&pban_l4,       0x33, 0x33, 0xff);
    store_color(&pban_l5,       0xff, 0x99, 0xff);
    store_color(&pban_lerr,     0x00, 0xff, 0xcc);
#endif

}


void
Clock_update(XtPointer client_data, XtIntervalId *iid)
{
    xp_periodical_rec *prof;
    char buf[BUFSIZ];
    time_t now;
    XFontStruct *xf;

    prof = (xp_periodical_rec*) client_data;
    
    now = time(NULL);
    mk_timestr(buf, "current", &now);

#if 0
    XtVaGetValues(pcan_time, XtNfont, &xf, NULL);
    printf("xf %p\n", xf);
    XtVaSetValues(pcan_time, XtNlabel, buf, XtNfont, time_font_info, NULL);
#endif
    XtVaSetValues(pcan_time, XtNlabel, buf, NULL);

    prof->id = XtAppAddTimeOut(app_context,
        prof->interval, Clock_update, prof);
}

do_quit(Widget w, XtPointer client_data, XtPointer call_data)
{
    exit(0);
}

fd_set  InputFdMask;


int
main(int argc, char **argv)
{
    int skey;
    XGCValues values;
    int valuemask;
    int tv;
    int ch;
    int zwid;

    zwid = 160*4;
    
    clear_paths();

    toplevel = XtVaAppInitialize(
        &app_context,                   /* Application context */
        "Sentry",                       /* Application class */
        options, XtNumber(options),     /* option list */
        &argc, argv,                    /* command line args */
        fallback_resources,             /* for missing app-defaults file */
        XtNtitle,       "sfview",
        XtNiconName,    "sfview",
        NULL);                          /* terminate varargs list */

    XtVaGetApplicationResources(toplevel,
        (XtPointer) &sentry_resource, resources, XtNumber(resources), NULL);

    XtAppAddActions(app_context, main_action_list, XtNumber(main_action_list));



    if(getenv("SENTRY_SERVER")) {
        strncpy(server_host, getenv("SENTRY_SERVER"), 127);
    }
    else {
        strncpy(server_host, sentry_resource.server_name, 127);
    }
    server_host[127]='\0';

    server_port = atoi(sentry_resource.server_port);
    data_port   = atoi(sentry_resource.data_port);
    server_rate = atoi(sentry_resource.req_rate);
    debug = atoi(sentry_resource.debug_level);
    show_help = atoi(sentry_resource.show_help);
    if(atoi(sentry_resource.shm_key)<=0)
        use_shm_flag = 0;
    else {
        use_shm_flag = 1;
        skey = atoi(sentry_resource.shm_key);
    }

    tv = assoc_n(matching_type_alist, sentry_resource.compare_func);
    if(debug) {
        fprintf(stderr, "compare type %d %x '%s' <= '%s'\n",
            tv, tv, assoc_v(matching_type_alist, tv),
            sentry_resource.compare_func);
    }
/*
    cmp_type = tv;
*/

    if(debug) {
        fprintf(stderr, "severname '%s'\n", sentry_resource.server_name);
    }

    tv = assoc_n(format_alist, sentry_resource.req_format);
    server_format = tv;

    tv = assoc_n(size_alist, sentry_resource.req_size);
    server_size = tv;

    expand_paths(sentry_resource.path);

    scan_range = atoi(sentry_resource.scan_range);


    if(debug>=10) {
    fprintf(stderr, "format %d, '%s' => '%s'\n",    
        tv, sentry_resource.req_format, assoc_v(format_alist, tv));
    fprintf(stderr, "size %d, '%s' => '%s'\n",    
        tv, sentry_resource.req_size, assoc_v(size_alist, tv));
    }


    if(show_help) {
        usage();
        exit(0);
    }


    time_font_info = XLoadQueryFont(XtDisplay(toplevel), "a14");
    if(!time_font_info) {
        time_font_info = XLoadQueryFont(XtDisplay(toplevel), "fixed");
        if(!time_font_info) {
            printf("not found font 'a14' and 'fixed'\n");
        }
    }
    if(debug)
        printf("time_font_info %p\n", time_font_info);


    fflush(stdout);
    fflush(stderr);
    SetupVisual(XtDisplay(toplevel), fvisual, svisual);
    fflush(stdout);
    fflush(stderr);

    output_format = SEN_F_RGB;
#if 0
    /* corder is solved by SetupVisual() */
    output_corder = SEN_F_RGB;
#endif


    pshell = XtVaAppCreateShell(
        "sentry", "Sentry",
        topLevelShellWidgetClass, XtDisplay(toplevel),
        XtNvisual, xp_foil_visual,
        XtNdepth, xp_foil_depth,
        XtNcolormap, xp_foil_cmap,
        NULL);

    pbase = XtVaCreateManagedWidget(
        "base", formWidgetClass, pshell,
        XtNvisual,  xp_foil_visual,
        XtNdepth,   xp_foil_depth,
        XtNcolormap,xp_foil_cmap,
        NULL);
    plabel = XtVaCreateManagedWidget(
        "label", labelWidgetClass, pbase,
        XtNlabel, "boot...",
        XtNjustify, XtJustifyLeft,
        XtNwidth,  zwid,
        XtNresize,  False,
        NULL);

    pwin_base = XtVaCreateManagedWidget(
        "Pcan", formWidgetClass, pbase,
        XtNfromVert, plabel,
        NULL);
    pwin_label = XtVaCreateManagedWidget(
        "label", labelWidgetClass, pwin_base,
        XtNlabel, "frames",
        XtNjustify, XtJustifyLeft,
        XtNresize,  False,
        XtNborderWidth, 0,
        NULL);
    pwin = XtVaCreateManagedWidget(
        "Pwin", lwwWidgetClass, pwin_base,
/*
        XtNfromVert, plabel,
*/
        XtNfromVert, pwin_label,
        XtNvisual,  xp_foil_visual,
        XtNdepth,   xp_foil_depth,
        XtNcolormap,xp_foil_cmap,
        XtNwidth,  zwid,
        XtNheight, (120+20)*npath,
        NULL);


    scbase  = XtVaCreateManagedWidget(
        "Pscbase", formWidgetClass, pbase,
        XtNfromVert, pwin_base,
        XtNwidth,  zwid,
        XtNborderWidth, 0,
        NULL);

    /* begin rev play ff end */

    sc_quit  = XtVaCreateManagedWidget(
        "Pscquit", commandWidgetClass, scbase,
        XtNlabel, "quit",
        NULL);
    XtAddCallback(sc_quit, XtNcallback, do_quit, NULL);

    sc_begin = XtVaCreateManagedWidget(
        "Pscstart", commandWidgetClass, scbase,
        XtNfromHoriz, sc_quit,
        XtNlabel, "|<",
        NULL);
    XtAddCallback(sc_begin, XtNcallback, bar_hand, (void*)-2);

    sc_rev = XtVaCreateManagedWidget(
        "Pscrev", commandWidgetClass, scbase,
        XtNfromHoriz, sc_begin,
        XtNlabel, "<<",
        NULL);
    XtAddCallback(sc_rev, XtNcallback, bar_hand, (void*)-1);

    sc_play = XtVaCreateManagedWidget(
        "Pscplay", commandWidgetClass, scbase,
        XtNfromHoriz, sc_rev,
        XtNlabel, ">",
        NULL);
    XtAddCallback(sc_play, XtNcallback, bar_hand, (void*)0);

    sc_ff = XtVaCreateManagedWidget(
        "Pscff", commandWidgetClass, scbase,
        XtNfromHoriz, sc_play,
        XtNlabel, ">>",
        NULL);
    XtAddCallback(sc_ff, XtNcallback, bar_hand, (void*)1);

    sc_end = XtVaCreateManagedWidget(
        "Pscend", commandWidgetClass, scbase,
        XtNfromHoriz, sc_ff,
        XtNlabel, ">|",
        NULL);
    XtAddCallback(sc_end, XtNcallback, bar_hand, (void*)2);

    sc_skip = XtVaCreateManagedWidget(
        "Pscskip", commandWidgetClass, scbase,
        XtNfromHoriz, sc_end,
        XtNlabel, "skip",
        NULL);
    XtAddCallback(sc_skip, XtNcallback, ToggleSkip, NULL);


    pcan_base = XtVaCreateManagedWidget(
        "Pban", formWidgetClass, pbase,
        XtNfromVert, scbase,
        NULL);
    pcan_label = XtVaCreateManagedWidget(
        "label", labelWidgetClass, pcan_base,
        XtNlabel, "Short",
        XtNjustify, XtJustifyLeft,
        XtNborderWidth, 0,
        XtNresize,  False,
        NULL);
    pcan_time = XtVaCreateManagedWidget(
        "label", labelWidgetClass, pcan_base,
        XtNfromHoriz, pcan_label,
        XtNfont,    time_font_info,
        XtNlabel,   "current 00/00/00 00:00:00",
        XtNjustify, XtJustifyLeft,
        XtNresize,  True,
        XtNwidth,   zwid/2,
        XtNborderWidth, 0,
        NULL);
    pcan = XtVaCreateManagedWidget(
        "Pban", formWidgetClass, pcan_base,
        XtNfromVert, pcan_label,
        XtNwidth,   zwid, XtNheight, 80,
        NULL);

    pban_base = XtVaCreateManagedWidget(
        "Pban", formWidgetClass, pbase,
        XtNfromVert, pcan_base,
        NULL);
    pban_label = XtVaCreateManagedWidget(
        "label", labelWidgetClass, pban_base,
        XtNlabel, "Long",
        XtNjustify, XtJustifyLeft,
        XtNresize,  False,
        XtNborderWidth, 0,
        NULL);
    pban = XtVaCreateManagedWidget(
        "Pban", formWidgetClass, pban_base,
        XtNfromVert, pban_label,
        XtNwidth,  zwid, XtNheight, 120,
        NULL);

#if 1
    pban_dn = XtVaCreateManagedWidget( "label", labelWidgetClass, pban_base,
        XtNlabel, "N/A", XtNfromHoriz, pban_label, NULL);
    pban_d0 = XtVaCreateManagedWidget( "label", labelWidgetClass, pban_base,
        XtNlabel, "0", XtNfromHoriz, pban_dn, NULL);
    pban_d1 = XtVaCreateManagedWidget( "label", labelWidgetClass, pban_base,
        XtNlabel, "<10", XtNfromHoriz, pban_d0, NULL);
    pban_d2 = XtVaCreateManagedWidget( "label", labelWidgetClass, pban_base,
        XtNlabel, "<50", XtNfromHoriz, pban_d1, NULL);
    pban_d3 = XtVaCreateManagedWidget( "label", labelWidgetClass, pban_base,
        XtNlabel, "<100", XtNfromHoriz, pban_d2, NULL);
    pban_d4 = XtVaCreateManagedWidget( "label", labelWidgetClass, pban_base,
        XtNlabel, "<500", XtNfromHoriz, pban_d3, NULL);
    pban_d5 = XtVaCreateManagedWidget( "label", labelWidgetClass, pban_base,
        XtNlabel, "more", XtNfromHoriz, pban_d4, NULL);
#endif

    ptscr = XtVaCreateManagedWidget(
        "Ptscr", scrollbarWidgetClass, pbase,
        XtNfromVert, pban_base,
        XtNorientation, XtorientHorizontal,
        XtNwidth,  zwid,
        NULL);
    XtAddCallback(ptscr, XtNscrollProc, Scroll_canvas,  NULL);
    XtAddCallback(ptscr, XtNjumpProc, Jump_canvas,  NULL);

    XtRealizeWidget(pshell);

    pwin_gc = XtGetGC(pwin, NULL, NULL);

    XtAppAddActions(app_context,
        pwin_action_list, XtNumber(pwin_action_list));
    XtOverrideTranslations(pwin,    XtParseTranslationTable(pwin_trans));
    XtOverrideTranslations(pcan,    XtParseTranslationTable(pcan_trans));
    XtOverrideTranslations(pcan,    XtParseTranslationTable(pothers_trans));
    XtOverrideTranslations(pban,    XtParseTranslationTable(pban_trans));
    XtOverrideTranslations(pban,    XtParseTranslationTable(pothers_trans));
    XtOverrideTranslations(pbase,   XtParseTranslationTable(pothers_trans));
    XtOverrideTranslations(pshell,  XtParseTranslationTable(pothers_trans));

    pick_colors();
    
#if 1
#if 0
    if(time_font_info) {
        XtVaSetValues(pban_d0, XtNfont, time_font_info, XtNlabel, "XC", NULL);
    }
#endif

    XtVaSetValues(pban_dn, XtNbackground, pban_grayC, NULL);
    XtVaSetValues(pban_d0, XtNbackground, pban_l0, NULL);
    XtVaSetValues(pban_d1, XtNbackground, pban_l1, NULL);
    XtVaSetValues(pban_d2, XtNbackground, pban_l2, NULL);
    XtVaSetValues(pban_d3, XtNbackground, pban_l3, NULL);
    XtVaSetValues(pban_d4, XtNbackground, pban_l4, NULL);
    XtVaSetValues(pban_d5, XtNbackground, pban_l5, NULL);
#endif


PASS;
    if(use_shm_flag) {
        init_shm(skey);
#if 0
        if(clear) {
            clear_shm();
        }
#endif
    }
    else {
#if 0
        ctrlfd = build_ctrl_port(server_host, server_port);
        if (ctrlfd < 0) {
            printf("fail setup control fd with port# %d\n", server_port);
            exit (1);
        }

PASS;
        datafd = build_data_port(data_port);
        if (datafd < 0) {
            printf("fail setup data fd with port# %d\n", data_port);
            exit (1);
        }
#endif
    }
PASS;

#if 0

    if(use_shm_flag) {
        fprintf(stderr, "sentry image viewer upon shared memory\n");

        win1.interval = 300; /* by ms */

        win1.id = XtAppAddTimeOut(app_context,
                   win1.interval, KickWin, &win1);
    }
    else {
        fprintf(stderr, "sentry image viewer with direct network access\n");

        init_ibuf();

        init_request();
        request_call();
        signal(SIGALRM, request_call);

PASS;

        FD_ZERO(&InputFdMask);
        FD_SET(ctrlfd, &InputFdMask);
        FD_SET(datafd, &InputFdMask);

        iid = XtAppAddInput(app_context, datafd,
                (caddr_t)XtInputReadMask, x_get_data, NULL);
#if 0
        ciid = XtAppAddInput(app_context, ctrlfd,
                (caddr_t)XtInputReadMask, retry_connect, NULL);
#endif

    }
#endif

    store_images(scan_range);

    XtVaSetValues(plabel,
        XtNlabel, "sfview - sentry file viewer",
        XtNborderWidth, 0,
        NULL);

    clock_timer.interval = 500;
    clock_timer.id = XtAppAddTimeOut(app_context,
                        clock_timer.interval, Clock_update, &clock_timer);


PASS;

    XtAppMainLoop(app_context);

}



