/*
 * sentry.c - image viewer for sentry protocol
 *      k-chinen@is.aist-nara.ac.jp, 1995-2001, 2003
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <signal.h>
#include <syslog.h>
#ifdef TIME_WITH_SYS_TIME
 #include <sys/time.h>
 #include <time.h>
#else
 #ifdef HAVE_SYS_TIME_H
  #include <sys/time.h>
 #else
  #include <time.h>
 #endif
#endif
#include <errno.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xproto.h>
#include <X11/cursorfont.h>

#include <X11/Xmu/WinUtil.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/Scrollbar.h>

/* private TrueColor Widget */
#include "Lww.h"

#include "alist.h"
#include "sentry.h"
#include "img.h"
#include "pool.h"
#include "client.h"

#include "rtr.h"

#include "fig.h"

#ifndef MIN
#define MIN(x,y)    ((x)<(y) ? (x):(y))
#endif

#ifndef MAX_IMG_LEN
#define MAX_IMG_LEN     (640*480*4)
#endif
#ifndef TMP_IMGBUF_LEN
#define TMP_IMGBUF_LEN  (320*240*4)
#endif

extern apair_t buflen_xrgb_alist;

extern  void request_call();

apair_t anchor_short_alist[] = {
    {"LT",  SEN_A_LT},
    {"L",   SEN_A_L},
    {"LB",  SEN_A_LB},
    {"RT",  SEN_A_RT},
    {"R",   SEN_A_R},
    {"RB",  SEN_A_RB},
    {"T",   SEN_A_T},
    {"C",   SEN_A_C},
    {"B",   SEN_A_B},
    {NULL,  -1},
};


typedef struct {
    /* id */
    char *name;
    char *serv_host;
    int   serv_port;
    int   recv_port;
    char *desc;

    struct sockaddr_in addr;
    int   cntl_fd;
    int   data_fd;
    subrecord_t *recbuf;

    /* request */
    int   type;
    int   size;
    int   color;
    int   format;
    int   rate;
    int   zoom;

    /* presentation */
    int   x;
    int   y;
    int   anchor;

    int rx;
    int ry;
    int gx;
    int gy;

    Pixmap  pix;

    sentry_imgrec inimg;
    sentry_imgrec outimg;

} cnode_t;

show_cnode(FILE *fp, char *msg, cnode_t *cn)
{
    fprintf(fp, "cnode %p : %s\n", cn, msg);
    fprintf(fp, "  name '%s'\n", cn->name);
    fprintf(fp, "  server '%s' port %d\n", cn->serv_host, cn->serv_port);
    fprintf(fp, "  receive port %d\n", cn->recv_port);
    fprintf(fp, "  description '%s'\n", cn->desc);
    fprintf(fp, "  addr -\n");
    fprintf(fp, "  fd cntl %d data %d\n", cn->cntl_fd, cn->data_fd);
    fprintf(fp, "  req. type %s(%d) size %s(%d) color %s(%d) format %s(%d)\n",
        assoc_v(req_type_alist, cn->type), cn->type,
        assoc_v(size_alist, cn->size), cn->size,
        assoc_v(color_alist, cn->color), cn->color,
        assoc_v(format_alist, cn->format), cn->format);
    fprintf(fp, "    rate %d, zoom %d\n",
        cn->rate, cn->zoom);
    fprintf(fp, "  position %d,%d, anchor %s(%d)\n",    
        cn->x, cn->y, assoc_v(anchor_alist, cn->anchor), cn->anchor);

    fflush(fp);
}

int
clear_cnode(cnode_t *cn)
{
    if(!cn) {
        return -1;
    }

    cn->name = NULL;
    cn->serv_host = NULL;
    cn->serv_port = -1;
    cn->recv_port = -1;
    cn->desc = NULL;
    memset(&cn->addr, 0, sizeof(cn->addr));
    cn->cntl_fd = -1;
    cn->data_fd = -1;
    cn->type    = -1;
    cn->size    = -1;
    cn->color   = -1;
    cn->format  = -1;
    cn->zoom    = -1;
    cn->rate    = -1;
    cn->x       = -1;
    cn->y       = -1;
    cn->anchor  = -1;

    cn->recbuf  = NULL;

    cn->inimg.body_len = 0;
    cn->outimg.body_len = 0;
}

int
setdefault_cnode(cnode_t *cn)
{
    int buflen;

    if(!cn) {
        return -1;
    }

    cn->name = NULL;
    cn->serv_host = NULL;
    cn->serv_port = SENTRY_CTRL_PORT;
    cn->recv_port = SENTRY_DATA_PORT;
    cn->desc = NULL;
    memset(&cn->addr, 0, sizeof(cn->addr));
    cn->cntl_fd = -1;
    cn->data_fd = -1;
    cn->type    = OPEN_UDP_UNICAST;
    cn->size    = SIZE_160x120;
    cn->color   = SEN_C_NONE;
    cn->format  = SEN_F_JPEG;
    cn->zoom    = 1;
    cn->rate    = 1000;
    cn->x       = 0;
    cn->y       = 0;
    cn->anchor  = SEN_A_LT;

    cn->recbuf  = NULL;

    buflen = 160*120*4;

    cn->inimg.body      = malloc(buflen);
    cn->inimg.body_len  = buflen;
    cn->outimg.body     = malloc(buflen);
    cn->outimg.body_len = buflen;

#if 0 
    cn->pix = XCreatePixmap(XtDisplay(gw),
        DefaultRootWindow(XtDisplay(gw)),
        160, 120, depth);
#endif
}


#ifndef MAX_CNODE
#define MAX_CNODE   (32)
#endif

cnode_t cnode_list[MAX_CNODE];
int cnode_num;

cnode_t*
find_cnode(char *name) {
    int i;
    for(i=0;i<cnode_num;i++) {
        if(strcmp(cnode_list[i].name, name)==0) {
            return &cnode_list[i];
        }
    }
    return NULL;
}


extern u_char client_type; 
extern int server_zoom;
extern int server_size;
extern int server_color;
extern int server_format; 
extern int server_rate;

static int show_help=0;

static int beep_mute=0;
static int ghost_mode=0;


typedef struct {
    String server_name;
    String server_port;
    String data_port;
    String req_rate;
    String req_type;
    String req_size;
    String req_zoom;
    String req_color;
    String req_format;
    String hivision_flag;
    String voffset;
    /* */
    String shm_key;
    /* */
    String compare_func;
    /* */
    String debug_level;
    String show_help;
    String img_history;
} sentry_resource_t;

sentry_resource_t sentry_resource;

#define Offset(field) (XtOffsetOf(sentry_resource_t, field))
static XtResource resources[] = {
  {"servername", "ServerName",  XtRString,  sizeof(String),
                    Offset(server_name), XtRString,  NULL},
  {"serverport", "ServerPort",  XtRString,  sizeof(String),
                    Offset(server_port), XtRString,  NULL},
  {"dataport",   "DataPort",  XtRString,  sizeof(String),
                    Offset(data_port), XtRString,  NULL},
  {"reqsize",   "ReqSize",  XtRString,  sizeof(String),
                    Offset(req_size), XtRString,  NULL},
  {"reqrate",   "ReqRate",  XtRString,  sizeof(String),
                    Offset(req_rate), XtRString,  NULL},
  {"reqformat", "ReqFormat",  XtRString,  sizeof(String),
                    Offset(req_format), XtRString,  NULL},
  {"reqtype",   "ReqType",  XtRString,  sizeof(String),
                    Offset(req_type), XtRString,  NULL},

  {"shmkey",    "ShmKey",  XtRString,  sizeof(String),
                    Offset(shm_key), XtRString,  NULL},

  {"comp",      "comp",  XtRString,  sizeof(String),
                    Offset(compare_func), XtRString,  NULL},

  {"debug",     "Debug",  XtRString,  sizeof(String),
                    Offset(debug_level), XtRString,  NULL},
  {"help",      "Help",  XtRString,  sizeof(String),
                    Offset(show_help), XtRString,  NULL},
  {"history",   "History",  XtRString,  sizeof(String),
                    Offset(img_history), XtRString,  NULL},
};




String fallback_resources[] = {
    "*reqsize:      160x120",
    "*reqrate:      200",
    "*reqformat:    JPEG",
    "*servername:   localhost",
    "*serverport:   7001",
    "*dataport:     7000",

    "*shmkey:       -1",

    "*comp:         none",

    "*debug:        0",

    "*help:         0",

    "*foreground:   black",
    "*background:   white",

    "*history:      16",
    NULL};

static XrmOptionDescRec options[] = {
{"-fg",         "*Foreground",  XrmoptionSepArg,    NULL},
{"-bg",         "*Background",  XrmoptionSepArg,    NULL},
{"-servername", "*servername",  XrmoptionSepArg,    NULL},
{"-serverport", "*serverport",  XrmoptionSepArg,    NULL},
{"-dataport",   "*dataport",    XrmoptionSepArg,    NULL},

{"-shmkey",     "*shmkey",      XrmoptionSepArg,    NULL},
{"-debug",      "*debug",       XrmoptionSepArg,    NULL},
{"-rate",       "*reqrate",     XrmoptionSepArg,    NULL},
{"-format",     "*reqformat",   XrmoptionSepArg,    NULL},
{"-size",       "*reqsize",     XrmoptionSepArg,    NULL},

{"-history",    "*history",     XrmoptionSepArg,    NULL},

{"-comp",       "*comp",        XrmoptionSepArg,    NULL},

{"-help",       "*help",        XrmoptionNoArg,     "1"},
};

Widget   toplevel;
XtAppContext app_context;

XtActionsRec main_action_list[] = {
    NULL,
};


Visual   *xp_slide_visual,   *xp_foil_visual;
int      xp_slide_depth,     xp_foil_depth;
Colormap xp_slide_cmap,      xp_foil_cmap;
int      xp_use_special_cmap;

String fvisual, svisual;


Pixel black, white;

Widget pwinmsg;
Widget pshell;
Widget pbase;
Widget pcpanel;
Widget plabel;
Widget pcmd;
Widget bquit;
Widget bmute, bghost;
GC  pwin_gc;

Pixmap diff_pix;

char *bg_fname=NULL;
xp_fig_elem_t *bg_fig=NULL;
int bg_wid=500, bg_ht=500;
int sc_wid=160, sc_ht=120;
Pixmap bg_pix;


typedef struct {
    XtIntervalId    id;
    time_t          interval;
} xp_periodical_rec;

xp_periodical_rec win1;


void periodical_update(XtPointer client_data, XtIntervalId *iid);


extern sentry_shmrec   *sentry_shm_chunk;


show_chunk()
{
    int i;
    int pos;

    for(i=0;i<sentry_shm_chunk_radix;i++) {
        pos = i+1;
        printf(" %3d: %3d %12d %4d\n", i, pos,
            sentry_shm_chunk[pos].entry.header.recv_sec,
            sentry_shm_chunk[pos].entry.header.length);
    }
}

int 
find_pos(time_t reftime)
{
    int i;
    int pos;
    int ret;

    ret = -1;

    for(i=0;i<sentry_shm_chunk_radix;i++) {
        pos = i+1;
#if 0
        printf(" %3d: %3d %12d %4d\n", i, pos,
            sentry_shm_chunk[pos].entry.header.recv_sec,
            sentry_shm_chunk[pos].entry.header.length);
#endif
        if(sentry_shm_chunk[pos].entry.header.recv_sec==reftime) {
            if(debug>=10) {
                printf(
" %3d: %3d time issue %d.%06d recv %d.%06d len %4d (%dx%d) pos %p [%d]\n",
                    i, pos,
                    sentry_shm_chunk[pos].entry.header.send_sec,
                    sentry_shm_chunk[pos].entry.header.send_usec,
                    sentry_shm_chunk[pos].entry.header.recv_sec,
                    sentry_shm_chunk[pos].entry.header.recv_usec,
                    sentry_shm_chunk[pos].entry.header.length,
                    sentry_shm_chunk[pos].entry.header.width,
                    sentry_shm_chunk[pos].entry.header.height,
                    sentry_shm_chunk[pos].entry.header.body,
                    sentry_shm_chunk[pos].entry.header.body_len
                    );
            }
            ret = pos;
        }
    }

    return ret;
}


SenDrawImage_xy(Widget w, int x, int y, sentry_imgrec *inimg, sentry_imgrec *outimg)
{
    int wid, ht;
    int xwid, xht;
    unsigned char *raw_data;
#if 0
    unsigned char *data_rgb=NULL, *data_bgrx=NULL;
    int buflen=0;
#endif
#if 1
    static unsigned char *data_rgb=NULL, *data_bgrx=NULL;
    static int buflen=0;
#endif
    int raw_len;
    int raw_format, raw_color;
    int chk;
    int cur_format=SEN_F_NONE;
    int cur_corder=SEN_C_NONE;
    int success=0;
    int outlen=-1;
    GC cur_gc;

    if(inimg==NULL)
        return;

#if 0
    if(debug>=10) {
        printf("SenDrawImage: output format '%s', color order '%s'\n",
            assoc_v(format_alist, output_format),   
            assoc_v(color_alist, output_corder));
    }
#endif


    raw_len     = inimg->length;
    raw_data    = (unsigned char*)inimg->body;
    raw_format  = inimg->format;
    raw_color   = inimg->corder;
    wid = inimg->width;
    ht  = inimg->height;

    if(!data_rgb || !data_bgrx) {
        buflen = (int)assoc_p(&buflen_xrgb_alist, MAX_DISPLAY_SIZE);
        if(buflen<=0) {
            buflen = TMP_IMGBUF_LEN;
        }
fprintf(stderr, "#buflen %d\n", buflen);

        data_bgrx = (unsigned char*)malloc(buflen);
        memset(data_bgrx, 0xff, buflen);

        data_rgb = (unsigned char*)malloc(buflen);
        memset(data_rgb, 0xff, buflen);
    }

#if 0
    if(debug>=10) {
        printf("SenDrawImage: %p %d '%s''%s' %dx%d, 3B/p=%d 4B/p=%d\n",
            raw_data, raw_len, assoc_v(format_alist, raw_format),
            assoc_v(color_alist, raw_color), wid, ht, wid*ht*3, wid*ht*4);
    }
#endif
    if(debug>=10) {
        printf(
"draw: got format %s, color %s, ptr %p, size %dx%d, len %d\n\
      out format %s, color %s\n",
        assoc_v(format_alist, raw_format), assoc_v(color_alist, raw_color), 
        raw_data, wid, ht, raw_len,
        assoc_v(format_alist, output_format),
        assoc_v(color_alist, output_corder));
    }

    switch(raw_format) {
    case SEN_F_JPEG:
        chk = decode_jpeg(raw_data, raw_len, data_rgb, &xwid, &xht);
        if(xwid!=wid && xht!=ht) {
            fprintf(stderr,
                "SenDrawImage: wid %d != xwid %d, ht %d != xht %d\n",
                wid, xwid, ht, xht);
        }
        cur_format = SEN_F_RGB;
        cur_corder = SEN_C_RGB;
        if(debug>=20){
            printf("  convert to '%s' (wid %d, ht %d)\n",
                assoc_v(color_alist, cur_corder), xwid, xht);
        }
        break;

    case SEN_F_NONE: /* XXX */
        memcpy(data_rgb, raw_data, raw_len);
        break;

    case SEN_F_RGB: /* XXX */
        memcpy(data_rgb, raw_data, raw_len);
        cur_format = raw_format; /* SEN_F_RGB */
        cur_corder = raw_color;
        break;
    }

    if(debug>=10) {
        printf(
"draw: cur format %s, color %s\n\
      out format %s, color %s\n",
        assoc_v(format_alist, cur_format), assoc_v(color_alist, cur_corder),
        assoc_v(format_alist, output_format),
        assoc_v(color_alist, output_corder)
        );
    }
    if(cur_format != output_format) {
        if(debug>=1) {
            fprintf(stderr, "image format miss-match cur=%s out=%s\n",
                assoc_v(format_alist, cur_format),
                assoc_v(format_alist, output_format));
        }
        goto put_fail;
    }

    if(output_format==SEN_F_RGB) {
        cur_gc = XtGetGC(w, NULL, NULL);
        if(output_corder==SEN_C_XBGR) {
            convert_RGB2XBGR(data_bgrx, data_rgb, wid * ht);

            outlen = wid*ht*4;
            put_image_xy(XtDisplay(w), xp_foil_visual,
                xp_foil_depth, cur_gc, XtWindow(w), x, y, wid, ht, data_bgrx);

            success = 1;

            copyattr_img(outimg, inimg);
            outimg->format = output_format;
            outimg->corder = output_corder;
            outimg->length = outlen;
            memcpy(outimg->body, data_bgrx, outlen);
        }
        else
        if(output_corder==SEN_C_XRGB) {
/*
            convert_RGB2XRGB(data_bgrx, data_rgb, wid * ht);
*/
            convert_RGB2BGRX(data_bgrx, data_rgb, wid * ht);

            outlen = wid*ht*4;
            put_image_xy(XtDisplay(w), xp_foil_visual,
                xp_foil_depth, cur_gc, XtWindow(w), x, y, wid, ht, data_bgrx);
            success = 1;

            copyattr_img(outimg, inimg);
            outimg->format = output_format;
            outimg->corder = output_corder;
            outimg->length = outlen;
            memcpy(outimg->body, data_bgrx, outlen);
        }
        else
        if(output_corder==SEN_C_RGB) {
            outlen = wid*ht*3;
            put_image_xy(XtDisplay(w), xp_foil_visual,
                xp_foil_depth, cur_gc, XtWindow(w), x, y, wid, ht, data_rgb);
            success = 1;

            copyattr_img(outimg, inimg);
            outimg->format = output_format;
            outimg->corder = output_corder;
            outimg->length = outlen;
            memcpy(outimg->body, data_rgb, outlen);
        }
        XtReleaseGC(w, cur_gc);
    }

put_fail:
    if(!success) {
        if(debug>=1) {
            fprintf(stderr, "fail some reason, format cur=%s out=%s\n",
                assoc_v(format_alist, cur_format),
                assoc_v(format_alist, output_format));
        }
    }
    else {
#if 0
        copyattr_img(outimg, inimg);
        memcpy(outimg->body, inimg->body, outlen);
#endif
    }

out:
}

SenDrawImage(Widget w, sentry_imgrec *inimg, sentry_imgrec *outimg)
{
    return SenDrawImage_xy(w, 0, 0, inimg, outimg);
}

static int call_count=0, lost_count=0, got_count=0, c_got_count=0;
static int refresh_flag = 1;


#if 0
int
sub_rgb(unsigned char *abuf, unsigned char *bbuf,
    unsigned char *cbuf, int wid, int ht)
{
    unsigned char *p, *q, *s;
    unsigned char *pr, *pg, *pb, *qr, *qg, *qb;
    unsigned char gp, gq;
    int iv;
    unsigned char *pend;
    unsigned char *vb, *ve;
    int dv;
    int len;
    
    len = wid*ht*3;

    p = abuf;
    q = bbuf;
    s = cbuf;
    pend = p+len;
    vb = p;
    ve = p + len;

    while(p<pend) {
        pr = p; pg = p+1; pb = p+2;
        qr = q; qg = q+1; qb = q+2;
        iv = (*pr+*pg+*pb)/3;
        gp = (iv>255? 255: iv);
        iv = (*qr+*qg+*qb)/3;
        gq = (iv>255? 255: iv);

        dv = (*pr-*qr)*(*pr-*qr)+(*pg-*qg)*(*pg-*qg)+(*pb-*qb)*(*pb-*qb);

#if 0
        if(p>=vb&&p<=ve) {
            fprintf(stderr, "p %08x %5d %03dx%03d: %02x %02x %02x : %6d\n",
                p, p-abuf, ((p-abuf)/3)%wid, ((p-abuf)/3)/wid,
                *pr, *pg, *pb, *qr, *qg, *qb, dv);
        }
#endif
    
/*DIFF*/
        if(dv<800) {
            *s++ = gp;
            *s++ = gp;
            *s++ = gp;
        }
        else {
            *s++ = 0xff;
            *s++ = (gp+gq)/2;
            *s++ = (gp+gq)/2;
        }
        p+=3;
        q+=3;
    }
}


int
sub_bgrx(unsigned char *abuf, unsigned char *bbuf,
    unsigned char *cbuf, int wid, int ht)
{
    unsigned char *p, *q, *s;
    unsigned char *pr, *pg, *pb, *qr, *qg, *qb;
    unsigned char gp, gq;
    int iv;
    unsigned char *pend;
    unsigned char *vb, *ve;
    int dv;
    int len;
    
    len = wid*ht*4;

    p = abuf;
    q = bbuf;
    s = cbuf;
    pend = p+len;
    vb = p;
    ve = p + len;

    while(p<pend) {
        pr = p; pg = p+1; pb = p+2;
        qr = q; qg = q+1; qb = q+2;
        iv = (*pr+*pg+*pb)/3;
        gp = (iv>255? 255: iv);
        iv = (*qr+*qg+*qb)/3;
        gq = (iv>255? 255: iv);

        dv = (*pr-*qr)*(*pr-*qr)+(*pg-*qg)*(*pg-*qg)+(*pb-*qb)*(*pb-*qb);

#if 0
        if(p>=vb&&p<=ve) {
            fprintf(stderr, "p %08x %5d %03dx%03d: %02x %02x %02x : %6d\n",
                p, p-abuf, ((p-abuf)/3)%wid, ((p-abuf)/3)/wid,
                *pr, *pg, *pb, *qr, *qg, *qb, dv);
        }
#endif
    
/*DIFF*/
        if(dv<800) {
            *s++ = gp;
            *s++ = gp;
            *s++ = gp;
            s++;
        }
        else {
#if 0
            *s++ = 0xff;
            *s++ = gp;
            *s++ = gq;
#endif
#if 1
            *s++ = gp;
            *s++ = gq;
            *s++ = (gp+gq)/2;
#endif
#if 0
            *s++ = 0xff;
            *s++ = (gp+gq)/2;
            *s++ = (gp+gq)/2;
#endif
            s++;
        }
        p+=4;
        q+=4;
    }
}

#endif


#if 0
static int img_his_radix=-1;
static sentry_imgrec *img_his=NULL;
static int img_his_pos=0;

void
show_image_history(FILE *fp)
{
    int i, u;

    for(i=0;i<img_his_radix;i++) {
        fprintf(fp, "%2d: %p %6d/%6d %3dx%3d\n",
            i,
            &img_his[i], img_his[i].length, img_his[i].body_len,
            img_his[i].width, img_his[i].height);
    }
#if 0
    for(u=img_his_pos;u<img_his_pos+img_his_radix;u++) {
        i = u%img_his_radix;
        fprintf(fp, "%2d: %c %2d : %p %6d/%6d %3dx%3d\n",
            u, (i==img_his_pos ? '*': '-'),  i,
            &img_his[i], img_his[i].length, img_his[i].body_len,
            img_his[i].width, img_his[i].height);
    }
#endif
}

init_image_history(int n)
{
    int i;
    int w;

    img_his_radix = n;
    img_his = (sentry_imgrec*) malloc(sizeof(sentry_imgrec)*img_his_radix);

    for(i=0;i<img_his_radix;i++) {
        w = (int)assoc_p(&buflen_xrgb_alist, MAX_DISPLAY_SIZE);
        img_his[i].body = (uint8_t*)malloc(w);
        img_his[i].body_len = w;
        img_his[i].length = -1;
    }

    if(debug) {
        fprintf(stderr, "img_his_radix %d\n", img_his_radix);
    }
}


int
RefreshHist(int n)
{
    sentry_imgrec *cur;
    int o;
    sentry_imgrec *last;
    int wid, ht;
    char msg[BUFSIZ];
    struct tm htm;
    char data_body[MAX_IMG_LEN];
    GC cur_gc;

    void DoDiffExpose(Widget w, XEvent *ev, String *s, Cardinal *c);

    if(n<0 || n>=img_his_radix) {
        return -1;
    }

    cur = &img_his[n];
    if(!cur) {
        return -1;
    }

    o = (img_his_radix + n - 1) % img_his_radix;
    last = &img_his[o];
    if(!last) {
        return -1;
    }
#if 0
    fprintf(stderr, "n %d, o %d\n", n, o);
#endif
    
    wid = cur->width;
    ht  = cur->height;

    if(cur->length>0) {

        localtime_r(&cur->recv_sec, &htm);
        sprintf(msg, "%02d %2d-%02d %02d:%02d:%02d.%02d",
            n, htm.tm_mon+1, htm.tm_mday,
            htm.tm_hour, htm.tm_min, htm.tm_sec, cur->recv_usec/10000);
        RefreshMsg(msg);

        cur_gc = XtGetGC(pwin2, NULL, NULL);

        if(ghost_mode && last->length>0) {
#if 0
fprintf(stderr,  "#corder %d output %d SEN_C_XRGB %d\n",
    cur->corder, output_corder, SEN_C_XRGB);
#endif
            if(output_corder=SEN_C_XRGB) {
                sub_bgrx(cur->body, last->body, data_body, wid, ht);
            }
            else
            if(output_corder=SEN_C_RGB) {
                sub_rgb(cur->body, last->body, data_body, wid, ht);
            }

            put_image(XtDisplay(pwin2), xp_foil_visual,
                xp_foil_depth, cur_gc, diff_pix, wid, ht, data_body);
        }
        else {
            put_image(XtDisplay(pwin2), xp_foil_visual,
                xp_foil_depth, cur_gc, diff_pix, wid, ht, cur->body);
        }
        XtReleaseGC(pwin2, cur_gc);

        DoDiffExpose(NULL, NULL, NULL, NULL);
    }
    else {
        fprintf(stderr, "#show n %d, length %d\n", n, cur->length);
    }
}


int
store_image_history(sentry_imgrec *cur, sentry_imgrec *last)
{
    char data_bgrx[MAX_IMG_LEN];
    char data_rgb[MAX_IMG_LEN];
    int wid, ht;
    int cnum;
    void DoDiffExpose(Widget w, XEvent *ev, String *s, Cardinal *c);
    
    wid = cur->width;
    ht  = cur->height;

    if(debug)
        show_image_history(stderr);

    cnum = img_his_pos;

    if(debug)
        fprintf(stderr, "#store to %d\n", cnum);

    copyattr_img(&img_his[img_his_pos], cur);
    memcpy(img_his[img_his_pos].body, cur->body, cur->length);
    img_his[img_his_pos].length = cur->length;

    img_his_pos = (img_his_pos+1) % img_his_radix;

    if(debug)
        show_image_history(stderr);

    RefreshHist(cnum);
}

#endif

int diff_capacity = 0x30;
int cmp_type = SEN_M_NONE;


RefreshMsg(char *msg)
{
    Dimension wid;

#if 0
    XtVaGetValues(pwinlabel, XtNwidth, &wid, NULL);
#endif
    XtVaSetValues(pwinmsg,
        XtNlabel, msg,
        XtNwidth, wid,
        NULL);
}

static int img_his_view = 0;


void
DoResize(Widget w, XEvent *ev, String *s, Cardinal *c)
{
    refresh_flag = 1;
}


void
ExposePanel(Widget w, XEvent *ev, String *s, Cardinal *c)
{
    GC tmp_gc;
    XExposeEvent *xee;

    xee = (XExposeEvent*)ev;
    
    if(!ev)
        fprintf(stderr, "ExposePanel:\n");
    else {
        fprintf(stderr, "ExposePanel: send %d, geo %dx%d+%d+%d, more %d\n",
            xee->send_event,
            xee->width, xee->height, xee->x, xee->y, xee->count);
    }
    

    if(!bg_fig) {
        fprintf(stderr, "bg_fig is empty\n");
        return;
    }

    tmp_gc = XtGetGC(pcpanel, NULL, NULL);
    
#if 0
    XSetForeground(XtDisplay(pcpanel), tmp_gc, 0x000000);
    DrawFigure(XtDisplay(pcpanel), bg_pix, tmp_gc,
        xp_foil_visual, xp_foil_depth, xp_foil_cmap,
        0, 0, &bg_wid, &bg_ht, bg_fig, 0, 0, 0);
#endif

    fprintf(stderr, "copy background pixmap with %dx%d\n",
        bg_wid, bg_ht);

    XCopyArea(XtDisplay(pcpanel), bg_pix, XtWindow(pcpanel),
        tmp_gc, 0, 0, bg_wid, bg_ht, 0, 0);

    XtReleaseGC(pcpanel, tmp_gc);
}



#if 0
static String pwin_trans="#override\n\
";

static String pwin2_trans="#override\n\
    <Expose>:           diff-expose()\n\
    <Key>Up:            diff-rotate-up()\n\
    <Key>Down:          diff-rotate-down()\n\
    <Key>Left:          diff-rotate-down()\n\
    <Key>Right:         diff-rotate-up()\n\
";

static String common_trans="#override\n\
    <Configure>:        live-resize()\n\
";

static XtActionsRec pwin_action_list[] = {
    {"live-expose",     DoLiveExpose},
    {"live-resize",     DoResize},
    {"diff-expose",     DoDiffExpose},
    {"diff-rotate-up",  DoDiffRotateUp},
    {"diff-rotate-down",DoDiffRotateDown},
};
#endif

static String common_trans="#override\n\
";

static String panel_trans="#override\n\
    <Expose>:           expose-panel()\n\
";

static XtActionsRec pwin_action_list[] = {
    {"expose-panel", ExposePanel},
};


void
quit_sentry(Widget w, XtPointer client_data, XtPointer call_data)
{
    exit(0);
}

void
toggle_mute(Widget w, XtPointer client_data, XtPointer call_data)
{
    beep_mute = 1 - beep_mute;
    fprintf(stderr, "beep_mute %d\n", beep_mute);
    fflush(stderr);

    if(beep_mute) {
        XtVaSetValues(bmute, XtNforeground, black, NULL);
        XtVaSetValues(bmute, XtNbackground, white, NULL);
    }
    else {
        XtVaSetValues(bmute, XtNforeground, white, NULL);
        XtVaSetValues(bmute, XtNbackground, black, NULL);
    }
}

void
toggle_ghost(Widget w, XtPointer client_data, XtPointer call_data)
{
}


void
usage()
{
        printf("sentry - image viewer of sentry system\n\
usage: sentry [options]\n\
\n\
option:\n\
    -help\n\
    -servername host    default host is localhost\n\
    -serverport port    \n\
    -dataport port      \n\
    -reqrate num        frame rate within per 100sec\n\
    -format format      frame format (e.g., NONE, JPEG, PNG, etc.)\n\
    -shmkey key         \n\
    -debug num          debug leve\n\
\n\
note:\n\
  * you should change dataport when you start multipul program\n\
  * frame rate does not changed by this program, if you access\n\
    with shared memory\n\
\n\
example:\n\
  %% sentry -severname camera1 -dataport 723\n\
  %% sentry -shmkey 129\n\
");
}


store_color(Pixel *sa, int r, int g, int b)
{
    if(output_corder==SEN_C_XRGB) {
        *sa  = r*0x010000;
        *sa += g*0x000100;
        *sa += b*0x000001;
    }
    else 
    if(output_corder==SEN_C_XBGR) {
        *sa  = b*0x010000;
        *sa += g*0x000100;
        *sa += r*0x000001;
    }
    
}



int
recv_image_multi(int datafd, subrecord_t **recvlist, char **buff,
    sentry_shmrec *info, sentry_imgrec *dest);

void
update_cwin(caddr_t client_data, int *fd, XtInputId *id)
{
    cnode_t *cn;
    char msg[BUFSIZ];
    int nread;
    sentry_shmrec info;
    int chk;
    char *dummy;
    struct tm htm;
    GC tmp_gc;

    cn = (cnode_t*)client_data;

    if(!cn) {
        return;
    }

#if 0
    fprintf(stderr, "cn %p: host %s, id %p %d : x,y = %d,%d\n",
        cn, cn->serv_host, id, *id, cn->x, cn->y);
#endif

    tmp_gc = XtGetGC(pcpanel, NULL, NULL);
    
    chk = recv_image_multi(*fd, &(cn->recbuf),
            &dummy, &info, &cn->inimg);
    
    SenDrawImage_xy(pcpanel, cn->x, cn->y, &cn->inimg, &cn->outimg);

    if(debug) {
    XDrawLine(XtDisplay(pcpanel), XtWindow(pcpanel), tmp_gc,
        cn->x-2, cn->y, cn->x+2, cn->y);
    XDrawLine(XtDisplay(pcpanel), XtWindow(pcpanel), tmp_gc,
        cn->x, cn->y-2, cn->x, cn->y+2);
    }

    XtReleaseGC(pcpanel, tmp_gc);

#if 0
    localtime_r(&(cn->inimg.recv_sec), &htm);
    sprintf(msg, "%2d-%02d %02d:%02d:%02d.%02d",
        htm.tm_mon+1, htm.tm_mday,
        htm.tm_hour, htm.tm_min, htm.tm_sec, cn->inimg.recv_usec/10000);

    XtVaSetValues(cn->Wmsg,
        XtNlabel, msg,
        NULL);
#endif

} 

void
project_cam(int an, int *dx, int *dy, int gx, int gy)
{
    int w, h;
    w = sc_wid;
    h = sc_ht;
    switch(an) {

    case SEN_A_LT:  *dx = gx;       *dy = gy;       break;
    case SEN_A_L:   *dx = gx;       *dy = gy-h/2;   break;
    case SEN_A_LB:  *dx = gx;       *dy = gy-h;     break;

    case SEN_A_RT:  *dx = gx-w;     *dy = gy;       break;
    case SEN_A_R:   *dx = gx-w;     *dy = gy-h/2;   break;
/*
    case SEN_A_RB:  *dx = gx-w;     *dy = gy-h;     break;
*/

    case SEN_A_T:   *dx = gx-w/2;   *dy = gy;       break;
    case SEN_A_C:   *dx = gx-w/2;   *dy = gy-h/2;   break;
    case SEN_A_B:   *dx = gx-w/2;   *dy = gy-h;     break;

    default:
                    *dx = gx;       *dy = gy;       break;
    }
}

void
split_cam_attr(char *src, char *d_name, char *d_anchor)
{
    int c;
    char *p;
    char *q;
    
    c = 0;
    p = src;
    p += 6;
#if 0
    fprintf(stderr, "src '%s', p '%s'\n", src, p);
    fflush(stderr);
#endif
    
    q = d_name;
    while(*p) {
        if(*p==':') {
            break;
        }
        *q++ = *p++;
    }
    *q = '\0';

    p++;
    q = d_anchor;
    while(*p) {
        if(*p==':') {
            break;
        }
        *q++ = *p++;
    }
    *q = '\0';
}

bind_cam(xp_fig_elem_t *fl)
{
    int i;
    fig_list_t *cur;
    fig_text_t *ftext;
    char n_str[BUFSIZ], an_str[BUFSIZ];
    int a;
    cnode_t *cn;

    i = 0;
    cur = fl->list;
    while(cur) {
        if(FIG_OBJTYPE(cur->cont)==FIG_TEXT) {
            ftext = (fig_text_t*)cur->cont;
#if 0
            fprintf(stderr, " text %2d: %d,%d '%s'\n",
                i, ftext->pos.x, ftext->pos.y, ftext->value);
#endif

            if(ftext->value[0]
                && strncmp(ftext->value,"::sen:",5)==0) {
                split_cam_attr(ftext->value, n_str, an_str);
                a = assoc_n(anchor_short_alist, an_str);

                /* clear label contents */
                ftext->value[0] = '\0';

#if 0
                fprintf(stderr, "  name %s anchor %s %d\n",
                    n_str, an_str, a);
#endif
                if(a==-1) {
                    a = SEN_A_LT;
                }

                cn = find_cnode(n_str);
                if(cn) {
                    cn->rx      = ftext->pos.x;
                    cn->ry      = ftext->pos.y;
                    cn->anchor  = a;
                }

            }

            i++;
        }
        cur = cur->next;
    }
    
}



int
make_cwin(cnode_t *cn)
{
    int i;
    int c;
    static int last_port=-1;

#if 1
    if(cn->recv_port==last_port) {
        cn->recv_port = (cn->recv_port + rand())%(1L<<15);
    }
#endif
    
    c = 0;
    do {
        cn->data_fd = build_data_port(cn->recv_port);
        fprintf(stderr, "    port %d -> fd %d\n",
            cn->recv_port, cn->data_fd);
        if(cn->data_fd<0) {
#if 0
            cn->recv_port++;
#else
            cn->recv_port = (cn->recv_port + rand())%(1L<<15);
#endif
        }
        c++;
    } while(cn->data_fd<0 && c<10);
    last_port = cn->recv_port;


    cn->cntl_fd = build_ctrl_port(cn->serv_host, cn->serv_port);
    if (cn->cntl_fd < 0) {
        printf("fail setup control fd with port# %d (%d)\n",
            cn->serv_port, errno);
    }

#if 0
    show_cnode(stderr, "wake", &ncn);
#endif
    show_cnode(stderr, "wake", cn);


 {
    int rw, rh;
    int gx, gy;

    rw = bg_fig->lrx - bg_fig->ulx;
    rh = bg_fig->lry - bg_fig->uly;

    gx = (cn->rx-bg_fig->ulx)*bg_wid/rw;
    gy = (cn->ry-bg_fig->uly)*bg_ht/rh;

    project_cam(cn->anchor, &cn->x, &cn->y, gx, gy);
#if 0
    printf("hot-point %d,%d anchor %s (%d) projected %d, %d\n",
        gx, gy, 
        assoc_v(anchor_alist, cn->anchor), cn->anchor,
        cn->x, cn->y);
#endif
 }


#if 0
    XawFormDoLayout(pcpanel, TRUE);
#endif


 {
    sentry_protocol3_request_hdr req;
    int nwrite;
    int tmpiid;

/*
    tmpiid = XtAppAddInput(app_context, tmp->data_fd,
                (caddr_t)XtInputReadMask, eventical_update, (void*)tmp);
*/
    tmpiid = XtAppAddInput(app_context, cn->data_fd,
                (caddr_t)XtInputReadMask, update_cwin, (void*)cn);

    fprintf(stderr, "tmpiid %d with fd %d\n", tmpiid, cn->data_fd);

    req.hd_ver  = SENTRY_PROTOCOL_VERSION;
    req.type    = OPEN_UDP_UNICAST;
    req.port    = htons(cn->recv_port);
    req.size    = htonl(cn->size);
    req.zoom    = htonl(cn->zoom);
    req.color   = htonl(cn->color);
    req.format  = htonl(cn->format);
    req.rate    = htonl(cn->rate);

    nwrite = write(cn->cntl_fd, (void*)&req, sizeof(req));

 }


goto out;

    close(cn->cntl_fd);
    cn->cntl_fd = -1;

    close(cn->data_fd);
    cn->data_fd = -1;
    

out:
    return 0;
}


void
build_bgpix()
{
    GC tmp_gc;
    int i;
    cnode_t *cn;

    fprintf(stderr, "plot background figure in %dx%d\n",
        bg_wid, bg_ht);

    bg_pix = XCreatePixmap(XtDisplay(pcpanel),
        DefaultRootWindow(XtDisplay(pcpanel)),
        bg_wid, bg_ht, xp_foil_depth);

    tmp_gc = XtGetGC(pcpanel, NULL, NULL);
    XSetForeground(XtDisplay(pcpanel), tmp_gc, 0xffffff);
    XFillRectangle(XtDisplay(pcpanel), bg_pix, tmp_gc,
        0, 0, bg_wid, bg_ht);

    if(!bg_fig) {
        fprintf(stderr, "bg_fig is empty\n");
    }
    else {
        XSetForeground(XtDisplay(pcpanel), tmp_gc, 0x000000);
        DrawFigure(XtDisplay(pcpanel), bg_pix, tmp_gc,
            xp_foil_visual, xp_foil_depth, xp_foil_cmap,
            0, 0, &bg_wid, &bg_ht, bg_fig, 0, 0, 0);

        for(i=0;i<cnode_num;i++) {
            cn = &cnode_list[i];
            XDrawString(XtDisplay(pcpanel), bg_pix, tmp_gc,
                cn->x, cn->y, cn->name, strlen(cn->name));
            XFillRectangle(XtDisplay(pcpanel), bg_pix, tmp_gc,
                cn->x, cn->y, 160, 80);
        }

        fprintf(stderr, "copy background pixmap with %dx%d\n",
            bg_wid, bg_ht);

        XCopyArea(XtDisplay(pcpanel), bg_pix, XtWindow(pcpanel), 
            tmp_gc, 0, 0, bg_wid, bg_ht, 0, 0);

    }

    XtReleaseGC(pcpanel, tmp_gc);
}







#define MARG    (10)

int
split_args(char *line, char *arg[])
{
    register char *p, *q;
    int k, c;

#if 0 
    fprintf(stderr, "line :%s\n", line);
#endif
    p = line;
    while(*p && (*p==' ' || *p=='\t')) {
        p++;
    }

    k = 0;
    do {
        if(*p=='#'||*p=='\r'||*p=='\n') {
            break;
        }
#if 0 
 fprintf(stderr, "  arg %2d: %s", k, p);
#endif
        q = arg[k];
        while(*p && *p!=' ' && *p!='\t' && *p!='\r' && *p!='\n') {
            *q++ = *p++;
        }
        *q = '\0';
        k++;

        while(*p && (*p==' ' || *p=='\t')) {
            p++;
        }

    } while(*p);

    return k;
}

int
read_config(char *cfname)
{
    FILE *cfp;
    char line[BUFSIZ];
    char argbuf[MARG*(BUFSIZ+1)];
    char *arg[MARG];
    int n;
    int i;
    cnode_t *cn;

    char *xbg_fname;
    int xbg_wid = -1, xbg_ht = -1;
    int xsc_wid = -1, xsc_ht = -1;

    cfp = fopen(cfname, "r");
    if(cfp==NULL)
        return -1;

    for(i=0;i<MARG;i++) {
        arg[i] = &argbuf[i*(BUFSIZ+1)];
    }

    while(fgets(line, BUFSIZ, cfp)) {
#if 0
        fprintf(stderr, "line>> %s", line);
#endif
        for(i=0;i<MARG;i++) {
            arg[i][0] = '\0';
        }

        n = split_args(line, arg);
        if(n<=0)
            continue;
    
#if 0
        fprintf(stderr, "line>> %s", line);
        for(i=0;i<n;i++){
            fprintf(stderr, "    %2d: '%s'\n", i, arg[i]);
        }
#endif

        if(strcasecmp(arg[0], "background")==0) {
            if(n>1) xbg_fname   = strdup(arg[1]);
            if(n>2) xbg_wid     = atoi(arg[2]);
            if(n>3) xbg_ht      = atoi(arg[3]);

#if 0
fprintf(stderr, "@fig '%s' %dx%d\n",
    xbg_fname, xbg_wid, xbg_ht);
#endif
        }
        else
        if(strcasecmp(arg[0], "scenesize")==0) {
            if(n>1) xsc_wid  = atoi(arg[1]);
            if(n>2) xsc_ht   = atoi(arg[2]);
#if 0
fprintf(stderr, "@scene %dx%d\n",
    xsc_wid, xsc_ht);
#endif
        }
        else
        if(strcasecmp(arg[0], "camera")==0) {
            cn = &cnode_list[cnode_num];
            setdefault_cnode(cn);
            if(n>1) cn->name = strdup(arg[1]);
            if(n>2) cn->serv_host = strdup(arg[2]);
            if(n>3) cn->serv_port = atoi(arg[3]);

#if 0
fprintf(stderr, "@cam '%s' '%s' %d\n",
    cn->name, cn->serv_host, cn->serv_port);
#endif
            cnode_num++;
        }

fflush(stderr);
    }

    fclose(cfp);

    /*
     * verify values
    */
    if(xbg_wid<0)    xbg_wid  = 500;
    if(xbg_ht<0)     xbg_ht   = 500;

    if(xsc_wid<0)    xsc_wid  = 160;
    else {
        switch(xsc_wid) {
        case 640:   xsc_wid = 640;   xsc_ht = 480;    break;
        case 320:   xsc_wid = 320;   xsc_ht = 240;    break;
        case 212:   xsc_wid = 212;   xsc_ht = 160;    break;
        case 160:   xsc_wid = 160;   xsc_ht = 120;    break;
        default:
                    xsc_wid = 160;   xsc_ht = 120;    break;
        }
    }

#if 0
    fprintf(stderr, "scene %dx%d\n", xsc_wid, xsc_ht);
#endif


    bg_fname    = xbg_fname;
    bg_wid      = xbg_wid;
    bg_ht       = xbg_ht;
    sc_wid      = xsc_wid;
    sc_ht       = xsc_ht;

    bg_fig = read_fig_file(bg_fname);
#if 0
    printf(" %d,%d-%d,%d\n",
        bg_fig->ulx, bg_fig->uly, bg_fig->lrx, bg_fig->lry);
#endif

    if(bg_fig)
        bind_cam(bg_fig);

}


#if 0
fd_set  InputFdMask;
#endif
subrecord_t *recvlist=NULL;


/* MAIN */

int
main(int argc, char **argv)
{
    XGCValues values;
    int valuemask;
    int tv;
    int nhistory;
    char *conf_fname;
    int i;


#if 0
    rtr_set_global_param(RTR_P_VERBOSE, NULL);
    if(debug) {
        rtr_set_global_param(RTR_P_DEBUG, NULL);
    }
#endif
    rtr_set_global_param(RTR_P_LENFREELIST, 512);
    rtr_set_global_param(RTR_P_QTIME, 2);
    rtr_initmem();

    toplevel = XtVaAppInitialize(
        &app_context,                   /* Application context */
        "Xp",                           /* Application class */
        options, XtNumber(options),     /* option list */
        &argc, argv,                    /* command line args */
        fallback_resources,             /* for missing app-defaults file */
        XtNtitle,       "XP",
        XtNiconName,    "XP",
        NULL);                          /* terminate varargs list */

    XtVaGetApplicationResources(toplevel,
        (XtPointer) &sentry_resource, resources, XtNumber(resources), NULL);

    XtAppAddActions(app_context, main_action_list, XtNumber(main_action_list));

    if(argc>=2) {
        fprintf(stderr, "argv[1] '%s'\n", argv[1]);
        conf_fname = strdup(argv[1]);
    }
    else {
        conf_fname = strdup("smap.conf");
    }
    read_config(conf_fname);


#if 0
    nhistory = atoi(sentry_resource.img_history);
    init_image_history(nhistory);
#endif

    tv = assoc_n(matching_type_alist, sentry_resource.compare_func);
    if(debug) {
        fprintf(stderr, "compare type %d %x '%s' <= '%s'\n",
            tv, tv, assoc_v(matching_type_alist, tv),
            sentry_resource.compare_func);
    }
    cmp_type = tv;

    if(debug) {
        fprintf(stderr, "severname '%s'\n", sentry_resource.server_name);
    }

    tv = assoc_n(format_alist, sentry_resource.req_format);
    server_format = tv;

    tv = assoc_n(size_alist, sentry_resource.req_size);
    server_size = tv;

    if(debug>=10) {
    fprintf(stderr, "format %d, '%s' => '%s'\n",    
        tv, sentry_resource.req_format, assoc_v(format_alist, tv));
    fprintf(stderr, "size %d, '%s' => '%s'\n",    
        tv, sentry_resource.req_size, assoc_v(size_alist, tv));
    }

    if(show_help) {
        usage();
        exit(0);
    }

    fflush(stdout);
    fflush(stderr);
    SetupVisual(XtDisplay(toplevel), fvisual, svisual);
    fflush(stdout);
    fflush(stderr);

    output_format = SEN_F_RGB;
#if 0
    /* corder is solved by SetupVisual() */
    output_corder = SEN_F_RGB;
#endif

    store_color(&black, 0x00, 0x00, 0x00);
    store_color(&white, 0xff, 0xff, 0xff);

    diff_pix =  XCreatePixmap(XtDisplay(toplevel),
                DefaultRootWindow(XtDisplay(toplevel)),
                640, 480, xp_foil_depth);

    pshell = XtVaAppCreateShell(
        "sentry", "Sentry",
        topLevelShellWidgetClass, XtDisplay(toplevel),
        XtNvisual,  xp_foil_visual,
        XtNdepth,   xp_foil_depth,
        XtNcolormap, xp_foil_cmap,
        NULL);

    pbase = XtVaCreateManagedWidget(
        "base", formWidgetClass, pshell,
        XtNvisual,  xp_foil_visual,
        XtNdepth,   xp_foil_depth,
        XtNcolormap, xp_foil_cmap,
        XtNresizable, TRUE,
        NULL);

    plabel = XtVaCreateManagedWidget(
        "label", labelWidgetClass, pbase,
        XtNlabel, "multiple cameras",
        XtNjustify, XtJustifyCenter,
        XtNborderWidth, 0,
        XtNresize,  True,
        NULL);

    pcmd = XtVaCreateManagedWidget(
        "cmd", formWidgetClass, pbase,
        XtNfromVert, plabel,
        XtNresizable, FALSE,
        XtNborderWidth, 0,
        NULL);

    bquit = XtVaCreateManagedWidget(
        "quit", commandWidgetClass, pcmd,
        NULL);
    XtAddCallback(bquit, XtNcallback, quit_sentry, NULL);

#if 0
    bmute = XtVaCreateManagedWidget(
        "mute", commandWidgetClass, pcmd,
        XtNfromHoriz, bquit,
        NULL);
    XtAddCallback(bmute, XtNcallback, toggle_mute, NULL);

    bghost = XtVaCreateManagedWidget(
        "ghost", commandWidgetClass, pcmd,
        XtNfromHoriz, bmute,
        NULL);
    XtAddCallback(bghost, XtNcallback, toggle_ghost, NULL);
#endif

    pcpanel = XtVaCreateManagedWidget(
        "cpanel", boxWidgetClass, pbase,
#if 0
        "cpanel", formWidgetClass, pbase,
        XtNorientation, XtEhorizontal,
#endif
        XtNfromVert, pcmd,
        XtNvisual,  xp_foil_visual,
        XtNdepth,   xp_foil_depth,
        XtNcolormap,xp_foil_cmap,
        XtNwidth,   bg_wid,
        XtNheight,  bg_ht,

        XtNresizable,   TRUE,
        NULL);

    pwinmsg = XtVaCreateManagedWidget(
        "difflabel", labelWidgetClass, pbase,
        XtNlabel, "",
        XtNfromVert, pcpanel,
        XtNborderWidth, 0,
        XtNwidth,       160,
        XtNresize,      1,
        NULL);

    XtRealizeWidget(pshell);

    XtAppAddActions(app_context,
        pwin_action_list, XtNumber(pwin_action_list));
    XtOverrideTranslations(pshell,
        XtParseTranslationTable(common_trans));
    XtOverrideTranslations(pcpanel,
        XtParseTranslationTable(panel_trans));

    XtRealizeWidget(pshell);

    for(i=0;i<cnode_num;i++) {
        make_cwin(&cnode_list[i]);
    }

    build_bgpix();
PASS;
    XtAppMainLoop(app_context);
}
