/*
 * get_data.c - get data (image header and body)
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <string.h>
#ifdef __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#include <errno.h>

#include "hook.h"
#include "alist.h"
#include "sentry.h"
#include "img.h"
#include "pool.h"

#ifndef K
#define K(x)    ((x)*1024)
#define M(x)    ((x)*1024*1024)
#endif


#ifdef  HAVE_TYPEDEF_FD_SET
typedef fd_set          fdset_t;
#else
typedef struct fd_set   fdset_t;
#endif  /* HAVE_FDSET_T */


/*
 * isreadable - check specified FD is readable or not.
 *      if sec<0        wait with blocking
 *
 *
 * return:
 *      1   readable
 *      0   not readable
 *     -1   error
 *
 */

int
isreadable_via_select(int fd, int sec, int usec)
{
    fdset_t fdset, rfdset;
    struct timeval timeout;
    struct timeval *ptimeout;
    int chk;

    FD_ZERO(&fdset);
    FD_SET(fd, &fdset);

    if(sec<0) {
        ptimeout = NULL;
    }
    else {
        ptimeout = &timeout;

        timeout.tv_sec  =  sec;
        if(usec<0) {
            timeout.tv_usec = 0;
        }
        else {
            timeout.tv_usec = usec;
        }
    }

retry:
    rfdset = fdset;
    chk = select(fd+1, &rfdset, NULL, NULL, ptimeout);

#if 0
    dump_fdset(fd+1, &fdset, &rfdset, NULL, NULL);
#endif
    if(chk<0) {
        if(errno==EINTR) {
            goto retry;
        }
        else {
            Error(" #isreadable:fd# %d, errno %d# ", fd, errno);
            return -1;
        }
    }

    if(FD_ISSET(fd, &rfdset)) {
#if 0
        Trace("readable ");
#endif
        return 1;
    }
    else {
#if 0
        Trace("unreadable ");
#endif
        return 0;
    }
}


int 
bind_socket(int sockfd, int port) 
{
  struct sockaddr_in serv_addr;

  bzero((char *)&serv_addr, sizeof(serv_addr));
  serv_addr.sin_family = AF_INET;
  serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  serv_addr.sin_port = htons(port);

  if (bind(sockfd, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) < 0) {
    fprintf(stderr, "server: cannot bind port %d. (%d)\n", port, errno);
    close(sockfd);
    return -1;
  }
  return 0;
}

int 
build_listen_port(int port, int type)
{
  int sockfd;
  int flag=1;

  while (1) {
    if ((sockfd = socket(AF_INET, type, 0)) < 0) {
      fprintf(stderr, "server: cannot make new TCP socket.\n");
      return -1;
    }
    
    if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, (char *)&flag, sizeof(flag)) < 0) {
      fprintf(stderr, "server: cannot set sockopt REUSEADDR.\n");
      /* return -1; */
    }

    if (bind_socket(sockfd, port) < 0) {
      fprintf(stderr, "build_listen_port: fail bind fd# %d, port# %d\n",
            sockfd, port);
      close(sockfd);
      if(errno==EACCES||errno==EBADF||errno==ENOTSOCK) {
        return -1;
      }

      fprintf(stderr, "server: please wait several seconds. (%d)\n", errno);
      sleep(3);
      continue;
    }
    break;
  }

  listen(sockfd, -1);
#if 0
  if (verb)
    fprintf(stderr,
        "server: new socket# %d, type# %d, port#, %d.\n",
        sockfd, type, port);
#endif

  return sockfd;
}

int
build_udp_port(int port) 
{
  int sockfd;
  int flag=1;
  int flaglen = sizeof(flag);

  if ((sockfd = socket(PF_INET, SOCK_DGRAM, 0)) < 0) {
    fprintf(stderr, "server: cannot make new UDP socket.\n");
    return -1;
  }

  if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR,
                 (char *)&flag, sizeof(flag)) < 0) {
    fprintf(stderr, "server: cannot setsockopt REUSEADDR\n");
    return -1;
  }


  /*
   * set SNDBUF size:
   *   challenge 16M, 8M, ...., 16K, 8K.
   */
  flag = M(16);
  while(flag >= K(8)) {
    if (setsockopt(sockfd, SOL_SOCKET, SO_SNDBUF,
                 (char *)&flag, flaglen) < 0) {
#if 0
      if (debug) 
    fprintf(stderr, "server: fail set SNDBUF to %9d bytes.\n", flag);
#endif
      flag = flag / 2 ;
    }
    else {
      fprintf(stderr, "server: success set SNDBUF size [%9d bytes].\n", flag);
      break;
    }
  }

#if 0
  if (verb)
    if (getsockopt(sockfd, SOL_SOCKET, SO_SNDBUF,
                   (char *)&flag, &flaglen) < 0)
      fprintf(stderr, "server: cannot get SNDBUF size.\n");
    else
      fprintf(stderr, "server: SNDBUF size = %d\n", flag);
#endif
  

  if (bind_socket(sockfd, port) < 0) {
    fprintf(stderr, "build_udp_port: fail bind fd# %d, port# %d\n",
        sockfd, port);
    close(sockfd);
    return -1;
  }

#if 0
  if (verb)
    fprintf(stderr,
        "server: new udp socket# %d, port# %d.\n",
        sockfd, port);
#endif

  return sockfd;
}



/*
 * connect_host_addr(): create a connect-mode socket.
 *   addr: network byte order address
 *   port: network byte order port# 
 *   type: SOCK_DGRAM/SOCK_STREAM, now only supported SOCK_DGRAM.
 *   return value: socket discripter
 */
int
connect_host_addr(struct in_addr addr, int port, int type)
{
  int sockfd;
  struct sockaddr_in peer;

  switch(type) {
  case SOCK_DGRAM:
    if ((sockfd = build_udp_port(0)) < 0) {
        fprintf(stderr, "cannot bind UDP port\n");
        return -1;
    }
    break;
  case SOCK_STREAM:
    sockfd = socket(AF_INET, SOCK_STREAM, 0);
    if(sockfd<0) {
        fprintf(stderr, "fail socket (%d)\n", errno);
        return -1;
    }
    break;
  default:
    return -1;
  }

  peer.sin_family = AF_INET;
  peer.sin_addr.s_addr = addr.s_addr;
  peer.sin_port = htons(port);
  if (connect(sockfd, (struct sockaddr *)&peer, sizeof(peer)) < 0) {
    fprintf(stderr, "connect_host_addr: fail to connect to %s:%d<%d> type %d\n",
        inet_ntoa(peer.sin_addr.s_addr),
        ntohs(peer.sin_port), port, type);
    close(sockfd);
    return -1;
  }

  return sockfd;
}

int
connect_host(char *hostname, int port, int type)
{
    int sockfd;
    struct sockaddr_in addr;
    struct hostent *hp;

    if ((hp = gethostbyname(hostname)) == NULL) {
        fprintf(stderr, "fail gethostbyname '%s' (%d)\n",
            hostname, errno);
        return -1;
    }
                    
    memcpy(&addr.sin_addr, hp->h_addr, hp->h_length);
    sockfd = connect_host_addr(addr.sin_addr, port, type);
    if(sockfd<0) {
        fprintf(stderr, "cannot connect '%s' %d\n", hostname, port);
    }

    return sockfd;
}


#ifdef __STDC__
int fdprintf(const int fd, const char *fmt, ...)
#else
int fdprintf(va_alist)
va_dcl
#endif
{
    va_list args;
#ifndef __STDC__
    int fd;
    char *fmt;
#endif
    char buffer[BUFSIZ];
    int nwrite;

#ifdef __STDC__
    va_start(args,fmt);
#else
    va_start(args);
    fd  = va_arg(args, int);
    fmt = va_arg(args, char *);
#endif

#ifdef HAVE_VSNPRINTF
    vsnprintf(buffer, BUFSIZ, fmt, args);
#else
    vsprintf(buffer, fmt, args);
#endif
#if 0
    Trace("<<<%s", buffer);
#endif
try:
    nwrite = write(fd, buffer, strlen(buffer));
#if 0  
    if(pipe_broken) {
        pipe_broken = 0;
    }
#endif
    if(nwrite<0) {
        if(errno==EINTR) {
            goto try;
        }
        else {
            Trace("fdprintf: fail write() fd# %d (%d)\n", fd, errno);
        }
    }

    va_end(args);

    return nwrite;
}

