/*==
 ==*/
/*
 * ssf - sentry stream filter program
 *      by k-chinen@is.aist-nara.ac.jp, 2003.
 * $Id$
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <signal.h>
#include <syslog.h>
#ifdef TIME_WITH_SYS_TIME
 #include <sys/time.h>
 #include <time.h>
#else
 #ifdef HAVE_SYS_TIME_H
  #include <sys/time.h>
 #else
  #include <time.h>
 #endif
#endif
#include <errno.h>

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/signal.h>
#include <sys/wait.h>
#include <time.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#ifdef USE_SYSLOG
#include <syslog.h>
#endif
#include <errno.h>

#include "alist.h"
#include "sentry.h"
#include "server.h"

#include "alist.h"
#include "sentry.h"
#include "img.h"
#include "pool.h"
#include "client.h"

#include "video.h"

#include "rtr.h"
#include "sf.h"

int video_sync_type = SENV_SYNC_UNDEF;

#ifdef PASS
#undef PASS
#endif
#if 0 
#define PASS \
    fprintf(stderr,"PASS %s:%d\n", __FILE__, __LINE__); fflush(stderr)
#else
#define PASS    (void)1
#endif


typedef struct {
    char *hostname;
    char recept_conn_id[BUFSIZ];
    int recept_fd;
    int recept_port;

    int recv_fd;
    int recv_port;

    sentry_imgrec *imgbuf;

    int recv_count;
    int recv_lasttime;
} upnode_t;


#ifdef USE_LOCKFILE
char    lockfile[] = "/var/tmp/sentryc.pid";
#endif

int write_out(char *prefix, int pathtype, sentry_imgrec *img);
int isreadable_via_select(int fd, int sec, int usec);


#ifndef DEFAULT_PREFIX
#define DEFAULT_PREFIX "/var/tmp/sen_bank"
#endif
#ifndef DEFAULT_PATHTYPE
#define DEFAULT_PATHTYPE (1)
#endif

int  save_file_flag = 0;
char prefix[BUFSIZ] = DEFAULT_PREFIX;
int  pathtype = DEFAULT_PATHTYPE;



#ifndef TMP_IMGBUF_LEN
#define TMP_IMGBUF_LEN  (384*1024)
#endif


void
init_sentryc()
{
#ifdef USE_LOCKFILE
  int chk;
  FILE *fp;
  char mesg[80];

  if (debug)
    fprintf(stderr, "init_sentryc: lockfile=%s\n", lockfile);

  while (1) {
    chk = access(lockfile, F_OK);
    if (chk < 0) break;
    
    if (errno == EINTR)
      continue;

    sprintf(mesg, "init_sentryc: locckfile(%s) is found", lockfile);
    perror(mesg);
    exit (1);
  }

  if ((fp = fopen(lockfile, "w")) == NULL) {
    sprintf(mesg, "init_sentryc: cannot open lockfile(%s)", lockfile);
    perror(mesg);
    exit (1);
  }

  sprintf(mesg, "%d\n", getpid());
  fprintf(stderr, "pid# %d\n", getpid());
  
  if (fputs(mesg, fp) == EOF) {
    perror("init_sentryc: cannot fputs()");
    exit (1);
  }
  fclose(fp);
#endif /* USE_LOCKFILE */
}
  
void
goodbye(void)
{
    if(debug)
        fprintf(stderr, "goodbye\n");

#ifdef DELETE_IPC
    delete_pool(sentry_shmid);
#endif

#if 0
    syslog(LOG_NOTICE, "sentryc closed key=%d (%#x)\n",
        sentry_shmkey, sentry_shmkey);
#endif
    if(use_shm_flag) {
        syslog(LOG_NOTICE, "sentryc closed key=%d (%#x)",
            sentry_shmkey, sentry_shmkey);
    }
    else {
        syslog(LOG_NOTICE, "sentryc exit");
    }
    closelog();

#ifdef USE_LOCKFILE
    unlink(lockfile);
#endif
}


int *bgbuf;
int last_bg_ht  = -1;
int last_bg_wid = -1;
int bg_count    = -1;

int
memory_bg(int op, sentry_imgrec *inimg, sentry_imgrec *outimg)
{
    int en = 3;
    int len;
    register unsigned char *p;
    register unsigned char *q;
    register unsigned int *u;
    unsigned char *z;

    len = inimg->width*inimg->height*en;
    if(bg_count<0) {
        bgbuf = (unsigned int*)malloc(len*sizeof(int));
        bg_count = 0;
        memset(bgbuf, 0, len*sizeof(int));
        fprintf(stderr, "init bgbuf first\n");
    }
    else {
        if(inimg->width!=last_bg_wid || inimg->height!=last_bg_ht) {
            free(bgbuf);
            bgbuf = (unsigned int*)malloc(len*sizeof(int));
            bg_count = 0;
            memset(bgbuf, 0, len*sizeof(int));
            fprintf(stderr, "reinit bgbuf size changed\n");
        }
    }

    
    if(bg_count%op==op-1) {

        p = inimg->body;
        u = bgbuf;
        z = p+len;
        while(p<z) {
            /* p used only for counter */
            *u = *u/2;
            p++;
            u++;
        }
        fprintf(stderr, "clear at count %d\n", bg_count);

        bg_count = bg_count/2;
    }

    p = inimg->body;
    q = outimg->body;
    u = bgbuf;
    z = p+len;
    while(p<z) {
        *u = *u + *p;
        if(*u/(bg_count+1)> 0xff) {
            *q = 0xff;
        }
        else {
            *q = *u / (bg_count+1);
        }
        p++;
        q++;
        u++;
    }

    last_bg_wid = inimg->width;
    last_bg_ht  = inimg->height;
    bg_count++;

    if(bg_count%100==0) {
        fprintf(stderr, "count %d\n", bg_count);
    }
#if 0
    copyattr_img(outimg, inimg);
#endif
    outimg->recv_sec = inimg->recv_sec;
    outimg->recv_usec = inimg->recv_usec;

    return 0;
}


static sentry_shmrec bk_info;
static sentry_imgrec bk_inframe;
static unsigned char *inimgbuf=NULL;
static sentry_imgrec bk_drframe[2];
static unsigned char *drimgbuf0=NULL;
static unsigned char *drimgbuf1=NULL;
static sentry_imgrec bk_outframe;
static unsigned char *outimgbuf=NULL;
static sentry_imgrec bk_cookframe;
static unsigned char *cookimgbuf=NULL;
static sentry_imgrec bk_recframe;
static unsigned char *recimgbuf=NULL;


extern apair_t buflen_xrgb_alist;


alloc_inimgbufs()
{
    inimgbuf = (unsigned char*)malloc(TMP_IMGBUF_LEN);
    bk_inframe.body = inimgbuf;
    bk_inframe.body_len = TMP_IMGBUF_LEN;
}

alloc_outimgbufs()
{
    outimgbuf = (unsigned char*)malloc(TMP_IMGBUF_LEN);
    bk_outframe.body = outimgbuf;
    bk_outframe.body_len = TMP_IMGBUF_LEN;
}

alloc_cookimgbufs()
{
    cookimgbuf = (unsigned char*)malloc(TMP_IMGBUF_LEN);
    bk_cookframe.body = cookimgbuf;
    bk_cookframe.body_len = TMP_IMGBUF_LEN;
}

alloc_recimgbufs()
{
    recimgbuf = (unsigned char*)malloc(TMP_IMGBUF_LEN);
    bk_recframe.body = recimgbuf;
    bk_recframe.body_len = TMP_IMGBUF_LEN;
}

alloc_drimgbufs()
{
    int buflen;

#if 0
    if(output_format!=SEN_F_RGB) {
        fprintf(stderr, "no idea for no RGB device\n");
        exit(1);
    }
#endif

#if 0
    if(output_corder==SEN_C_RGB||
        output_corder==SEN_C_BGR) {
        buflen = TMP_IMGBUF_LEN;
    }
    else {
        buflen = TMP_IMGBUF_LEN;
    }
#endif
    buflen = (int)assoc_p(&buflen_xrgb_alist, MAX_DISPLAY_SIZE);
 if(debug) {
    fprintf(stderr, "buflen %d; TMP_IMGBUF_LEN %d, MAX_DISPLAY_SIZE %d\n",
        buflen, TMP_IMGBUF_LEN, MAX_DISPLAY_SIZE);
 }
    if(buflen==0) {
        buflen = TMP_IMGBUF_LEN;
    }

    drimgbuf0 = (unsigned char*)malloc(buflen);
    bk_drframe[0].body = drimgbuf0;
    bk_drframe[0].body_len = buflen;

    drimgbuf1 = (unsigned char*)malloc(buflen);
    bk_drframe[1].body = drimgbuf1;
    bk_drframe[1].body_len = buflen;
}


int cmp_type = SEN_M_9POINTS;
int diff_capacity = 0x20;
subrecord_t *recvlist=NULL;

int debug;
int verb;

apair_t qos_attr_alist[] = {
    {"Size",    QOS_SIZE},
    {"Zoom",    QOS_ZOOM},
    {"Color",   QOS_COLOR},
    {"Format",  QOS_FORMAT},
    {"Rate",    QOS_RATE},
    {NULL,      -1},
};

apair_t qos_desc_alist[] = {
    {size_alist,    QOS_SIZE},
    {color_alist,   QOS_COLOR},
    {format_alist,  QOS_FORMAT},
    {NULL, -1},
};

int capture_format=SEN_C_NONE;
int capture_corder=SEN_C_NONE;

#ifdef LOGTIME
time_t logtime;
struct tm *log_tm;
#endif


int  conn_no=0;
char down_listen_conn_id[BUFSIZ]="";
#if 0
char s_conn_id[BUFSIZ]="";
#endif


#ifndef MAX_LOWERNODES
#define MAX_LOWERNODES (10)
#endif

int  nupnode=0;
upnode_t upnodes[MAX_LOWERNODES];

#if 0
int data_port = SENTRY_DATA_PORT;
#endif
int ctrl_port = SENTRY_CTRL_PORT;

int tcpsock=-1;            /* socket(ctrl/data) for tcp connection */
int udpsock=-1;            /* socket(ctrl only) for udp connection */

#if 0
int unicast_datasock;   /* data socket for udp uniicast */
#endif
int multicast_datasock; /* data socket for udp multicast */

int maxreadfds;
fd_set readmask;

int maxwritefds;
fd_set writemask;

char vserver[80] = "localhost";
int  require_encode_type = SEN_F_NONE;

int filter_mode = FILTER_THRU;

extern int xl_offset, xr_offset, yt_offset, yb_offset;

void
reaper()
{
  union wait status;
  while(wait3((int *)&status, WNOHANG, (struct rusage *)0) >= 0)
    ;
}

void
finish_exit()
{
#ifdef LOGTIME
  time_t logtime;
  struct tm *log_tm;
#endif

  if (tcpsock >= 0) {
    shutdown(tcpsock, 2);
    close(tcpsock);
    tcpsock = -1;
  }
  if (udpsock >=0) {
    shutdown(udpsock, 2);
    close(udpsock);
    udpsock = -1;
  }
  if (multicast_datasock >= 0) {
    shutdown(multicast_datasock, 2);
    close(multicast_datasock);
    multicast_datasock = -1;
  }

  if (verb) {
#ifdef LOGTIME
    logtime = time(NULL);
    log_tm = localtime(&logtime);
    fprintf(stderr, "[%4d/%02d/%02d %02d:%02d:%02d] ",
        log_tm->tm_year + 1900, log_tm->tm_mon+1, log_tm->tm_mday,
        log_tm->tm_hour, log_tm->tm_min, log_tm->tm_sec);
#endif
    fprintf(stderr, "finish\nBye.\n");
  }

#ifdef USE_SYSLOG
  syslog(LOG_NOTICE, "ssf stopped.\n");
  closelog();
#endif

  exit (0);
}


upnode_t*
find_upnode(char *id)
{
    int i;

    for(i=0;i<nupnode;i++) {
        if(strcmp(upnodes[i].recept_conn_id, id)==0) {
            return &upnodes[i];
        }
    }

    return NULL;
}


#if 0
int
show_uppers(FILE *ofp, char *msg)
{
    fprintf(ofp, "conn. ID '%s' for unid\n", s_conn_id);
    fprintf(ofp, "server '%s' with port# %d as fd# %d\n",
                    server_host, server_port, ctrlfd);
    fprintf(ofp, "I am type-%s(%d) client with port# %d as fd# %d\n",
                    assoc_v(req_type_alist, client_type),
                    client_type, data_port, datafd);
    fprintf(ofp, "request size %s(%d), zoom %d, color %s(%d), format %s(%d), rate %4.1f f/s\n",
                    assoc_v(size_alist, server_size), server_size,
                    server_zoom,
                    assoc_v(color_alist, server_color), server_color,
                    assoc_v(format_alist, server_format), server_format,
                    (double)server_rate/100);
    fprintf(ofp, "interaction request interval %d sec\n", request_interval);
}
#endif

int
show_uppers(FILE *ofp, char *msg)
{
    int i;

    fprintf(ofp, "have %d upnodes\n", nupnode);
    for (i=0;i<nupnode;i++) {
        fprintf(ofp,
            " %3d: '%s' %-16s port# %5d fd# %3d; fd# %3d port# %5d; %p\n",
            i, upnodes[i].recept_conn_id,
            upnodes[i].hostname,
            upnodes[i].recept_fd, upnodes[i].recept_port,
            upnodes[i].recv_fd, upnodes[i].recv_port,
            upnodes[i].imgbuf
            );
    }
    fflush(ofp);
}

int
show_uppers_stat(FILE *ofp, char *msg)
{
    int i;
    int now;

    fprintf(ofp, "have %d upnodes\n", nupnode);
    now = time(NULL);
    for (i=0;i<nupnode;i++) {
        fprintf(ofp,
            " %3d: %-16s %7d %6d\n",
            i, upnodes[i].hostname,
            upnodes[i].recv_count, 
            now - upnodes[i].recv_lasttime);
    }
    fflush(ofp);
}

init_upnodetable()
{
    int i;
    
    for(i=0;i<MAX_LOWERNODES;i++) {
        upnodes[i].hostname = NULL;
        upnodes[i].recept_conn_id[0] = '\0';
        upnodes[i].recept_fd    = -1;
        upnodes[i].recept_port  = -1;
        upnodes[i].recv_fd      = -1;
        upnodes[i].recv_port    = -1;

        upnodes[i].imgbuf       = NULL;

        upnodes[i].recv_count    = 0;
        upnodes[i].recv_lasttime = 0;
    }
}

int
alloc_upnode_imgbuf(upnode_t *upp)
{
    sentry_imgrec *imgrec;
    
    if(!upp ||upp->imgbuf) {
        return -1;
    }

    imgrec = (sentry_imgrec*)malloc(sizeof(sentry_imgrec));
    if(!imgrec) {
        return -1;
    }
    upp->imgbuf = imgrec;
    
    imgrec->body = (unsigned char*)malloc(TMP_IMGBUF_LEN);
    if(imgrec->body)
        imgrec->body_len = TMP_IMGBUF_LEN;

    return 0;
}


int
show_lowers(FILE *ofp, char *msg)
{
    fprintf(ofp, "conn. ID '%s' for lnid\n", down_listen_conn_id);
    fprintf(ofp, "listening addr '%s' port# %d with TCP as fd# %d\n",
                    "", ctrl_port, tcpsock);
    fprintf(ofp, "listening addr '%s' port# %d with UDP as fd# %d\n",
                    "", ctrl_port, udpsock);
    show_client(ofp, "");
    fflush(ofp);
}


#if 1
#define CMD_QUIT        (1)
#define CMD_HELP        (2)
#define CMD_OPEN        (3)
#define CMD_CLOSE       (4)
#define CMD_CONNECT     (5)
#define CMD_DISCONNECT  (6)
#define CMD_LIST        (7)
#define CMD_ULIST       (8)
#define CMD_LLIST       (9)
#define CMD_STAT       (10)
#define CMD_FILTER     (30)
#define CMD_SYSTEM     (50)

apair_t cmdal[] = {
    {"QUIT",        CMD_QUIT},
    {"HELP",        CMD_HELP},
    {"OPEN",        CMD_OPEN},
    {"CLOSE",       CMD_CLOSE},
    {"CONNECT",     CMD_CONNECT},
    {"DISCONNECT",  CMD_DISCONNECT},
    {"LIST",        CMD_LIST},
    {"ULIST",       CMD_ULIST},
    {"LLIST",       CMD_LLIST},
    {"STAT",        CMD_STAT},
    {"FILTER",      CMD_FILTER},
    {"SYSTEM",      CMD_SYSTEM},
    {NULL,          -1},
};
#endif

/*
 */
void
send_request()
{
    int i;

    signal(SIGALRM, SIG_IGN);

    fprintf(stderr, "have %d upnodes\n", nupnode);
    for (i=0;i<nupnode;i++) {
        if(upnodes[i].recept_conn_id[0]) {
            fprintf(stderr, " %3d: fd# %d, port# %d\n", 
                i, upnodes[i].recept_fd, upnodes[i].recv_port);
            _request_call(upnodes[i].recept_fd, upnodes[i].recv_port);
        }
    }
    signal(SIGALRM, send_request);
}

int
store_conn_id(char *ids, int idn)
{
    int mid;

    *ids++ = 'C';

    mid = idn;
    while(mid) {
        *ids = 'A' + (mid%26);
        ids++;
        mid -= mid%26;
    }
    *ids = '\0';
}

help()
{
    fprintf(stdout,"100\r\n\
supported commands:\r\n\
  HELP\r\n\
  QUIT\r\n\
\r\n\
  OPEN <host> <port>\r\n\
  CLOSE <#id>\r\n\
  CONNECT <host> <port>\r\n\
  DISCONNECT <#id>\r\n\
\r\n\
  FILTER <mode>\r\n\
        THRU, GRAY, FLIP, FLOP, EDGE and DIFF is available\r\n\
.\r\n");
    fflush(stdout);
}

int
do_cmd(char *cmdbuf)
{
    char xcmd[BUFSIZ];
    char arg1[BUFSIZ], arg2[BUFSIZ], arg3[BUFSIZ], arg4[BUFSIZ];
    int cmd;
    int xport, rport;
    int xtcpsock=-1, xudpsock=-1, xdatafd=-1, xctrlfd=-1;
    int chk;
    int tc;
    int xfm;
    upnode_t *upp;

#if 0
    chk = split_cmd(cmdbuf, xcmd, arg);
#endif
    chk = split_arg4(cmdbuf, xcmd, arg1, arg2, arg3, arg4);

    cmd = assoc_n(cmdal, xcmd);
    fprintf(stderr, "cmd %d [%s] arg [%s] [%s]\n",
        cmd, xcmd, arg1, arg2);
    
    switch(cmd) {
    case CMD_HELP:
            help();
            break;
    case CMD_QUIT:
            fprintf(stdout, "200\r\n");
            fflush(stdout);
            exit(0);
            break;
    case CMD_SYSTEM:
            fprintf(stdout, "201 %d %s %s\n",
                SENTRY_PROTOCOL_VERSION, __DATE__, __TIME__);
            break;
    case CMD_ULIST:
            fprintf(stdout, "100\r\n");
            show_uppers(stdout, "I have");
            fprintf(stdout, ".\r\n");
            break;
    case CMD_LLIST:
            fprintf(stdout, "100\r\n");
            show_lowers(stdout, "I have");
            fprintf(stdout, ".\r\n");
            break;
    case CMD_LIST:
            fprintf(stdout, "100\r\n");
            show_uppers(stdout, "I have");
            show_lowers(stdout, "I have");
            fprintf(stdout, ".\r\n");
            break;
    case CMD_OPEN:
            if(down_listen_conn_id[0]) {
                fprintf(stderr, "listening port is already opend\n");
                fprintf(stdout, "400\r\n");
            }
            else
            if(arg1[0] && arg2[0]) {
                xport = atoi(arg2);
                xtcpsock = build_listen_port(xport, SOCK_STREAM);
                xudpsock = build_listen_port(xport, SOCK_DGRAM);
                if(xtcpsock>=0 && xudpsock>=0) {
                    ctrl_port = xport;
                    tcpsock = xtcpsock;
                    udpsock = xudpsock;

                    FD_SET(tcpsock, &readmask);
                    maxreadfds = MAX(maxreadfds, tcpsock);
                    FD_SET(udpsock, &readmask);
                    maxreadfds = MAX(maxreadfds, udpsock);
                    conn_no++;
                    store_conn_id(down_listen_conn_id, conn_no);
                    fprintf(stdout, "201 %s\r\n", down_listen_conn_id);
                }
                else {
                    if(xtcpsock>=0) {
                        close(xtcpsock);
                        xtcpsock = -1;
                    }
                    if(xudpsock>=0) {
                        close(xudpsock);
                        xudpsock = -1;
                    }
                    fprintf(stdout, "400\r\n");
                }
            }
            else {
                fprintf(stderr, "ignore argument(s)\n");
                fprintf(stdout, "400\r\n");
            }
            break;
    case CMD_CLOSE:
            fprintf(stderr, "'%s' vs '%s'\n", arg1, down_listen_conn_id);
            if(strcmp(arg1, down_listen_conn_id)==0) {
#if 1
                show_client(stderr, "all-stars");
#endif
                chk = clear_all_client();

                if(tcpsock>=0) {
                    FD_CLR(tcpsock, &readmask);
                    shutdown(tcpsock, 2);
                    close(tcpsock);
                    tcpsock = -1;
                }

                if(udpsock>=0) {
                    FD_CLR(udpsock, &readmask);
                    shutdown(udpsock, 2);
                    close(udpsock);
                    udpsock = -1;
                }

                down_listen_conn_id[0] = '\0';
                fprintf(stdout, "200\r\n");
            }
            else {
                fprintf(stdout, "400\r\n");
            }
            break;

    case CMD_CONNECT:
#if 0
            /* excusive; unid node is limited to one */
            if(s_conn_id[0]) {
                fprintf(stderr, "connection for unid node is already connected\n");
                fprintf(stdout, "400\r\n");
            }
            else
#endif
            if(arg1[0] && arg2[0]) {
                upnode_t *upp;

#if 0
                xport = atoi(arg2);
                xctrlfd = build_ctrl_port(arg1, xport);
                if(xctrlfd < 0) {
                    fprintf(stderr, "fail setup ctrl fd\n");
                }
                ctrlfd = xctrlfd;
                strcpy(server_host, arg1);
#endif

#if 0
                tc = 0;
                do {
                    xport = rand()%65000;
                    xdatafd = build_data_port(xport);
                    if (xdatafd < 0) {
                        fprintf(stderr, "fail setup data fd\n");
                    }
                    tc++;
                } while(xdatafd<0 && tc<5);

                data_port = xport;
                datafd = xdatafd;
                fprintf(stderr, "listen data with fd# %d, port# %d\n",  
                    datafd, data_port); 

                FD_SET(datafd, &readmask);
                maxreadfds = MAX(maxreadfds, datafd);

                request_call();
#endif

                if(nupnode>=MAX_LOWERNODES-1) {
                    fprintf(stdout, "400\r\n");
                    break;
                }

                upp = &upnodes[nupnode];

                xport = atoi(arg2);
                xctrlfd = build_ctrl_port(arg1, xport);
                if(xctrlfd < 0) {
                    fprintf(stderr, "fail setup ctrl fd\n");
                }

                xdatafd = build_data_port_try(-1, &rport);
                if(xdatafd<0) {
                    fprintf(stderr, "cannot open listening port\n");
                }
                data_port = rport;
                datafd = xdatafd;

                upp->recept_fd = xctrlfd;
                upp->recept_port = xport;
                upp->hostname = strdup(arg1);
                upp->recv_fd = xdatafd;
                upp->recv_port = rport;

                FD_SET(xdatafd, &readmask);
                maxreadfds = MAX(maxreadfds, xdatafd);

                conn_no++;
                store_conn_id(upp->recept_conn_id, conn_no);

                alloc_upnode_imgbuf(upp);

                nupnode++;

                fprintf(stdout, "201 %s\r\n", upp->recept_conn_id);

                send_request();
            }
            else {
                fprintf(stderr, "ignore argument(s)\n");
                fprintf(stdout, "400\r\n");
            }
            fflush(stdout);

            break;

    case CMD_DISCONNECT:
#if 0
            fprintf(stderr, "'%s' vs '%s'\n", arg1, s_conn_id);
            if(strcmp(arg1, s_conn_id)==0) {
                shutdown(ctrlfd, 2);
                close(ctrlfd);
                ctrlfd = -1;

                FD_CLR(datafd, &readmask);
                shutdown(datafd, 2);
                close(datafd);
                datafd = -1;

                s_conn_id[0] = '\0';

                fprintf(stdout, "200\r\n");
            }
#endif
            upp = find_upnode(arg1);
            if(upp) { 
                shutdown(upp->recept_fd, 2);
                close(upp->recept_fd);
                upp->recept_fd = -1;
    
                FD_CLR(upp->recv_fd, &readmask);
                shutdown(upp->recv_fd, 2);
                close(upp->recv_fd);
                upp->recv_fd = -1;

                upp->recept_conn_id[0] = '\0';
                free(upp->hostname);
                upp->hostname = NULL;

                fprintf(stdout, "200\r\n");
            }
            else {
                fprintf(stdout, "400\r\n");
                fflush(stdout);
            }

            break;

    case CMD_FILTER:
            xfm = assoc_n(filter_name_al, arg1);
            if(xfm>=0) {
                filter_mode = xfm;
                fprintf(stderr, "filter set %s -> %d\n", arg1, filter_mode);
                fprintf(stdout, "200\r\n");
            }
            else {
                fprintf(stdout, "400\r\n");
            }
            break;
    default:
            fprintf(stderr, "sorry, not implemented\n");
            fprintf(stdout, "400\r\n");
        
            break;
    }
    fflush(stdout);
}


extern void init_req_val(), inti_client();
extern int engray_rgb(unsigned char*, unsigned char*, int, int, int);
extern int flip_rgb(unsigned char*, unsigned char*, int, int, int);
extern int flop_rgb(unsigned char*, unsigned char*, int, int, int);
extern int vline_rgb(unsigned char*, unsigned char*, int, int, int);
extern int brightcircle_rgb(unsigned char*, unsigned char*,
            int, int, int, int);
extern int dope_rgb(unsigned char*, unsigned char*, int, int, int, int);
extern int sub3_rgb(unsigned char*, unsigned char*, unsigned char*,
            int, int, int);
extern int sub3hc_rgb(unsigned char*, unsigned char*, unsigned char*,
            int, int, int);

static int  got_count=0;
static int  sent_count=0;

static int  buffer_allocated=0;


int
handle_image(upnode_t *upp)
{
    char packet[TMP_IMGBUF_LEN+BUFSIZ];
    /*
     * header + data(image) =
     *      TMP_IMGBUF_LEN + sizeof(sentry_protocol3_request_hdr)
     *      < TMP_IMGBUF_LEN + BUFSIZ;
     */
    sentry_protocol3_image_hdr *hdr;

    sentry_shmrec *info;
    sentry_imgrec *inframe, *outframe;
    sentry_imgrec *rec;
    sentry_imgrec *bang;
    static sentry_imgrec *curimg=NULL, *lastimg=NULL;

    char *tmp_buff;
    int drfno;
    int fd;
    int cid;
    struct timeval curtime, lasttime;

    int  nsend;
    int  int_c=100;
    char *rimg_pos;

    int chk;
    int hdiff;

#if 0
    fflush(stdout);
    fprintf(stderr, "upp %p\n", upp);
    fflush(stderr);
#endif

PASS;

    info = &bk_info;
    if(!buffer_allocated) {
        alloc_inimgbufs();
        alloc_drimgbufs();
        alloc_outimgbufs();

        alloc_cookimgbufs();
        alloc_recimgbufs();

        buffer_allocated = 1;
    }


    if(!upp) {
        return -1;
    }
        
    {
#if 0
        fprintf(stderr,
            "upp conn '%s' host %s port %d fd %d; port %d fd %d\n",
            upp->recept_conn_id,
            upp->hostname, upp->recept_port, upp->recept_fd,
            upp->recv_port, upp->recv_fd);
#endif

#if 0
fprintf(stderr, "got_count %d\n", got_count);
fflush(stderr);
#endif

        outframe = NULL;
        bang = NULL;

        gettimeofday(&curtime, NULL);

        inframe = &bk_inframe;
        chk = recv_image(upp->recv_fd, &tmp_buff, info, inframe);
        if(chk) {
            goto out;
        }

#if 0
        show_imgattr(stderr, inframe);
#endif

        drfno = got_count%2;
        curimg = &bk_drframe[drfno];
PASS;

#if 1
{ 
    static int ct=0;
    static char *fname[] = {"/tmp/ssf0.jpg","/tmp/ssf1.jpg"};
   
    dump_to_file(fname[ct%2], inframe->body, inframe->length);
    ct++;
}
#endif


        copyattr_img(curimg, inframe);
        if(inframe->format==SEN_F_JPEG) {
            chk = decode_jpeg(inframe->body, inframe->length,
                    curimg->body, curimg->body_len, &curimg->width, &curimg->height);
            if(chk==0) {
                curimg->length = curimg->width * curimg->height *3;
                curimg->format = SEN_F_RGB;
                curimg->corder = SEN_C_RGB;

PASS;
#if 0
                show_imgattr(stderr, curimg);
#endif
                if(upp->imgbuf) {
                    copyattr_img(upp->imgbuf, curimg);
                    memcpy(upp->imgbuf->body, curimg->body, curimg->length);
                    upp->recv_count++;
                    upp->recv_lasttime = curtime.tv_sec;
                }
            }
        }

        if(got_count%report_interval==0) {
            syslog(LOG_NOTICE, "sentryc key=%d; got %d data\n",
                sentry_shmkey, got_count);
        }
PASS;

        switch(filter_mode) {
        case FILTER_THRU:
            outframe = inframe;
            break;
        case FILTER_GRAY:
            bang = &bk_cookframe;
            copyattr_img(bang, curimg);
            engray_rgb(bang->body, curimg->body,
                curimg->width, curimg->height, 3);
            break;
        case FILTER_FLIP:
            bang = &bk_cookframe;
            copyattr_img(bang, curimg);
            flip_rgb(bang->body, curimg->body,
                curimg->width, curimg->height, 3);
            break;
        case FILTER_FLOP:
            bang = &bk_cookframe;
            copyattr_img(bang, curimg);
            flop_rgb(bang->body, curimg->body,
                curimg->width, curimg->height, 3);
            break;
        case FILTER_EDGE:
            bang = &bk_cookframe;
            copyattr_img(bang, curimg);
            hvline_rgb(bang->body, curimg->body, curimg->width, curimg->height, 3);
            break;
        case FILTER_A1:
#if 0
            bang = &bk_cookframe;
            copyattr_img(bang, curimg);
            brightcircle_rgb(bang->body, curimg->body,
                curimg->width, curimg->height, 3, curimg->height/3);
#endif

#if 1
            rec = &bk_recframe;
            copyattr_img(rec, curimg);
            dope_rgb(rec->body, curimg->body, curimg->width, curimg->height, 3, 10);
            bang = &bk_cookframe;
            copyattr_img(bang, curimg);
            /* check how many diff */
            hdiff = sub3hc_rgb(bang->body, curimg->body, rec->body,
                curimg->width, curimg->height, 3);
            if(hdiff<2048) {
                bang = NULL;
            }
#endif

            break;
        case FILTER_A2:

#if 1
            rec = &bk_recframe;
            copyattr_img(rec, curimg);
            dope_rgb(rec->body, curimg->body, curimg->width, curimg->height, 3, 10);
            bang = &bk_cookframe;
            copyattr_img(bang, curimg);
            /* check how many diff */
            hdiff = sub3hc_rgb(bang->body, curimg->body, rec->body,
                curimg->width, curimg->height, 3);
            if(hdiff<100) {
                bang = NULL;
            }
#endif

#if 0
            if(lastimg) {
                change = img_cmp(curimg, lastimg, cmp_type, (void*)&diff_capacity);
                if(change) {
                    /* write like ghost */
                    bang = &bk_cookframe;
                    copyattr_img(bang, curimg);
                    sub_rgb(curimg->body, lastimg->body, bang->body,
                        curimg->width, curimg->height);
                }
            }
            /* first time */
            else {
                bang = curimg;
            }
#endif
            break;

        case FILTER_CLIPCIRCLE:
            bang = &bk_cookframe;
            copyattr_img(bang, curimg);
            clipcircle_rgb(bang->body, curimg->body,
                curimg->width, curimg->height, 3, curimg->height/3);
            break;

        case FILTER_ZOOM:
            bang = &bk_cookframe;
            copyattr_img(bang, curimg);
            zoomcenter_rgb(bang->body, curimg->body,
                curimg->width, curimg->height, 3);
            break;

        case FILTER_MEMBG:
#if 0
    if(curimg)
    fprintf(stderr, "membg; cur %dx%d\n", curimg->width, curimg->height);
    else 
    fprintf(stderr, "membg; lost cur %p\n", curimg);
#endif
            bang = &bk_cookframe;
            copyattr_img(bang, curimg);
            chk = memory_bg(1000, curimg, bang);
            break;

        case FILTER_MEMBGSHORT:
            bang = &bk_cookframe;
            copyattr_img(bang, curimg);
            chk = memory_bg(100, curimg, bang);
            break;

        case FILTER_MEMBGLONG:
            bang = &bk_cookframe;
            copyattr_img(bang, curimg);
            chk = memory_bg(10000, curimg, bang);
            break;

        case FILTER_ADD:
        case FILTER_SUB:
            {
                int i;
                sentry_imgrec *cur1, *cur2;
                upnode_t *p;

                cur1 = NULL;
                cur2 = NULL;
                for(i=0;i<nupnode;i++) {
                    p = &upnodes[i];
                    if(p->recept_conn_id[0]) {
                        if(cur1 && p->imgbuf) {
                            cur2 = p->imgbuf;
                            break;
                        }
                        else
                        if(!cur1 && p->imgbuf) {
                            cur1 = p->imgbuf;
                        }
                    }
                }

                bang = &bk_cookframe;
                copyattr_img(bang, curimg);
                if(cur1 && cur2) {
                    if(filter_mode==FILTER_ADD) {
                        add3_rgb(bang->body, cur1->body, cur2->body,
                            curimg->width, curimg->height, 3);
                    }
                    else {
                        subghost_rgb(cur1->body, cur2->body, bang->body,
                            curimg->width, curimg->height);
                    }
                }
                else
                if(cur1) {
                    memcpy(bang->body, cur1->body, cur1->length);
                }

            }
            break;


        case FILTER_WIPEFG:
        case FILTER_WIPEBG:
        case FILTER_GHOST:
            {
                int i;
                sentry_imgrec *cur1, *cur2;
                upnode_t *p;

                cur1 = NULL;
                cur2 = NULL;
                for(i=0;i<nupnode;i++) {
                    p = &upnodes[i];
                    if(p->recept_conn_id[0]) {
                        if(cur1 && p->recv_count>0 && p->imgbuf) {
                            cur2 = p->imgbuf;
                            break;
                        }
                        else
                        if(!cur1 && p->recv_count>0 && p->imgbuf) {
                            cur1 = p->imgbuf;
                        }
                    }
                }

                bang = &bk_cookframe;
                if(cur1 && cur2) {
#if 0
    fprintf(stderr, "ghost|wipe; cur1 %dx%d\n", cur1->width, cur1->height);
#endif
                    copyattr_img(bang, cur1);
                    if(filter_mode==FILTER_WIPEFG) {
                        subwipebg_rgb(cur2->body, cur1->body, bang->body,
                            curimg->width, curimg->height);
                    }
                    else
                    if(filter_mode==FILTER_WIPEBG) {
                        subwipebg_rgb(cur1->body, cur2->body, bang->body,
                            curimg->width, curimg->height);
                    }
                    else {
                        subghost_rgb(cur1->body, cur2->body, bang->body,
                            curimg->width, curimg->height);
                    }
                }
                else
                if(cur1) {
                    copyattr_img(bang, cur1);
#if 0
fprintf(stderr, "have cur1(%p), not have cur2(%p)\n", cur1, cur2);
#endif
                    memcpy(bang->body, cur1->body, cur1->length);
                }
                else {
                    bang = NULL;
                }

            }
            break;


        }

PASS;

        /*
         * if 'bang', make outframe.
         * if 'outframe' is presented, convert and send it as JPEG
         * to lnid node.
         * Then, when your filter have JPEG, you may entry 'outframe'.
         * otherwise, if your filter have RGB, you should set 'bang'.
         */
        if(bang) {
            int tlen;

            outframe = &bk_outframe;
            copyattr_img(outframe, bang);
            tlen = convert_jpeg(bang->body,
                    bang->length, bang->width, bang->height,
                    &outframe->width, &outframe->height, &rimg_pos);
#if 0
            fprintf(stderr, "length %d\n", tlen);
#endif
            if(tlen<=0) {
                fprintf(stderr, "ignore length\n");
                outframe = NULL;
            }
            else {
                outframe->format = SEN_F_JPEG;
                outframe->length = tlen;
                memcpy(outframe->body, rimg_pos, tlen);
            }
        }

PASS;

        if(outframe) {
            hdr = (sentry_protocol3_image_hdr*)packet;

            hdr->hd_ver = SENTRY_PROTOCOL_VERSION;
            hdr->hd_ts[0] = htonl(outframe->send_sec);
            hdr->hd_ts[1] = htonl(outframe->send_usec);
            hdr->format = htonl(outframe->format);
            hdr->offset = htonl(0);
            hdr->size   = htonl(outframe->length);
            hdr->width  = htonl(outframe->width);
            hdr->height = htonl(outframe->height);
            hdr->color  = htonl(outframe->corder);

            memcpy(&packet[sizeof(*hdr)],
                outframe->body, outframe->body_len);

            for(fd=0; fd<=maxwritefds; fd++) {
                if (FD_ISSET(fd, &writemask)) {
                    nsend = rtr_send(fd, packet,
                                sizeof(*hdr)+outframe->length);
                    if(nsend<0) {
                        cid = find_client_by_datafd(fd);
                        if(cid>=0) {
                            fprintf(stderr, "clear cid# %d\n", cid);
                            clear_client(cid);
                            continue;
                        }
                    }
                    sent_count++;
                }
            }
        }

PASS;

        lastimg = curimg;
        got_count++;

        if(got_count<=1) {
            lasttime = curtime;
        }
        if(got_count%int_c==0) {
            fprintf(stderr,
                "f# %6d [%d], time %5d ms, rate %5.1f ms/f %4.1f f/s; sent %d\n",
                got_count, int_c,
                SUB_TIME(curtime, lasttime),
                ((double)SUB_TIME(curtime, lasttime))/int_c,
                ((double)int_c*1000/SUB_TIME(curtime, lasttime)),
                sent_count);
            lasttime = curtime;

            show_uppers(stderr,"");
            show_uppers_stat(stderr,"");
        }

#if 0
        FD_CLR(datafd, &readfds);
#endif
    }


out:
PASS;
    return 0;
}


/* server */
void
mainloop()
{
    fd_set readfds;
    int nfds;
    int newsockfd;
    struct sockaddr_in from;
    int fromlen;
    sentry_protocol3_request_hdr req;
    char cmdbuf[BUFSIZ];
    int  nreadcmd;
    int chk;
    int i;
    upnode_t *upp;

#if 0
    char sysmesg[256];
    char buf[8192];
    int nr;
#endif

#if 0
    sentry_protocol3_image_hdr *hdr;
    char packet[TMP_BUF_LEN];
    char packet[TMP_IMGBUF_LEN];
#endif

#if 0
    sentry_shmrec *info;
    sentry_imgrec *inframe, *outframe;
    sentry_imgrec *cur, *last;
    sentry_imgrec *rec;
    sentry_imgrec *bang;
    char *tmp_buff;
    int drfno;
    int pos;
    int fd;
    int change;
    int nsend;
    struct timeval curtime, lasttime;
    int cid;
#endif

#if 0
    int  got_count=0;
    int  sent_count=0;
    int  int_c=10;
    char *rimg_pos;
    int hdiff;
#endif



    (void)init_req_val();
    (void)init_client();

    fromlen = sizeof(from);

    FD_ZERO(&readmask);
    maxreadfds = 0;
    FD_SET(0, &readmask); 
    fprintf(stdout,"100 welcome. this is ssf.\r\n.\r\n");
    fflush(stdout);

    FD_ZERO(&writemask);
    maxwritefds = 0;

    while (1) {
        readfds = readmask;

        nfds = select(maxreadfds+1, &readfds,
                (fd_set *)NULL, (fd_set *)NULL,
                (struct timeval *)NULL);
        if (nfds <= 0) {
#ifdef USE_SYSLOG
            sprintf(sysmesg, "ssf: select: %s", sys_errlist[errno]);
            syslog(LOG_ERR, sysmesg);
#endif
            continue;
        }

        if(FD_ISSET(0, &readfds)) {
            chk = get_cmdline(0, cmdbuf);
            if(chk==0) {
                break;
            }
            if(chk>0) {
                do_cmd(cmdbuf);
            }
            nreadcmd = strlen(cmdbuf);
#if 0
            fprintf(stderr, "++++++\n");
            fprintf(stderr, " %4d:%s\n", nreadcmd, cmdbuf);
            fflush(stderr);
#endif
            continue;
        }

        upp = NULL;
        if(nupnode>0) {
            for(i=0;i<nupnode;i++) {
                if(upnodes[i].recv_fd>=0 &&
                    FD_ISSET(upnodes[i].recv_fd, &readfds)) {
                    upp = &upnodes[i];
                    handle_image(upp);
                }
            }
            if(upp) {
                continue;
            }
        }

        /* connection request arrived at tcp port */
        if(tcpsock>0 && FD_ISSET(tcpsock, &readfds)) {
            FD_CLR(tcpsock, &readfds);

            newsockfd = accept(tcpsock,
                (struct sockaddr *)&from, &fromlen);
            if (newsockfd < 0)
                goto udp_check;

            if (verb) 
                fprintf(stderr, "TCP connection is arrived.\n");

            if (debug) {
                fprintf(stderr, 
                    "from->sin_addr = %s\n", inet_ntoa(from.sin_addr));
                fprintf(stderr, "from->sin_port = %d\n", 
                    ntohs(from.sin_port));
            }

          /* get_tcp_request(newsockfd) */

          if (add_client(0xf2, &from, newsockfd) < 0) {
            fprintf (stderr, "close connection port# %d.\n", newsockfd);
            shutdown(newsockfd, 2);
            close(newsockfd);
            continue;
          }

          FD_SET(newsockfd, &readmask);
          if (maxreadfds < newsockfd) maxreadfds = newsockfd;

          FD_SET(newsockfd, &writemask);
          if (maxwritefds < newsockfd) maxwritefds = newsockfd;
          
          fprintf(stderr,
            "ssf: tcp sock# %d is connected from %s:%d.\n", 
              newsockfd,
              inet_ntoa((struct in_addr *)ntohl(from.sin_addr.s_addr)),
              ntohs(from.sin_port));
          continue;
        }

udp_check:
        /* request arrived at udp port */
        if (udpsock>0 && FD_ISSET(udpsock, &readfds)) { 
    PASS;
            if (verb)
                fprintf(stderr,
                    "UDP connection/request is arrived.\n");

            get_udp_request(&from, &fromlen, &req);
            process_request(&from, &fromlen, &req, 0);

            if (debug) {
                u_long addr = ntohl(from.sin_addr.s_addr);

                fprintf(stderr, "addr# %x -> %x, %s\n", 
                    ntohl(from.sin_addr.s_addr), addr,
                    inet_ntoa((struct in_addr *)ntohl(addr)));
            }

            fprintf(stderr,
    "ssf: QoS request are arrived to UDP port# %d from %s:%d\n", 
             udpsock, inet_ntoa(from.sin_addr), ntohs(from.sin_port));

        }
        else {
            /* QoS request arrived at TCP port */
            /*
             * QoS requests are arrived from TCP port.
             *   not implemented yet... :)
            */
            if (verb) 
                fprintf(stderr, "some data is arrived.\n");

            fprintf(stderr,
                "this routine has not implemented yet.\n");
            continue;
            /*
            get_tcp_request();
            */
        }
    } /* end-of-while */
  /* NOT REACHED */
}


int
main(int argc, char **argv)
{
  int c;
  extern char *optarg;
  extern void usage(), finish_exit();
#ifdef LOGTIME
  time_t logtime;
  struct tm *log_tm;
#endif /* LOGTIME */
#ifdef DEBUG
  debug = 1;
#endif /* DEBUG */
#ifdef VERBOSE
  verb = 1;
#endif /* VERBOSE */

#ifdef USE_SYSLOG
  openlog("ssf", LOG_PID, LOG_DAEMON);
  syslog(LOG_NOTICE, "ssf starts.");
#endif

  if (verb) {
#ifdef LOGTIME
    logtime = time(null);
    log_tm = localtime(&logtime);
    fprintf(stderr, "[%4d/%02d/%02d %02d:%02d:%02d] ",
        log_tm->tm_year+1900, log_tm->tm_mon+1, log_tm->tm_mday,
        log_tm->tm_hour, log_tm->tm_min, log_tm->tm_sec);
#endif
    fprintf(stderr, "ssf starts.\n");
  }

  init_upnodetable();

  signal(SIGQUIT, finish_exit);
  signal(SIGTERM, finish_exit);
  signal(SIGINT, finish_exit);

  while ((c = getopt(argc, argv, "hDd:vjp:l:r:t:b:")) != -1) {
    switch(c) {
    case 'h':
      usage();
      exit (1); /* not reached */
    case 'D':
      debug = 100;
      break;
    case 'd':
      debug = atoi(optarg);
      break;
    case 'v':
      verb = 1;
      break;
    case 'j':
      require_encode_type = SEN_F_JPEG;
      break;
    case 'p':
      data_port = atoi(optarg);
      ctrl_port = data_port + 1;
      break;
    }
  }
  
#if 0
  /*
   * video setup
   */
  if (check_video(vserver)) {
    fprintf(stderr, "ssf: sorry, video(%s) is not avaliable.\n", "localhost");
    exit (2);
  }
#endif

#if 0
    if(verbose) {
        rtr_set_global_param(RTR_P_VERBOSE, NULL);
    }
    if(debug) {
        rtr_set_global_param(RTR_P_DEBUG, NULL);
        rtr_set_global_param(RTR_P_VERBOSE, NULL);
    }
#endif

#if 0
    rtr_set_global_param(RTR_P_VERBOSE, NULL);
#endif
#if 1
    rtr_set_global_param(RTR_P_DEBUG, NULL);
#endif

    rtr_set_global_param(RTR_P_LENFREELIST, 1024);
    rtr_set_global_param(RTR_P_QTIME, 2);

    rtr_initmem();


#if 0
    tcpsock = build_listen_port(ctrl_port, SOCK_STREAM);
    udpsock = build_listen_port(ctrl_port, SOCK_DGRAM);

    multicast_datasock = build_udp_port(0);

    if (tcpsock < 0 || udpsock < 0)
        exit (1);
#endif
  
    (void)mainloop();
    exit (0);
}

void
usage() {
  printf("usage: ssf [options]\n\n");

  printf("option:\n");
  printf("       -h       help\n");
  printf("       -j       force jpeg encoding (default=disable)\n");
  printf("       -p port  data port(default=%d)\n", SENTRY_DATA_PORT);
  printf("       -D       debug mode\n");
  printf("       -d num   debug level\n");
  printf("       -v       verbose mode\n");
  printf("\n");

  exit (1);
}

int
change_video_qos(char*X)
{
    fprintf(stderr, "change_video_qos:\n");
}
int
check_video(char*X)
{
    fprintf(stderr, "check_video:\n");
}
void
init_video(char*X)
{
    fprintf(stderr, "init_video:\n");
}

void
send_image(int signo)
{
#if 0
    fprintf(stderr, "send_image:\n");
#endif
    fprintf(stderr, "!");
    fflush(stderr);
}

