
/*
 * udprecv.c - receive udp packets and reassemble them.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <sys/types.h>
#include <netinet/in.h>
#ifdef TIME_WITH_SYS_TIME
 #include <sys/time.h>
 #include <time.h>
#else
 #ifdef HAVE_SYS_TIME_H
  #include <sys/time.h>
 #else
  #include <time.h>
 #endif
#endif
#include <errno.h>

#include "alist.h"
#include "sentry.h"


extern int debug;
extern int verbose;

#define OLIST_SIZE      256
struct image_offset {
  int offset;
  int length;
  struct image_offset *next;
} offset_list[OLIST_SIZE];
struct image_offset *olist_free;

#define IBUF_SIZE       32
struct {
#define IBUF_FREE       0
#define IBUF_USED       1
  int valid;                    /* valid flag (0:free, 1:used) */
  struct timeval ts;            /* timestamp */
#if 0
  long long ts;                 /* timestamp */
#endif
  struct image_offset *olist;
  caddr_t pimage;               /* pointer to image buffer */
  int size;                     /* image buffer size */
  int sum;                      /* summing of store */
} ibuf[IBUF_SIZE];
int max_ibuf_id;


/*
 * image buffer list for udp packet
 */
void clear_ibuf(int id)
{
  struct image_offset *p;

  ibuf[id].valid = 0;
  /* ibuf[id].size = 0;*/
  ibuf[id].sum = 0;

  /* 
   * clear offset list: change link to free list 
   */
  if (ibuf[id].olist != NULL) {
    for(p=ibuf[id].olist; p->next; p=p->next);
    p->next = olist_free;
    olist_free = ibuf[id].olist;
    ibuf[id].olist = NULL;
  }

  if (debug>=100)
    fprintf(stderr, "ibuf[%d] is clear.\n", id);

#if 0
  if (debug) {
    fprintf(stderr, "--- ibuf[%d] ---\n", id);
    fprintf(stderr, "  valid  = %d\n", ibuf[id].valid);
    fprintf(stderr, "  ts     = %d\n", ibuf[id].ts);
    fprintf(stderr, "  olist  = %x\n", ibuf[id].olist);
    fprintf(stderr, "  piamge = %x\n", ibuf[id].pimage);
    fprintf(stderr, "  size   = %d\n", ibuf[id].size);
    fprintf(stderr, "  sum    = %d\n", ibuf[id].sum);
  }
#endif

}

void init_ibuf()
{
  int i;

  for(i=0; i<IBUF_SIZE; i++)
    clear_ibuf(i);
  max_ibuf_id = -1;

  for(i=0; i<OLIST_SIZE; i++)
    offset_list[i].next = &offset_list[i+1];
  olist_free = offset_list;
}

#if 0
int find_ibuf(long long *t) 
{
  int ibuf_id;

  for(ibuf_id=0 ; ibuf_id<IBUF_SIZE; ibuf_id++)
    if ( ibuf[ibuf_id].valid == 1 && (*t) == ibuf[ibuf_id].ts )
      return ibuf_id;

  return -1;
}
#endif /* 0 */

#if 0
int new_ibuf()
{
  int ibuf_id;

  for(ibuf_id=0; ibuf_id<IBUF_SIZE; ibuf_id++)
    if (ibuf[ibuf_id].valid == 0) {
      if (max_ibuf_id < ibuf_id) max_ibuf_id = ibuf_id;
      return ibuf_id;
    }

  max_ibuf_id = (max_ibuf_id + 1) % IBUF_SIZE;
  return max_ibuf_id;

  fprintf(stderr, "fail to allocate new ibuf.\n");
  return -1;
}
#endif /* 0 */

struct image_offset *
new_olist(int p_off, int p_len, struct image_offset *p)
{
  struct image_offset *off_p;

  if (olist_free==NULL) {
    perror("Cannot allocate new offset list");
    return p;
  }

  off_p = olist_free;
  olist_free = olist_free->next;

  off_p->offset = p_off;
  off_p->length = p_len;
  off_p->next = p;

  return off_p;
}

void
free_olist(struct image_offset *p)
{
  p->next = olist_free;
  olist_free = p;
}

int
join_ibuf(int id, int part_off, int part_len)
{
  struct image_offset *p, *prev, *next;

  prev = NULL;
  for(p=ibuf[id].olist; p; p=p->next) {
    if (part_off < p->offset) {
      if (part_off + part_len < p->offset) {
        if (prev==NULL) {
          if (debug)
            fprintf(stderr, "part_off+part_len<p->offset; prev=NULL\n");
          ibuf[id].olist = new_olist(part_off, part_len, p);
        }
        else 
          prev->next = new_olist(part_off, part_len, p);
        break;
      } else if (part_off + part_len == p->offset) {
        p->offset = part_off;
        p->length += part_len;
        break;
      } else { /* part_off + part_len > p->offset */
        fprintf(stderr, "offset of the received packet is something wrong.\n");
        return -1;
      }
    } 
    else if (part_off == p->offset) {
      if (part_len > p->length)
        fprintf(stderr, "length of the receveid packet is something wrong.\n");
      return -1;
    }
    else if (part_off > p->offset) {
      if (part_off > p->offset + p->length) {
        prev = p;
        continue;
      } else if (part_off == p->offset + p->length) {
        p->length += part_len;
        break;
      } else { /* part_off < p->offset + p->length */
        fprintf(stderr, "offset of the received packet is something wrong.\n");
        return -1;
      }
    }
  } /* end-of-for loop */

  if (p==NULL) /* add to the list tail */
    if (prev==NULL) {
      ibuf[id].olist = new_olist(part_off, part_len, p);
    }
    else 
      prev->next = new_olist(part_off, part_len, p);

  /* shrink */
  for(p=ibuf[id].olist; p->next; p=p->next) {
    next = p->next;
    if (p->offset + p->length == next->offset) {
        p->length += next->offset;
        p->next = next->next;
        free(next);
    }
  }
  return 0;
#if 0
  p = olist; 
  while (1) {
    next = p->next;
    if (next == NULL) break;
    if (p->offset + p->length == next->offset) {
      p->length += next->length;
      p->next =  next->next;
      free_olist(next);
      p = p->next;
    }
  }
#endif

  /* end of join_ibuf */
#if 0
  struct image_offset *olist = ibuf[id].olist;
  struct image_offset *p;
  struct image_offset *pos1, *pos2, *free;
  int i;

  if (part_off == 0) { /* top packet */
    olist->offset = part_off;
    for(i=1; i<OLIST_SIZE; i++)
      if (olist->offset + olist->length == olist[i].offset) {
        olist->length = part_len + olist[i].length;
        olist[i].offset = 0;
        olist[i].length = 0;
      }
  }
  else { /* not top packet */
    pos1 = pos2 = free = -1;
    for(i=0; i<OLIST_SIZE; i++) {
      if (olist[i].offset + olist[i].length == part_off)
        pos1 = i;
      if (part_off + part_len == olist[i].offset)
        pos2 = i;
      if (i > 0 && olist[i].offset == 0 && olist[i].length == 0)
        free = i;
      if (pos1 >= 0 && pos2 >= 0 && free > 0)
        break;
    }
  }

  /*** part2 ***/
  if (part_off == 0) { /* top packet */
    off_t->offset = part_off; /* = 0 */
    off_t->length = part_len;

    /* shrink */
    return;
  }

  /* part_off > 0 */

  /* search empty cell */
  pos = -1;
  for(i=1; i<OLIST_SIZE; i++)
    if (off_t[i].offset == 0 && off_t[i].length == 0)
      break;
  pos = i;

  /* shrink list */
  p = off_t;
  if (off_t->length == 0) {
    p = &off_t[1];
    prev = off_t;
  }
  for(; p; p->next) {
    if (part_off < p->offset) {
      if (part_off + part_length == p->offset) {
        p->offset = part_off;
        p->length += part_len;
      } else is (part_off + part_length < p->offset) {
        if (prev->offset = prev->length == part_off)
          prev->length += part_len;
        else {
          prev->next = &off_t[pos];
          off_t[pos].offset = part_off;
          off_t[pos].length = part_len;
          off_t[pos].next = p;
        }
      }
    } else if (part_off == p->offset) {
      /* normally, nothring occured. */
    }
  }
#endif
}

/*
 * find ibuf_list, and shrink offset list
 */
int
udp_reassemble(sentry_protocol3_image_hdr *h, char *data, int datalen, char **image) 
{
  int id;
#if 0
  long long *tsp;
  long long oldest;
#endif
  struct timeval tstamp, oldest;
  int free_id, oldest_id;
  int image_size, data_offset;
  int chk, sum;

  free_id = oldest_id = -1;
  oldest.tv_sec = oldest.tv_usec = 0;

#if 0
  tsp = (long long *)&h->hd_ts;
#endif
  tstamp.tv_sec = ntohl(h->hd_ts[0]);
  tstamp.tv_usec = ntohl(h->hd_ts[1]);

  /* find ibuf_id */
  for(id=0; id<IBUF_SIZE; id++) {
    if (ibuf[id].valid == IBUF_USED) {
      if (tstamp.tv_sec == ibuf[id].ts.tv_sec &&
          tstamp.tv_usec == ibuf[id].ts.tv_usec)
        break; /* ibuf_id was found */
      if ( oldest.tv_sec > ibuf[id].ts.tv_sec ||
           oldest.tv_sec == ibuf[id].ts.tv_sec &&
           oldest.tv_usec > ibuf[id].ts.tv_usec ||
           oldest.tv_sec == 0 && oldest.tv_usec == 0 ) {
        oldest.tv_sec = ibuf[id].ts.tv_sec;
        oldest.tv_usec = ibuf[id].ts.tv_usec;
      }
#if 0
      if ( (*tsp) == ibuf[id].ts ) break; /* ibuf_id was found */
      if ( oldest > ibuf[id].ts || oldest == 0 ) { /* for garbage */
        oldest = ibuf[id].ts;
        oldest_id = id;
      }
#endif

      continue;
    }
    if ( free_id == -1 && ibuf[id].valid == IBUF_FREE) /* search free */
      free_id = id;
  }
  if (id >= IBUF_SIZE) { /* not found in ibuf_list */
    if (free_id == -1)
      id = oldest_id;
    else 
      id = free_id;

    ibuf[id].valid = IBUF_USED;
    ibuf[id].ts.tv_sec = tstamp.tv_sec;
    ibuf[id].ts.tv_usec = tstamp.tv_usec;
    ibuf[id].sum = 0;
#if 0
    ibuf[id].ts = *tsp;
#endif
  }

  /* id has been found/allocated. */

#if 0
  if (ibuf[0].valid == 1 && (*tsp) == ibuf[0].ts) 
    id = 0;
  else {
    id = find_ibuf(tsp);
    if (id < 0)
      if (ibuf[0].valid == 0) 
        id = 0;
    else 
      id = new_ibuf(tsp);
    if (id < 0)
      return -1;

    /* dump ibuf */
    ibuf[id].valid = 1;
    ibuf[id].ts = *tsp;
  }
#endif

  image_size = ntohl(h->size);
  if (ibuf[id].size == 0 || ibuf[id].size < image_size) {

        if(debug>=10){
        fprintf(stdout, "new buffer %2d (new %ld, old %ld).\n", 
                id, (long)image_size, (long)ibuf[id].size);
        }

    if (ibuf[id].pimage != NULL) {
                if(debug>=10){
                fprintf(stdout, "resize buffer %2d.\n", id);
                }

        free(ibuf[id].pimage);
        ibuf[id].size = 0;
    }
    ibuf[id].size = image_size;
    ibuf[id].pimage = (caddr_t)malloc(ibuf[id].size);
  }

  (*image) = ibuf[id].pimage;
  
  data_offset = ntohl(h->offset);
  if (data_offset + datalen > image_size) {
    fprintf(stderr, "Waring: offset is wrong.\n");
    datalen = image_size - data_offset;
  }

  /* 
   * store image: copy packets to stored.
   */
  bcopy(data, (char *)&(*image)[data_offset], datalen);
  ibuf[id].sum += datalen;

  
  chk = join_ibuf(id, data_offset, datalen);
  if (debug>=100) {
    fprintf(stderr, "=== ibuf[%d] chk# %d ===\n", id, chk);
    fprintf(stderr, "  valid:     %d\n", ibuf[id].valid);
    fprintf(stderr, "  ts:        %d.%06d\n", ibuf[id].ts.tv_sec, ibuf[id].ts.tv_usec);
    fprintf(stderr, "  olist:     %x\n", ibuf[id].olist);
    if (ibuf[id].olist != NULL) {
      fprintf(stderr, "    offset   %d\n", ibuf[id].olist->offset);
      fprintf(stderr, "    length   %d\n", ibuf[id].olist->length);
    }
    fprintf(stderr, "  pimage:    %x\n", ibuf[id].pimage);
    fprintf(stderr, "  buff_size: %d\n", ibuf[id].size);
    fprintf(stderr, "  sum:       %d\n", ibuf[id].sum);
    fprintf(stderr, "  image_size %d\n\n", image_size);
  }

  if (ibuf[id].olist != NULL &&
      ibuf[id].olist->offset == 0 &&
      ibuf[id].olist->length == image_size) {
    sum = ibuf[id].sum;
    clear_ibuf(id);
    return sum;
  }
  else 
    return 0;
}
