/*
 * Sentry Video System for PC with video4linux
 *   is controlled QoS according to clients' requests.
 *
 *        original by akira-y@is.aist-nara.ac.jp, 1994,1995
 *        modified by k-chinen@is.aist-nara.ac.jp, 1996
 *        modified by satosi-f@is.aist-nara.ac.jp, 1997,2000
 *        modified by atsush-m@is.aist-nara.ac.jp, 2002
 *
 *   $Id:$
 */

/* for FreeBSD with 'bktr' device */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

/*
 * video.c - video (chunk of images) handling procedures
 */
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <signal.h>

/*
 * NOTE:
 *  According to 'man meteor', METEOR_GEO_RGB24 in this device means
 *  RGBX color order (4 bytes per pixel.)
 */
/*
 *  The manual describes capturring with ring buffer mode.
 *  however, it is strange.  the sample is not work.
 *  then, I written with signal.
 */

#include <machine/ioctl_meteor.h>
#if 0
#include <machine/ioctl_bt848.h>
#endif

#include "alist.h"
#include "sentry.h"
#include "server.h"
#include "video.h"

#define VDEV_NAME "/dev/bktr0"

#define FRAME_MAX   (32)


#define fioctlmsg(x)    fprintf(stderr, "fail ioctl "#x" (%d) %s:%d\n",\
                            errno, __FILE__, __LINE__)

int video_sync_type = SENV_SYNC_BY_SIGNAL;

extern int debug;
extern int verb;
extern int errno;
extern int xsize, ysize;
extern int capture_xsize, capture_ysize;
extern int xl_offset, xr_offset, wide_offset;
extern int yt_offset, yb_offset;
extern _req_val req_val;

int n, vfd=-1;
unsigned char *cur=NULL, *prev=NULL, *mapbase=NULL;
extern int frameSize;
int last_frameSize=-1;
char *vdev_name = VDEV_NAME;
static unsigned char *caparea;

static int last_ssize=SIZE_UNKNOWN;
static int cur_ssize=SIZE_UNKNOWN;


int framesize;
int allframesize;
int _nframe;
struct meteor_geomet geo;
struct meteor_video fb;
struct metror_mem *common_mem=NULL;

struct flock lock;

int
check_video(char *vserver)
{
  /* check if it can open device */
  if(vfd == -1){
    if((vfd = open(vdev_name, O_RDONLY)) < 2){
      perror("open video");
      return 1;
    }else{
      close(vfd);
      vfd = -1;
      return 0;
    }
  }

  /* check if device is running */
  fcntl(vfd, F_GETLK, lock);
  if(lock.l_type == F_UNLCK)
    return 0;
  else
    return 1;
}

void 
init_video(char *vserver)
{
  extern void setup_video();
  struct meteor_geomet geo;
  int c;

#if 0
    debug = 1001;
#endif

  /* connect to the daemon */
  if((vfd = open(vdev_name, O_RDONLY)) < 2){
    perror("open video");
    exit(21);
  }

  /* device lock */
  lock.l_type = F_WRLCK;
  lock.l_start = 0;
  lock.l_whence = SEEK_SET;
  lock.l_len = 0;
  fcntl(vfd, F_SETLK, &lock);

#if 0
  /* get divice parameter */
  if(ioctl(vfd, VIDIOCGCAP, &vd) < 0){
    fioctlmsg(VIDOCGCAP);
    perror("ioctl(VIDOCGCAP)");
    exit(22);
  } 

  /* check device parameter */
  fprintf(stderr, "vd.name: \"%s\"\n", vd.name);
  fprintf(stderr, "vd.type=0x%08x", vd.type);
  if(vd.type&VID_TYPE_CAPTURE) fprintf(stderr, " CAPTURE");
  if(vd.type&VID_TYPE_TUNER) fprintf(stderr, " TUNER");
  if(vd.type&VID_TYPE_TELETEXT) fprintf(stderr, " TELETEXT");
  if(vd.type&VID_TYPE_OVERLAY) fprintf(stderr, " OVERLAY");
  if(vd.type&VID_TYPE_CHROMAKEY) fprintf(stderr, " CHROMAKEY");
  if(vd.type&VID_TYPE_CLIPPING) fprintf(stderr, " CLIPPING");
  if(vd.type&VID_TYPE_FRAMERAM) fprintf(stderr, " FRAMERAM");
  if(vd.type&VID_TYPE_SCALES) fprintf(stderr, " SCALES");
  if(vd.type&VID_TYPE_MONOCHROME) fprintf(stderr, " MONOCHROME");
  if(vd.type&VID_TYPE_SUBCAPTURE) fprintf(stderr, " SUBCAPTURE");
  fprintf(stderr, "\n");
  fprintf(stderr, "vd.channels=%d\n", vd.channels);
  fprintf(stderr, "vd.audios=%d\n", vd.audios);
  fprintf(stderr, "vd.maxwidth=%d\n", vd.maxwidth);
  fprintf(stderr, "vd.maxheight=%d\n", vd.maxheight);
  fprintf(stderr, "vd.minwidth=%d\n", vd.minwidth);
  fprintf(stderr, "vd.minheight=%d\n", vd.minheight);
  fprintf(stderr, "\n");
#endif

#if 0
  /* get channel parameters */
  for(n=0; n<vd.channels; n++){
    vc[n].channel = n;
    if(ioctl(vfd, VIDIOCGCHAN, &vc[n]) < 0){
      fioctlmsg(VIDIOCGCHAN);
      perror("ioctl(VIDIOCGCHAN)");
      exit(23);
    }

    fprintf(stderr, "vc[%d].channel=%d\n", n, vc[n].channel);
    fprintf(stderr, "vc[%d].name=\"%s\"\n", n, vc[n].name);
    fprintf(stderr, "vc[%d].tuners=%d\n", n, vc[n].tuners);
    fprintf(stderr, "vc[%d].flags=0x%08x", n, vc[n].flags);
    if(vc[n].flags&VIDEO_VC_TUNER) fprintf(stderr, " TUNER");
    if(vc[n].flags&VIDEO_VC_AUDIO) fprintf(stderr, " AUDIO");
    fprintf(stderr, "\n");
    fprintf(stderr, "vc[%d].type=0x%08x", n, vc[n].type);
    if(vc[n].type&VIDEO_TYPE_TV) fprintf(stderr, " TV");
    if(vc[n].type&VIDEO_TYPE_CAMERA) fprintf(stderr, " CAMERA");
    fprintf(stderr, "\n");
    fprintf(stderr, "vc[%d].norm=%d\n", n, vc[n].norm);
    fprintf(stderr, "\n");
  }
#endif
#if 0
  /* choice "Composite1" */
  vc[1].norm = 1;    /* 0:Television 1:Composite1 2:S-Video */  
  if(ioctl(vfd, VIDIOCSCHAN, &vc[1]) < 0){
    fioctlmsg(VIDIOCSCHAN);
    perror("ioctl(VIDIOCSCHAN)");
    exit(24);
  }
#endif

  c =  METEOR_INPUT_DEV0;

  if (ioctl(vfd, METEORSINPUT, &c) < 0) {
    fioctlmsg(METEORSINPUT);
    fprintf(stderr, "ioctl METEORSINPUT failed: %d\n", errno);
    exit(1);
  }

#if 0
  c = BT848_IFORM_F_AUTO;
  if (ioctl(vfd, BT848SFMT, &c) < 0) {
    fioctlmsg(BT848SFMT);
    fprintf(stderr, "ioctl BT848SFMT failed: %d\n", errno);
    exit(1);
  }
#endif

  (void)setup_video();

#if 0
    SET_QOS(QOS_SIZE, SIZE_160x120);
    set_size();
#endif
    start_capture();
}



void
setup_video(void)
{
  extern void set_color(), set_zoom(), set_size();
  int c;


  framesize = 640*480*4;
    allframesize = 160*120*4*FRAME_MAX;
    allframesize = ((allframesize+4095)/4096)*4096;
    _nframe = FRAME_MAX;
    
#if 0
  caparea = mmap((caddr_t) 0, framesize, PROT_READ,
                MAP_SHARED, vfd, (off_t) 0);
#else
  caparea = mmap((caddr_t) 0, allframesize+4096, PROT_READ,
                MAP_SHARED, vfd, (off_t) 0);
#endif
    if(caparea==(unsigned char *)MAP_FAILED) {
        fprintf(stderr, "fail mmap %d\n", errno);
    }
    fprintf(stderr, "mmap map to %p <size %d>\n", caparea, allframesize);

    common_mem = (struct meteor_mem*) (caparea+allframesize);

  fb.addr     = 0;
  fb.width    = 0;
  fb.banksize = 0;
  fb.ramsize =  0;

  if(ioctl(vfd, METEORSVIDEO, &fb)<0) {
    fioctlmsg(METEORSVIDEO);
    fprintf(stderr, "ioctl METEORSVIDEO failed: %d\n", errno);
    exit(41);
  }

  capture_format = SEN_F_RGB;
  capture_corder = SEN_C_RGB;
    /*
     * color order at outputting in this module is RGB.
     * but, raw color order is RGBX.
     */


#if 1
  /* set up the capture type and default size */
  geo.columns = 160;
  geo.rows    = 120;
#if 0
  geo.frames  = 1;
#else
  geo.frames  = _nframe;
#endif
  geo.oformat = METEOR_GEO_RGB24;

    if(debug>=10) {
        fprintf(stderr, "set geo %d x %d, %d, %d\n",
            geo.columns, geo.rows, geo.frames, geo.oformat);
    }

  if(cur_ssize!=SIZE_UNKNOWN) {
    fprintf(stderr, "cur_ssize is strange (%d)\n", cur_ssize);
  }
  last_ssize = cur_ssize;
  cur_ssize = SIZE_160x120;


  fprintf(stderr, "%s: try set size %d x %d frame %d, format %#x\n",
    __FILE__,
    geo.columns, geo.rows, geo.frames, geo.oformat);
  if (ioctl(vfd, METEORSETGEO, &geo) < 0) {
    fioctlmsg(METEORSETGEO);
    fprintf(stderr, "ioctl METEORSETGEO failed: %d\n", errno);
#if 0
    exit(1);
#endif
    if(errno==EBUSY) {
        fprintf(stderr, "bussy, skip\n");
    }
    else {
        exit(9);
    }
  }
  fprintf(stderr, "set size %d x %d frame %d, format %#x\n",
    geo.columns, geo.rows, geo.frames, geo.oformat);
#endif

    frameSize = geo.columns * geo.rows * 4;
    fprintf(stderr, "frameSize %d, '%s' ==== \n", frameSize,
        assoc_v(color_alist, capture_corder));

#if 0
  if (CHK_CHANGE_FLAG(QOS_SIZE))    set_size();
  if (CHK_CHANGE_FLAG(QOS_COLOR))   set_color();
  if (CHK_CHANGE_FLAG(QOS_ZOOM))    set_zoom();
#endif

#if 0
  c = 4;
  if(ioctl(vfd, METEORSACTPIXFMT, &c)<0) {
    fioctlmsg(METEORSACTPIXFMT);
    fprintf(stderr, "# METEORSACTPIXFMT %d\n", c);
  }
  fprintf(stderr, "set METEORSACTPIXFMT %d\n", c);
#endif

#if 0
  memset((void*)&cliplist, 0, sizeof(cliplist));
  if(ioctl(vfd, BT848SCLIP, &cliplist)<0) {
    fioctlmsg(BT848SCLIP);
    fprintf(stderr, "# BT848SCLIP\n");
  }
#endif

}

int _sig_no = SIGUSR2;

static
void
print_params()
{
    if(!common_mem) {
        fprintf(stderr, "common_mem is not available\n");
        return;
    }
    fprintf(stderr, "common_mem %p\n", common_mem);
    fprintf(stderr, "  frame_size %d\n",
             ((struct meteor_mem*)common_mem)->frame_size);
    fprintf(stderr, "  num_bufs   %d\n",
             ((struct meteor_mem*)common_mem)->num_bufs);
    fprintf(stderr, "  lowat      %d\n",
             ((struct meteor_mem*)common_mem)->lowat);
    fprintf(stderr, "  hiwat      %d\n",
             ((struct meteor_mem*)common_mem)->hiwat);
    fprintf(stderr, "  active     %d\n",
             ((struct meteor_mem*)common_mem)->active);
    fprintf(stderr, "  num_active %d\n",
             ((struct meteor_mem*)common_mem)->num_active_bufs);
}

void
sig_catcher()
{
    time_t now;

    now = time(NULL);
    fprintf(stderr, "called %d\n", now);
#if 0
    fprintf(stderr, "common_mem %p\n", common_mem);
    fprintf(stderr, "  frame_size %d\n", common_mem->frame_size);
    fprintf(stderr, "  frame_size %d\n", common_mem->num_bufs);
#endif
    print_params();

    send_image(311);
}

static int capture_count=0;
static int capture_inited=0;

char *
start_capture()
{
    int c, ii;
    struct meteor_capframe capframe;
    int nread;
    extern void convert_color();
    extern void send_image(int);

#if 0
    fprintf(stderr, "start_capture: %d\n", capture_inited);
#endif

    if (debug>=200) {
        fprintf(stderr,
            "start_capture: caparea %p, frameSize %d (%dx%d)\n",
            caparea, frameSize, capture_xsize, capture_ysize);
        fprintf(stderr,
            "  buffers cur %p, prev %p\n", cur, prev);
    }

    if(capture_inited>0) {
        /* already capturing, do nothing */
        goto out;
    }

#if 0
    ii = METEOR_CAP_CONTINOUS;
    if (ioctl(vfd, METEORCAPTUR, &ii) < 0) {
        fioctlmsg(METEORCAPTUR);
        fprintf(stderr, "ioctl CAP_CONTINOUS failed: %d\n", errno);
        exit(1);
    }
#endif

    if(cur)
        free(cur);
#if 0
    cur = (unsigned char*)malloc(capture_xsize*capture_ysize*3);
#endif
    cur = (unsigned char*)malloc(frameSize);
    last_frameSize = frameSize;

    if(debug>=10) {
        fprintf(stderr, "start_capture: new cur %p len %d\n",
            cur, frameSize);
    }

#if 0
    memcpy(cur, caparea, frameSize);
#endif

    if(prev)
        free(prev);

    if(ioctl(vfd, METEORSSIGNAL, &_sig_no)) {
        fioctlmsg(METEORSSIGNAL);
        abort();
    }


#if 0
    capframe.command=METEOR_CAP_N_FRAMES;
#if 0
    capframe.signal=SIGUSR2;
#endif
    capframe.lowat=_nframe/2;       /* must be < hiwat */
    capframe.hiwat=_nframe;         /* must be < FRAME_MAX */

    if(ioctl(vfd, METEORCAPFRM, &capframe) < 0) {
        fioctlmsg(METEORCAPFRM);
        abort();
    }

#endif

#if 0
    signal(SIGUSR2, sig_catcher);
#endif
    signal(SIGUSR2, send_image);

#if 0
    nread = read(vfd, cur, frameSize);
    if(nread<0) {
        fprintf(stderr, "read ret %d with %d (%d)\n",
            nread, frameSize, errno);
        abort();
    }
#endif
#if 1
    c = METEOR_CAP_CONTINOUS;
    if(ioctl(vfd, METEORCAPTUR, &c) < 0) {
        fioctlmsg(METEORCAPTUR);
        abort();
    }
#endif

    capture_inited++;

    if (debug>=200)
        fprintf(stderr, "start_capture: inited %d; return %d, %#x\n",
            capture_inited, cur, cur);

out:

    return cur;
}

char *
capture_frame()
{
    extern void convert_color();
    int c, ii;
    int nread;
    struct timeval tbegin, tend;

    if (debug>=200)
        fprintf(stderr,
"capture_frame: caparea %p, cur %p, prev %p, frameSize %d (%dx%d)\n",
            caparea, cur, prev, frameSize, capture_xsize, capture_ysize);

    if(capture_inited==0) {
        start_capture();
    }

    if(frameSize!=last_frameSize) {
        free(cur);
        cur = (char*)malloc(frameSize);
    }

#if 1
    memcpy(cur, caparea, frameSize);
#endif
#if 0
    gettimeofday(&tbegin, NULL);
    nread = read(vfd, cur, frameSize);
/*
    nread = read(vfd, cur, 160*120*3);
*/
    gettimeofday(&tend, NULL);
    fprintf(stderr, "-- %d (%d)\n%d.%06d\n%d.%0d\n",
        nread, errno,
        tbegin.tv_sec, tbegin.tv_usec,
        tend.tv_sec, tend.tv_usec);
#endif
    last_frameSize = frameSize;

    if(cur != prev) {
        if(debug) {
            fprintf(stderr,
                "capture_frame: prev %p -> cur %p\n", prev, cur);
        }
        if(prev)
            free(prev);
        prev = cur;
    }

#if 0
    if(cur!=NULL)
        convert_color();
#endif
#if 0
    if(cur!=NULL)
        convert_BGRtoRGB();
#endif
    convert_BGRA2RGB_s(cur, capture_xsize, capture_ysize);

    capture_count++;

    if (debug>=200)
        fprintf(stderr, "capture_frame: %dth; return %d, %#x\n",
            capture_count, cur, cur);

    return cur;
}


int
release_frame()
{
  extern unsigned char *cur;
      
  /*
   * Finished with frame, unlocked the buffer
   */
  if (debug>=200)
    fprintf(stderr, "release_frame:\n");

  return 0;
}


/*
 * set video capture parameters 
 */
void
set_size(void)
{
  int req;

    /* clear wide_offset */
    wide_offset = 0;

    req = GET_QOS(QOS_SIZE);
    if(req!=cur_ssize) {
        if(debug>=100) {
            fprintf(stderr, "set_size: req %s, cur %s, last %s\n",
                assoc_v(size_alist, req),
                assoc_v(size_alist, cur_ssize),
                assoc_v(size_alist, last_ssize));
        }
    }


  /* Set the frame size
   *   640x480, 320x240, 212x160, 160x120: normal vision (w:h =  4:3)
   *   640x360, 320x180, 212x120, 160x90: wide vision (w:h = 16:9 )
   */
  switch (req) 
  {
  case SIZE_640x480:
    xsize = 640;
    ysize = 480;
    break;
  case SIZE_320x240:
    xsize = 320;
    ysize = 240;
    break;
  case SIZE_212x160:
  default:
    xsize = 212;
    ysize = 160;
    break;
  case SIZE_160x120:
    xsize = 160;
    ysize = 120;
    break;
    
  case SIZE_640x360:
    xsize = 640;
    ysize = 360;
    wide_offset = 60;        /* 60 = (480 - 360) / 2 */
    break;
  case SIZE_320x180:
    xsize = 320;
    ysize = 180;
    wide_offset = 30;        /* 30 = (240 - 180) / 2 */
    break;
  case SIZE_212x120:
    xsize = 212;
    ysize = 120;
    wide_offset = 20;        /* 20 = (160 - 120) / 2 */
    break;
    
  case SIZE_160x90:
    xsize = 160;
    ysize = 90;
    wide_offset = 15;        /* 15 = (120 - 90) / 2 */
    break;
  }


  capture_xsize = xsize + xl_offset + xr_offset;
  capture_ysize = ysize + (wide_offset * 2) + yt_offset + yb_offset;  


#if 1
  /* set up the capture type and default size */
  geo.columns = capture_xsize;
  geo.rows    = capture_ysize;
#if 0
  geo.frames  = 1;
#else
  geo.frames  = _nframe;
#endif
    if(capture_corder==SEN_C_RGB) {
        geo.oformat = METEOR_GEO_RGB24;
    }
    else {
        geo.oformat = METEOR_GEO_RGB24;
    }

    if(debug>=10) {
        fprintf(stderr, "set geo %d x %d, %d, %d\n",
            geo.columns, geo.rows, geo.frames, geo.oformat);
    }

  if(debug>=100) {
    fprintf(stderr, "cur_ssize %s(%d) -> %s(%d)\n",
        assoc_v(size_alist, cur_ssize), cur_ssize,
        assoc_v(size_alist, req), req);
  }
  last_ssize = cur_ssize;
  cur_ssize = req;

  fprintf(stderr, "%s: try set size %d x %d frame %d, format %#x\n",
    __FILE__,
    geo.columns, geo.rows, geo.frames, geo.oformat);
  if (ioctl(vfd, METEORSETGEO, &geo) < 0) {
    fioctlmsg(METEORSETGEO);
    fprintf(stderr, "ioctl METEORSETGEO failed: %d\n", errno);
#if 0
    exit(1);
#endif
    if(errno==EBUSY) {
        fprintf(stderr, "busy, skip\n");
    }
    else {
        exit(9);
    }
  }
  fprintf(stderr, "set size %d x %d frame %d, format %#x\n",
    geo.columns, geo.rows, geo.frames, geo.oformat);
#endif


    last_ssize = cur_ssize;
    cur_ssize = req;
    
#if 0
    if(capture_corder==SEN_C_RGB || capture_corder==SEN_C_BGR) {
        frameSize = xsize*ysize*3;
    }
    if(capture_corder==SEN_C_RGBA || capture_corder==SEN_C_XRGB) {
        frameSize = xsize*ysize*4;
    }
    else {
        frameSize = xsize*ysize*4;
    }
#endif
    /* constant */
    frameSize = xsize*ysize*4;

    if(debug>=200) {
        fprintf(stderr, "new frameSize %d\n", frameSize);
    }

    cur = NULL;

    CLR_CHANGE_FLAG(QOS_SIZE);
}

void
set_zoom(void)
{
  CLR_CHANGE_FLAG(QOS_ZOOM);
}

void 
set_color(void)
{
  PASS;
 
  /* set up the capture type and size */
  geo.columns = capture_xsize;
  geo.rows    = capture_ysize;
  geo.frames  = 1;
  geo.oformat = METEOR_GEO_RGB24;

  fprintf(stderr, "try set size %d x %d frame %d, format %#x\n",
    geo.columns, geo.rows, geo.frames, geo.oformat);
  if (ioctl(vfd, METEORSETGEO, &geo) < 0) {
    fioctlmsg(METEORSETGEO);
    fprintf(stderr, "ioctl METEORSETGEO failed: %d\n", errno);
#if 0
    exit(1);
#endif
    if(errno==EBUSY) {
        fprintf(stderr, "bussy, skip\n");
    }
    else {
        exit(9);
    }
  }
  fprintf(stderr, "set size %d x %d frame %d, format %#x\n",
    geo.columns, geo.rows, geo.frames, geo.oformat);

  CLR_CHANGE_FLAG(QOS_COLOR);
}

int 
close_video()
{
    if (verb)
      fprintf(stderr, "close_video:\n");

#if 0
    munmap(cur, framesize);
#else
    munmap(cur, allframesize);
#endif

    return 0;
}

/*
 * change_video_qos: change capturing parameter.
 *   this function is realized by re-setup_video(),
 *   which may cause the memory leak. (future work:-p)
 */
static u_char vqflag = 
(QOS_REQUEST_SIZE | QOS_REQUEST_ZOOM | QOS_REQUEST_COLOR | QOS_REQUEST_FORMAT);

void
change_video_qos()
{
#if 0
  if (debug)
    fprintf(stderr, "change_video_qos: change_flag# %x, vqflag# %x\n",
        req_val.change_flag, vqflag);
#endif

  if (req_val.change_flag & vqflag) {
    close_video();
    setup_video();
  }
}

#if 0
void
convert_BGRAtoRGB()
{
  unsigned char *ptr1, *ptr2, tmp;

  /* covert RGB lines. BGRa -> RGB */
  for(n=0, ptr1=cur, ptr2=cur; n<capture_xsize*capture_ysize; n++, ptr1+=4, ptr2+=3){
    tmp=ptr1[0]; ptr2[0]=ptr1[2]; ptr2[1]=ptr1[1]; ptr2[2]=tmp;
  }
}

void
convert_color()
{
  unsigned char *ptr1, tmp;

  /* covert BGR to RGB */
  n=0;
  ptr1=cur; 
ptr2=0;
  while(n<capture_xsize*capture_ysize) {
    tmp=ptr1[0];
ptr2++;
    ptr1[0]=ptr1[2];
ptr2++;
    ptr1[2]=tmp;
ptr2++;
    n++;
    ptr1+=3;
  }
}

void
convert_BGRtoRGB(void)
{
    unsigned char *p, *pend;
    unsigned char tmp;
    
    p = cur;
    pend = cur + capture_xsize*capture_ysize*3;
    while(p<pend) {
        tmp = *p;
        *p = *(p+2);
        p++;
        p++;
        *p = tmp;
        p++;
    }
}

#endif
