/*
 * Sentry Video System
 *   is controlled QoS according to clients' requests.
 *
 *        original by akira-y@is.aist-nara.ac.jp, 1994,1995
 *        modified by k-chinen@is.aist-nara.ac.jp, 1996
 *        modified by satosi-f@is.aist-nara.ac.jp, 1997,2000
 *
 *   $Id: $
 */


/*
 * video.c - video (chunk of images) handling procedures
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <sys/time.h>
#include <dmedia/vl.h>
#include <errno.h>

#include "alist.h"
#include "sentry.h"
#include "server.h"



extern int debug;
extern int verb;

extern _req_val   req_val;

extern int xsize, ysize;
extern int capture_xsize, capture_ysize;
extern int xl_offset, xr_offset, wide_offset;
extern int frameSize;

static VLServer   svr;
static VLPath     path;
static VLNode     src, drn;
static VLControlValue  val;
static VLBuffer   buffer;


int
check_video(char *vserver)
{
  /* check if vl is running */
  if ((svr = vlOpenVideo(NULL)) == NULL)
    return 1;

  vlCloseVideo(svr);
  return 0;
}

void 
init_video(char *vserver)
{
  extern void setup_video();

  /* connect to the daemon */
  if ((svr = vlOpenVideo(NULL)) == NULL)
    vlPerror("open video");

  /* set up a drain node in memory */
  drn = vlGetNode(svr, VL_DRN, VL_MEM, VL_ANY);
  
  /* set up a source node on any video sources */
  src = vlGetNode(svr, VL_SRC, VL_VIDEO, VL_ANY);
  
  (void)setup_video();

}

void
setup_video(void)
{
  extern void set_color(), set_zoom(), set_size();

  /* create a path using the first device that will support it */
  path = vlCreatePath(svr, VL_ANY, src, drn);

  /* set up the hardware for and define the usage of the path */
  if ((vlSetupPaths(svr, (VLPathList)&path, 1, VL_SHARE, VL_SHARE)) < 0)
    vlPerror("vlSetupPaths");

  if (CHK_CHANGE_FLAG(QOS_SIZE)) set_size();
  if (CHK_CHANGE_FLAG(QOS_ZOOM)) set_zoom();
  if (CHK_CHANGE_FLAG(QOS_COLOR)) set_color();

  frameSize = vlGetTransferSize(svr, path);

  /* Get the video size */
  vlGetControl(svr, path, drn, VL_SIZE, &val);
  xsize = val.xyVal.x;
  ysize = val.xyVal.y;

  fprintf(stderr, "open_video: frameSize = %d (%d x %d)\n",
     frameSize, xsize, ysize);

  /* create and register a buffer for 1 frame */
  buffer = vlCreateBuffer(svr, path, drn, 1);
  if (buffer == NULL)
    vlPerror("vlCreateBuffer");

  vlRegisterBuffer(svr, path, drn, buffer);


  /* begin the data tranfer */
  if (vlBeginTransfer(svr, path, 0, NULL))
    vlPerror("vlBeginTransfer");
}

char *
capture_frame()
{
  VLInfoPtr  info;
  static char *cur=NULL, *prev=NULL;

  if (debug)
    fprintf(stderr, "capture_frame:\n");

  do {
    /* sginap is dependent of SGI machine */
    sginap(1);             /* wait a tick */

    /* info = vlGetNextValid(svr, buffer); */
    info = vlGetLatestValid(svr, buffer);
  } while (!info);

  /* Get a pointer to the frame */
  cur = (char *)vlGetActiveRegion(svr, buffer, info);
  if (cur != prev) {
    if (debug) {
      fprintf(stderr, "capture_frame: prev %d -> cur %d\n", prev, cur);
    }
    /*if (prev) free(prev);*/
    prev = cur;
  }
  return cur;
}

int
release_frame()
{
  /*
   * Finished with frame, unlocked the buffer
   */
  if (debug)
    fprintf(stderr, "release_frame:\n");

  vlPutFree(svr, buffer);
  return 0;
}


/*
 * set video capture parameters 
 */
void
set_size(void)
{
  VLControlValue offset,offset2;

  offset.xyVal.x = offset.xyVal.y = 0;
  offset2.xyVal.x = offset2.xyVal.y = 0;
  wide_offset = 0;

  /* Set the frame size
   *   640x480, 320x240, 212x160, 160x120: normal vision (w:h =  4:3)
   *   640x360, 320x180, 212x120, 160x90: wide vision (w:h = 16:9 )
   */
  switch (GET_QOS(QOS_SIZE)) {
  case SIZE_640x480:
    val.xyVal.x = 640;
    val.xyVal.y = 480;
    break;
  case SIZE_320x240:
    val.xyVal.x = 320;
    val.xyVal.y = 240;
    break;
  case SIZE_212x160:
  default:
    val.xyVal.x = 212;
    val.xyVal.y = 160;
    break;
  case SIZE_160x120:
    val.xyVal.x = 160;
    val.xyVal.y = 120;
    break;
    
  case SIZE_640x360:
    val.xyVal.x = 640;
    val.xyVal.y = 360;
    offset.xyVal.x = 0;
    offset.xyVal.y = 60;         /* 60 = (480 - 360) / 2 */
    offset2.xyVal.y = 60;
    break;
  case SIZE_320x180:
    val.xyVal.x = 320;
    val.xyVal.y = 180;
    offset.xyVal.x = 0;
    offset.xyVal.y = 30;         /* 30 = (240 - 180) / 2 */
    offset2.xyVal.y = 30;
    break;
  case SIZE_212x120:
    val.xyVal.x = 212;
    val.xyVal.y = 120;
    offset.xyVal.x = 0;
    offset.xyVal.y = 20;         /* 20 = (160 - 120) / 2 */
    offset2.xyVal.y = 20;
    break;
    
  case SIZE_160x90:
    val.xyVal.x = 160;
    val.xyVal.y = 90;
    offset.xyVal.x = 0;
    offset.xyVal.y = 15;         /* 0 = (120 - 90) / 2 */
    offset2.xyVal.y = 15;
    break;
  }

  if (verb)
    fprintf(stderr, "set_size: %dx%d+%d+%d\n", val.xyVal.x, val.xyVal.y, offset.xyVal.x, offset.xyVal.y);

  if (vlSetControl(svr, path, drn, VL_OFFSET, &offset) < 0) {
    vlPerror("VL_OFFSET");
    wide_offset = offset.xyVal.y;
    xsize = val.xyVal.x;
    ysize = val.xyVal.y;

    val.xyVal.x += xr_offset + xl_offset + offset2.xyVal.x;
    val.xyVal.y += wide_offset + offset2.xyVal.y;
  }

  if (vlSetControl(svr, path, drn, VL_SIZE, &val) < 0)
    vlPerror("VL_SIZE");

  /*CLR_CHANGE_FLAG(QOS_SIZE);*/
  req_val.change_flag &= ~(0x01 << QOS_SIZE); 
#if 0
  if (debug)
    fprintf(stderr, "set_size: change_flag# %x\n", req_val.change_flag);
#endif
}

void
set_zoom(void)
{

  /* Set the zoom ratio */
  val.fractVal.numerator = 1;
  val.fractVal.denominator = GET_QOS(QOS_ZOOM);

  if (vlSetControl(svr, path, drn, VL_ZOOM, &val) < 0)
    vlPerror("VL_ZOOM");

  /*CLR_CHANGE_FLAG(QOS_ZOOM);*/
  req_val.change_flag &= ~(0x01 << QOS_ZOOM); 
#if 0
  if (debug)
    fprintf(stderr, "set_zoom: change_flag# %x\n", req_val.change_flag);
#endif
}

void 
set_color(void)
{
  /* set the packing the RGB */
  switch (GET_QOS(QOS_COLOR)) {
  case FULL_COLOR:
    val.intVal = VL_PACKING_RGB_8;
    capture_format = SEN_F_RGB;
    capture_corder = SEN_C_RGB;
    break;
  case MONOCHROME:
  default:
    val.intVal = VL_PACKING_Y_8_P;
    capture_format = SEN_F_RGB;
    capture_corder = SEN_C_GRAY;
    break;
  }

  if (vlSetControl(svr, path, drn, VL_PACKING, &val) < 0) 
    vlPerror("VL_PACKING");
  
  /*CLR_CHANGE_FLAG(QOS_COLOR);*/
  req_val.change_flag &= ~(0x01 << QOS_COLOR); 
#if 0
  if (debug)
    fprintf(stderr, "set_color: change_flag# %x\n", req_val.change_flag);
#endif
}

int 
close_video()
{
    if (verb)
      fprintf(stderr, "close_video:\n");
    vlEndTransfer(svr, path);
    vlDeregisterBuffer(svr, path, drn, buffer);
    vlDestroyBuffer(svr, buffer);
    vlDestroyPath(svr, path);
    return 0;
}

/*
 * change_video_qos: change capturing parameter.
 *   this function is realized by re-setup_video(),
 *   which may cause the memory leak. (future work:-p)
 */
static u_char vqflag = 
(QOS_REQUEST_SIZE | QOS_REQUEST_ZOOM | QOS_REQUEST_COLOR | QOS_REQUEST_FORMAT);

void
change_video_qos()
{
#if 0
  if (debug)
    fprintf(stderr, "change_video_qos: change_flag# %x, vqflag# %x\n",
        req_val.change_flag, vqflag);
#endif

  if (req_val.change_flag & vqflag) {
    close_video();
    setup_video();
  }
}
